/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#ifndef ArmeChunk_h
#include "ArmeChunk.h"
#endif //ArmeChunk_h

#ifndef ArmeArchiveHandler_h
#include "ArmeArchiveHandler.h"
#endif //ArmeArchiveHandler_h

#ifndef ArmeDataFileRef_h
#include "ArmeDataFileRef.h"
#endif //ArmeDataFileRef_h

#ifndef TArmeDoc_h
#include "TArmeDoc.h"
#endif //TArmeDoc_h

ArmeChunk::~ArmeChunk()
{
	if (fDataFileRef)
		fDataFileRef->Release();
}


void ArmeChunk::WriteToArchive(ArmeArchiveHandler* archiver) const
{
	archiver->WriteCPPObject(fDataFileRef);
	archiver->WriteArmeTime(fStartTimeInFile);
	archiver->WriteArmeTime(fDuration);
	//from version 0x0001000C
	archiver->WriteArmeTime(fDurationInFile);
}

void ArmeChunk::ReadFromArchive(ArmeArchiveHandler* unarchiver)
{
	fDataFileRef = dynamic_cast<ArmeDataFileRef*>(unarchiver->ReadCPPObject());
	fDataFileRef->AddRef();  //the ArmeDataFileRef doesn't save its reference counter so, we have to add a ref

	fStartTimeInFile = unarchiver->ReadArmeTime();
	fDuration = unarchiver->ReadArmeTime();

	fDurationInFile = unarchiver->ReadArmeTime();
}

ArmeChunk::ArmeChunk(ArmeDataFileRef* dataFileRef, ArmeTime startTimeInFile, ArmeTime duration)
{
	fDataFileRef = dataFileRef;
	fDataFileRef->AddRef();
	fStartTimeInFile = startTimeInFile;
	fDuration = duration;
	fDurationInFile = duration;
}

ArmeChunk::ArmeChunk()
{
	fDataFileRef = NULL;
	fStartTimeInFile = ARME_TIME_UNDEF;
	fDuration = ARME_TIME_UNDEF;
	fDurationInFile = ARME_TIME_UNDEF;
}

ArmeChunk& ArmeChunk::operator=(const ArmeChunk& arg)
{
	fDataFileRef = arg.fDataFileRef;
	fDataFileRef->AddRef();
	fStartTimeInFile = arg.fStartTimeInFile;
	fDurationInFile = arg.fDurationInFile;
	fDuration = arg.fDuration;
	
	return *this;
}

void ArmeChunk::HasMoved(ArmeTime newStartTime)
{
	// when the chunk is moved, we have to make sure that its boundaries remains on frame boundaries.
	// so, we recompute its duration
	ArmeTime endTime = newStartTime + fDuration;
	fDataFileRef->fDoc->RoundArmeTimeToNearestVideoFrame(endTime);
	
	fDuration = endTime - newStartTime;
}

int ArmeChunk::Split(ArmeChunk* result, ArmeTime localSplitTime) const  //returns the number of ArmeChunk created
{
	int nbChunksCreated = 0;

	nbChunksCreated += ExtractSlice(result, ARME_TIME_ZERO, localSplitTime);
	nbChunksCreated += ExtractSlice(result, localSplitTime, fDuration);

	return nbChunksCreated;
}

int ArmeChunk::DeleteSlice(ArmeChunk* result, ArmeTime localSliceStartTime, ArmeTime localSliceEndxTime) const  //returns the number of ArmeChunk created
{
	int nbChunksCreated = 0;

	nbChunksCreated += ExtractSlice(result, ARME_TIME_ZERO, localSliceStartTime);
	nbChunksCreated += ExtractSlice(result, localSliceEndxTime, fDuration);

	return nbChunksCreated;
}

int ArmeChunk::ExtractSlice(ArmeChunk* result, ArmeTime localSliceStartTime, ArmeTime localSliceEndxTime) const  //returns the number of ArmeChunk created
{
	int nbChunksCreated = 0;
	if (localSliceStartTime < ARME_TIME_ZERO)
		localSliceStartTime = ARME_TIME_ZERO;
	if (localSliceEndxTime > fDuration)
		localSliceEndxTime = fDuration;
	if (localSliceEndxTime < ARME_TIME_ZERO)
		localSliceEndxTime = ARME_TIME_ZERO;
	if (localSliceStartTime > fDuration)
		localSliceStartTime = fDuration;
	
	if (localSliceStartTime != localSliceEndxTime)
	{
		result[nbChunksCreated] = *this;
		result[nbChunksCreated].fDuration = localSliceEndxTime - localSliceStartTime;
		result[nbChunksCreated].fDurationInFile = _ArmeTime(long(double(fDurationInFile.data) * double (localSliceEndxTime.data - localSliceStartTime.data) / double (fDuration.data)));
		result[nbChunksCreated].fStartTimeInFile += _ArmeTime(long(double(fDurationInFile.data) * double (localSliceStartTime.data) / double (fDuration.data)));
	
		nbChunksCreated++;
	}
	return nbChunksCreated;
}

int ArmeChunk::TimeStretch(ArmeChunk* result, ArmeTime localStartTime, ArmeTime localEndxTime, ArmeTime localNewEndxTime) const //returns the number of ArmeChunk created
{
	int nbChunksCreated = 0;

	if (localStartTime < ARME_TIME_ZERO)
		localStartTime = ARME_TIME_ZERO;
	if (localEndxTime > fDuration)
		localEndxTime = fDuration;
	if (localEndxTime < ARME_TIME_ZERO)
		localEndxTime = ARME_TIME_ZERO;
	if (localStartTime > fDuration)
		localStartTime = fDuration;
	
	if (localStartTime != localEndxTime)
	{
		//extract the part that is before the part that is stretched
		nbChunksCreated += ExtractSlice(result, ARME_TIME_ZERO, localStartTime);
	
		//extract the part that is stretched
		nbChunksCreated += ExtractSlice(result, localStartTime, localEndxTime);
		//stretch it
		result[nbChunksCreated-1].fDuration = localNewEndxTime - localStartTime;
	
		//extract the part that is after the part that is stretched
		nbChunksCreated += ExtractSlice(result,localEndxTime, fDuration);
	}
	
	return nbChunksCreated;
}


PicHandle TVideoChunk::GetPict(ArmeTime localTime) const
{
	//$$TODO : some caching should be done here

	return fDataFileRef->fDoc->fMasterMovie->GetOriginalTrack()->GetPict((const ArmeVideoDataFileRef*)fDataFileRef, _ArmeTime(fStartTimeInFile.data + localTime.data*fDurationInFile.data/fDuration.data));
}

TVideoChunk::~TVideoChunk()
{
}

void TVideoChunk::WriteToArchive(ArmeArchiveHandler* archiver) const
{
	ArmeChunk::WriteToArchive(archiver);
}

void TVideoChunk::ReadFromArchive(ArmeArchiveHandler* unarchiver)
{
	ArmeChunk::ReadFromArchive(unarchiver);
}

TVideoChunk::TVideoChunk(ArmeDataFileRef* dataFileRef, ArmeTime startTimeInFile, ArmeTime duration) : ArmeChunk(dataFileRef, startTimeInFile, duration)
{
}

TVideoChunk::TVideoChunk() : ArmeChunk()
{
}

