/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#define ArmeChunk_h

#ifndef ArmeChunk_h
#include "ArmeChunk.h"
#endif //ArmeChunk_h

#ifndef ZzTypes_h
#include "ZzTypes.h"
#endif //ZzTypes_h

#ifndef ZzVerbose_h
//#include "ZzVerbose.h"
#endif //ZzVerbose_h

#ifndef ArmeBasicTypes_h
#include "ArmeBasicTypes.h"
#endif //ArmeBasicTypes_h

#ifndef ArmeDataFileRef_h
#include "ArmeDataFileRef.h"
#endif //ArmeDataFileRef_h

#ifndef TArmeDoc_h
#include "TArmeDoc.h"
#endif //TArmeDoc_h

class ArmeArchiveHandler;

/*
An ArmeChunk is a reference on a part of data file; this means that it has a DataFileRef, a fStartTimeInFile and a duration.  For video chunks, the fStartTimeInFile is not used.
*/
class ArmeChunk
{
public :	
	void		WriteToArchive(ArmeArchiveHandler* archiver) const;
	void 		ReadFromArchive(ArmeArchiveHandler* unarchiver);
	ArmeChunk(ArmeDataFileRef* dataFileRef, ArmeTime startTimeInFile, ArmeTime duration);
	ArmeChunk(const ArmeChunk& arg) {*this = arg;};
	ArmeChunk();
	virtual ~ArmeChunk();
public :	
	ArmeDataFileRef*	fDataFileRef; 
	ArmeTime			fStartTimeInFile;  
	ArmeTime			fDurationInFile;
	ArmeTime			fDuration;
	
	virtual int Split(ArmeChunk* result, ArmeTime splitTime) const;  //returns the number of ArmeChunk created
	virtual int DeleteSlice(ArmeChunk* result, ArmeTime sliceStartTime, ArmeTime sliceEndxTime) const;  //returns the number of ArmeChunk created
	virtual int ExtractSlice(ArmeChunk* result, ArmeTime sliceStartTime, ArmeTime sliceEndxTime) const;  //returns the number of ArmeChunk created
	virtual int TimeStretch(ArmeChunk* result, ArmeTime localStartTime, ArmeTime localEndxTime, ArmeTime localNewEndxTime) const; //returns the number of ArmeChunk created

	void	HasMoved(ArmeTime newStartTime);
	ArmeChunk& operator=(const ArmeChunk& arg);

	bool				IsVideo() const {return fDataFileRef->IsVideoRef();};
};


class TVideoChunk : public ArmeChunk
{
public :
	void		WriteToArchive(ArmeArchiveHandler* archiver) const;
	void 		ReadFromArchive(ArmeArchiveHandler* unarchiver);
	TVideoChunk(ArmeDataFileRef* dataFileRef, ArmeTime startTimeInFile, ArmeTime duration);
	TVideoChunk(const TVideoChunk& arg) {*this = arg;};
	TVideoChunk();
	virtual ~TVideoChunk();

	PicHandle GetPict(ArmeTime localTime) const;
};


/*
The ArmeChunkList is a set of chunks.  You can extract, delete or insert chunks in it.  Note that, there is one new member to the ArmeChunkList (that you should not consider for audio) : it's the fStartTime (absolute start time of the list in the doc).  Thanks to this member, I can use absolute coordinates when dealing with video chunk list and you can use relative coordinates when dealing with audio.
*/
template <class T> class TArmeChunkList
{
public :
	/*! @function ctor() : creates an empty chunk list */
				TArmeChunkList();

	/*! @function ctor(const ArmeChunk* chunks, SInt32 nbChunks) : creates an chunk list with a list of chunks */
				TArmeChunkList(const T* chunks, SInt32 nbChunks, ArmeTime listStartTime = ARME_TIME_ZERO);

	/*! @function ctor(const ArmeChunk& chunk) : creates an chunk list with one chunk */
				TArmeChunkList(const T& chunk, ArmeTime listStartTime = ARME_TIME_ZERO);

	/*! @function ctor(other) : we don't allow initialisation */
				TArmeChunkList(const TArmeChunkList& other) { ZzAssert(0); }

	/*! @function ctor(other) : we don't allow initialisation */
				TArmeChunkList(TArmeChunkList& other) { ZzAssert(0); }

				~TArmeChunkList();

	void		Init(const T* chunks, SInt32 nbChunks, ArmeTime listStartTime = ARME_TIME_ZERO); 
				
	/*! @function operator= : we don't allow assignment */
	TArmeChunkList& operator=(const TArmeChunkList& other);

	/*! @function operator= : we don't allow assignment */
	TArmeChunkList& operator=(TArmeChunkList& other);

	/*! @function InsertChunk Inserts a chunk at the given position in the chunk list. */
	void		InsertChunk(const T& chunk, ArmeTime pos);

	/*! @function InsertChunk Inserts a chunk list at the given position in the chunk list */
	void		InsertChunkList(const TArmeChunkList& chunkList, ArmeTime pos);

	/*! @function Split splits the chunk at splitTime*/
	void		Split(ArmeTime splitTime);
	
	/*! @function RemoveUnnescessarySplits removes all unnecessary splits (i.e. : if two consecutive chunks reference two consecutive segments of a same movie, we merge them)*/
	void		RemoveUnnescessarySplits();
	/*! @function ExtractSlide Returns a chunk list made with a subrange of this list*/
	TArmeChunkList ExtractSlice(ArmeTime sliceStart, ArmeTime len) const;

	TArmeChunkList* GetSliceCopy(ArmeTime startTime, ArmeTime endTime) const;
	
	/*! @function ReplaceRangeWithChunk Replaces a range of the list with another chunk.*/
	void		ReplaceRangeWithChunk(const T& chunk, ArmeTime pos, ArmeTime len);
																																
	/*! @function DeleteSlice Removes a slice outta the chunk list. If the slice is greater than- or
	              equal to the whole list, then the function returns true */
	bool		DeleteSlice(ArmeTime sliceStart, ArmeTime length);
	
	void		TimeStretch(ArmeTime startTime, ArmeTime endTime, ArmeTime newEndTime);	
	
	/*! @function GetNbChunks The number of chunks in the list */
	SInt32		GetNbChunks() const;

	TArmeChunkList*				GetCopy() const;

	/*! @function GetChunk Returns a const ref to a chunk. */
	const T&			GetChunk(SInt32 i) const;
	
	const T*			GetChunkAtTime(ArmeTime absoluteTime) const;
	
	ArmeTime					GetStartTimeOfChunk(const T* chunk) const;
	ArmeTime					GetStartTime() const {return fStartTime;};
	ArmeTime					GetEndxTime() const {return GetStartTime() + GetDuration();};
	
	/*! @function GetDuration() Returns the duration of the chunk list */
	ArmeTime					GetDuration() const;

	void						Move(ArmeTime newStartTime);

	// arch stuff
	void						ReadFromArchive(ArmeArchiveHandler* unarchiver);
	void						WriteToArchive(ArmeArchiveHandler* archiver) const;

	bool						IsVideo() const {return GetChunk(0).IsVideo();};


private :
	void		UpdateCachedDuration();

//	virtual T*	CreateChunkArray(int nb);

	T*				fChunks;
	SInt32			fNbChunks;
	ArmeTime		fCachedDuration;
	ArmeTime		fStartTime;		//absolute start position of the chunk list (for audio chunk list, since you work in relative time, set it to ARME_TIME_ZERO (default)
};


template<class T> inline SInt32 TArmeChunkList<T>::GetNbChunks() const
{
	return fNbChunks;
}

template<class T> inline const T& TArmeChunkList<T>::GetChunk(SInt32 i) const
{
	ZzAssert(i >= 0 && i < fNbChunks);
	return fChunks[i];
}

template<class T> inline ArmeTime TArmeChunkList<T>::GetDuration() const 
{
	return fCachedDuration;
}

template<class T> inline TArmeChunkList<T>::~TArmeChunkList()
{
	delete [] fChunks;
}

template<class T> inline TArmeChunkList<T>::TArmeChunkList() :
	fChunks(NULL),
	fNbChunks(0)
{
	Init(NULL, 0, ARME_TIME_ZERO);
}

template<class T> inline TArmeChunkList<T>::TArmeChunkList(const T& chunk, ArmeTime listStartTime)
{
	Init(&chunk, 1, listStartTime);
}

template<class T> inline TArmeChunkList<T>::TArmeChunkList(const T* chunks, SInt32 nbChunks, ArmeTime listStartTime)
{
	Init(chunks, nbChunks, listStartTime);
}

template<class T> inline void TArmeChunkList<T>::Init(const T* chunks, SInt32 nbChunks, ArmeTime listStartTime)
{
	fStartTime = listStartTime;
	fNbChunks = nbChunks;
	fChunks = new T[fNbChunks];
	for(SInt32 i = 0; i < fNbChunks; i++)
	{
		fChunks[i] = chunks[i];
	}
	UpdateCachedDuration();
}

template<class T> inline void TArmeChunkList<T>::InsertChunk(const T& chunk, ArmeTime pos)
{
	TArmeChunkList<T> insertedList(chunk);
	InsertChunkList(insertedList,pos);
}

template<class T> inline void TArmeChunkList<T>::InsertChunkList(const TArmeChunkList<T>& chunkList, ArmeTime pos)
{
	// the slice needs to be totally contained in the list
	// this should be handled by the parent object
		
	// the resulting new chunk list cannot be longer than the sum of 
	// both lists + 2, in case the new list is inserted in the middle of
	// a chunk
	SInt32 maxNewNbChunks = fNbChunks+chunkList.GetNbChunks()+2; 
	T* newChunks = new T[maxNewNbChunks];
	SInt32 newNbChunks = 0;
	
	TArmeChunkList<T> listBeforeInsertPoint = ExtractSlice(fStartTime, pos - fStartTime);
	TArmeChunkList<T> listAfterInsertPoint = ExtractSlice(pos, fCachedDuration - (pos - fStartTime));
	
	
	for(SInt32 i = 0; i < listBeforeInsertPoint.GetNbChunks(); i++)
		newChunks[newNbChunks++] = listBeforeInsertPoint.GetChunk(i);

	// insert the list
	for( SInt32 j = 0; j < chunkList.GetNbChunks(); j++)
		newChunks[newNbChunks++] = chunkList.GetChunk(j);
	
	for(SInt32 i = 0; i < listAfterInsertPoint.GetNbChunks(); i++)
		newChunks[newNbChunks++] = listAfterInsertPoint.GetChunk(i);
	
	ZzAssert(newNbChunks <= maxNewNbChunks);
	delete [] fChunks;
	fNbChunks = newNbChunks;
	fChunks = newChunks;
	UpdateCachedDuration();
}

template<class T> inline void TArmeChunkList<T>::ReplaceRangeWithChunk(const T& chunk, ArmeTime pos, ArmeTime len)
{
	DeleteSlice(pos,len);
	InsertChunk(chunk,pos);
}

template<class T> inline bool TArmeChunkList<T>::DeleteSlice(ArmeTime sliceStart, ArmeTime length)
{
	// the slice needs to be totally contained in the list
	// this should be handled by the parent object
		
	if (length <= ARME_TIME_ZERO)
		return true;
	
	// the resulting new chunk list cannot be longer than the old one + 1 (in case the slice is deleted in the middle of a chunk) 
	SInt32 maxNewNbChunks = fNbChunks+1;
	T* newChunks = new T[maxNewNbChunks];
	SInt32 newNbChunks = 0;
	
	ArmeTime sliceEndx = sliceStart + length;
	ArmeTime curChunkEndx = fStartTime;
	ArmeTime curChunkStartx = fStartTime;

	
	for(SInt32 i = 0; i < fNbChunks; i++)
	{
		const ArmeChunk& curChunk = fChunks[i];
		curChunkEndx += curChunk.fDuration;
		
		newNbChunks += curChunk.ExtractSlice(&(newChunks[newNbChunks]), ARME_TIME_ZERO, sliceStart - curChunkStartx);
		newNbChunks += curChunk.ExtractSlice(&(newChunks[newNbChunks]), sliceEndx - curChunkStartx, curChunk.fDuration);					
		curChunkStartx += curChunk.fDuration;
	}
	
	ZzAssert(newNbChunks <= maxNewNbChunks);

	// delete old array of chunks
	delete [] fChunks;
	fNbChunks = newNbChunks;

	fChunks = newChunks;
	UpdateCachedDuration();
	ZzAssert(fCachedDuration >= ARME_TIME_ZERO);
	return (fCachedDuration == ARME_TIME_ZERO);
}

template<class T> inline void TArmeChunkList<T>::ReadFromArchive(ArmeArchiveHandler* unarchiver)
{
	fStartTime = unarchiver->ReadArmeTime();
	fNbChunks = unarchiver->ReadUInt32();
	fChunks = new T[fNbChunks];
	for (int i=0; i < fNbChunks; ++i)
	{
		fChunks[i].ReadFromArchive(unarchiver);
	}
	fCachedDuration = unarchiver->ReadArmeTime();
}

template<class T> inline void TArmeChunkList<T>::WriteToArchive(ArmeArchiveHandler* archiver) const
{
	archiver->WriteArmeTime(fStartTime);
	archiver->WriteUInt32(GetNbChunks());
	for (int i=0; i < GetNbChunks(); ++i)
	{
		fChunks[i].WriteToArchive(archiver);
	}
	archiver->WriteArmeTime(fCachedDuration);
}


template<class T> inline void TArmeChunkList<T>::UpdateCachedDuration()
{
	fCachedDuration = ARME_TIME_ZERO;
	for(SInt32 i = 0; i < fNbChunks; i++)
	{
		fCachedDuration += fChunks[i].fDuration;
	}
}


template<class T> inline void TArmeChunkList<T>::TimeStretch(ArmeTime startTime, ArmeTime endTime, ArmeTime newEndTime)
{
	// the slice needs to be totally contained in the list
	// this should be handled by the parent object

	// the resulting new chunk list cannot be longer than the old one + 2 
	SInt32 maxNewNbChunks = fNbChunks+2;
	T* newChunks = new T[maxNewNbChunks];
	SInt32 newNbChunks = 0;

	TArmeChunkList<T> listBeforeSliceToStretch = ExtractSlice(fStartTime, startTime - fStartTime);
	TArmeChunkList<T> listSliceToStretch = ExtractSlice(startTime, endTime - startTime);
	TArmeChunkList<T> listAfterSliceToStretch = ExtractSlice(endTime, fStartTime + fCachedDuration - endTime);
	
	
	for(SInt32 i = 0; i < listBeforeSliceToStretch.GetNbChunks(); i++)
		newChunks[newNbChunks++] = listBeforeSliceToStretch.GetChunk(i);

	// insert the list
	for( SInt32 j = 0; j < listSliceToStretch.GetNbChunks(); j++)
		newNbChunks += listSliceToStretch.GetChunk(j).TimeStretch(&(newChunks[newNbChunks]), ARME_TIME_ZERO, listSliceToStretch.GetChunk(j).fDuration, _ArmeTime(long(double(listSliceToStretch.GetChunk(j).fDuration.data) * double(newEndTime.data - startTime.data) / double(endTime.data - startTime.data))) );
	
	for(SInt32 i = 0; i < listAfterSliceToStretch.GetNbChunks(); i++)
		newChunks[newNbChunks++] = listAfterSliceToStretch.GetChunk(i);

	ZzAssert(newNbChunks <= maxNewNbChunks);

	// delete old array of chunks
	delete [] fChunks;
	fNbChunks = newNbChunks;

	fChunks = newChunks;
	UpdateCachedDuration();
	ZzAssert(fCachedDuration >= ARME_TIME_ZERO);
	return;
}

template<class T> inline TArmeChunkList<T> TArmeChunkList<T>::ExtractSlice(ArmeTime sliceStart, ArmeTime len) const
{
	// the slice needs to be totally contained in the list
	if(sliceStart <= fStartTime)
		sliceStart = fStartTime;
	if (sliceStart + len > (fStartTime + fCachedDuration))
		len = fStartTime + fCachedDuration - sliceStart;

	// the extracted slice cannot have more chunks than the list, obviously
	SInt32 maxNewNbChunks = fNbChunks;
	
	T* newChunks = new T[maxNewNbChunks];
	SInt32 newNbChunks = 0;
	if (len >= ARME_TIME_ZERO)
	{
		ArmeTime sliceEndx = sliceStart + len;
		ArmeTime curChunkEndx = fStartTime;
		ArmeTime curChunkStartx = fStartTime;	
		
		for(SInt32 i = 0; i < fNbChunks; i++)
		{
			const T& curChunk = fChunks[i];
			curChunkEndx += curChunk.fDuration;
	
			newNbChunks += curChunk.ExtractSlice(&(newChunks[newNbChunks]), sliceStart - curChunkStartx, sliceEndx - curChunkStartx);
			curChunkStartx += curChunk.fDuration;
		}
	}
	
	ZzAssert(newNbChunks <= maxNewNbChunks);
	TArmeChunkList<T> newChunkList(newChunks,newNbChunks);
	newChunkList.fStartTime = sliceStart;
	delete [] newChunks;
	return newChunkList;
}

template<class T> inline void TArmeChunkList<T>::Split(ArmeTime splitTime)
{
	// the splitTime needs to be in the list
	if ((splitTime <= fStartTime) || (splitTime >= (fStartTime + fCachedDuration)))
		return;

	// the new list can have maximum one chunk more thant the original list
	SInt32 maxNewNbChunks = fNbChunks + 1;
	
	T* newChunks = new T[maxNewNbChunks];
	SInt32 newNbChunks = 0;

	newNbChunks += ExtractSlice(&(newChunks[newNbChunks]), fStartTime, splitTime - fStartTime);
	newNbChunks += ExtractSlice(&(newChunks[newNbChunks]), splitTime, fCachedDuration - (splitTime - fStartTime));

	ZzAssert(newNbChunks <= maxNewNbChunks);

	// delete old array of chunks
	delete [] fChunks;
	fNbChunks = newNbChunks;

	fChunks = newChunks;
	UpdateCachedDuration();
	ZzAssert(fCachedDuration >= ARME_TIME_ZERO);
	return;
}

template<class T> inline void TArmeChunkList<T>::RemoveUnnescessarySplits()
{
	if (fNbChunks < 1)
		return;

	// the new list can have maximum the same number of chunks as the original
	SInt32 maxNewNbChunks = fNbChunks;
	
	T* newChunks = new T[maxNewNbChunks];
	SInt32 newNbChunks = 0;

	ArmeChunk& previousChunk = fChunks[0];
	ArmeChunk& currentChunk = fChunks[1];

	newChunks[0] = previousChunk;
	
	for (int i=1; i < fNbChunks; ++i)
	{
		currentChunk = fChunks[i];
		//if they both reference the same file
		if (previousChunk.fDataFileRef == currentChunk.fDataFileRef)
		{
			//if the previous chunks finishes where the current one starts
			if ((previousChunk.fStartTimeInFile + previousChunk.fDurationInFile) == currentChunk.fStartTimeInFile)
			{
				// and, if they play at the same rate
				if (fabs(double(previousChunk.fDurationInFile)/double(previousChunk.fDuration) - double(currentChunk.fDurationInFile)/double(currentChunk.fDuration)) < 0.001)
				{
					newChunks[newNbChunks].fDuration += currentChunk.fDuration;
					newChunks[newNbChunks].fDurationInFile += currentChunk.fDurationInFile;
				}
			}
		}
		else
			newChunks[++newNbChunks] = currentChunk;

		previousChunk = fChunks[i];
	}
}


template<class T> inline TArmeChunkList<T>*	TArmeChunkList<T>::GetSliceCopy(ArmeTime startTime, ArmeTime endTime) const
{
	TArmeChunkList<T> list = ExtractSlice(startTime, endTime - startTime);

	TArmeChunkList<T>* ret = new TArmeChunkList<T>(list.fChunks, list.fNbChunks, list.fStartTime);

	return ret;
}


template <> inline void TArmeChunkList<TVideoChunk>::Move(ArmeTime newStartTime)
{
	fStartTime = newStartTime;
	if (GetNbChunks() <= 0)
		return;
	//ensure that the chunkList start on the limit of a frame
	GetChunk(0).fDataFileRef->fDoc->RoundArmeTimeToNearestVideoFrame(fStartTime);
	ArmeTime endTime = fStartTime + fCachedDuration;
	//ensure that it will end on the limit of a frame
	GetChunk(0).fDataFileRef->fDoc->RoundArmeTimeToNearestVideoFrame(endTime);
	fCachedDuration = endTime - fStartTime;
	
	ArmeTime tmpStartTime = fStartTime;
	//ensure that all chunks have boundaries on frame boundaries
	for (int i=0; i < GetNbChunks(); ++i)
	{
		fChunks[i].HasMoved(tmpStartTime);
		tmpStartTime += fChunks[i].fDuration;
	}
}

template<class T> inline void TArmeChunkList<T>::Move(ArmeTime newStartTime)
{
	fStartTime = newStartTime;
}

template<class T> inline TArmeChunkList<T>* TArmeChunkList<T>::GetCopy() const
{
	return new TArmeChunkList<T>(fChunks, fNbChunks, fStartTime);
}


template<class T> inline TArmeChunkList<T>& TArmeChunkList<T>::operator=(const TArmeChunkList<T>& other)
{
	ZzAssert(0);
	return *this;
}

template<class T> inline TArmeChunkList<T>& TArmeChunkList<T>::operator=(TArmeChunkList<T>& other)
{
	ZzAssert(0);
	return *this;
}

template<class T> inline const T* TArmeChunkList<T>::GetChunkAtTime(ArmeTime absoluteTime) const
{
	ArmeTime curChunkEndx = fStartTime;
	ArmeTime curChunkStartx = fStartTime;	
	
	for(SInt32 i = 0; i < fNbChunks; i++)
	{
		const T& curChunk = fChunks[i];
		curChunkEndx += curChunk.fDuration;

		if ((absoluteTime >= curChunkStartx) && (absoluteTime < curChunkEndx))
			return &(fChunks[i]);

		curChunkStartx += curChunk.fDuration;
	}
	
	return NULL;
}

template<class T> inline ArmeTime TArmeChunkList<T>::GetStartTimeOfChunk(const T* chunk) const
{
	ArmeTime curChunkStartx = fStartTime;	
	
	for(SInt32 i = 0; i < fNbChunks; i++)
	{
		if (chunk == &(fChunks[i]))
			return curChunkStartx;

		curChunkStartx += fChunks[i].fDuration;
	}
	
	return ARME_TIME_MIN;
}
