/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#define ArmeTimeRange_h

#ifndef ArmeTimeRange_h
#include "ArmeTimeRange.h"
#endif //ArmeTimeRange_h

#ifndef ArmeBasicTypes_h
#include "ArmeBasicTypes.h"
#endif //ArmeBasicTypes_h

/*
A class to maintain information on a time range or a list of time range.
You can compute the union of two time ranges (or list of time ranges) with the Add function.

Exemple : 
ArmeTimeRange A(1, 3);
ArmeTimeRange B(6, 7);
   0123456789
A = ***
B =      **

A.Add(&B);
ArmeTimeRange C(5,6);
   0123456789
A = ***  **
C =     **

A.Add(&C);
ArmeTimeRange D(2,6);
   0123456789
A = *** ***
D =  *****

A.Add(&D);
   0123456789
A = *******


*/
class ArmeTimeRange
{
public :

	ArmeTime		fStart;
	ArmeTime		fEnd;
	ArmeTimeRange*	fNext;
	
	ArmeTimeRange(ArmeTime start, ArmeTime end) {fStart = start; fEnd = end; fNext = NULL;};
	~ArmeTimeRange() {delete fNext;};

	inline ArmeTimeRange* Add(ArmeTimeRange* arg);

protected :
	inline ArmeTimeRange* AddOneRangeToList(ArmeTimeRange* firstOfTheList, ArmeTimeRange* arg);
	inline int AddOneRangeToOneRange(ArmeTimeRange* arg1, ArmeTimeRange* arg2);

private : 
	ArmeTimeRange(const ArmeTimeRange& src) {ZzAssert(0);};
	ArmeTimeRange(ArmeTimeRange& src) {ZzAssert(0);};
};


ArmeTimeRange* ArmeTimeRange::AddOneRangeToList(ArmeTimeRange* firstOfTheList, ArmeTimeRange* arg)
{
	ArmeTimeRange* iter = firstOfTheList;
	bool bHasUpdatedOneRange = false;

	// update each of the range separately
	while (iter)
	{
		if (AddOneRangeToOneRange(iter, arg) == -1)
			bHasUpdatedOneRange = true;
		iter = iter->fNext;
	}

	if (!bHasUpdatedOneRange)
	{
                //We don't need the arg range (other ranges were updated already)
                delete arg;
        
		// the new range to add did not had any area in common with any of the range in the list
		// we simply insert it in the list at the correct position and return
		if (arg->fStart < firstOfTheList->fStart)
		{
			arg->fNext = firstOfTheList;
			return arg; //arg is now the first of the list
		}
		else
		{
			ArmeTimeRange* iter = firstOfTheList;
			while (iter->fNext && (iter->fNext->fStart < arg->fStart)) 
			{
				iter = iter->fNext;
			}
			arg->fNext = iter->fNext;
			iter->fNext = arg;
		}
	}
	else
	{
		// we need to remove all the common area of the ranges that are in the list
		ArmeTimeRange* iter = firstOfTheList;
		while (iter)
		{
			ArmeTimeRange* iter2 = iter->fNext;
			ArmeTimeRange* prevIter2 = iter;
			while (iter2)
			{
				if (AddOneRangeToOneRange(iter, iter2) == -1)
				{
					prevIter2->fNext = iter2->fNext;
					iter2->fNext = NULL;
					delete iter2;
					iter2 = prevIter2->fNext;
				}
				else
				{
					iter2 = iter2->fNext;
				}
			}
			
			iter = iter->fNext;
		}
	}
	
	return firstOfTheList;
}

int ArmeTimeRange::AddOneRangeToOneRange(ArmeTimeRange* arg1, ArmeTimeRange* arg2)
{
	if ((arg1->fStart >= arg2->fStart) && (arg2->fEnd >= arg1->fStart))
	{
		arg1->fStart = arg2->fStart;
		// arg1 has been updated and now covers arg2, arg2 can be removed by the calling function
		return -1;
	}
	if ((arg1->fEnd <= arg2->fEnd) && (arg2->fStart <= arg1->fEnd))
	{
		arg1->fEnd = arg2->fEnd;
		// arg1 has been updated and now covers arg2, arg2 can be removed by the calling function
		return -1;
	}
	if ((arg1->fStart <= arg2->fStart) && (arg2->fEnd <= arg1->fEnd))
	{
		// arg2 is already inside arg1, calling function can remove arg2
		return -1;
	}

	return +1; // arg2 has no common area with ar1, it can be added to the list
}

ArmeTimeRange* ArmeTimeRange::Add(ArmeTimeRange* arg)
{	
	ArmeTimeRange* iterArg = arg;
	ArmeTimeRange* firstOfTheList = this;
	
	// add every member of arg one by one
	while (iterArg)
	{
		ArmeTimeRange* tempNext = iterArg->fNext;
		firstOfTheList = AddOneRangeToList(firstOfTheList, iterArg);
		iterArg = tempNext;	
	}

	return firstOfTheList;
}