/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/
 
#define TArmeDoc_h
#ifndef TArmeDoc_h
#include "TArmeDoc.h"
#endif //TArmeDoc_h

#ifndef ArmeBasicTypes_h
#include "ArmeBasicTypes.h"
#endif //ArmeBasicTypes_h

#ifndef TArmeTLC_h
#include "TArmeTLC.h"
#endif //TArmeTLC_h

#ifndef ZzPtrArray_h
#include "ZzPtrArray.h"
#endif //ZzPtrArray_h

#ifndef ZzLocker_h
#include "ZzLocker.h"
#endif //ZzLocker_h

#ifndef ZzBroadcaster_h
#include "ZzBroadcaster.h"
#endif //ZzBroadcaster_h

#ifndef ArmeArchivable_h
#include "ArmeArchivable.h"
#endif //ArmeArchivable_h

#ifndef MEDocPlayer_h
#include "MEDocPlayer.h"
#endif //MEDocPlayer_h

#ifndef MEWSObjAudioPlayer_h
#include "MEWSObjAudioPlayer.h"
#endif //MEWSObjAudioPlayer_h

#ifndef ZzTimer_h
#include "ZzTimer.h"
#endif ZzTimer_h

#include <map>

class TArmeDoc;
class TArmeWSObj;
class TArmeWSObjPos;
class TArmeMasterMovie;
class MEAudioClip;
class MEMixingInfo;
class MEDocPlayer;
class MEWSObjAudioPlayer;
class ZzSplitTask;
class ArmeDataFileRef;
class ArmeVideoDataFileRef;
class ZzString;
class TVideoXFadeEffectDesc;

/*! 

	@class TArmeDoc
	
	@abstract This class represents the model data on the MediaEngine side.
	
	@discussion The TArmeDoc class encaspulates all the document model data on the
	MediaEngine side.
	
	There can be data outside the MediaEngine that is still considered part of the
	document such as visual representation parameters or command stack for actions
	that modify data NOT in the MediaEngine model.
	
	This document class also features a player object, which is responsible for playing
	the document. This player object is not serialized.
	
	This class is thus linked in a 1:1 fashion with a player object. 
	
	It also inherits from ZzRwLocker, because different threads need to read and write
	from and to the document data. (ZzRwLocker is a read/write locker. Data can be locked read-only
	(where several threads can get this lock) or locked read-write (where only ONE thread can 
	get this lock)).
	
	It also inherits from ZzBroadcaster, which means it can broadcast changes of its data to
	listeners (such as the View layer)

*/

#define kMaxHierarchyDepth 64
#define kMaxWSObjInstances 262144

class TArmeTLC;
class ArmeArchiverHandler;
class MEAudioChunkFileManager;

// some file reading function take a calling thread parameter. For example,
// when reading audio files in the disk read thread and reading them from the
// appkit thread at the same time, we need to have separate file handles in order
// to be able to have different seek positions in the same file.
enum MECallingThreadType
{
	kCallingThread_AppKit = 0,
	kCallingThread_DRT,
	kCallingThread_OfflineAudioSplitTaskThread,
	kCallingThread_OfflineRenderer,
	kCallingThread_Count //KEEP THIS @ LAST POSITION
};

class MIProgressBarDlgManager;

class TArmeDoc : public ZzBroadcaster, public ArmeArchivable
{
public : 
	/*! @function TArmeDoc This function builts an empty TArmeDoc and creates a default player for it */
									TArmeDoc(const char* path, bool bNewFile, WorkingCodecType workingCodec);
									TArmeDoc(ArmeArchiveHandler* unarchiver);
	virtual							~TArmeDoc();


public : // ArmeArchivable interface
	virtual bool			WriteToArchive(ArmeArchiveHandler* archiver) const;
	bool					MakeProjectSelfContained(MIProgressBarDlgManager* progressMngr);

public :
	/*! @function TArmeDoc This function builts a TArmeDoc from a flattened file and creates a default player for it */
	ArmeErr							InitWithFile(const char* path);
	
	/*! @function NotifyModelChange This function sends a generic "model has changed" notification to all listeners. */
	void							NotifyModelChange();
	
	/*! @function NotifyAudioClipDataHasChanged This function tells the world the data of an audio clip has changed. Please do 
                  what you need to do to update the view of the model */
	void							NotifyAudioClipDataHasChanged(MEAudioClip* clip);
	
	/*! @function SecondsToArmeTime This function converts a number of seconds (float value) into a number of ArmeTime units. 
		It depends upon the frame rate so it depends upon the document. */
	ArmeTime						SecondsToArmeTime( float sec ) const;

	/*! @function SecondsToArmeTime This function converts a number of ArmeTime units into a number of seconds (float value)
		It depends upon the frame rate so it depends upon the document. */
	float							ArmeTimeToSeconds( ArmeTime time ) const;
	
	/*! @function ReadFloatFrame Copies the audio data in a float buffer */
	void							ReadFloatFrames(float* dst, SInt64* posInDoc, SInt64* actualFrames );

	/*! @function ReadFlatFloatFrame Copies the audio data in a float buffer */
	void 							ReadFlatFloatFrames(SInt64 posInDoc, SInt64 nbFrames);
	
	/*! @function GetDuration Returns the length of the top level container */
	ArmeTime						GetDuration() const;
	
/*

	these function are shortcuts for the GetTLCPlayer() (the TLC player) functions
	with the same name

*/

	/*! @function Play This method tells the associated graph manager to start playing the document */
	void							Play();
	
	/*! @function Stop stops playback */
	void							Stop();
	
	/*! @function Seek moves the play position (result is not immediate due to buffering). Not thread-safe ! */
	void							Seek(ArmeTime newTime);

	/*! @function SeekToStart moves the play position (result is not immediate due to buffering). Not thread-safe ! */
	void							SeekToStart();
	
	/*! @function SeekToEnd moves the play position (result is not immediate due to buffering). Not thread-safe ! */
	void							SeekToEnd();
	
	/*! @function SetLLocator moves the left locator. Not thread-safe ! */
	void							SetLLocator(ArmeTime leftTime);

	/*! @function SeRLLocator moves the right locator. Not thread-safe !  */
	void							SetRLocator(ArmeTime rightTime);
	
	/*! @function GetLLocator Returns the left locator pos. Not thread-safe ! */
	ArmeTime 						GetLLocator() const;

	/*! @function GetRLocator Returns the right locator pos. Not thread-safe ! */
	ArmeTime 						GetRLocator() const;
	
	/*! @function IsPlaying Returns whether the doc is playing or not */
	bool							IsPlaying() const;
	
	/*! @function ToggleLoop Switches the player state to/from looped */
	bool							ToggleLoop();

	/*! @function IsLooping Returns whether the doc player is in the "looping between locators" mode */
	bool							IsLooping() const;
	
	/*! @function GetPlayPos Returns the current play position in the doc (meaningful even if paused or stopped) */
	ArmeTime						GetPlayPos() const;
	
	/*! @function Reload the buffers so that playback from position "pos" is instantaneous */
	void							Preroll(SInt64 pos);
	void 							Preroll(ArmeTime pos);
	
	void							SetDoesUpdateMovieArmeTime(bool update);
	bool							DoesUpdateMovieArmeTime() const;
	
	/*! @function CancelPreroll clears the buffer in case of seeking to avoid sticky buffers */
//	void							CancelPreroll();
	
public : 
	/*! @function GetTopContainer This function returns
	the topmost container OF THE DOCUMENT (a window may have a different Top Level Container (TLC)) */
	TArmeTLC*						GetTopContainer() const;
	
	/*! @function CreateWSObjFromFile UNUSED (but still working ok if uncommented) : creates a single clip or a group containing several
	    clips (with the same start time) from a single file (that can contain several tracks). */
/*
	TArmeWSObj* 		CreateWSObjFromFile( 
							const ZzString* 			zzPath, 
							const TArmeWSObjPos& 		pos,
							MIProgressBarDlgManager*	progressMngr,
							ArmeErr*					err);
*/
	/*! @function CreateWSObjsFromFile This function, being given a path and a TArmeWSObjPos,
		creates clips, inserts them in the doc and adds them to the retObjects array. */
	ArmeErr		 		CreateWSObjsFromFile( 
							ZzArchPtrArray*				theObjects,
						//	const char*					path, 
							const ZzString*				cstrPath,
							ArmeDataFileRef*			fileRef,
							const TArmeWSObjPos& 		pos,
							MIProgressBarDlgManager*	progressMngr);
					
	/*! @function GetZeroPosition This function is there for convenience. It returns a (0,0) pos in the document just under the TLC */
	TArmeWSObjPos		GetZeroPos() const;
	
	/*! @function GetAndIncrNextIndexForAudioPluginClass Sends the next index for a given plugin class. Every time a plugin of this class
	    is created, the index for the corresponding plugin is increased. They start at 1. */
	UInt32						GetAndIncrNextIndexForAudioPluginClass(VARPIPluginClass* pluginClass);
	void						ReadPluginClassInstanceIndices(ArmeArchiveHandler* unarchiver);
	void						WritePluginClassInstanceIndices(ArmeArchiveHandler* archiver) const;
		
	void				RemoveVideoDataFileRef(ArmeDataFileRef* fileRef);
	ArmeDataFileRef*	GetNewVideoDataFileRef(const char* sourceFile = NULL, bool* isNewRef = NULL);
//	ArmeDataFileRef*	GetNewAudioDataFileRef(); /* OBSOLETE */

	MEAudioChunkFileManager*		AudioChunkFileManager();
	const MEAudioChunkFileManager*	AudioChunkFileManager() const;

	/*! @function SetAudioChunkFileManagerDidDirtifyDoc The MEAudioChunkFileManager ReadFromArchive method
		calls this TArmeDoc method when the user is prompted for a new location when audio files are not found. This
		tells the doc that it is dirty. later in the unarchiving process, the MIMontageDocument checks this and 
		dirtifies the doc accordingly */
	void							SetAudioChunkFileManagerDidDirtifyDoc();
	
	/*! @function GetAudioChunkFileManagerDidDirtifyDoc Allows the MIMontageDocument to know whether the
	doc did get dirty during unarchiving because of relocated external audio files */
	bool							GetAudioChunkFileManagerDidDirtifyDoc() const;

protected :
	MEAudioChunkFileManager*	fAudioChunkFileManager;

	mutable ZzArray<ArmeDataFileRef*>*	fDataFileRefArray;
	
public :
	void				RemoveReferencesOnUnusedFilesInMasterMovie();
	void				DeleteUnusedFiles();
	void				CleanOldTransitionsDesc();
public :

	UInt32				GetStillImageIndex() {return ++fStillImageIndex;};
protected :
	UInt32				fStillImageIndex;
public :
	void 				LockRead();
	void 				UnlockRead();
	void 				LockWrite();
	void 				UnlockWrite();
	
	/*! @function GetProjectFolderPath Returns the current path of the project */
	const char*			GetProjectFolderPath() const;

	/*! @function GetProjectFolderPath Sets the current path of the project (when unarchiving) */
	void 				SetProjectFolderPath(const char* folderPath);

	/*! @function RecreateNonArchivedCachedData This function tells the various objects to 	
		          recreate their cached data (for instance : thumbnails, cached length, etc...).
				  There can be lengthy operations so  */
	void				RecreateNonArchivedCachedData(MIProgressBarDlgManager* progressMngr);
		
	UInt32				GetAudioFrameRate() const;
	float				GetVideoFrameRate() const { return float(GetVideoTimeScale())/float(GetVideoFrameDuration());};
	UInt32				GetVideoTimeScale() const { return fMasterMovie->GetVideoTimeScale();};
	UInt32				GetVideoFrameDuration() const { return fMasterMovie->GetVideoFrameDuration();};
	TimeValue			RoughTimeValueFromArmeTime(ArmeTime time) const;
	TimeValue			TimeValueFromArmeTime(ArmeTime time) const;
	ArmeTime			ArmeTimeFromTimeValue(TimeValue time) const;
	ArmeTime			RoundArmeTimeToNearestVideoFrame(ArmeTime &time) const;

	/*! @function GetSoloObject Returns the address of the soloed object and children,
	    or NULL if there are no soloed objects */
//	TArmeWSObj*		GetSoloObject();
	TArmeWSObj*		GetVideoSoloObject();
	TArmeWSObj*		GetAudioSoloObject();
	
	/*! @function SetVideoSoloObject Tells the doc which is the video soloed object */
	void			SetVideoSoloObject(TArmeWSObj* obj);
	/*! @function SetAudioSoloObject Tells the doc which is the audio soloed object */
	void			SetAudioSoloObject(TArmeWSObj* obj);
		
public :
	TArmeWSObj*					fVideoSoloObject;
	TArmeWSObj*					fAudioSoloObject;

//	ZzArray<TVideoXFadeEffect*>*		fOuttaHierarchyTransitionsPool;
	ZzArray<TVideoXFadeEffect*>*		fNewlyCreatedTransitions;
	ZzArray<TVideoXFadeEffect*>*		fTransitionsToRemove;
	ZzArray<TVideoXFadeEffectDesc*>*	fOldTransitionsDesc;

	void	AddToOldTransitionsDesc(TVideoXFadeEffect* xFade);
	void	RemoveFromOldTransitionsDesc(TVideoObject* obj);
	void	RemoveFromOldTransitionsDesc(TVideoObject* obj1, TVideoObject* obj2);
	TVideoXFadeEffectDesc*	FindEffectDescInOldTransitions(TVideoObject* clip1, TVideoObject* clip2);

	/*! @var fMasterMovie A big movie in which all the video clips are stored*/
	TArmeMasterMovie*			fMasterMovie;
	TArmeMasterMovie*			GetMasterMovie() {return fMasterMovie;};

	bool						fViewerIsVisible;

	std::map<std::string,int>	fPluginClassInstanceIndices;

public :
	/*! @function GetMixingInfo Returns the MEMixingInfo instance responsible to store the dynamic and
		static information needed for the routing and the mixer. The VAO -> RAO mapping, the attenuation
		level for each bus, the current instantaneous value for VU Meter display, etc... 
	*/
	MEMixingInfo* 				GetMixingInfo();
	const MEMixingInfo* 		GetMixingInfo() const;

	/*! @function GetPlayer Returns the MEDocPlayer instance responsible for playing back this document.
		This player, in turn, has references to the discrete players (at least one for the TLC). */
	MEDocPlayer* 				GetPlayer();
	const MEDocPlayer* 			GetPlayer() const;

	/*! @function GetTLCPlayer Returns the MEWSObjAudioPlayer that is created by default and plays back the TLC.
	     Other subplayers can be created for containers, but the main player seen is the TLC one */
	MEWSObjAudioPlayer* 		GetTLCPlayer();
	const MEWSObjAudioPlayer* 	GetTLCPlayer() const;
	
private :
	/*! @var fPlayer The object responsible for playing back this document */
	MEDocPlayer*		fPlayer;
	
	/*! @var fMixingInfo This object stores the VAO -> RAO routing info */
	MEMixingInfo*		fMixingInfo;
		
	/*! @var fTopContainer The top level container for this document */
	TArmeTLC*			fTopContainer;
	
	/*! @var fAudioChunkFileNextIndex The counter that's used to generate new names for audio chunk files */
	SInt32				fMEAudioChunkFileNextIndex;
	ZzArray<UInt32>		fMEAudioChunkFileRefCounts;

	/*! @var fAudioChunkNextIndex The counter that's used to generate new names for audio chunk files */
	UInt32				fAudioChunkNextIndex; /*OBSOLETE*/

	/*! @var fVideoChunkNextIndex The counter that's used to generate new names for video chunk files */
	UInt32				fVideoChunkNextIndex;

	/*! @var fCurrentProjectFolderPath the path to the project folder (NOT SERIALIZED, OF COURSE) */
	char				fCurrentProjectFolderPath[kMaxPathLength];
	
#ifdef DEBUG
	ZzTimer				fLockTimer;
#endif//DEBUG
	ZzRwLocker			fRwLocker;
	
	bool				fDoesUpdateMovieArmeTime;
	
	bool				fWasDocDirtifiedDuringAudioChunkManagerUnarchiving;
/*protected :
	UInt32				fNextVideoFileIndex;
	ZzArray<UInt32>		fVideoFileRefCount;
	ZzArray<char*>		fVideoFileNames;
public :
	void				GetVideoFileNameAndRef(char* filename, UInt32* index);
	void				AddRefToVideoFile(UInt32 index);
	void				RemoveRefToVideoFile(UInt32 index);*/
};
	
inline MEDocPlayer* TArmeDoc::GetPlayer()
{
	return fPlayer;
}

inline const MEDocPlayer* TArmeDoc::GetPlayer() const
{
	return fPlayer;
}

inline MEWSObjAudioPlayer* TArmeDoc::GetTLCPlayer()
{
	return fPlayer->GetWSObjAudioPlayer(0); // the first player is always the TLC player
}

inline const MEWSObjAudioPlayer* TArmeDoc::GetTLCPlayer() const
{
	return fPlayer->GetWSObjAudioPlayer(0); // the first player is always the TLC player
}

	
inline TArmeTLC* TArmeDoc::GetTopContainer() const
{
	return fTopContainer;
}

inline ArmeTime TArmeDoc::SecondsToArmeTime( float sec ) const
{
	// here we implicitly agree that ArmeTime == frame rate
	return _ArmeTime( (sample_t)(sec*(float)GetAudioFrameRate()+0.5f) );  //0.5f is there to round to the closest int
}

inline float TArmeDoc::ArmeTimeToSeconds( ArmeTime time ) const
{
	return (float)(GET_VAL(time))/(float)GetAudioFrameRate();
}

inline TArmeWSObj* TArmeDoc::GetVideoSoloObject()
{
	return fVideoSoloObject;
}

inline TArmeWSObj* TArmeDoc::GetAudioSoloObject()
{
	return fAudioSoloObject;
}

inline void TArmeDoc::Play()
{
	StartMovie(fMasterMovie->GetQTMovie());
	GetTLCPlayer()->Play();
}

inline void TArmeDoc::Stop()
{
	StopMovie(fMasterMovie->GetQTMovie());
	GetTLCPlayer()->Stop();
}

inline void TArmeDoc::SetLLocator(ArmeTime leftTime)
{
	LockWrite();
	GetTLCPlayer()->SetLLocator(leftTime);
	UnlockWrite();
}

inline void TArmeDoc::SetRLocator(ArmeTime rightTime)
{
	LockWrite();
	GetTLCPlayer()->SetRLocator(rightTime);
	UnlockWrite();
}

inline ArmeTime TArmeDoc::GetLLocator() const
{
	return GetTLCPlayer()->GetLLocator();
}

inline ArmeTime TArmeDoc::GetRLocator() const
{
	return GetTLCPlayer()->GetRLocator();
}

inline bool TArmeDoc::IsPlaying() const
{
	return GetTLCPlayer()->IsPlaying();
}

inline bool TArmeDoc::ToggleLoop()
{
	return GetTLCPlayer()->ToggleLoop();
}

inline bool TArmeDoc::IsLooping() const
{
	return GetTLCPlayer()->IsLooping();
}

inline ArmeTime TArmeDoc::GetPlayPos() const
{
	return GetTLCPlayer()->GetPlayPos();
}

//inline void TArmeDoc::CancelPreroll()
//{
//	LockWrite();
//	GetTLCPlayer()->CancelPreroll();
//	UnlockWrite();
//}

inline MEMixingInfo* TArmeDoc::GetMixingInfo()
{
	return fMixingInfo;
}

inline const MEMixingInfo* TArmeDoc::GetMixingInfo() const
{
	return fMixingInfo;
}

inline bool	TArmeDoc::DoesUpdateMovieArmeTime() const
{
	return fDoesUpdateMovieArmeTime;
}


inline MEAudioChunkFileManager* TArmeDoc::AudioChunkFileManager()
{
	ZzRtAssert(fAudioChunkFileManager);
	return fAudioChunkFileManager;
}

inline const MEAudioChunkFileManager* TArmeDoc::AudioChunkFileManager() const
{
	ZzRtAssert(fAudioChunkFileManager);
	return fAudioChunkFileManager;
}
