/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#define TVideoObject_h

#ifndef TVideoObject_h
#include "TVideoObject.h"
#endif //TVideoObject_h

#ifndef TArmeMasterMovie_h
#include "TArmeMasterMovie.h"
#endif //TArmeMasterMovie_h

#include <Quicktime/Quicktime.h>

#ifndef TArmeWSObj_h
#include "TArmeWSObj.h"
#endif //TArmeWSObj_h

#ifndef ZzArray_h
#include "ZzArray.h"
#endif //ZzArray_h

#define VERBOSE_VIDEO_OBJECT

//define some options used in GetTrackRep functions
enum {
	kConsiderOnlyVideoTransition,
	kDontConsiderVideoTransition,
	kNoOption
};


class TVideoEffect;
class TVideoXFadeEffect;
class ArmeTimeRange;

class TVideoObject : public TArmeWSObj
{
public :
    TArmeMasterMovie*	fMasterMovie;

protected :
	bool				fHasBeenEdited;
	ArmeTime			fOldStartTime;

    // fields for UI-based stretching and cropping
    ArmeTime			fEffectiveDuration;
    ArmeTime			fBeginOffset;
    ArmeTime			fEndOffset;

	SInt32					fRefCount;

public :
	virtual void			AddRef();
	virtual void			Release();
	virtual bool			IsInMasterMovie() const;
	SInt32					GetRefCount() const {return fRefCount;};

	virtual void			SetStartTrim(ArmeTime startTrim);
	virtual void			SetEndTrim(ArmeTime endTrim);
	virtual ArmeTime		GetStartTrim() const;
	virtual ArmeTime		GetEndTrim() const;
	
	ArmeTime				fStartTrim;
	ArmeTime				fEndTrim;

	virtual void			RoundTimingInfo();

	/*! @function GranulateTime returns the time where the nearest media sample is*/
	virtual ArmeTime		GranulateTime(ArmeTime t) const;
	
	TVideoXFadeEffect*		FindTransitionBetweenChildren(const TVideoObject* clip1, const TVideoObject* clip2);

	virtual void			TimeStretch(ArmeTime startTime, ArmeTime endTime, ArmeTime newEndTime);

	/*! @function ContainsVideoAtTime*/
	virtual bool			ContainsVideoAtTime(ArmeTime t) const = 0;

	/*! @function ContainsVideoDuring*/
	virtual bool			ContainsVideoDuring(ArmeTime startTime, ArmeTime endTime, bool bConsiderAlsoMutedObjects=false, bool bConsiderAlsoTransparentTextClips = false) const = 0;

	/*! @function Dump Display debug information about the object*/
	virtual void			Dump(int indentLevel = 0) const;

	/*! @function AddListOfStartAndEndTimeOfVideoClip add to the list all the time at which a video child starts or ends*/
	virtual void 			AddListOfStartAndEndTimeOfVideoClip(ArmeTime startTime, ArmeTime endTime, ZzArray<ArmeTime> &list, bool bConsiderAlsoMutedObjects) const;

    /*! @function GetPict	Extract a picture from the tracks at the specified time (timeSec is expressed in seconds in track time)*/
    virtual PicHandle		GetPict(ArmeTime time) const {return NULL;};

	TVideoObject*			GetTransitionFor(TVideoObject* obj1, TVideoObject* obj2);
	void					RemoveUnusedTransitions();

    // functions for UI-based stretching and cropping
    void			SetEffectiveDuration (ArmeTime inEffectiveDuration);
    void			SetBeginOffset (ArmeTime inBeginOffset);
    void			SetEndOffset (ArmeTime inEndOffset);

protected :
	mutable ArmeTime		fCachedAbsoluteStartTimeOfVideoClip;
	mutable ArmeTime		fCachedAbsoluteEndTimeOfVideoClip;
public :
	virtual void			InvalidateCachedTimeInfo();

public :
	virtual ArmeTime		GetAbsoluteStartTimeOfVideoClip(bool bConsiderAlsoMutedObjects = false) const;
	virtual ArmeTime		GetAbsoluteEndxTimeOfVideoClip(bool bConsiderAlsoMutedObjects = false) const;

	/*! @function SetPos Brute-force setter for the position. to be used called by the interface */
	virtual void			SetPos( const TArmeWSObjPos& pos );

    /*! @function Move Move all the tracks*/
	virtual void			Move() = 0;

    /*! @function GetQTMasterMovie Get a regerence on the QT movie of the document*/
    Movie					GetQTMasterMovie() const {return fMasterMovie->GetQTMovie();};

    /*! @function GetMasterMovie Get a regerence on the master movie of the document*/
    TArmeMasterMovie*		GetMasterMovie() const {return fMasterMovie;};

    TVideoObject(const char* path, TArmeDoc* doc);
    virtual			~TVideoObject();
	
	// arch stuff
	TVideoObject(ArmeArchiveHandler* handler);
	virtual bool WriteToArchive(ArmeArchiveHandler* handler) const;
	
	/*! @function Constructor This constructor creates a stand-alone video object (mainly used to be put in the clipboard)*/
	TVideoObject(const char* name);
	
	virtual void 					SetMemento(TArmeWSObjMemento* memento);
	
public  :
	virtual void		SetOuttaHierarchy();
	virtual void		SetBackInHierarchy(TArmeWSObj* parent);

/*	virtual void		_NotifyParentOuttaHierarchy();
	virtual void		_NotifyParentBackInHierarchy();*/

public :
	virtual void 		SetMuted(bool muted);
protected :
	void				_SetMuted(bool muted);
	bool				fIsMutedForSolo;
public :
	void				SetMutedForSolo(bool muted);
	bool				IsVisible() const;
	
	virtual void		Init();

protected :
	virtual void _FillMemento(TArmeWSObjMemento* memento) const;
public :
	virtual TArmeWSObjMemento* GetMemento() const;

};

class TVideoObjectMemento : public TArmeWSObjMemento
{
public :
	TArmeMasterMovie*		fMasterMovie;
	bool					fIsMutedForSolo;
	ArmeTime							fStartTrim;
	ArmeTime							fEndTrim;
							
							TVideoObjectMemento();
							TVideoObjectMemento(ArmeArchiveHandler* unarchiver);
	virtual bool			WriteToArchive(ArmeArchiveHandler* archiver) const;
	virtual ~TVideoObjectMemento() {};
};


