/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/
 
#define MEAudioChunkList_h
#ifndef MEAudioChunkList_h
#include "MEAudioChunkList.h"
#endif //MEAudioChunkList_h

#ifndef ZzTypes_h
#include "ZzTypes.h"
#endif //ZzTypes_h

#ifndef ZzVerbose_h
//#include "ZzVerbose.h"
#endif //ZzVerbose_h

#ifndef ME_AudioChunks_h
#include "ME_AudioChunks.h"
#endif //ME_AudioChunks_h

#ifndef TArmeDoc_h
#include "TArmeDoc.h"
#endif //TArmeDoc_h

class ArmeArchiveHandler;

class MEAudioChunkList
{
	friend std::ostream& operator<<(std::ostream& s, const MEAudioChunkList& audioChunkList);
public :
	/*! @function ctor() : creates an empty chunk list */
						MEAudioChunkList();

	/*! @function ctor() : creates an empty chunk list */
						~MEAudioChunkList();

	/*! @function ctor(const MEAudioChunk* chunks, SInt32 nbChunks) : creates an chunk list with a list of chunks */
						MEAudioChunkList(const MEAudioChunk* chunks, SInt32 nbChunks);

	/*! @function ctor(const MEaudioChunk& chunk) : creates an chunk list with one chunk */
						MEAudioChunkList(const MEAudioChunk& chunk);

	/*! @function ctor(other) : we don't allow initialisation */
						MEAudioChunkList(const MEAudioChunkList& other) { ZzAssert(0); }

	/*! @function IsEmpty : returns whether the list is a totally empty chunk list */
	bool 				IsEmpty() const;

	/*! @function ctor(other) : copies the other list chunks, one by one */
						MEAudioChunkList(MEAudioChunkList& other);
				
	/*! @function operator= : deletes the current chunk list and copies the other list chunks, one by one */
	MEAudioChunkList& 	operator=(const MEAudioChunkList& other);

	/*! @function InsertChunk Inserts a chunk at the given position in the chunk list. */
	void				InsertChunk(const MEAudioChunk& chunk, SInt64 pos);

	/*! @function InsertChunk Inserts a chunk list at the given position in the chunk list */
	void				InsertChunkList(const MEAudioChunkList& chunkList, SInt64 pos);
	
	/*! @function ExtractSlide Returns a chunk list made with a subrange of this list*/
	MEAudioChunkList	ExtractSlice(SInt64 sliceStart, SInt64 len) const;
	
	/*! @function Clears Clears the whole list*/
	void				Clear(); 
	
	/*! @function ReplaceRangeWithChunk Replaces a range of the list with another chunk.*/
	void				ReplaceRangeWithChunk(const MEAudioChunk& chunk, SInt64 pos, SInt64 len);
																																
	/*! @function DeleteSlice Removes a slice outta the chunk list. If the slice is greater than- or
	              equal to the whole list, then the function returns true */
	bool				DeleteSlice(SInt64 sliceStart, SInt64 length);
	
	/*! @function GetNbChunks The number of chunks in the list */
	SInt32				GetNbChunks() const;

	/*! @function GetChunk Returns a const ref to a chunk. */
	const MEAudioChunk&	GetChunk(SInt32 i) const;
	
	/*! @function GetDuration() Returns the duration of the chunk list */
	SInt64				GetDuration() const;

	/*! @function WriteToArchive Write self to archive */
	void				WriteToArchive(ArmeArchiveHandler* archiver) const;

	/*! @function ReadFromArchive Reads self from archive */
	void				ReadFromArchive(ArmeArchiveHandler* archiver);

private :
	void				UpdateCachedDuration();

	MEAudioChunk*	fChunks;
	SInt32			fNbChunks;
	SInt64			fCachedDuration;
};

inline std::ostream& operator<<(std::ostream& s, const MEAudioChunkList& audioChunkList)
{
	s << "(nb=" << audioChunkList.fNbChunks << " -> ";
	for(SInt32 i = 0; i < audioChunkList.fNbChunks-1; i++)
	{
		s << audioChunkList.fChunks[i] << ",";
	}
	s << audioChunkList.fChunks[audioChunkList.fNbChunks-1] << " )";
	return s;
}

inline SInt32 MEAudioChunkList::GetNbChunks() const
{
	return fNbChunks;
}

inline const MEAudioChunk& MEAudioChunkList::GetChunk(SInt32 i) const
{
	ZzAssert(i >= 0 && i < fNbChunks);
	return fChunks[i];
}

inline SInt64 MEAudioChunkList::GetDuration() const
{
	return fCachedDuration;
}

