/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#ifndef MEAudioClip_h
#include "MEAudioClip.h"
#endif //MEAudioClip_h

#ifndef MEAudioClipData_h
#include "MEAudioClipData.h"
#endif //MEAudioClipData_h

#ifndef ZzHelpers_h
#include "ZzHelpers.h"
#endif //ZzHelpers_h

#ifndef TArmeMediaEngine_h
#include "TArmeMediaEngine.h"
#endif //TArmeMediaEngine_h

#ifndef ZzRABI_h
#include "ZzRABI.h"
#endif //ZzRABI_h

#ifndef ME_AudioExportFunctions_h
#include "ME_AudioExportFunctions.h"
#endif //ME_AudioExportFunctions_h

#include <assert.h>
#include <sys/syslimits.h>
//#ifndef ZzDisableVerbose_h
//#include "ZzDisableVerbose.h"
//#endif //ZzDisableVerbose_h

//#ifndef MESplitTaskMergeObjectsInNewMTGAFile_h
//#include "MESplitTaskMergeObjectsInNewMTGAFile.h"
//#endif //MESplitTaskMergeObjectsInNewMTGAFile_h

#undef VER_DEF_TOPIC
#define VER_DEF_TOPIC MEAudioClip

#ifndef ArmeNotifications_h
#include "ArmeNotifications.h"
#endif //ArmeNotifications_h

#ifndef ZzHeapWalker_h
#include "ZzHeapWalker.h"
#endif ZzHeapWalker_h

#ifndef ZzSplitTask_h
#include "ZzSplitTask.h"
#endif //ZzSplitTask_h

#ifndef MIProgressBarDlgManager_h
#include "MIProgressBarDlgManager.h"
#endif //MIProgressBarDlgManager_h

#ifndef ZzException_h
#include "ZzException.h"
#endif //ZzException_h

class MEAudioClipData;
class ZzSplitTask;

MEAudioClip::MEAudioClip(TArmeDoc* doc, MEAudioClipData* data) : TArmeWSObj("",doc)
{
	SetNameCopy("");
	fAudioClipData = data;
	fAudioClipData->Retain(this); // we need to inc refcount
	fVAORouting.index = 0;
	if(data->GetNbChannels() == 2)
		fVAORouting.bStereo = true;
	else
		fVAORouting.bStereo = false;
	NotifyVAORoutingChange();
	InvalidateTrackIndices();
};

MEAudioClip::MEAudioClip(TArmeDoc* doc, MEAudioClipData* data, bool bCopyOrAlias) : TArmeWSObj("",doc)
{
	// 99% of the exceptions here will happen in the MEAudioClipData
	// allocation thus it will be freed ok (provided MEAudioClipData::MEAudioClipData 
	// is well done !)
	// so we don't really need any resource releaser in here except for the 
	// possible mem exception in Retain( ) but this is very unlikely.
	
	SetNameCopy("");
	
	if(bCopyOrAlias)
	{
		fAudioClipData = new MEAudioClipData(data); // ref count is set to 1 by ctor
		fAudioClipData->Retain(this); // tell data we own it
		fAudioClipData->Release(NULL); // release initial ref count
	}
	else
	{	
		ZzAssert(0);
		fAudioClipData = data; // that's not enough since an alias needs to be created on the clip board, too....
		fAudioClipData->Retain(this); // we need to inc refcount
	}
	
	fVAORouting.index = 0;
	if(data->GetNbChannels() == 2)
		fVAORouting.bStereo = true;
	else
		fVAORouting.bStereo = false;
	NotifyVAORoutingChange();
	InvalidateTrackIndices();
};

MEAudioClip::MEAudioClip(TArmeDoc* doc, ZzArchPtrArray* objSet, MIProgressBarDlgManager* progressMngr) : TArmeWSObj("",doc)
{
	// sets the indices to -1
	InvalidateTrackIndices(); // doesn't throw
	
	// initialize the routing . it will be changed shortly because it will use the one
	// of the first object in the selection
	fVAORouting.index = 0;
	fVAORouting.bStereo = true;
	
	// create an empty audio clip data that will hold the rendered audio clip	
	fAudioClipData = new MEAudioClipData(doc); // creates an empty clip data. ref count set to 1
	// since an exception can happen later, let's create a RABI object
	// which will release the fAudioClipData with a NULL param whatever happens
	// if everything is ok, we can simply call Retain(this) at the end
	// of this ctor
	ZzRABI_AudioClipDataRelease fAudioClipDataReleaser(fAudioClipData,NULL); // this means the dtor will call fAudioClipData->Release(NULL)
					
	// first, we create a new file index to store the mixdown result :
	// let's create a new chunk
	MEAudioChunk chunk = GetDoc()->AudioChunkFileManager()->AllocateNewZeroRefMTGAChunk();
	// please note the initial chunk refcount is 0, so we really don't need to do anything
	// at this point. it will be deleted UNLESS we retain it in a chunk list
			
	// retrieve the path of the chunk
	// PLEASE NOTE this file does not exist at this point !
	char path[PATH_MAX];
	chunk.GetPath(path, PATH_MAX);
	
	// now let's create the zzsplittask that merges the objects into the file
	UInt32 outputChannels = 0;
	MEVAORouting newRouting;// the new vao routing the clip needs to be assigned to. this is the routing of the first object in the selection
		
	// this function will raise an exception if something goes wrong
	MEAudioExportInMTGAFile(path,objSet,false,0,0,&newRouting, progressMngr); // false,0,0 means no slice
	
	// creare a RABI that deletes the file if something goeswrong
	ZzRABI_RemoveFile mtgaFileRemover(path);
	// if we reach this line, it means the file has been created successfully.
	// otherwise it will have been deleted automatically when leaving the 
	// MEAudioExportInMTGAFile scope. that's how it should work
	
	// PLEASE NOTE the audiodata being Released in case of troubles means 
	// the chunkfile will be deleted someday anyway but we delete it right
	// here with mtgaFileRemover because it can be a big file (we don't 
	// postpone the deletion in TArmeDoc::RemoveUnusedFiles()

	// MEAudioExportInMTGAFile tells the new clip which vao it is routed to
	fVAORouting = newRouting;
	
	// some vao routings have changed. warn everyone
	NotifyVAORoutingChange();
	
	if(newRouting.bStereo)
		outputChannels = 2;
	else
		outputChannels = 1;
		
	// let's open the file, check its duration and fill the chunk "duration"
	MECompoundAudioFile file;
	file.OpenReadMTGA(path);// this can throw. no problem
	chunk.duration = file.GetNbFrames();
	if(file.GetNbChannels() != outputChannels)
	{
		file.Close();
		ZZ_THROW_1ARG(ZzAudioClipImportException,"Serious rendering problem");
	}

	file.Close();
	
	// create a chunk list that only contains this chunk
	MEAudioChunkList list(chunk);
	
	// assign chunk list to audio data
	GetData()->AdoptNewChunkList(list);
	GetData()->SetNbChannels(outputChannels);
	
	// let's compute the clip data main preview
	ZzSplitTask* splitTaskComputeMainPreview;
	fAudioClipData->ComputeMainPreviewSplitTask(&splitTaskComputeMainPreview); // this can throw, too
	ZzSplitTask::ExecuteInProgressManagerContext(splitTaskComputeMainPreview,progressMngr); // this can throw
	
	mtgaFileRemover.SaveResource(); // do NOT delete the file
	fAudioClipData->Retain(this); // tell data we own it. now it won't be deleted anymore
}

MEAudioClip::MEAudioClip(const MEAudioChunk& chunk, SInt32 nbChannels, MIProgressBarDlgManager* progressMngr) : TArmeWSObj("",chunk.doc)
{
	SetNameCopy("");
	
	try
	{
		fAudioClipData = NULL;
		fAudioClipData = new MEAudioClipData(chunk.doc, MEAudioChunkList(chunk), nbChannels, progressMngr); // ref count set to 1
		fAudioClipData->Retain(this); // tell the data we own it
		fAudioClipData->Release(NULL); // release the original refcount
	}
	catch( ZzException& e)
	{
		// no need to do anything. the MEAudioClip dtor will handle this.
		e.DefaultCatcher();
		throw(e);
		//err->type = kArmeErrUnknown; // this will ensure the audio clip IS deleted
	}
		
	InvalidateTrackIndices();
}

MEAudioClip::MEAudioClip(const ZzString& name, const FSSpec& movieFile, int trackIndex, TArmeDoc* doc, MIProgressBarDlgManager* progressMngr) :
	TArmeWSObj("",doc)
{
	SetNameCopy(name.getUTF8String());
	
	// this this throws and exception, both the clip data and the clip itself will be deleted
	fAudioClipData = new MEAudioClipData(doc, movieFile, trackIndex, progressMngr); // this can throw an exception
	
	// if we reach this point, the data has been created. "delete" can't be used in case of exceptions 
	// since it's a "refcounter" object
	// thus we need to use a rabi object that releases the data. it needs to be released(NULL) because
	// the first owner is always NULL.
	ZzRABI_AudioClipDataRelease	audioClipDataReleaser(fAudioClipData,NULL); // will call Release(NULL) on the clip data
	
	// assign the default routing
	fVAORouting.index = 0;
	if(fAudioClipData->GetNbChannels() == 2)
		fVAORouting.bStereo = true;
	else
		fVAORouting.bStereo = false;
	NotifyVAORoutingChange();

	InvalidateTrackIndices();
	
	// erverything was ok. we can now retain the object. if an exception is thrown in Retain, it still works .
	fAudioClipData->Retain(this); // tell data we own it
}

MEAudioClip::MEAudioClip(ArmeArchiveHandler* unarchiver) :
	TArmeWSObj(unarchiver)
{
	ArmeArchivable* archObj;
	archObj = unarchiver->ReadCPPObject();
	fAudioClipData = dynamic_cast<MEAudioClipData*>(archObj);
	fAudioClipData->Retain(this);
	if(archObj)
	{
		ZzAssert(fAudioClipData);
	}
	InvalidateTrackIndices();
}

bool MEAudioClip::WriteToArchive(ArmeArchiveHandler* archiver) const
{
	TArmeWSObj::WriteToArchive(archiver);
	archiver->WriteCPPObject(fAudioClipData);
	return true;
}

/*
MEAudioClip::MEAudioClip(const char* path, TArmeDoc* doc, MIProgressBarDlgManager* progressMngr, ArmeErr* err) :
	TArmeWSObj("",doc)
{
	try
	{
		// if fAudioClipData ctor can't finish, let's make sure the MEAudioClip dtor will behave correctly
		fAudioClipData = NULL;
	
		// set name from path
		char clipName[64];
		ZzGetFileNameFromPath(clipName,63,path);
		SetNameCopy(clipName);

		heapwalker(fAudioClipData = new MEAudioClipData(doc,path,progressMngr,err)); // this can throw
		
		// at this point, fAudioClipData is well created
		fAudioClipData->Retain(this); // tell data we own it
		fAudioClipData->Release(NULL); // release initial ref count
		if ( err->type != kArmeOK ) 
		{
			ZZ_THROW(ZzException); // better have the MEAudioClipData 
		}
		
		fVAORouting.index = 0;
		if(fAudioClipData->GetNbChannels() == 2)
			fVAORouting.bStereo = true;
		else
			fVAORouting.bStereo = false;
		NotifyVAORoutingChange();
		
		InvalidateTrackIndices();
	}
	catch (ZzException* e)
	{
		e->DefaultCatcher();
		delete e;
		err->type = kArmeErrUnknown;
	}
}
*/

void MEAudioClip::InvalidateTrackIndices()
{
	for(UInt32 playerIndex = 0; playerIndex < kMEDocPlayer_MaxNbAudioPlayers; playerIndex++)
	{
		fTrackIndex[playerIndex] = -1;
	}
}

MEAudioClip::~MEAudioClip()
{
	// decr the ref cnt of clip data
	fAudioClipData->Release(this);
}

TArmeWSObj* MEAudioClip::GetSliceCopy(ArmeTime sliceStartTime, ArmeTime sliceEndxTime, const char* tempDirectoryName, MIProgressBarDlgManager* progressMngr) const
{
	vernl(6,"MEAudioClip::GetSliceCopy() (1)");
	// if the slice is outside of the clip, let's return 0.
	if(sliceStartTime >= GetAbsoluteEndxTime())
		return NULL;
	if(sliceEndxTime <= GetAbsoluteStartTime())
		return NULL;

	bool bCopyTheWholeStuff = false; // whether the whole object gets copied

	ArmeTime copyStartTime;  //the relative pos of the copied object
	ArmeTime absoluteStartTime = GetAbsoluteStartTime();
	ArmeTime absoluteEndxTime = GetAbsoluteEndxTime();

	if ((sliceStartTime == ARME_TIME_UNDEF) || (sliceStartTime == ARME_TIME_MIN))
		copyStartTime = GetPos().fStartTime;
	else if (sliceStartTime < absoluteStartTime)
	{
		if (sliceStartTime < fPos.fParent->GetAbsoluteStartTime())
			copyStartTime = absoluteStartTime - fPos.fParent->GetAbsoluteStartTime();
		else
			copyStartTime = absoluteStartTime - sliceStartTime;
	}
	else
		copyStartTime = ARME_TIME_ZERO;
	
	if ((sliceStartTime == ARME_TIME_UNDEF) || (sliceStartTime < absoluteStartTime))
		sliceStartTime = absoluteStartTime;
	
	if (sliceEndxTime == ARME_TIME_UNDEF || sliceEndxTime > absoluteEndxTime)
		sliceEndxTime = absoluteEndxTime;
		
	if(sliceStartTime == absoluteStartTime && sliceEndxTime == absoluteEndxTime)
		bCopyTheWholeStuff = true;

	if (sliceStartTime == sliceEndxTime)
		return NULL;

	// let's convert to absolute
	SInt64 sliceStartTime64 = GET_VAL(sliceStartTime) -GET_VAL(absoluteStartTime);
	if(sliceStartTime64 < 0)
		sliceStartTime64 = 0;
			
	SInt64 sliceEndxTime64 = GET_VAL(sliceEndxTime) - GET_VAL(absoluteStartTime);
	if(sliceEndxTime64 > GET_VAL(GetDuration()))
		sliceEndxTime64 = GET_VAL(GetDuration());

	if(sliceStartTime64 > sliceEndxTime64)
	{
		SInt64 tmp = sliceStartTime64;
		sliceStartTime64 = sliceEndxTime64;
		sliceEndxTime64 = tmp;
	}
	if(sliceStartTime64 == sliceEndxTime64)
	{
		vernl(1,"WARNING : MEAudioClip::GetSliceCopy() with a zero-length slice on this clip. Returning NULL");
		return NULL;
	}
	
	SInt64 sliceLen64 = sliceEndxTime64 - sliceStartTime64;
	ZzAssert(sliceLen64 > 0);
	
	// THESE TIMES ARE ABSOLUTE TIMES !

	MEAudioClipData* clipData;
	if(bCopyTheWholeStuff)
	{
		// we simply copy the original data
		clipData = new MEAudioClipData(fAudioClipData);
	}
	else
	{

		MEAudioChunkList slicedList = GetData()->GetChunkList().ExtractSlice(sliceStartTime64,sliceLen64);
		// now create a MEAudioClipData for clipboard with the sliced list
		clipData = new MEAudioClipData(GetDoc(),slicedList,fAudioClipData->GetNbChannels(),progressMngr); // ref cnt becomes 0
		ZzAssert(clipData->GetDuration() > 0);
	}

	// and, eventually, create an audio clip that references this data, but with a NULL document (it's on the clip board)
	MEAudioClip* clip = new MEAudioClip(NULL,clipData,true); // the "true" param means a copy is created, not an alias
	clipData->Release(NULL); // release the original ref count --> deletion! 
	clip->SetMuted(GetMuted());
	clip->SetNameCopy(GetName());
	clip->fPos.fStartTime = copyStartTime;
	clip->fPos.fParent = NULL;
	clip->fCachedAbsoluteStartTime = sliceStartTime;
	clip->SetVAORouting(GetVAORouting());

	vernl(6,"MEAudioClip::GetSliceCopy() (4)");
	return clip;
}

ArmeTime MEAudioClip::InsertSlice(ArmeTime insertTime, const TArmeWSObj* obj, MIProgressBarDlgManager* progressMngr)
{
	// call SetNewDocForClipboardPaste on the pasted data if needed
	const MEAudioClip* audioObj = dynamic_cast<const MEAudioClip*>(obj);
	
	if(audioObj->GetData()->GetDoc() != GetDoc())
	{
		vernl(0,"FATAL ERROR : interdoc cut/copy/paste is not supported yet.");
	}
	
	ZzAssert(audioObj->GetData()->GetDoc() == GetDoc());
	
	ZzSplitTask* splitTask;
	fAudioClipData->InsertChunkList(audioObj->GetData()->GetChunkList(), GET_VAL(insertTime) - GET_VAL(GetAbsoluteStartTime()), &splitTask );
	ArmeErr err = ZzSplitTask::ExecuteInProgressManagerContext(splitTask,progressMngr);
	// ***marcp*** il faut grer les erreurs
	if(err.type != kArmeOK)
	{
		ZzAssert(0);
	}
	return _ArmeTime(GET_VAL(insertTime) + GET_VAL(obj->GetDuration()));
}

ArmeTime MEAudioClip::GetDuration() const
{
	return _ArmeTime(fAudioClipData->GetDuration());
}

bool MEAudioClip::DeleteSlice(ArmeTime startTime, ArmeTime endTime, MIProgressBarDlgManager* progressMngr)
{
	bool bDeleteWholeObject = true;
	bool bDeleteStartTrim = false;
	bool bDeleteEndTrim = false;
	ArmeTime absoluteStartTime = GetAbsoluteStartTime();
	ArmeTime absoluteEndxTime = GetAbsoluteEndxTime();

	if ((startTime == ARME_TIME_UNDEF) || (startTime == ARME_TIME_MIN))
		startTime = absoluteStartTime;
	if ((endTime == ARME_TIME_UNDEF) || (endTime == ARME_TIME_MAX))
		endTime = absoluteEndxTime;

	ArmeTime cutStartTime = startTime;

	if (cutStartTime <= absoluteStartTime)
		cutStartTime = absoluteStartTime;
	else
		bDeleteWholeObject = false;

	if (endTime >= absoluteEndxTime)
		endTime = absoluteEndxTime;
	else
		bDeleteWholeObject = false;

	if ((cutStartTime < absoluteEndxTime) && (endTime >= absoluteEndxTime))
		bDeleteEndTrim = true;
	if ((cutStartTime <= absoluteStartTime) && (endTime > absoluteStartTime))
		bDeleteStartTrim = true;

	if (bDeleteWholeObject)
	{
		return true;  // the object will be deleted outside this routine
	}
	else
	{
		ZzAssert(! (bDeleteStartTrim && bDeleteEndTrim) ); // 
		ArmeTime startTrimToDelete = ARME_TIME_ZERO;  
		
		// if we zero the start trim, then first move the clip to the left
		// by the trim amount
		if(bDeleteStartTrim)
		{
			startTrimToDelete = fAudioClipData->GetStartTrim();
			cutStartTime -= startTrimToDelete;
			fPos.fStartTime -= startTrimToDelete;
			InvalidateCachedTimeInfo();
		}
		
		// if we zero the end trim, we need to increase the slice deleted 
		// endTime by the end trim amount
		if(bDeleteEndTrim)
			endTime += fAudioClipData->GetEndTrim();
			
		// the zero the start trims if needed
		if(bDeleteStartTrim)
			fAudioClipData->SetStartTrim(ARME_TIME_ZERO,false);
		if(bDeleteEndTrim)
			fAudioClipData->SetEndTrim(ARME_TIME_ZERO,false);
			
		ZzSplitTask* splitTask = NULL;
		if (endTime > cutStartTime)
			fAudioClipData->DeleteSlice(GET_VAL(cutStartTime)-GET_VAL(GetAbsoluteStartTime()),GET_VAL(endTime)-GET_VAL(cutStartTime),&splitTask);
		// if we have deleted 'something' before the video clip, the start of the clip must be moved to the start of what has been deleted
		if (cutStartTime != startTime)
		{
			ZzAssert(fPos.fParent);

//			fPos.fStartTime = startTime - fPos.fParent->GetAbsoluteStartTime();
/*			if (endTime > absoluteStartTime)
				fPos.fStartTime -= (absoluteStartTime - startTime) - startTrimToDelete; //startTime - fPos.fParent->GetAbsoluteStartTime();
			else
				fPos.fStartTime -= (endTime - startTime); //startTime - fPos.fParent->GetAbsoluteStartTime();*/

			if (endTime > absoluteStartTime)
			{
				if (startTime < fPos.fParent->GetAbsoluteStartTime())
					fPos.fStartTime = ARME_TIME_ZERO;
				else
				{
					ArmeTime newAbsoluteStartTime = startTime; //startTime - fPos.fParent->GetAbsoluteStartTime();
					fPos.fStartTime = newAbsoluteStartTime - fPos.fParent->GetAbsoluteStartTime();
				}
			}
			else
			{
				if (startTime < fPos.fParent->GetAbsoluteStartTime())
				{
					if (endTime > fPos.fParent->GetAbsoluteStartTime())
						fPos.fStartTime -= (endTime - fPos.fParent->GetAbsoluteStartTime());
				}
				else			
					fPos.fStartTime -= (endTime - startTime); //startTime - fPos.fParent->GetAbsoluteStartTime();
			
			}
			fPos.fParent->_NotifyChildPosHasChanged();
			InvalidateCachedTimeInfo();
		}
		if (splitTask)
		{
			ArmeErr err = ZzSplitTask::ExecuteInProgressManagerContext(splitTask,progressMngr);
			// ***marcp*** il faut grer les erreurs
			if(err.type != kArmeOK)
			{
				ZzAssert(0);
			}
		}
		return false;
	}
}

void MEAudioClip::_SetLengh(ArmeTime len)
{
	ZzAssert(0);
}

SInt32 MEAudioClip::GetType() const
{
	return kAudioClip;
}

MEAudioClip::MEAudioClip(MEAudioClip&) : TArmeWSObj("",NULL)
{
	ZzAssert(0);
}

MEAudioClip::MEAudioClip(const MEAudioClip&) : TArmeWSObj("",NULL)
{
	ZzAssert(0);
}

void MEAudioClip::Move()
{	
	TArmeWSObj::Move();
//	GetDoc()->Notify(kNotification_AudioClipMoved,(void*)this);
}

void MEAudioClip::CountPostUnarchivingLengthyTasks(UInt32* nbOfTasks)
{
	*nbOfTasks = *nbOfTasks + 1;
}

void MEAudioClip::RecreateNonArchivedCachedData(MIProgressBarDlgManager* progressMngr)
{
	ZzSplitTask* splitTask;
	fAudioClipData->ComputeMainPreviewSplitTask(&splitTask);
	progressMngr->EnterSubTask();
	if(splitTask)
	{
		ArmeErr err = ZzSplitTask::ExecuteInProgressManagerContext(splitTask,progressMngr);
		// ***marcp*** il faut grer les erreurs
		if(err.type != kArmeOK)
		{
			ZzAssert(0);
		}
	}
	progressMngr->LeaveSubTask();
}

SInt32 MEAudioClip::GetAudioTrackForPlayer(UInt32 playerIndex) const
{	
	ZzAssert(playerIndex < kMEDocPlayer_MaxNbAudioPlayers);
	return fTrackIndex[playerIndex];
}

bool MEAudioClip::ContainsAudioDuring(ArmeTime startTime, ArmeTime endTime, bool bConsiderAlsoMutedObjects) const
{
	if (bConsiderAlsoMutedObjects || !IsMuted())
	{
		if ((startTime != ARME_TIME_UNDEF) && (endTime != ARME_TIME_UNDEF))
		{
			if (GetAbsoluteStartTime() >= startTime) 
			{
				if (GetAbsoluteStartTime() <= endTime)
					return true;
			}
			else if (GetAbsoluteEndxTime() >= startTime)
			{
				return true;
			}
		}
		else if ((startTime == ARME_TIME_UNDEF) && (endTime != ARME_TIME_UNDEF))
		{
			if (GetAbsoluteStartTime() <= endTime)
				return true;
		}
		else if ((startTime != ARME_TIME_UNDEF) && (endTime == ARME_TIME_UNDEF))
		{
			if (GetAbsoluteEndxTime() >= startTime)
				return true;
		}
		else
			return true; //both startTime and endTime are undef
	}
	return false;
}

#ifdef OLD_VAO_RAO
bool MEAudioClip::ContainsAudioObjectsRoutedTo(bool bStereo, SInt32 vaoIndex) const
{
	return( fVAORouting.index == vaoIndex && fVAORouting.bStereo == bStereo);
}

void MEAudioClip::GebruikenVAOs(ZzArray<SInt32>* monoVAOsUsed, ZzArray<SInt32>* stereoVAOsUsed) const
{
	if(fVAORouting.bStereo)
	{
		if(stereoVAOsUsed->FindIndexForItem(fVAORouting.index) == -1)
			stereoVAOsUsed->AddItem(fVAORouting.index);
	}
	else
	{
		if(monoVAOsUsed->FindIndexForItem(fVAORouting.index) == -1)
			monoVAOsUsed->AddItem(fVAORouting.index);
	}
};
#endif //#ifdef OLD_VAO_RAO

void MEAudioClip::_NotifyParentPosHasChanged()
{
	InvalidateCachedTimeInfo();
//	GetDoc()->Notify(kNotification_AudioClipMoved,(void*)this);
}

void MEAudioClip::SetMuted(bool muted)
{
	TArmeWSObj::SetMuted(muted);
	if(muted)
		fAudioGain = 0.0;
	else
		fAudioGain = 1.0;
}


void MEAudioClip::SetStartTrim(ArmeTime startTrim)
{
	fAudioClipData->SetStartTrim(startTrim);
//	GetDoc()->Notify(kNotification_AudioClipMoved,(void*)this);
}

void MEAudioClip::SetEndTrim(ArmeTime endTrim)
{
	fAudioClipData->SetEndTrim(endTrim);
//	GetDoc()->Notify(kNotification_AudioClipMoved,(void*)this);
}

ArmeTime MEAudioClip::GetStartTrim() const
{
	return fAudioClipData->GetStartTrim();
}

ArmeTime MEAudioClip::GetEndTrim() const
{
	return fAudioClipData->GetEndTrim();
}

TArmeWSObjMemento* MEAudioClip::GetMemento() const
{
	TArmeWSObjMemento* memento = new MEAudioClipMemento();
	_FillMemento(memento);
	return memento;
}

void MEAudioClip::_FillMemento(TArmeWSObjMemento* memento) const
{
	MEAudioClipMemento* audioMemento = dynamic_cast<MEAudioClipMemento*>(memento);
	ZzAssert(audioMemento);

	TArmeWSObj::_FillMemento(memento);
	
	audioMemento->fAudioClipDataMemento = fAudioClipData->CreateMemento();
}

MEAudioClipMemento::MEAudioClipMemento()
{
	// dont have anything to do.
}

MEAudioClipMemento::~MEAudioClipMemento()
{
	delete fAudioClipDataMemento;
}

MEAudioClipMemento::MEAudioClipMemento(ArmeArchiveHandler* handler) :
	TArmeWSObjMemento(handler)
{
	ArmeArchivable* archObj = handler->ReadCPPObject();
	fAudioClipDataMemento = dynamic_cast<MEAudioClipDataMemento*>(archObj);
	ZzAssert(fAudioClipDataMemento);
}

bool MEAudioClipMemento::WriteToArchive(ArmeArchiveHandler* handler) const
{
	TArmeWSObjMemento::WriteToArchive(handler);
	handler->WriteCPPObject(fAudioClipDataMemento);
	return true;
}

void MEAudioClip::SetMemento(TArmeWSObjMemento* memento)
{
	MEAudioClipMemento* audioMemento = dynamic_cast<MEAudioClipMemento*>(memento);
	ZzAssert(audioMemento);
	
	ZzSplitTask* splitTask;
	fAudioClipData->SetMemento(audioMemento->fAudioClipDataMemento,&splitTask);
	MIProgressBarDlgManager mngr("Applying changes");
	ArmeErr err = ZzSplitTask::ExecuteInProgressManagerContext(splitTask,&mngr);
	// ***marcp*** il faut grer les erreurs
	if(err.type != kArmeOK)
	{
		ZzAssert(0);
	}
	TArmeWSObj::SetMemento(memento);
}

void MEAudioClip::SetOuttaHierarchy()
{
	GetDoc()->LockWrite();
	InvalidateTrackIndices();
	GetDoc()->UnlockWrite();
	
	TArmeWSObj::SetOuttaHierarchy();
}

void MEAudioClip::CheckAudioDataNeedsPreviewComputingRecursive()
{
	if(fAudioClipData->GetMainPreview() == NULL)
		fAudioClipData->ComputeMainPreview();
}
	
	
	
