/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#ifndef MEAudioObjRenderer_h
#include "MEAudioObjRenderer.h"
#endif //MEAudioObjRenderer_h

#ifndef ZzHelpers_h
#include "ZzHelpers.h"
#endif //ZzHelpers_h

#ifndef MEMTGAAudioFile_h
#include "MEMTGAAudioFile.h"
#endif //MEMTGAAudioFile_h

#ifndef MEAudioClip_h
#include "MEAudioClip.h"
#endif //MEAudioClip_h

#ifndef MIProgressBarDlgManager_h
#include "MIProgressBarDlgManager.h"
#endif //MIProgressBarDlgManager_h

#ifndef DSP_Globals_h
#include "DSP_Globals.h"
#endif //DSP_Globals_h

const SInt32 kOfflineRenderingBufferNbFrames = 65536;


MEAudioObjRenderer::MEAudioObjRenderer(
	ZzArchPtrArray* objSet,
	bool			bSliceSel,
	SInt64			sliceStartTime,
	SInt64			sliceEndxTime,
	UInt32			maxRequestedBufferFrames,
	bool			bApplyRAOVAOGain,
	UInt32			nbChannels)
{
	UInt32 i;

	fMaxRequestedBufferFrames 	= maxRequestedBufferFrames;
	fObjSet 					= objSet;
	fApplyRAOVAOGain 			= bApplyRAOVAOGain;
	fNbChannels 				= nbChannels;
	
	SInt64 absStartTime = 0;
	SInt64 absEndxTime = 0;
	
	if(!bSliceSel)
	{
		GetObjectSetTimeSpan(fObjSet,&absStartTime,&absEndxTime); // static call
	}
	else
	{
		absStartTime = sliceStartTime;
		absEndxTime = sliceEndxTime;
	}
			
	fCurrentAbsRenderPos = absStartTime;
	
	// now let's allocate the buffers that are used for depth hierarchy rendering
	// allocate them stereo (it's enough)
	for(i = 0; i < kCoreAudioPlayer_MaxHierarchyDepth; i++)
	{
		fOfflineTmpBuffers[i] = NULL; // for exception handling
	}
	
	try
	{
		for(i = 0; i < kCoreAudioPlayer_MaxHierarchyDepth; i++)
		{
			try
			{
				fOfflineTmpBuffers[i] = new float [kOfflineRenderingBufferNbFrames*2]; // INTERNAL buffers are always stereo 
			}
			catch(...)
			{
				throw;
			}
		}
	}
	catch(...)
	{
		for(i = 0; i < kCoreAudioPlayer_MaxHierarchyDepth; i++)
		{
			delete [] fOfflineTmpBuffers[i];
		}
		throw;
	}

	// create the offline plugins for all the array objects
	try
	{
		RecursivelyCreateOfflinePluginsForAllObjsInArray(fObjSet);
	}
	catch(...)
	{
		throw;
	}
}

MEAudioObjRenderer::~MEAudioObjRenderer()
{
	for(UInt32 i = 0; i < kCoreAudioPlayer_MaxHierarchyDepth; i++)
	{
		delete [] fOfflineTmpBuffers[i];
	}
	RecursivelyDeleteOfflinePluginsForAllObjsInArray(fObjSet);
}

void MEAudioObjRenderer::ReadNextFloatFrames(float* dest, SInt32 nbFrames)
{
//	printf("MEAudioObjRenderer::ReadNextFloatFrames(nbFrames = %d)\n",nbFrames);
//	printf("this->fCurrentAbsRenderPos = %lld\n",fCurrentAbsRenderPos);
	// zero the dest buffer
	memset(dest,0,nbFrames*sizeof(float)*fNbChannels);
	
	bool bStereo = false;
	if(fNbChannels == 2)
		bStereo = true;

	// loop on every obj in the array
	for(SInt32 i = 0; i < fObjSet->CountItems(); i++)
	{
		// request the current TArmeWSObj*
		ArmeArchivable* archObj = fObjSet->ItemAt(i);
		TArmeWSObj* wsobj = dynamic_cast<TArmeWSObj*>(archObj);
		TArmeDoc* theDoc = wsobj->GetDoc();
		ZzAssert(wsobj);
//		printf("wsobj(%p)->GetName() = %s   wsobj->GetAbsoluteStartTime() = %lld\n",(void*)wsobj,wsobj->GetName(),GET_VAL(wsobj->GetAbsoluteStartTime()));
	
		// compute its absStartTime
		SInt64 objAbsStartTime = GET_VAL(wsobj->GetAbsoluteStartTime());
		
		// if it's outside the currently rendered slice, just go on to the
		// next object
		if(objAbsStartTime >= fCurrentAbsRenderPos + nbFrames)
		{
//			printf("(objAbsStartTime >= fCurrentAbsRenderPos + nbFrames) --> continuing to next object\n");
			continue;
		}
		
		SInt64 objAbsEndxTime = objAbsStartTime + GET_VAL(wsobj->GetDuration());
		if(fCurrentAbsRenderPos >= objAbsEndxTime)
		{
//			printf("(fCurrentAbsRenderPos >= objAbsEndxTime) --> continuing to next object\n");
			continue;
		}
		
		// otherwise let's render it
		// the current "top set level" object always writes in fOfflineTmpBuffers[0]
		// zero the buffer where THIS object will write into
		memset(fOfflineTmpBuffers[0],0,nbFrames*sizeof(float)*2); // potentially stereo
		
		ArmeTime startTimeInObj = _ArmeTime( fCurrentAbsRenderPos - objAbsStartTime);
			
		const MEVAORouting& thisVAORouting = wsobj->GetTopParentVAORouting();

		if(!fApplyRAOVAOGain) // if we don't need to apply the VAORAO gain/pan, we simply mix
		{
			// we don't need any offset in the dest buffer....
			// the AddObjectDataIntoBuffer() will clip start address and duration if we request something outside of the object
			
			// we ask the object to render in its native nb of channels
			AddObjectDataIntoBuffer(wsobj,thisVAORouting.bStereo,fOfflineTmpBuffers,0,0,startTimeInObj,nbFrames);

			if(fNbChannels == 1) // everything is mono
			{
				MixMonoBufferToMonoBuffer(dest,fOfflineTmpBuffers[0],nbFrames);
			}
			else // everything is stereo
			{
				MixStereoBufferToStereoBuffer(dest,fOfflineTmpBuffers[0],nbFrames);
			}
		}
		else // otherwise we need to take into account the current object vaorao gain
		{
			memset(fOfflineTmpBuffers[1],0,nbFrames*sizeof(float)*2); // potentially stereo
			// we don't need any offset in the dest buffer....
			// the AddObjectDataIntoBuffer() will clip start address and duration if we request something outside of the object
			
			// please note the object is rendered in its final routing.
			AddObjectDataIntoBuffer(wsobj,thisVAORouting.bStereo,fOfflineTmpBuffers,1,0,startTimeInObj,nbFrames);

			// first we attenuate the buffer according to its routing
			// in place VAO attenuation (mono->mono and stereo->stereo        panning is used for stereo only)
			theDoc->GetMixingInfo()->AttenuateVAOBuffer_Offline(thisVAORouting.bStereo,thisVAORouting.index,fOfflineTmpBuffers[1],nbFrames);
			
			if(fNbChannels == 1)
			{
				// if it's mono then we RAO-render from fOfflineTmpBuffers[1] to fOfflineTmpBuffers[0]
				// then we stereo -> mono mix down from fOfflineTmpBuffers[0] to dest
				if(thisVAORouting.bStereo)
					theDoc->GetMixingInfo()->MixMapPanStereoVAOIntoStereoBuffer_Offline(fOfflineTmpBuffers[0], thisVAORouting.index, fOfflineTmpBuffers[1], nbFrames);
				else
					theDoc->GetMixingInfo()->MixMapPanMonoVAOIntoStereoBuffer_Offline(fOfflineTmpBuffers[0], thisVAORouting.index, fOfflineTmpBuffers[1], nbFrames);
				MixStereoBufferToMonoBuffer(dest,fOfflineTmpBuffers[0],nbFrames);
			}
			else
			{
				if(thisVAORouting.bStereo)
					theDoc->GetMixingInfo()->MixMapPanStereoVAOIntoStereoBuffer_Offline(dest, thisVAORouting.index, fOfflineTmpBuffers[1], nbFrames);
				else
					theDoc->GetMixingInfo()->MixMapPanMonoVAOIntoStereoBuffer_Offline(dest, thisVAORouting.index, fOfflineTmpBuffers[1], nbFrames);
			}
		}
	}
	fCurrentAbsRenderPos += nbFrames;				
}

// this code is a duplicate of the playback code, but much simpler (without dezipping and the like)
void MEAudioObjRenderer::AddObjectDataIntoBuffer(
	TArmeWSObj*	wsobj,
	bool 		bStereo, 
	float** 	buffers,
	SInt32		level,
	SInt32		frameOffsetInDstBuffer,
	ArmeTime	startRelTime, 
	SInt32		duration)
{
	// if the object is muted, we simply discard this call
	if(wsobj->GetMuted())
		return;
		
	ZzRtAssert(frameOffsetInDstBuffer < kOfflineRenderingBufferNbFrames);

	// if we are requested some frames before clip start, let's skip that part
	if(GET_VAL(startRelTime) < 0)
	{
		duration += GET_VAL(startRelTime);
		frameOffsetInDstBuffer += (SInt32)(-GET_VAL(startRelTime));
		GET_VAL(startRelTime) = 0;
		vernl(6," modifying frameOffsetInDstBuffer = " << frameOffsetInDstBuffer << " duration = " << duration);
	}
	
	// now if the duration is negative, it means we don't have anything to add
	if(duration <= 0)
		return;
	
	// if duration is too big, let's clip it
	if(GET_VAL(startRelTime) + duration > GET_VAL(wsobj->GetDuration()))
	{
		duration = GET_VAL(wsobj->GetDuration()) - GET_VAL(startRelTime);
		vernl(6," wsobj->GetDuration() = " << wsobj->GetDuration() << " -> modifying duration = " << duration);
	}

	// now if the duration is negative, it means we are PAST the object and shouldn't do anything
	if(duration <= 0)
		return;
	ZzAssert(frameOffsetInDstBuffer < kOfflineRenderingBufferNbFrames);
		
	float* dst;
	if(bStereo)
	{
		dst = buffers[level] + 2*frameOffsetInDstBuffer;
	}
	else
	{
		dst = buffers[level] + frameOffsetInDstBuffer;
	}

	switch(wsobj->GetType())
	{
	case TArmeWSObj::kAudioClip :
		{
			MEAudioClip* audioClip = dynamic_cast<MEAudioClip*>(wsobj);
			ZzAssert(audioClip);
			if(bStereo)
			{
				if( audioClip->IsStereo() ) 
				{
					// audio clip is stereo and dest is stereo !
					audioClip->GetData()->ReadFloatFramesToFlatBuffer(buffers[level+1],GET_VAL(startRelTime),duration,kCallingThread_OfflineRenderer);
					MixStereoBufferToStereoBuffer(dst,buffers[level+1],duration);
				}
				else
				{
					// audio clip is mono and dest is stereo !
					audioClip->GetData()->ReadFloatFramesToFlatBuffer(buffers[level+1],GET_VAL(startRelTime),duration,kCallingThread_OfflineRenderer);
					MixMonoBufferToStereoBuffer(dst,buffers[level+1],duration);
				}
			}
			else
			{
				if( audioClip->IsStereo() ) 
				{
					// audio clip is stereo and dest is mono !
					audioClip->GetData()->ReadFloatFramesToFlatBuffer(buffers[level+1],GET_VAL(startRelTime),duration,kCallingThread_OfflineRenderer);
					MixStereoBufferToMonoBuffer(dst,buffers[level+1],duration);
				}
				else
				{
					// audio clip is mono and dest is mono !
					audioClip->GetData()->ReadFloatFramesToFlatBuffer(buffers[level+1],GET_VAL(startRelTime),duration,kCallingThread_OfflineRenderer);
					MixMonoBufferToMonoBuffer(dst,buffers[level+1],duration);
				}
			}
		}
		break;
	case TArmeWSObj::kAudioProcessor :
		{
			SInt32 nbChannels;
			if(bStereo)
				nbChannels = 2;
			else
				nbChannels = 1;
				
			memset(buffers[level+1],0,sizeof(float)*duration*nbChannels);
			
			for(SInt32 iChild = 0; iChild < wsobj->GetChildrenCount(); iChild++)
			{
				AddObjectDataIntoBuffer(wsobj->GetChild(iChild),bStereo,buffers,level+1,0,
				_ArmeTime(GET_VAL(startRelTime)-GET_VAL(wsobj->GetChild(iChild)->GetStartTime())),duration);
			}
			
			wsobj->GetOfflineRenderingPlugin()->SetInputChannels(0, nbChannels);
			wsobj->GetOfflineRenderingPlugin()->SetOutputChannels(0, nbChannels);
			wsobj->GetOfflineRenderingPlugin()->SetInputBuffer(0, buffers[level+1]);
			wsobj->GetOfflineRenderingPlugin()->SetOutputBuffer(0, dst);
//			wsobj->GetOfflineRenderingPlugin()->ProtectedProcessFramesOverwrite((SInt32)GET_VAL(duration));
			wsobj->GetOfflineRenderingPlugin()->ProtectedProcessFramesAdd((SInt32)(duration));
		}
		break;
	default : // it's a regular container or another unknown objet....
		{
			for(SInt32 iChild = 0; iChild < wsobj->GetChildrenCount(); iChild++)
			{
				AddObjectDataIntoBuffer(
					wsobj->GetChild(iChild),
					bStereo,
					buffers,
					level,
					frameOffsetInDstBuffer,
					_ArmeTime(
						GET_VAL(startRelTime) - 
						GET_VAL(wsobj->GetChild(iChild)->GetStartTime())),
					duration
					);
			}
		}
		break;
	}
}		


void MEAudioObjRenderer::DeleteOfflineTmpBuffers(float** offlineTmpBuffers)
{
	for(UInt32 i = 0; i < kCoreAudioPlayer_MaxHierarchyDepth; i++)
	{
		delete [] offlineTmpBuffers[i];
	}
}

void MEAudioObjRenderer::MixStereoBufferToStereoBuffer(float* dst, const float* src, SInt32 nbFrames, float leftGain, float rightGain)
{
	float* _dst = dst;
	const float* _src = src;
	for(SInt32 frame = 0; frame < nbFrames; frame++)
	{
		*_dst += leftGain*(*_src);
		_dst++;
		_src++;
		*_dst += rightGain*(*_src);
		_dst++;
		_src++;
	}
}

void MEAudioObjRenderer::MixStereoBufferToMonoBuffer(float* dst, const float* src, SInt32 nbFrames)
{
	float* _dst = dst;
	const float* _src = src;
	for(SInt32 frame = 0; frame < nbFrames; frame++)
	{
		*_dst += 0.7071067811865476*(*_src); // sqrt(2) for equal power mono rendering
		_src++;
		*_dst += 0.7071067811865476*(*_src);
		_dst++;
		_src++;
	}
}

void MEAudioObjRenderer::MixMonoBufferToStereoBuffer(float* dst, const float* src, SInt32 nbFrames, float gain, float pan)
{
	double curalpha = (pan+1.0)*kQuarterPi;
	double curcosalpha = cos(curalpha);
	double cursinalpha = sin(curalpha);
	float curTotalLeftFading = curcosalpha*gain;
	float curTotalRightFading = cursinalpha*gain;

	float* _dst = dst;
	const float* _src = src;
	for(SInt32 frame = 0; frame < nbFrames; frame++)
	{
		*_dst += curTotalLeftFading*(*_src);
		_dst++;
		*_dst += curTotalRightFading*(*_src);
		_dst++;
		_src++;
	}
}

void MEAudioObjRenderer::MixMonoBufferToMonoBuffer(float* dst, const float* src, SInt32 nbFrames)
{
	float* _dst = dst;
	const float* _src = src;
	for(SInt32 frame = 0; frame < nbFrames; frame++)
	{
		*_dst += *_src;
		_dst++;
		_src++;
	}
}

	
void MEAudioObjRenderer::RecursivelyDeleteOfflinePluginsForAllObjsInArray(ZzArchPtrArray* objSet)
{
	TArmeWSObj* wsobj;
//	ArmeArchivable* archobj;
	for(SInt32 i = 0; i < objSet->CountItems(); i++)
	{
		wsobj = dynamic_cast<TArmeWSObj*>(objSet->ItemAt(i));
		RecursivelyDeleteOfflinePlugin(wsobj);
	}
}

void MEAudioObjRenderer::RecursivelyCreateOfflinePluginsForAllObjsInArray(ZzArchPtrArray* objSet)
{
	TArmeWSObj* wsobj;
//	ArmeArchivable* archobj;
	for(SInt32 i = 0; i < objSet->CountItems(); i++)
	{
		wsobj = dynamic_cast<TArmeWSObj*>(objSet->ItemAt(i));
		RecursivelyCreateOfflinePlugin(wsobj);
	}
}

void MEAudioObjRenderer::RecursivelyCreateOfflinePlugin(TArmeWSObj* root)
{
	root->CreateOfflineAudioPlugin(kOfflineRenderingBufferNbFrames);
	for(SInt32 ichild = 0; ichild < root->GetChildrenCount(); ichild++)
	{
		RecursivelyCreateOfflinePlugin(root->GetChild(ichild));
	}
}

void MEAudioObjRenderer::RecursivelyDeleteOfflinePlugin(TArmeWSObj* root)
{
	root->DeleteOfflinePlugin();
	for(SInt32 ichild = 0; ichild < root->GetChildrenCount(); ichild++)
	{
		RecursivelyDeleteOfflinePlugin(root->GetChild(ichild));
	}
}

UInt32 MEAudioObjRenderer::GetMaxNbChannels(ZzArchPtrArray* objSet)
{	
	ZzAssert(objSet->CountItems() > 0);
	int i;
	for(i = 0; i < objSet->CountItems(); i++)
	{
		TArmeWSObj* wsobj = dynamic_cast<TArmeWSObj*>(objSet->ItemAt(i));
		ZzAssert(wsobj);
		if(wsobj->GetTopParentVAORouting().bStereo)
			return 2;
	}
	return 1;
}

void MEAudioObjRenderer::GetObjectSetTimeSpan(ZzArchPtrArray* objSet, SInt64* absObjSetStartTime, SInt64* absObjSetEndxTime)
{
	ArmeTime wholeSetAbsStartTime = ARME_TIME_MAX;
	ArmeTime wholeSetAbsEndxTime = ARME_TIME_MIN;
	
	ZzAssert(objSet->CountItems() > 0);

	int i;
	
	for(i = 0; i < objSet->CountItems(); i++)
	{
		TArmeWSObj* wsobj = dynamic_cast<TArmeWSObj*>(objSet->ItemAt(i));
		ZzAssert(wsobj);
		if(	wsobj->GetAbsoluteStartTime() < wholeSetAbsStartTime)
			wholeSetAbsStartTime = wsobj->GetAbsoluteStartTime();
		if(	wsobj->GetAbsoluteEndxTime() > wholeSetAbsEndxTime)
			wholeSetAbsEndxTime = wsobj->GetAbsoluteEndxTime();
	}
	*absObjSetStartTime = GET_VAL(wholeSetAbsStartTime);
	*absObjSetEndxTime	= GET_VAL(wholeSetAbsEndxTime);
}

/*

ArmeErr	MEAudioObjRenderer::Render(
	const char* 				mtgaPath,
	ZzArchPtrArray* 			objSet,
	MEVAORouting* 				newRouting,
	MIProgressBarDlgManager*	progressManager,
	bool						forExport)
{
	ArmeErr err;
	// there is a requirement on the objSet where an object in the set can not
	// be an ascendant of another object in the set!!!!!
	// $$$TODO enforce this (it should be ok if putInArrayAndRemoveDescendants has been called on the MIWSSelection)
	
	bool renderToStereo = true; // false if (forExport == false) && (objSet[0]->GetTopParentVAORouting().bStereo == false)
		
	ArmeTime wholeSetAbsStartTime = ARME_TIME_MAX;
	ArmeTime wholeSetAbsEndxTime = ARME_TIME_MIN;
	
	// let's take a snapshot of the VAO mapping of the first object
	ZzAssert(objSet->CountItems() > 0);
	TArmeWSObj* firstObject = dynamic_cast<TArmeWSObj*>(objSet->ItemAt(0));
	*newRouting = firstObject->GetTopParentVAORouting();
	TArmeDoc* theDoc = firstObject->GetDoc();
	
	// if it's an IN-VAO render, then we can render mono if the target VAO is mono, too
	if( (forExport == false) && ((dynamic_cast<TArmeWSObj*>(objSet->ItemAt(0)))->GetTopParentVAORouting().bStereo == false) )
		renderToStereo = false;
	
	bool selectionContainDifferentRoutings = false; 
	
	int i;
	
	for(i = 0; i < objSet->CountItems(); i++)
	{
		TArmeWSObj* wsobj = dynamic_cast<TArmeWSObj*>(objSet->ItemAt(i));
		
		// this is needed to check if we're trying to render objects from different
		// VAOs
		const MEVAORouting& thisVAORouting = wsobj->GetVAORouting();
		if( *newRouting != thisVAORouting)
			selectionContainDifferentRoutings = true;

		ZzAssert(wsobj);
		if(	wsobj->GetAbsoluteStartTime() < wholeSetAbsStartTime)
			wholeSetAbsStartTime = wsobj->GetAbsoluteStartTime();
		if(	wsobj->GetAbsoluteEndxTime() > wholeSetAbsEndxTime)
			wholeSetAbsEndxTime = wsobj->GetAbsoluteEndxTime();
	}
	
	SInt64 absStartTime = GET_VAL(wholeSetAbsStartTime);
	SInt64 wholeDuration = GET_VAL(wholeSetAbsEndxTime) - GET_VAL(wholeSetAbsStartTime);
	SInt64 currentRelFrameInWholeSet = 0;
	
	if(selectionContainDifferentRoutings && (forExport == false) )
		ZzWarningMessageBox("The selection contains objects routed to different internal outputs. Please note the result will use the first internal output selected.");

	// now let's allocate the buffers that are used for depth hierarchy rendering
	// allocate them stereo (it's enough)
	float*	fOfflineTmpBuffers[kCoreAudioPlayer_MaxHierarchyDepth];
	for(i = 0; i < kCoreAudioPlayer_MaxHierarchyDepth; i++)
	{
		fOfflineTmpBuffers[i] = new float [kOfflineRenderingBufferNbFrames*sizeof(float)*2];
	}
	
	// create the offline plugins for all the array objects
	RecursivelyCreateOfflinePluginsForAllObjsInArray(objSet);
	
	// dstFile
	//MEMTGAAudioFile dstFile;
	
	SInt32 nbChannels;

	if(renderToStereo)
	{
		nbChannels = 2;
	}
	else
	{
		nbChannels = 1;
	}
	
			
	while(currentRelFrameInWholeSet < wholeDuration)
	{
		// we loop on all the frames that span the objects
		// (from fAbsStartTime to fAbsStartTime+fWholeDuration) by chunks of kOfflineRenderingBuffersNbFrames
		
		// for each little chunk, we loop on all the objects in the set and ask them to add their data to the temp
		// buffer, then we write this piece of data to the MEMTGAAudioFile we've created for this vao.	
		try
		{
			progressManager->Update(((double)currentRelFrameInWholeSet/(double)wholeDuration));
		}
		catch(...)
		{
			RecursivelyDeleteOfflinePluginsForAllObjsInArray(objSet);
			DeleteOfflineTmpBuffers(fOfflineTmpBuffers);
			throw;
		}
					
		// let's compute the number of frames we'll render in this iteration
		SInt64 remainingFrames = wholeDuration - currentRelFrameInWholeSet;
		SInt32 duration; // the number of frames we render in this call
		if(remainingFrames > kOfflineRenderingBufferNbFrames)
			duration = kOfflineRenderingBufferNbFrames;
		else
			duration = remainingFrames;
		
		// zero the dest buffer
		memset(fOfflineTmpBuffers[0],0,kOfflineRenderingBufferNbFrames*sizeof(float)*nbChannels);
//		memset(fOfflineTmpBuffers[2],0,kOfflineRenderingBufferNbFrames*sizeof(float)*nbChannels);
		
		// for every object in the array, we request 
		for(SInt32 i = 0; i < objSet->CountItems(); i++)
		{
			// zero the buffer where THIS object will write into
			memset(fOfflineTmpBuffers[1],0,kOfflineRenderingBufferNbFrames*sizeof(float)*nbChannels);

			TArmeWSObj* wsobj = dynamic_cast<TArmeWSObj*>(objSet->ItemAt(i));
			ZzAssert(wsobj);
			ArmeTime startTimeInObj = _ArmeTime( absStartTime - GET_VAL(wsobj->GetAbsoluteStartTime()) + currentRelFrameInWholeSet); // we use GetAbsoluteStartTime() because the wsobj is perhaps not a direct child of the tlc....
			
			if(forExport == false) // if we are not exporting, we simply mix
			{
				// we don't need any offset in the dest buffer....
				// the AddObjectDataIntoBuffer() will clip start address and duration if we request something outside of the object
				AddObjectDataIntoBuffer(wsobj,renderToStereo,fOfflineTmpBuffers,1,0,startTimeInObj,duration);

				if(nbChannels == 1) // everything is mono
				{
					MixMonoBufferToMonoBuffer(fOfflineTmpBuffers[0],fOfflineTmpBuffers[1],duration);
				}
				else // everything is stereo
				{
					MixStereoBufferToStereoBuffer(fOfflineTmpBuffers[0],fOfflineTmpBuffers[1],duration);
				}
			}
			else // otherwise we need to take into account the current object vaorao gain
			{
				const MEVAORouting& thisVAORouting = wsobj->GetTopParentVAORouting();
				
				// we don't need any offset in the dest buffer....
				// the AddObjectDataIntoBuffer() will clip start address and duration if we request something outside of the object
				
				// please note the object is rendered in its final routing.
				AddObjectDataIntoBuffer(wsobj,thisVAORouting.bStereo,fOfflineTmpBuffers,1,0,startTimeInObj,duration);

				// first we attenuate the buffer according to its routing
				// in place VAO attenuation (mono->mono and stereo->stereo        panning is used for stereo only)
				theDoc->GetMixingInfo()->AttenuateVAOBuffer_Offline(thisVAORouting.bStereo,thisVAORouting.index,fOfflineTmpBuffers[1],duration);
				
				// now we need to  mix it down into the fake rao buffer which is here fOfflineTmpBuffers[0] which has nbChannels == 2
				// (always for export $$$EXPORT)
				if(thisVAORouting.bStereo)
					theDoc->GetMixingInfo()->MixMapPanStereoVAOIntoStereoBuffer_Offline(fOfflineTmpBuffers[0], thisVAORouting.index, fOfflineTmpBuffers[1], duration);
				else
					theDoc->GetMixingInfo()->MixMapPanMonoVAOIntoStereoBuffer_Offline(fOfflineTmpBuffers[0], thisVAORouting.index, fOfflineTmpBuffers[1], duration);
			}
		}
	
		(*fCallback)(offlineTmpBuffers[0],duration,fCookie);
			
		currentRelFrameInWholeSet += duration;				
	}
	
	ZzAssert(currentRelFrameInWholeSet == wholeDuration);
	
	// close file
	//dstFile.Close();
	
	// delete the offline plugin from all the wsobjs
	RecursivelyDeleteOfflinePluginsForAllObjsInArray(objSet);
	DeleteOfflineTmpBuffers(offlineTmpBuffers);

	return err;
}


void MEAudioObjRenderer::WriteFloat32Frames(float* data,UInt32 nbFrames)
{
	ZzRtAssert(0);
	throw;
}

*/
