/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/
#define MEAudioObjRenderer_h
#ifndef MEAudioObjRenderer_h
#include "MEAudioObjRenderer.h"
#endif //MEAudioObjRenderer_h

#ifndef ArmeErrors_h
#include "ArmeErrors.h"
#endif //ArmeErrors_h

/*!
	@class MEAudioObjRenderer
*/

typedef ArmeErr (*MEAudioObjRendererCallback)(float* buffer, UInt32 nbFrames, UInt32 nbChannels, void* cookie);

class MEAudioObjRenderer
{
public :
				MEAudioObjRenderer(
					ZzArchPtrArray* objSet,
					bool			bSliceSel,
					SInt64			startGlobalTime,
					SInt64			endxGlobalTime,
					UInt32			maxRequestedBufferFrames,
					bool			bApplyRAOVAOGain,
					UInt32			nbChannels);

	virtual		~MEAudioObjRenderer();
	
	void 		ReadNextFloatFrames(float* dest, SInt32 nbFrames);

public :									
	static void		GetObjectSetTimeSpan(ZzArchPtrArray* objSet, SInt64* absObjSetStartTime, SInt64* objSetDuration);
	static UInt32	GetMaxNbChannels(ZzArchPtrArray* objSet);
private :
	static void	DeleteOfflineTmpBuffers(float** offlineTmpBuffers);
	static void	MixStereoBufferToStereoBuffer(float* dst, const float* src, SInt32 nbFrames, float leftGain = 1.0, float rightGain = 1.0);
	static void	MixStereoBufferToMonoBuffer(float* dst, const float* src, SInt32 nbFrames);
	static void	MixMonoBufferToStereoBuffer(float* dst, const float* src, SInt32 nbFrames, float gain = 1.0, float pan = 0.0);
	static void	MixMonoBufferToMonoBuffer(float* dst, const float* src, SInt32 nbFrames);

	static void	RecursivelyDeleteOfflinePluginsForAllObjsInArray(ZzArchPtrArray* objSet); // deletes the offline plugin that the armeobj was told to create by us
	static void	RecursivelyDeleteOfflinePlugin(TArmeWSObj* root);
	static void	RecursivelyCreateOfflinePluginsForAllObjsInArray(ZzArchPtrArray* objSet);
	static void	RecursivelyCreateOfflinePlugin(TArmeWSObj* root);
	
	static void	AddObjectDataIntoBuffer(
					TArmeWSObj*	wsobj,
					bool 		bStereo, 
					float** 	buffers,
					SInt32		level,
					SInt32		frameOffsetInDstBuffer,
					ArmeTime	startRelTime, 
					SInt32		duration);
private :
	void		WriteFloat32Frames(float* data,UInt32 nbFrames);
	
private :
	/*! @var fCurrentAbsRenderPos the frame relative to the whole set start. */
	SInt64						fCurrentAbsRenderPos;
	
	/*! @var fApplyRAOVAOGain Whether each object needs to be applied is RAOVAO gain/panning
		when being rendered. Please note that this leads to the following behaviour :
		
		2ch output, stereo VAO, no gain applied : VAO is directly mixed in the dest buffer
		2ch output, mono VAO, no gain applied : VAO is rendered in tmp buffer, and each M sample is copied in both L and R channels of the dest buffer
		1ch output, stereo VAO, no gain applied : VAO is is rendered in tmp buffer, and L and R samples are added/mixed in the dest buffer
		1ch output, mono VAO, no gain applied : VAO is directly mixed in dest buffer
		
		2ch output, stereo VAO, gain applied : the RAO output is mixed
		2ch output, mono VAO, no gain applied : the RAO output is mixed
		1ch output, stereo VAO, no gain applied : the RAO output is mixed bu adding the L and R samples to each dest buffer M sample
		1ch output, mono VAO, no gain applied : the RAO output is mixed bu adding the L and R samples to each dest buffer M sample
	*/
	bool						fApplyRAOVAOGain;
	
	/*! @var fObjSet The array of objects we need to render */
	ZzArchPtrArray*				fObjSet;

	/*! @var fMaxRequestedBufferFrames The max number of frames we are requested in a ReadNextFloatFrames( ) call */
	UInt32						fMaxRequestedBufferFrames;
	
	/*! @var fNbChannels	The number of channels the renderer needs to use when spitting out frames... */
	UInt32						fNbChannels;

	/*! @var fOfflineTmpBuffers	The depth buffers used for hierarchy rendering... */
	float*						fOfflineTmpBuffers[kCoreAudioPlayer_MaxHierarchyDepth];
};


