/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#define MEAudioRecordManager_h

#ifndef MEAudioRecordManager_h
#include "MEAudioRecordManager.h"
#endif //MEAudioRecordManager_h

#ifndef TArmeDoc_h
#include "TArmeDoc.h"
#endif //TArmeDoc_h

#ifndef MEAudioChunkList_h
#include "MEAudioChunkList.h"
#endif //MEAudioChunkList_h

#import <CoreAudio/CoreAudio.h>

// this is a singleton class

class TArmeDoc;
class ZzRingBuffer;
class MECoreAudioRecorder;
class MEMTGAAudioFile;
class ZzQueue;
class MEAudioMonitoring;
class MEMTGAPlayer;

struct MEAudioRecordManagerTakeInfo
{
	ArmeTime					startTime;
	ArmeTime					duration;
	ZzArray<MEAudioChunkFileIndex>	fileIndices;
	ZzArray<SInt32>				fileNbChannels;
	SInt64						playerPosAtRest;
};

struct MEAudioRecordInputConfig
{
	bool			enabled;
	bool			joined;
	float			targetLinearGain;
	float			currentLinearGain;
	float			dezipGainIncrPerFrame;
	float			analyzeMaxSqAmpl;
	bool			analyzeSatur;
	SInt32			analyzeSaturResetCounter;
	float			analyzeDBFS;
	MEAudioRecordInputConfig();
};

inline MEAudioRecordInputConfig::MEAudioRecordInputConfig()
{
	enabled = false;
	joined = true;
	targetLinearGain = 1.0;
	currentLinearGain = 1.0;
	//analyzeMaxSqAmpl no need to init this one
	analyzeSatur = false;
	analyzeDBFS = -1000.0;
}

struct MEAudioRecordFileResources
{
	// code : 
	// CCFR means "created in CreateFileResources"
	// DIDY means deleted in DestroyDynamicRecordResources
	// DIST means deleted in DestroyStaticRecordResources

	SInt32				input; // CCFR
	SInt32				nbChannels; // CCFR
	ZzQueue*			queue; // CCFR DIDY
	ZzRingBuffer*		ringBuffer; // CCFR DIDY
	bool				somethingHasBeenWritten; // CCFR
	MEAudioChunkFileIndex	fileIndex; // CCFR
	MEMTGAAudioFile*		file; // CCFR DIDY
	
	// tmpBuffer is used from ring buffer to DRT
	float*				tmpBuffer; // CCFR DIDY
	SInt32				tmpBufferNbFrames; // CCFR DIDY
	
	// tmpBufferRT is used from RT to ring buffer
	float*				tmpBufferRT; // CCFR DIDY
	SInt32				tmpBufferRTNbFrames; // CCFR DIDY
	
	pthread_t			threadID;// CCFR DIDY
};


/*!

	@class MEAudioRecordManager
	
	Singleton class that manages recording sound data into files.
	
	Here's how it work : 
	
	When idle (fState == kIdle), the record manager contains a NULL fRecordResources pointer,
	a NULL fCurrentDoc and a NULL fHardwareRecorder. It's really an idle class
	
	****** When things are about to be recorded, the PrepareRecordingForDoc() method is called. This
	creates a hardware recorder that will send the buffers to us (but does NOT start it at this
	time). Then, this function creates the record input config array, which tells the inputs that
	are enabled, levels, etc.. Eventually, CreateRecordResources() is called. This function : 
	
	- parses the array of inputs and creates as many record resources as needed files.
	- creates a queue and a thread per file, and also two tmp buffers (one for the RTT and
	  one for the DRT)
	
	WARNING : fRecordResources must be null when calling CreateRecordResources(). It does NOT destroy
	the existing ones before creating the new ones.
	
	Then the recording thread is started
	
	****** StartRecording only enables the buffers to be recorded in the file
	
	
*/

class MEAudioRecordManager
{
	enum MEAudioRecordManagerState
	{
		kIdle,
		kReadyToRecord,
		kRecording
	};
	
public : 
	static MEAudioRecordManager*	SharedAudioRecordManager();
	
public : 
				MEAudioRecordManager();

	/*! @function PrepareRecordingForDoc Tells the recorder we are about to record data for the 'doc' document. This
	    sets up the default parameters for recording. StartRecording() should be called next, or EnableInput() and JoinInput() */
	void		PrepareRecordingForDoc(TArmeDoc* doc, AudioDeviceID inDevID);
	
	/*! @function PutInIdleState Tells the recorder we are done with this session */
	void 		PutInIdleState();
	
	/*! @function GetNbInputs returns the number of MONO audio inputs to display. this is an even number. Can be called anytime. */
	SInt32		GetNbInputs() const;
	
	/*! @function SetInputEnabled enables the (iInput)th mono input according to 'enabled'. 0-based */
	void		SetInputEnabled(SInt32 iInput, bool enabled);

	/*! @function SetInputEnabled enables the (iInput)th mono input. 0-based */
	bool		IsInputEnabled(SInt32 iInput);
	
	/*! @function GetInputDBFSValue Allow to retrieve the value that needs to be displayed in the vu meters */
	float 		GetInputDBFSValue(SInt32 input) const;

	/*! @function IsInputSaturated Whether an input has been saturated since PrepareRecordingForDoc or since the last call to SetInputNotSaturated for this input */
	bool 		IsInputSaturated(SInt32 input) const;
	
	/*! @function SetInputNotSaturated Resets the saturation state for this input */
	void 		SetInputNotSaturated(SInt32 input);

	/*! @function SetInputJoined After calling this, the (iInput)th mono input and  (iInput+1)th mono input will be joined or in a stereo file 
	    according to the 'joined' parameter */
	void		SetInputJoined(SInt32 iFirstInputOfThePair, bool joined);

	/*! @function IsInputJoined Returns true if the (iInput)th mono input and (iInput+1)th mono input are to be joined in a stereo file. false otherwise
		(for instance, it returns false even if (iFirstInputOfThePair-1) and (iFirstInputOfThePair) are joined. */
	bool		IsInputJoined(SInt32 iFirstInputOfThePair) const;
	
	/*! @function SetInputAttenuationDB Sets the attenuation in dB for the (iInput)th (should be less than or equal to zero) */
	void		SetInputAttenuationDB(SInt32 iInput, float dB);

	/*! @function GetInputAttenuationDB Sets the attenuation in dB for the (iInput)th (should be less than or equal to zero) */
	float		GetInputAttenuationDB(SInt32 iInput) const;
	
	/*! @function StartRecording Starts the actual recording */
	void		StartRecording();

	/*! @function StopRecording Starts the actual recording and returns the new take number */
	SInt32		StopRecording();
	
	/*! @function DisableMonitoring Starts the actual recording */
	void		DisableMonitoring();
	
	/*! @function EnableMonitoring Starts the actual recording */
	void		EnableMonitoring();
		
	/*! @function GetNbTakes Get the current number of recorded takes. A StartRecording() StopRecording() pair generates a new take */
	SInt32				GetNbTakes() const;

	/*! @function DiscardTake $$$TODO WARNING DO NOT DELETE THE FILE */
	void				DiscardTake(SInt32 take);

	/*! @function GetTakeStartTime The doc start time of the take. No supported for looped recording right now */
	ArmeTime			GetTakeStartTime(SInt32 take) const;

	/*! @function GetTakeDuration The doc duration of the take. No supported for looped recording right now */
	ArmeTime			GetTakeDuration(SInt32 take) const;
	
	/*! @function GetTakeEndxTime The doc end time of the take. No supported for looped recording right now */
	ArmeTime			GetTakeEndxTime(SInt32 take) const;
	
	/*! @function GetTakeNbAudioFiles The number of files in this take. There is one mono file generated per non-joined audio input and one mono file generated for each mono input */
	SInt32				GetTakeNbAudioFiles(SInt32 take) const;
	
	/*! @function GetTakeFile Returns the file index of a given file within a take. This index is a doc-wide way to identify audio files */
	MEAudioChunkFileIndex	GetTakeFile(SInt32 take, SInt32 file) const;

	/*! @function GetTakeFileNbChannels The number of channels in a given recorded file. only 1 or 2 at the moment */
	SInt32				GetTakeFileNbChannels(SInt32 take, SInt32 file) const;
	
	/*! @function GetTakePlayerPosAtRest Retrieves the player pos when it was las stopped */
	SInt64				GetTakePlayerPosAtRest(SInt32 take) const;

	/*! @function SetTakePlayerPosAtRest Retrieves the player pos when it is stopped (in order to restore it at re-creation) */
	void				SetTakePlayerPosAtRest(SInt32 take, SInt64 pos);
	
	/*! @function StartTakeMonitoring allows to monitor a take */
	void				StartTakeMonitoring(SInt32 take);

	/*! @function SeekTakeMonitoringPosition nsia */
	void				SeekTakeMonitoringPosition(SInt32 take, ArmeTime seekp);

	/*! @function IsTakeMonitoringPlaying nsia */
	bool				IsTakeMonitoringPlaying(SInt32 take);

	/*! @function GetTakeMonitoringPosition "where we are" in a take monitoring */
	ArmeTime			GetTakeMonitoringPosition(SInt32 take);

	/*! @function StopTakeMonitoring stops the monitoring of a take */
	void				StopTakeMonitoring(SInt32 take);

	
	/*! @function ProcessIncomingFrames REALTIME CONTEXT This function receives a buffer 
	    from the hardware, distributes it into the tmpBufferRT of each FILE, writes the 
		resulting buffer into the ring buffer and posts a message for each thread to know
		new data needs to be recorded */
	void				ProcessIncomingFrames(float* src, SInt32 nbFrames);
	
	/* @function AnalyzeInputBuffer This function analyzes the attenuated input buffer for DBFS values and
	    saturation measurement */
	void 				AnalyzeInputBuffer(float* srcBuffer);

	/* @function AttenuateInputBuffer This function applies the gain to those inputs that are enabled */
	void 				AttenuateInputBuffer(float* src, SInt32 nbFrames);
	
	UInt32				GetAudioFrameRate() const;
				
private :
	
	// for each audio file that needs to be recorded, this function creates 
	// - a ring buffer (ZzRingBuffer instance)
	// - a thread
	// - a message queue
	void				CreateRecordResources();
	
	//! @function DestroyDynamicRecordResources() destroys all the resources that got 
	// created in CreateRecordResources() and that are used by the DRT
	void				DestroyDynamicRecordResources();

	//! @function DestroyStaticRecordResources Destroys everything in the record resources
	// that hasn't been destroyed by DestroyDynamicRecordResources().
	// we need first to destroy threads, ring buffers, etc... then later  when the take is saved,
	// we can destroy the file index, etc... and even destroy the array itself.
	void				DestroyStaticRecordResources();

	// simply calls DestroyDynamicRecordResources() then DestroyStaticRecordResources()
	void 				DestroyRecordResources();
	

private :
	// the one and only instance of this class
	static MEAudioRecordManager* fsSingletonInstance;
	
	// the config for each input (enabled or not). The size of this array can change
	// when a new input device a created. Since the record manager is called from within
	// a modal dlg box, we only need to create this array at dlg init time. Thus the dlg init
	// function must call PrepareRecordingForDoc, which creates these configs
	MEAudioRecordInputConfig*				fInputConfig;

	// the info the DRT uses when recorded. there are as many items here
	// as there are files (mono or stereo) to created. Totally destroyed and recreated
	// each time an input param changes (except for for those [such as input level] than
	// can be modified DURING recording)
	MEAudioRecordFileResources*				fRecordResources;
	
	// the number of configured files infos
	SInt32									fNbRecordResources;
	
	// we need a state mostly for debugging purposes
	MEAudioRecordManagerState				fState;

	// inited in PrepareRecordingForDoc() each time because it can change between sessions.
	MECoreAudioRecorder*					fHardwareRecorder;
	
	// inited in PrepareRecordingForDoc()
	TArmeDoc*								fCurrentDoc;
	
	// once StopRecording() is called, a new take info is added to this array
	ZzArray<MEAudioRecordManagerTakeInfo*>	fTakeInfos;
	
	// the atomic var we use in the RT thread AND in the DRT thread AND in the AppKit thread
	bool									fSendBuffersToRecordThread;
	
	// the array of recorded takes. a new take is created for each StartRecording()/StopRecording() pair
	ZzArray<MEAudioRecordManagerTakeInfo>	fTakes;
	
	ZzLocker								fRealTimeThreadMutex;
	
	ArmeTime								fRecordStartTime;
	
	MEMTGAPlayer**							fMTGAPlayers;
	SInt32									fCurrentlyPlayedTake;
	
	float									fDBPerBufferAttenuation; // this is used in the dbfs values computed for the vu meters
	SInt32									fNbOfBuffersForSaturationReset;
	/*! @var fMonitoringStereoBuffer The buffer where the input channels are mixed down in stereo before being sent
		to the monitoring object */
	float*									fMonitoringStereoBuffer;
	MEAudioMonitoring*						fMonitoringObject;
};

	