/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#define MEAudioTrackPlayer_h

#ifndef MEAudioTrackPlayer_h
#include "MEAudioTrackPlayer.h"
#endif //MEAudioTrackPlayer_h

#ifndef ArmeBasicTypes_h
#include "ArmeBasicTypes.h"
#endif //ArmeBasicTypes_h

//#ifndef TAudioClip2_h
//#include "TAudioClip2.h"
//#endif //TAudioClip2_h

#ifndef MECompoundAudioFile_h
#include "MECompoundAudioFile.h"
#endif //#ifndef MECompoundAudioFile_h

#ifndef ZzQueue_h
#include "ZzQueue.h"
#endif //ZzQueue_h

#define AUDIOTRACKPLAYER_USE_ZZQUEUE

//#define MEAudioTrackPlayer_STATIC_CHUNK_ARRAY
#define kMEAudioTrackPlayer_MaxNbChunksInTrack 16384

#define kRingBufferNbBits 13

/* $$$TODO : put a minimum limit on the spacing between 
left and right locator otherwise the number of queued
chunks might become huuge */
#define kNbReadRequestsInRing 1024

// the feeder thread will queue at least
// kMinimumFramesForDiskQueuing frames at a the time
#define kMinimumFramesForDiskQueuing 4096 

/*!
	@class MEAudioTrackReadRequest
	decribes a piece of data that needs to be read
	off and audio file by a disk reading thread.
*/

struct MEAudioTrackReadRequest
{
	SInt64	 		start;
	SInt64	 		duration;
};


/*!
	@class MEAudioTrackPlayer
	Player for the TAudioClip2 model class.
	This class is not meant to be serialized
*/

class TArmeDoc;

class MEAudioTrackPlayer
{
	friend void* MEAudioTrackPlayer_FeederThread(void* cookie);
public :
				MEAudioTrackPlayer(MEWSObjAudioPlayer* parentPlayer);
				~MEAudioTrackPlayer(); // destructor : kills the queue and the thread
//	void 		CancelPreroll();
	void		Preroll(SInt64 pos);

	/*! @function MixFromRingBuffer Mixes a range from the ring buffer (possibly crossing the ring start)
		to a flat location in memory */
	void 						MixFromRingBuffer(float startGain, float endGain, float* dst, SInt64 srcTime, SInt32 nbFrames );

	/*! @function MixFromRingBuffer Mixes A PART OF NBFRAMES from the current read pos in the ring buffer 
		(possibly crossing the ring start) to a flat location in memory and adds zero padding before and after. 
		The total number of frames copied from the ring buffer is nbFrames-(blankNbFramesBefore+blankNbFramesAfter) 
		THis is used when a read buffer is not totally contained in an audio clip */
	void 						MixFromRingBuffer(float startGain, float endGain, float* dst, SInt32 nbFrames, SInt32 blankNbFramesBefore, SInt32 blankNbFramesAfter);

	/*! @function SetStereo We have to tell the track its number of channels so that it doesn't have to look in its chunks */
	void 						SetStereo(bool stereo);

	/*! @function IsStereo Whether the track references only stereo chunks or only mono chunks. Mixed tracks are not possible */
	bool						IsStereo() const;

	/*! @function QueueNextBit Tell the track player that we're done nagging it with read request and the she can safely increment
	    the read pointer and tell its DRT to queue nbFrames extra frames */
	void 						QueueNextBit(SInt64 curPos,SInt32 nbFrames);

private :
	/*! @function RingBufferNbFrames Returns the number of frames in the ring buffer */
	SInt32 						RingBufferNbFrames() const;
	
	/*! @function FillNextReadRequest Fills a chunk desc in the ring array of chunk descs with a set of values (moduloed index) */
	void 						FillNextReadRequest(SInt64 start, SInt64 duration);
	
	/*! @function GetChunkDesc Returns the address of a given chunk desc (moduloed index) so you can read it and modify it */
	MEAudioTrackReadRequest*	GetReadRequest(SInt32 index);

	static void					PutSilenceInRingBuffer(
									float*	ringBuffer, 
									SInt32	ringBufferNbBits,
									SInt64	writePosInRingBuf,
									SInt64	length,
									SInt32	nbChannels);
									
	void 						ReadFromChunkToRingBuffer(
									const MEAudioChunk& chunk,
									SInt64 writePosInRingBuf, 
									SInt64 startInChunk, 
									SInt32 length);

	/*! 
		@function ReadFloat32ContentToRingBuffer 
		This function reads "length" clip frames from "start" and copies them at moduloed
		index "writePosInRingBuf" in the ring buffer starting at "ringBuffer" and having 1<<"ringBufferNbBits" frames.
	*/
	void 						ReadFloat32ContentToRingBuffer(
									float*			ringBuffer, 
									SInt32			ringBufferNbBits,
									SInt64			writePosInRingBuf,
									SInt64			start, 
									SInt64			length);

	/*! @function StartFeederThread Starts the feeder thread */
	void 						StartFeederThread();

	/*! @function FeederThread The entry poiny for the thread that reads data off the clip data files (feeder thread) */
	void*						FeederThread();
	
	/*! 
		@function CreateFutureChunkList 
		
		This function answers the following question :
		
		What will be playing in the future, if I'm now playing time T1.
		
		More specifically, it answers :
		
		I'm playing the current N frames of the ringbuffer, and thus I 
		need to queue N new frames at the end of the ringbuffer. Please
		tell me what these frames are if the length if the read ahead
		time ("safe length" of the RB) is fReadAheadNbFrames.
		
		it is very easy to answer if the loop is disabled in the doc player.
		If I'm playing [T1,T1+N], I need to queue [T1+fReadAheadNbFrames,T1+fReadAheadNbFrames+N]
		
		However, if the right locator is between T1 and T1+fReadAheadNbFrames, 
		and if the loop (loop between locators) is enabled, then things can
		become pretty hairy... This function uses the document. 
		
		So, the parameters are :
		
		1) currentStart : the start of the range ____we've just been requested___ (and, if things
		went well, that's just what the player has been given).
		
		2) length : the length of the range we've given and, thus, the length of the
		range we need to queue.
		
		3) offset : the offset, in the future, where the queued range is.
		
		it's up to the reader to fill zeroes when the time is outside the clip.
		
		The function returns the "answer" as list of doc time chunks that get copied 
		into the chunk desc ring starting at index fCurChunkIndex (fCurChunkIndex and 
		fCurChunkBlockSize are automatically incremented by this function). 
	*/
	void 						CreateFutureChunkList(SInt64 currentStart,SInt32 length,SInt32 offset);
	
	/*! @function GetDoc Returns the document the clips belongs to */
	MEWSObjAudioPlayer*			GetParentPlayer();
	
	TArmeDoc*					GetDoc();
	
public :
	/*
		chunk management
	*/
	void						AddSilentChunk(SInt64 duration);
	void						AddChunk(const MEAudioChunk& chunk);
	SInt32						GetNbChunks() const;
	const MEAudioChunk&			GetChunk(SInt32 i) const;
	void						EmptyChunkList();
	/*! This is a tmp value we're using for track assignement */
	SInt64						fTmpEndxTimeOfLastInsertedClipInThisTrack;
	
private :
#ifdef MEAudioTrackPlayer_STATIC_CHUNK_ARRAY
	MEAudioChunk				fStaticChunks[kMEAudioTrackPlayer_MaxNbChunksInTrack];
	SInt32						fNextChunkIndex;
#else // ! MEAudioTrackPlayer_STATIC_CHUNK_ARRAY
	ZzArray<MEAudioChunk>		fChunks; // the chunks that are played by this track. 
								// the track doesn't deal with clips. it deals with chunks.
#endif // MEAudioTrackPlayer_STATIC_CHUNK_ARRAY

	MEWSObjAudioPlayer*			fParentPlayer;

	/*! @var fRingBuffer The ring buffer where data is stored */
	float*						fRingBuffer;
	
	/*! @var fRingBufferNbBits  1<<fRingBufferNbBits is the number of frames
	    in the ring buffer */
	SInt32						fRingBufferNbBits;
	SInt32						fReadAheadNbFrames;
	
	/*! @var fWritePosInRingBuf The current position (frame index) where the
		disk reading thread writes into the ring buffer */
	SInt64						fWritePosInRingBuf;

	/*! @var fReadPosInRingBuf The current position (frame index) where the
		audio out thread (indirectly) reads into the ring buffer */
	SInt64						fReadPosInRingBuf;
	
	/* @var fChunkDescRing The list of chunks the feeder thread and real time threads use to
	   communicate */
	MEAudioTrackReadRequest		fRingOfReadRequests[kNbReadRequestsInRing];
	
	/*! @var fReadRequestBlockStartIndex The read requesst that starts the block we'll send to the feeder thread later on*/
	SInt32						fCurReadRequestBlockStartIndex;
	
	/*! @var fCurReadRequestIndex The next read request we'll write to at each QUEUE NEXT BIT message */
	SInt32						fCurReadRequestIndex;
	
	/*! @var fCurReadRequestBlockSize The (ever increasing) nb of frames of the current read request block */
	SInt32						fCurReadRequestBlockSize;
	
	/*! @var fNbChannels all the chunks must have the same number of channels */
	SInt32						fNbChannels;
	
	/*! @var fBytesPerFrame Always 8 for stereo Float32 frames. Wait and see. */
//	SInt32						fBytesPerFrame;
	
#ifdef AUDIOTRACKPLAYER_USE_ZZQUEUE
	ZzQueue*					fFeederZzQueue;
#else //#ifdef AUDIOTRACKPLAYER_USE_ZZQUEUE
	/*! @var fFeederThreadQueueID The queue that the real time thread and feeder thread use to communicate. Chunk descs
	    requests are posted in this queue. */
	MPQueueID					fFeederThreadQueueID;
#endif //#ifdef AUDIOTRACKPLAYER_USE_ZZQUEUE
		
	/*! @var fFeederThread the thread that streams data off the audio files */
	pthread_t					fFeederThread;
	
	/*! @var fAudioFile the MEMTGAAudioFile the feeder thread uses to read data off the audio chunks */
	MECompoundAudioFile			fAudioFile;
};

inline SInt32 MEAudioTrackPlayer::RingBufferNbFrames() const
{
	return (1 << fRingBufferNbBits);
}

inline MEAudioTrackReadRequest* MEAudioTrackPlayer::GetReadRequest(SInt32 index)
{
	SInt32 realIndex = index % kNbReadRequestsInRing;
	return &(fRingOfReadRequests[realIndex]);
}
	
inline MEWSObjAudioPlayer* MEAudioTrackPlayer::GetParentPlayer()
{
	return fParentPlayer;
}

inline TArmeDoc* MEAudioTrackPlayer::GetDoc()
{
	return fParentPlayer->GetDoc();
}





