/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#include "MEBufferMachine.h"

#ifndef ZzRingBuffer_h
#include "ZzRingBuffer.h"
#endif //ZzRingBuffer_h

#ifndef ZzQueue_h
#include "ZzQueue.h"
#endif //ZzQueue_h

#ifndef ZzHelpers_h
#include "ZzHelpers.h"
#endif //ZzHelpers_h

#include <cmath>

MEBufferMachine::MEBufferMachine()
{
	fNbChannels = 0;
	fLatencyNbFrames = 0;
	fRingBuffer = NULL;
}

void MEBufferMachine::Initialize(SInt32 nbChannels, SInt32 log2LatencyNbFrames)
{
	fNbChannels = nbChannels;
	fLatencyNbFrames = 1 << (log2LatencyNbFrames); // half the buffer size
	
	// log2LatencyNbFrames+1 means the buffer size is twice the latency
	fRingBuffer = new ZzRingBuffer(fNbChannels*sizeof(float),log2LatencyNbFrames+1,fLatencyNbFrames);
	StartFeederThread();
}

void MEBufferMachine::ReadFrames(float* dst, SInt32 nbFrames)
{
//	printf("MEBufferMachine::ReadFrames into %p (%d frames)\n",dst,(int)nbFrames);
	fRingBuffer->ReadToFlatBuffer((UInt8*)dst, nbFrames);
	fRingBuffer->IncrReadPos(nbFrames);
	fNotifyQueue->Notify((void*)1,(void*)nbFrames,0,0);
}

/*
void MEBufferMachine::FeederFunction(float* dst, SInt64 startPos, SInt32 nbFrames)
{
	// let's say the period is 128 frames.
	// which is 2.9ms at 44.1KHz
	// thus f = 344Hz
	
	SInt32 phaseNbFrames = startPos & 128;
	double radIncrPerFrame = 2.0*M_PI/128.0;
	double radCurPhase = phaseNbFrames*radIncrPerFrame;
	
	if(fNbChannels == 1)
	{
		for(SInt32 i = 0; i < nbFrames; i++)
		{
			double val = sin(radCurPhase);
			*dst = val;
			dst++;
			radCurPhase += radIncrPerFrame;
		}
	}
	else if (fNbChannels == 2)
	{
		for(SInt32 i = 0; i < nbFrames; i++)
		{
			double val = sin(radCurPhase);
			*dst = val;
			dst++;
			*dst = val;
			dst++;
			radCurPhase += radIncrPerFrame;
		}
	}
}
*/

SInt32 MEBufferMachine::LatencyNbFrames() const
{
	return fLatencyNbFrames;
}

void* MEBufferMachine_FeederThread(void* cookie)
{
	MEBufferMachine* This = (MEBufferMachine*)cookie;
	return This->FeederThread();
}

void MEBufferMachine::StartFeederThread()
{
	// create the queue where the file in thread reads its message
	fNotifyQueue = new ZzQueue(10); // we store room for 1<<10 = 1024 elements
	// launch the thread
	pthread_create(&fFeederThread, NULL, MEBufferMachine_FeederThread, (void*)this);
	ZzSetDRTThreadPriority(fFeederThread);
}

SInt32 MEBufferMachine::GetNbChannels() const
{
	return fNbChannels;
}

void* MEBufferMachine::FeederThread()
{
	while(true)
	{
		SInt32 a,b,c,d;
		fNotifyQueue->Wait((void**)&a,(void**)&b,(void**)&c,(void**)&d);
//		printf("MEBufferMachine::FeederThread() received %d %d %d %d\n",(int)a,(int)b,(int)c,(int)d);
	
		// four NULL params means "thread must terminate". normally, it shouldn't be used since we kill the queue first
		if( a == 0 && b == 0 && c == 0 && d == 0)
			return NULL;

		UInt8* dest1;
		UInt8* dest2;
		SInt32 dest1Size;
		SInt32 dest2Size;

		fRingBuffer->GetWriteFromFlatBufferInfo(b, &dest1, &dest1Size, &dest2, &dest2Size );
		
		FeederFunction((float*)dest1, dest1Size);
		fRingBuffer->IncrWritePos(dest1Size);
		if(dest2)
		{
			FeederFunction((float*)dest2, dest2Size);
			fRingBuffer->IncrWritePos(dest2Size);
		}
	}
	return NULL;
}

