/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#ifndef MECompoundAudioFile_h
#include "MECompoundAudioFile.h"
#endif //#ifndef MECompoundAudioFile_h

#ifndef MEMTGAAudioFile_h
#include "MEMTGAAudioFile.h"
#endif //MEMTGAAudioFile_h

MECompoundAudioFile::MECompoundAudioFile()
{
	fUsesMTGA = false;
	fState = kNotOpen;
}

MECompoundAudioFile::~MECompoundAudioFile()
{
	if(fMTGAAudioFile.GetOpenMode() != MEMTGAAudioFile::kNotOpen)
		fMTGAAudioFile.Close();
	if(fSFPAudioFile.GetOpenMode() != MEMTGAAudioFile::kNotOpen)
		fSFPAudioFile.Close();
}

const MEAudioChunk& MECompoundAudioFile::GetOpenChunk()
{
	ZzAssert(fState == kOpenReadChunk || fState == kOpenWriteChunk);
	return fChunk;
}

ArmeErr	MECompoundAudioFile::OpenReadChunk(TArmeDoc* doc, const MEAudioChunk& chunk)
{
	ZzAssert(fState == kNotOpen);
	if(doc->AudioChunkFileManager()->IsChunkFileMTGA(chunk))
	{
		ArmeErr err = fMTGAAudioFile.OpenReadChunk(doc,chunk);
		if(err.type == kArmeOK)
		{
			fChunk = chunk;
			fState = kOpenReadChunk;
			fUsesMTGA = true;
		}
		else
		{
			return err; // for bp
		}
		return err;
	}
	else
	{
		ArmeErr err;
		err = fSFPAudioFile.OpenRead(chunk.GetSFPPath());
		if(err.type == kArmeOK)
		{
			fChunk = chunk;
			fState = kOpenReadChunk;
			fUsesMTGA = false;
		}
		else
		{
			return err; // for bp
		}
		return err;
	}
}

ArmeErr	MECompoundAudioFile::OpenSaveNewChunkFile(
			TArmeDoc* 		doc, 
			MEAudioChunk&	chunk,
			SInt32			channels)
{
	ZzAssert(fState == kNotOpen);
	if(doc->AudioChunkFileManager()->IsChunkFileMTGA(chunk))
	{
		ArmeErr err = fMTGAAudioFile.OpenSaveNewChunkFile(doc,chunk,channels);
		if(err.type == kArmeOK)
		{
			fChunk = chunk;
			fState = kOpenWriteChunk;
			fUsesMTGA = true;
		}
		return err;
	}
	else
	{
//		ArmeErr err = fSFPAudioFile.OpenSaveNewChunkFile(doc,chunk,channels);
		ArmeErr err;
		err.type = kArmeErrNYI;
		if(err.type == kArmeOK)
		{
			fChunk = chunk;
			fState = kOpenWriteChunk;
			fUsesMTGA = false;
		}
		return err;
	}
}

ArmeErr	MECompoundAudioFile::ReadFloat32Frames(float* samples, SInt64* nbFrames)
{
	ZzRtAssert(fState == kOpenRead || fState == kOpenReadChunk);
	if(fUsesMTGA)
	{
		return fMTGAAudioFile.ReadFloat32Frames(samples, nbFrames);
	}
	else
	{
		return fSFPAudioFile.ReadFloat32Frames(samples, nbFrames);
	}
}

ArmeErr	MECompoundAudioFile::ReadFloat32ContentToFlatBuffer(float* buffer, SInt64 start, SInt64 length )
{
	ZzRtAssert(fState == kOpenRead || fState == kOpenReadChunk)
	ArmeErr err = Seek(start);
	if(err.type != kArmeOK)
		return err;
	SInt64 length2 = length;
	err = ReadFloat32Frames(buffer, &length2);
	ZzRtAssert(length2 == length);
	return err;
}

ArmeErr	MECompoundAudioFile::ReadFloat32ContentToRingBuffer(
			float*	ringBuffer, 
			SInt32	ringBufferNbBits,
			SInt64	writePosInRingBuf,
			SInt64	start, 
			SInt64	length)
{
	//$$$TODO move common code here
	ZzRtAssert(fState == kOpenRead || fState == kOpenReadChunk);
	if(fUsesMTGA)
	{
		return fMTGAAudioFile.ReadFloat32ContentToRingBuffer(ringBuffer, ringBufferNbBits, writePosInRingBuf, start, length);
	}
	else
	{
		return fSFPAudioFile.ReadFloat32ContentToRingBuffer(ringBuffer, ringBufferNbBits, writePosInRingBuf, start, length);
	}
}

ArmeErr MECompoundAudioFile::WriteFloat32FramesFromFlatBuffer(const float* samples, SInt64 chunkStartTime, SInt64 dur)
{
	ZzRtAssert(fState == kOpenWrite || fState == kOpenWriteChunk);
	if(fUsesMTGA)
	{
		return fMTGAAudioFile.WriteFloat32FramesFromFlatBuffer(samples, chunkStartTime, dur);
	}
	else
	{
		ZzRtAssert(0);
		ArmeErr err; 
		return err;
		//return fSFPAudioFile->WriteFloat32FramesFromFlatBuffer(samples, chunkStartTime, dur);
	}
}

ArmeErr	MECompoundAudioFile::WriteFloat32Frames(const float* samples, SInt64 nbFrames)
{
	ZzRtAssert(fState == kOpenWrite || fState == kOpenWriteChunk);
	if(fUsesMTGA)
	{
		return fMTGAAudioFile.WriteFloat32Frames(samples, nbFrames);
	}
	else
	{
		ZzRtAssert(0);
		ArmeErr err; 
		return err;
		//return fSFPAudioFile->WriteFloat32FramesFromFlatBuffer(samples, nbFrames);
	}
}

ArmeErr	MECompoundAudioFile::Close()
{
	if(fState == kNotOpen)
		return ArmeErr(kArmeOK);
	ArmeErr err;
	if(fUsesMTGA)
	{
		fUsesMTGA = false; // set it to false doesn't really matter 
		err = fMTGAAudioFile.Close();
	}
	else
	{
		err = fSFPAudioFile.Close();
	}
	if(err.type == kArmeOK)
		fState = kNotOpen;
	return err;
}

SInt64	MECompoundAudioFile::GetNbFrames()
{
	ZzRtAssert(fState != kNotOpen);
	if(fUsesMTGA)
	{
		return fMTGAAudioFile.GetNbFrames();
	}
	else
	{
		//$$$$$$$ 
		//return fSFPAAudioFile.GetNbFrames();
		ZzRtAssert(0);
		return 0;
	}
}

UInt32 MECompoundAudioFile::GetNbChannels()
{
	ZzRtAssert(fState != kNotOpen);
	if(fUsesMTGA)
	{
		return fMTGAAudioFile.GetNbChannels();
	}
	else
	{
		//$$$$$$$ 
		//return fSFPAAudioFile.GetNbChannels();
		ZzRtAssert(0);
		//ArmeErr err; 
		return 0;
	}
}

ArmeErr	MECompoundAudioFile::Seek(SInt64 offset)
{
	if(fUsesMTGA)
	{
		return fMTGAAudioFile.Seek(offset);
	}
	else
	{
		return fSFPAudioFile.Seek(offset);
	}
}

UInt32	MECompoundAudioFile::GetOpenMode()
{
	return fState;
}

ArmeErr	MECompoundAudioFile::OpenReadMTGA(const char* path)
{
	ZzRtAssert(fState == kNotOpen);
	ArmeErr err = fMTGAAudioFile.OpenReadMTGA(path);
	if(err.type == kArmeOK)
	{
		fState = kOpenReadChunk;
		fUsesMTGA = true;
	}
	return err;
}





