/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/
#define MESFPAudioFile_h
#ifndef MESFPAudioFile_h
#include "MESFPAudioFile.h"
#endif //MESFPAudioFile_h

#ifndef MEAudioChunkList_h
#include "MEAudioChunkList.h"
#endif //MEAudioChunkList_h

#include <AudioToolbox/AudioFile.h>
#include <AudioToolbox/AudioConverter.h>

/*!
	@class MESFPAudioFile
	
	This class allows to read Single Packet Frame files
	
	(in short : files that are NOT temporally compressed).
	
	The conversion is done from whatever ---> 32 bit floating point
	
	Please note neither the sample rate nor the nb of channels are
	converted as of now	
	
*/

class MESFPAudioFile
{
public :
	enum{ kNotOpen, kOpenRead, kOpenWrite };

	static const UInt32 kTmpBufferNbFrames;

public :
	MESFPAudioFile();
	~MESFPAudioFile();
	
	OSStatus 				ACInputProc(
								AudioConverterRef	inAudioConverter,
								UInt32*				outDataSize,
								void**				outData);
	
	ArmeErr					ReadFloat32Frames(float* samples, SInt64* nbFrames);
	ArmeErr					ReadFloat32ContentToFlatBuffer(float* buffer, SInt64 start, SInt64 length );
	ArmeErr					ReadFloat32ContentToRingBuffer(
								float*	ringBuffer, 
								SInt32	ringBufferNbBits,
								SInt64	writePosInRingBuf,
								SInt64	start, 
								SInt64	length); // to be deleted and moved in the parent code
	ArmeErr					WriteFloat32FramesFromFlatBuffer(const float* samples, SInt64 chunkStartTime, SInt64 dur); // it's only a seek THEN a WriteFloat32Frames
	ArmeErr					WriteFloat32Frames(const float* samples, SInt64 nbFrames);
	ArmeErr					Close();
	SInt64					GetNbFrames() const;
	UInt32					GetNbChannels() const;
	ArmeErr					Seek(SInt64 offset);
	UInt32					GetOpenMode() const;
	ArmeErr					OpenRead(const char* path);
	ArmeErr					OpenRead(const ZzString& zzpath);
	ArmeErr					OpenRead(FSRef* fsref, bool isLittleEndian);
	ArmeErr					OpenSave(
								const char*		path, 
								UInt32			frameRate, 
								UInt32			nbChannels, 
								UInt32			bitDepth,
								UInt32			floatingPoint);

public :
	static ArmeErr 			IsFileSuitableForRealTime(const ZzString& path);

	static UInt64			GetFileNbFrames(const ZzString& zzpath);
	static UInt32			GetFileNbChannels(const ZzString& zzpath);
	
	/*! @function WrapAudioFileOpen Opens an audio file and returns an AudioFileID */
	static ArmeErr 			WrapAudioFileOpen(FSRef* fsref, AudioFileID* pAudioFileID);

	/*! @function WrapGetAudioStreamBasicDesc Retrieves the stream format of an open file */
	static ArmeErr 			WrapGetAudioStreamBasicDesc(AudioFileID audioFileID, AudioStreamBasicDescription* streamBasicDesc);

	/*! @function WrapGetMaximumPacketSize Retrieves the max packet size in bytes of an open file 
		please note that for real time operations this needs to be a SINGLE FRAME */
	static ArmeErr 			WrapGetMaximumPacketSize(AudioFileID fAudioFileID, UInt32* maximumPacketSize);
	static ArmeErr 			WrapGetPacketCount(AudioFileID audioFileID, UInt64* packetCount);

private :
//	MESFPAudioFileFmt	GetFmt();
	UInt32				GetFrameRate() const;
	double				GetDurationSec() const;
	SInt32 				SizeOfSample() const;
	SInt32 				SizeOfFrame() const;
	bool				IsOpenedForWriting() const;
	bool				IsOpenedForReading() const;
	UInt32 				GetSourceBytesPerFrame() const;
	UInt32 				GetTargetBytesPerFrame() const;

private :
	static bool 		TryToCheckIfLittleEndianFile(const ZzString& zzpath);

private :
	AudioFileID					fAudioFileID; //$$$todo change this on windows
	AudioStreamBasicDescription	fSourceStreamDesc;
	AudioStreamBasicDescription	fOutputStreamDesc;
	UInt64						fNbFrames; // frame == packet because (SFP means Single Frame Packet)
	UInt32						fOpenMode;
	AudioConverterRef			fTheConverter; // the obj that converts to floating point
	SInt64						fCurrentPos;
	float*						fTmpBuffer; // the buffer where raw data is read from the file
};

inline UInt32 MESFPAudioFile::GetOpenMode() const
{
	return fOpenMode;
}

inline bool MESFPAudioFile::IsOpenedForWriting() const
{
	return (fOpenMode == kOpenWrite);
}

inline bool MESFPAudioFile::IsOpenedForReading() const
{
	// writing this in one line fools cpp_precomp for some reason with gcc 2.95.2 on MacOS X (Dec 2002 Dev Tools)
	return (fOpenMode == kOpenRead);
}

inline UInt32 MESFPAudioFile::GetSourceBytesPerFrame() const
{
	return fSourceStreamDesc.mBytesPerFrame;
}

inline UInt32 MESFPAudioFile::GetTargetBytesPerFrame() const
{
	return fOutputStreamDesc.mBytesPerFrame;
}




