/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#ifndef MESplitTaskComputeAudioClipDataPreview_h
#include "MESplitTaskComputeAudioClipDataPreview.h"
#endif //MESplitTaskComputeAudioClipDataPreview_h

#ifndef MEAudioClipData_h
#include "MEAudioClipData.h"
#endif //MEAudioClipData_h

#ifndef ZzHeapWalker_h
#include "ZzHeapWalker.h"
#endif ZzHeapWalker_h

#ifndef ZzRABI_h
#include "ZzRABI.h"
#endif //ZzRABI_h

#ifndef ZzException_h
#include "ZzException.h"
#endif //ZzException_h

#undef VER_DEF_TOPIC
#define VER_DEF_TOPIC MESplitTaskComputeAudioClipDataPreview

class MEExceptionComputingPreview : public ZzException
{	
public :
#ifdef SHOW_DEBUG_INFO_IN_ERROR_MSG
	MEExceptionComputingPreview(const char* userFriendlyMessage, const char* sourceFile, int lineNumber) : ZzException(sourceFile, lineNumber) { fUserFriendlyMessage = userFriendlyMessage; };
#else
	MEExceptionComputingPreview(const char* userFriendlyMessage) : ZzException() { fUserFriendlyMessage = userFriendlyMessage; };
#endif SHOW_DEBUG_INFO_IN_ERROR_MSG
	virtual const char* ErrorMessage() const;
private :
	const char* fUserFriendlyMessage;
};

const char* MEExceptionComputingPreview::ErrorMessage() const
{
	return fUserFriendlyMessage;
}

MESplitTaskComputeAudioClipDataPreview::MESplitTaskComputeAudioClipDataPreview(MEAudioClipData* data, ArmeErr* err)
{
	fAudioClipData = data;
}

bool MESplitTaskComputeAudioClipDataPreview::CanBeCancelled() const
{
	return false;
}

const char* MESplitTaskComputeAudioClipDataPreview::GetDesc() const
{
	return "Computing audio clip overview...";
}

ArmeErr MESplitTaskComputeAudioClipDataPreview::PerformOnce()
{
	ArmeErr err;
	// cache it for easy access
	fChunkListDuration = fAudioClipData->GetChunkList().GetDuration();
	ZzAssert(fChunkListDuration > 0);
	
	// NRDIV in frames (negative remainder division)
	fPreviewSize = fAudioClipData->GetPreviewNbBins(); 
//	printf("MESplitTaskComputeAudioClipDataPreview::PerformOnce() fPreviewSize = %d\n", fPreviewSize);
																	
	// 2* because stereo
	fPreview = new MEAudioClipDataPreviewPoint[fAudioClipData->GetNbChannels()*fPreviewSize];
	ZzRABI_ArrayDeleter<MEAudioClipDataPreviewPoint> deletePreview(&fPreview);
	
	ZzAssert(fAudioClipData->GetNbChannels()==1||fAudioClipData->GetNbChannels()==2);
	fTmpBuffer = new float[kMainPreviewBinSize*fAudioClipData->GetNbChannels()];

	deletePreview.SaveResource(); // we won't throw until function ends
	// indices
	fCurFrame = 0;
	fPreviewPoint = 0;
	return err;
}

MESplitTaskComputeAudioClipDataPreview::~MESplitTaskComputeAudioClipDataPreview()
{
}

ArmeErr MESplitTaskComputeAudioClipDataPreview::PerformNextBit()
{	
	// by default, we'll delete this when exiting this
	// function. we'll only save them if everything is ok
	ZzRABI_ArrayDeleter<MEAudioClipDataPreviewPoint> deletePreview(&fPreview);
	ZzRABI_ArrayDeleter<float> deleteTmpBuffer(&fTmpBuffer);
		
	vernl(6,"MESplitTaskComputeAudioClipDataPreview::PerformNextBit() fCurFrame = " << fCurFrame << " fChunkListDuration = " << fChunkListDuration);
	ArmeErr err;
	err.type = kArmeOK;
	if(fCurFrame >= fChunkListDuration)
	{
		vernl(6,"****** MESplitTaskComputeAudioClipDataPreview::PerformNextBit() FINISHED !!!!!!! ***************");
		// not need to "delete fTmpBuffer"....... deleteTmpBuffer will take care of this
		
		// first trigger th
//		fAudioClipData->NotifyDataHasChanged();
		fAudioClipData->CloseTmpFile(kCallingThread_AppKit);
		fAudioClipData->AdoptMainPreview(fPreview,fPreviewSize); // this won't throw
		
		// the preview has been adopted, so let's unrabi it
		deletePreview.SaveResource();
		
		// the preview has been computed so now we can redraw the GUI
		// this can throw if some listener proc throws
		// but we don't care since we don't own any resource
		fAudioClipData->NotifyDataHasChanged(); 
				
		SetProgress(1.0);
		return err;
	}
	
	// we are going to make a rough magnitude estimate.
	// we want to update the progress bar between 2 and 5 times a second
	// otoh, we know the min/max stuff is so easy it is disk bound
	// disk : between 5Mb/second and 30Mb/second
	// stereo float audio data : sizeof(float)*2*48000 bytes/sec = 375Kb/sec
	// thus the disk reads audio files between 14 and 85 times faster
	// than real time
	// thus, if we want to update minimum 2 times/second, we need to make
	// packets of 0.5sec*14 = 7seconds. if disk is 5Mb/sec, the progress bar
	// will be updated at 2Hz. If it is 30Mb sec, it will be updated at 12Hz,
	// which is ok since it will be a faster computer/
	
	// thus, (let's assume sample rate is 48k stereo)
	// if we want packets of 7 seconds, it means 48k*7 = 336000 frames.
	// thus, since we process by kMainPreviewBinSize frames packets, we 
	// need to process 336k/kMainPreviewBinSize packets at a time.
	
	// $$$TODO take document sample rate into account, nb of channels, speed
	// of mac measured in an automatic benchmark, etc.. etc.. etc...
	
	int nbOfBigPacketsProcessed = 336000/kMainPreviewBinSize;
	int iBigPacket = 0;

	// set the progress
	float progress;
	if(fCurFrame == 0)
		progress = 0.0;
	else
		progress = (1.0/(float)((double)fChunkListDuration/(double)fCurFrame)); //because of SInt64 oddities
	// beware of rounding errors.... processing can't be finished now !
	if(progress >= 1.0)
		progress = 0.99;
	vernl(6,"MESplitTaskComputeAudioClipDataPreview::PerformNextBit() fCurFrame = " << (signed long long)fCurFrame << " fChunkListDuration = " << (signed long long)fChunkListDuration << " ABOUT TO CALL SetProgress(" << progress << ")");
	SetProgress(progress);

	for(iBigPacket = 0; iBigPacket < nbOfBigPacketsProcessed; iBigPacket++)
	{
		// make sure we don't read past the end. because our nbOfBigPacketsProcessed is
		// not very clever.
		if(fCurFrame >= fChunkListDuration)
		{	
			// this is NOT an error
			deletePreview.SaveResource();
			deleteTmpBuffer.SaveResource();
			return err;
		}

		SInt32 remainingFrames = fChunkListDuration- fCurFrame;
		SInt32 framesToRead = kMainPreviewBinSize;
		if(framesToRead > remainingFrames)
			framesToRead = remainingFrames;
		err = fAudioClipData->ReadFloatFramesFromAChunkListToFlatBuffer(
			fTmpBuffer, fCurFrame, framesToRead, fAudioClipData->GetChunkList(),kCallingThread_AppKit);
		if(err.type != kArmeOK)
		{
			ZZ_THROW_1ARG(MEExceptionComputingPreview,"Cannot read from the edit list...");
		}
			
		fCurFrame += framesToRead;
			
		if(fAudioClipData->IsStereo())
		{
			float Lmin, Rmin;
			float Lmax = Lmin = fTmpBuffer[0];
			float Rmax = Rmin = fTmpBuffer[1];
			for(SInt32 i = 0; i < framesToRead; i++)
			{
				float Lval = fTmpBuffer[(i << 1)];
				float Rval = fTmpBuffer[(i << 1)+1];
				
				if(Lval > Lmax)	Lmax = Lval;
				if(Rval > Rmax)	Rmax = Rval;
				if(Lval < Lmin)	Lmin = Lval;
				if(Rval < Rmin)	Rmin = Rval;
			}
			
			fPreview[fPreviewPoint].max = Lmax;
			fPreview[fPreviewPoint].min = Lmin;
			fPreviewPoint++;
			ZzAssert(fPreviewPoint < 2*fPreviewSize);
			fPreview[fPreviewPoint].max = Rmax;
			fPreview[fPreviewPoint].min = Rmin;
			fPreviewPoint++;
		}
		else
		{
			float Lmin;
			float Lmax = Lmin = fTmpBuffer[0];
			for(SInt32 i = 0; i < framesToRead; i++)
			{
				float Lval = fTmpBuffer[i];
				if(Lval > Lmax)	Lmax = Lval;
				if(Lval < Lmin)	Lmin = Lval;
			}
			
			ZzAssert(fPreviewPoint < fPreviewSize);
			fPreview[fPreviewPoint].max = Lmax;
			fPreview[fPreviewPoint].min = Lmin;
			fPreviewPoint++;
		}
	}
	deletePreview.SaveResource();
	deleteTmpBuffer.SaveResource();
	return err;
};
	
/*
MEAudioClipDataPreviewPoint* MEAudioClipData::ComputeMainPreview(
	const MEAudioChunkList&	chunkList, 
	ProgressTracker*	progressTracker)
{
	// we need as many sample points as there are samples / kMainPreviewBinSize
	// and this is stereo only
	SInt64 chunksSize = 0;
	SInt32 i;
	for(i = 0; i < nbChunks; i++)
	{
		chunksSize += chunks[i].length;
	}
				
	// NRDIV in frames (negative remainder division)
	SInt32 previewSize = (chunksSize-1)/kMainPreviewBinSize + 1; 
							
	// 2* because stereo
	MEAudioClipDataPreviewPoint* preview = new MEAudioClipDataPreviewPoint[2*previewSize]; 
	verntl(audioClipsBenjamin,5,"NEW : preview = " << preview);
	// the tmp buffer where we are going to read from the chunk set
	float* tmpBuffer = new float[kMainPreviewBinSize*2];
	verntl(audioClipsBenjamin,5,"NEW : tmpBuffer = " << tmpBuffer);
	
	// indices
	SInt64 curFrame = 0;
	SInt64 previewPoint = 0;

	while(curFrame < chunksSize)
	{
		if(progressTracker)
		{
			if(curFrame == 0)
				progressTracker->SetProgress(0.0);
			else
				progressTracker->SetProgress(1.0/(float)(chunksSize/curFrame));
		}

		ReadFloatFramesFromAChunkListToFlatBuffer(
			tmpBuffer, curFrame, kMainPreviewBinSize, chunks, nbChunks);
		curFrame += kMainPreviewBinSize;
		
		float Lmax = 0.0;
		float Rmax = 0.0;
		for(SInt32 i = 0; i < kMainPreviewBinSize; i++)
		{
			float Lval = tmpBuffer[(i << 1)];
			float Rval = tmpBuffer[(i << 1)+1];
			
			if(Lval > Lmax)
				Lmax = Lval;
			if(Rval > Rmax)
				Rmax = Rval;
		}
		preview[previewPoint].max = Lmax;
		previewPoint++;
		preview[previewPoint].max = Rmax;
		previewPoint++;
	}
	delete [] tmpBuffer;
	return preview;
};
*/
