/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/
 
#define ME_AudioChunks_h
#ifndef ME_AudioChunks_h
#include "ME_AudioChunks.h"
#endif //ME_AudioChunks_h

#ifndef ZzTypes_h
#include "ZzTypes.h"
#endif //ZzTypes_h

#ifndef ZzArray_h
#include "ZzArray.h"
#endif //ZzArray_h

#include <ostream>

class ArmeArchiveHandler;

/*************************************************************************************

MEAudioChunkFileIndex

*************************************************************************************/

struct MEAudioChunkFileIndex
{
	SInt32	data;
	void WriteToArchive(ArmeArchiveHandler* archiver) const;
	void ReadFromArchive(ArmeArchiveHandler* unarchiver);
	
	MEAudioChunkFileIndex(SInt32 _data) { data = _data; }
	MEAudioChunkFileIndex() { data = -1; }
};

inline bool operator==(const MEAudioChunkFileIndex& c1, const MEAudioChunkFileIndex& c2)
{
	return (c1.data == c2.data);
}

inline bool operator!=(const MEAudioChunkFileIndex& c1, const MEAudioChunkFileIndex& c2)
{
	return !(c1 == c2);
}

extern MEAudioChunkFileIndex kMEUninitializedChunkFileIndex;

/*************************************************************************************

MEAudioChunk

*************************************************************************************/
class TArmeDoc;

class MEAudioChunk
{
	friend std::ostream& operator<<(std::ostream& s, const MEAudioChunk& audioChunk);
public :
							MEAudioChunk(TArmeDoc* doc);
							MEAudioChunk(); // we need this for array ctor
							~MEAudioChunk();
							MEAudioChunk(const MEAudioChunk& other);
	MEAudioChunk&			operator=(const MEAudioChunk& other);
	void					WriteToArchive(ArmeArchiveHandler* archiver) const;
	void					ReadFromArchive(ArmeArchiveHandler* archiver);
	void					SetFileIndex(MEAudioChunkFileIndex _chunkFileIndex);
	void					Retain();
	void					Release();
	bool					IsSilentChunk() const;
	MEAudioChunkFileIndex	GetFileIndex() const; 
	const char*				GetSFPPath() const;
	void					GetPath(char* path, SInt32 bufferSize) const;

public :
	static MEAudioChunk	CreateSilentChunk(TArmeDoc* doc, SInt64 duration);

private :	
	// index of the file that contains this chunk in the main doc chunkfile index. file name can be computed from this number
	MEAudioChunkFileIndex	chunkFileIndex; 
public :	
	// starting frame of this chunk in the chunkfile
	SInt64				startInFile;
	// duration of this chunk in the chunkfile
	SInt64				duration;
	// the document where this chunk belongs too
	TArmeDoc*			doc;
};

extern bool operator==(const MEAudioChunk& c1, const MEAudioChunk& c2);
extern bool operator!=(const MEAudioChunk& c1, const MEAudioChunk& c2);

/*************************************************************************************

MEChunkFileInfo

*************************************************************************************/

class MEChunkFileInfo
{
public :
	bool 		bMTGA;
	char*		fPath; // only valid if bMTGA is false
	bool		bFileIsAlias;  //introduced in version 1.4
	UInt32		fRefCount;
	bool		fOwnedByMontage;

public :
	void		ReadFromArchive(ArmeArchiveHandler* unarchiver);
	void		WriteToArchive(ArmeArchiveHandler* archiver) const;
				MEChunkFileInfo();
				~MEChunkFileInfo();
	
	
private : 
	// make sure there's a central repository for the infos. don't leave them
	// anywhere
						MEChunkFileInfo(const MEChunkFileInfo&);
						MEChunkFileInfo(MEChunkFileInfo&);
	MEChunkFileInfo&	operator=(const MEChunkFileInfo&);
	MEChunkFileInfo&	operator=(const MEChunkFileInfo);
};

/*************************************************************************************

MEAudioChunk

*************************************************************************************/
class TArmeDoc;

class MEAudioChunkFileManager
{
public :
								MEAudioChunkFileManager(TArmeDoc* doc);
								~MEAudioChunkFileManager();
								
	/*! @function SetOldStyleRefsUpToExc initializes the whole stuff with MTGA refs owned
		by montage. This is needed for backwards compatibility */
	void						SetOldStyleRefsUpToExc(UInt32 audioChunkNextIndex);

	/*! @function GetNewAudioChunkIndex Generates a project-wide unique audio chunk index */
	MEAudioChunkFileIndex		GetNewIndex();
	
	void						SetChunkFileExternalSFP(MEAudioChunkFileIndex index, const char* path);
	

	/*! @function GetMTGAPath returns the internal name of a path the corresponds to "index". It's doesn't really
		matter if the index is correct (it shouldn't actually) because this function is meant to check whether
		there are actual files in the project folder that are outside the range of the correct indices. In this case,
		they get deleted. the static version needs a doc to work on, while the member version uses the 
		fDoc member of the audio chunk file manager */
	static void					GetMTGAPath(const TArmeDoc* doc, char* path, UInt32 bufferSize, MEAudioChunkFileIndex index);
	void						GetMTGAPath(char* path, UInt32 bufferSize, MEAudioChunkFileIndex index) const;
	
	/*! @function GetPath Retrieves the real path of a correct chunk file (that exists in this project) */
	void						GetPath(char* path, UInt32 bufferSize, MEAudioChunkFileIndex index) const;
	void						GetPath(char* path, UInt32 bufferSize, const MEAudioChunk& chunk) const;
	
	// this static method is 'shared' with the MEAudioChunk class
	static void 				GetPath(const TArmeDoc* doc, char* path, UInt32 bufferSize, MEAudioChunkFileIndex index);
	
	void						ReleaseFile(MEAudioChunkFileIndex index);
	void						RetainFile(MEAudioChunkFileIndex index);
	
	// these 2 functions are just handy shortcuts
	void						ReleaseFile(const MEAudioChunk& chunk);
	void						RetainFile(const MEAudioChunk& chunk);

	bool						IsChunkFileMTGA(const MEAudioChunk& chunk) const;
	bool						IsChunkFileMTGA(MEAudioChunkFileIndex index) const;

	void						ReadFromArchive(ArmeArchiveHandler* unarchiver);
	void						WriteToArchive(ArmeArchiveHandler* archiver);
	
	void						DeleteUnusedFiles();
	
	/*! @function GetChunkFileInfo Allows to read a particular chunk file info */
	const MEChunkFileInfo*		GetChunkFileInfo(MEAudioChunkFileIndex index) const;

	/*! @function GetNbChunkFileInfos The number of registered chunk files */
	SInt32 						GetNbChunkFileInfos() const;
	
	MEAudioChunk				AllocateNewZeroRefMTGAChunk();
	MEAudioChunkFileIndex 		AllocateNewZeroRefMTGAChunkIndex();

private :
	// this array contains the audio 
	ZzArray<MEChunkFileInfo*>	fAudioChunkFileInfos;				
	TArmeDoc*					fDoc;
};


