/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#ifndef ME_Audio_File_Helpers_h
#include "ME_Audio_File_Helpers.h"
#endif //ME_Audio_File_Helpers_h

#ifndef MEQTSCAudioReader_h
#include "MEQTSCAudioReader.h"
#endif //MEQTSCAudioReader_h

#ifndef MIProgressBarDlgManager_h
#include "MIProgressBarDlgManager.h"
#endif //MIProgressBarDlgManager_h

#ifndef ZzException_h
#include "ZzException.h"
#endif //ZzException_h

#ifndef ZzHelpers_h
#include "ZzHelpers.h"
#endif //ZzHelpers_h

#ifndef MEMTGAAudioFile_h
#include "MEMTGAAudioFile.h"
#endif //MEMTGAAudioFile_h

#ifndef ZzRABI_h
#include "ZzRABI.h"
#endif //ZzRABI_h

#include <memory>

void ME_ConvertQTAudioTrackToMTGAFile(
	const FSSpec& 				movieFile, 
	int 						trackIndex, 
	const ZzString&				mtgaFilePath,
	UInt32						targetFrameRate,
	MIProgressBarDlgManager*	progressDlg)
{
	ArmeErr err;
	// DEBUG CHECK : make sure we don't overwrite an existing MTGA file
	FILE* debugFile = fopen(mtgaFilePath.getUTF8String(),"rb");
	if(debugFile)
	{
		fclose(debugFile);
		ZzString txt = ZzString("Sorry, but file ") + mtgaFilePath + ZzString(" already exists. An exception will be thrown");
		ZzWarningMessageBox(txt.getUTF8String());
		ZZ_THROW_1ARG(ZzAudioClipImportException,"Target internal file already exists");
	}
	
	const UInt32 						kBufferNbFramesForConversion = 65536;
	float* tmpFrames = new float[kBufferNbFramesForConversion*2];
	ZzRABI_ArrayDeleter<float>					rabiTmpFrames(&tmpFrames);
	
	MEQTSCAudioReader* qtReader = new MEQTSCAudioReader(movieFile,trackIndex,targetFrameRate,kBufferNbFramesForConversion,&err);
	ZzRABI_ObjectDeleter<MEQTSCAudioReader> 	rabiQTReader(&qtReader);
	
	if(err.type != kArmeOK) //$$$TODO move the throw inside the MEQTSCAudioReader ctor
	{
		ZZ_THROW_1ARG(ZzAudioClipImportException,ArmeGetStringFromError(err));
	}
	
	if(qtReader->GetNbFrames() == 0)
	{
		ZZ_THROW_1ARG(ZzAudioClipImportException,"The audio track being imported is empty");
	}
	
	MEMTGAAudioFile dstFile;
	err = dstFile.OpenSaveMTGA(mtgaFilePath.getUTF8String(),targetFrameRate,qtReader->GetNbChannels(),kMEMTGAAudioFileFmt_Float32);
	if(err.type != kArmeOK)
	{
		ZZ_THROW_1ARG(ZzAudioClipImportException,"Cannot open the destination file in the document folder. Please make sure your hard disk is not full.");
	}
	
	SInt64					totalFramesRead = 0;
	bool 					finished = false;
	SInt32 					bufferNbFrames; // number of read frames
	float					progress;

	while(!finished)
	{
		try
		{
			err = qtReader->ReadFloatFrames(&finished,tmpFrames,&bufferNbFrames,&progress);
			if(err.type != kArmeOK)
			{
				ZZ_THROW_1ARG(ZzAudioClipImportException,"Error reading data from the audio track.");
			}
			if(progressDlg)
				progressDlg->Update(progress);
		}
		catch(...)
		{
			dstFile.Close();
			remove(mtgaFilePath.getUTF8String());
			throw;
		}
		
		if(bufferNbFrames > 0)
		{
			totalFramesRead += bufferNbFrames;
			try
			{
				err = dstFile.WriteFloat32Frames(tmpFrames,(SInt64)bufferNbFrames);
				if(err.type != kArmeOK)
				{
					ZZ_THROW_1ARG(ZzAudioClipImportException,"Error writing in the imported file.");
				}
			}
			catch(...)
			{
				dstFile.Close();
				remove(mtgaFilePath.getUTF8String());
				throw;
			}
		}
		else
		{
			finished = true; // either finished = true OR bufferNbFrames = 0 in the ReadFloatFrames call mean it's done
		}
	}
	dstFile.Close();
}

