/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#ifndef OfflineAudioDirtyResampler_h
#include "OfflineAudioDirtyResampler.h"
#endif //OfflineAudioDirtyResampler_h

#include "ABOTwavesync.h"
#include "ABOTinterp8.h"
#include "ABOTfilters.h"

#import "MIOfflineAudioDirtyResamplerController.h"
#import "MIOfflineAudioProcessingNormalizeDlgController.h"

bool OfflineAudioDirtyResampler::DisplayAudioProcessorConfiguration(
												OfflineAudioProcessorParams* offlineParams)
{
	NSModalSession session;
	NSWindow* 	win;
    SInt64 duration, maxDuration=0;
    OfflineAudioReader* reader;
    SInt32 nbClips = offlineParams->GetNbClips();
    
	id dlgController = [MIOfflineAudioDirtyResamplerController alloc];
	[dlgController initWithWindowNibName:@"MIOfflineAudioDirtyResampler" owner:dlgController];
	win = [dlgController window];
    
    for (int i=0; i<nbClips; i++)
    {
        reader = offlineParams->GetReader(i);
        duration = reader->GetDuration();
        if (duration > maxDuration) maxDuration = duration;
    }

    [dlgController setOrigDuration:_ArmeTime(maxDuration)];
    [dlgController setAlgoType:0];
	
	session = [NSApp beginModalSessionForWindow:win];
	while([NSApp runModalSession:session] == NSRunContinuesResponse)
	{
	}
	[NSApp endModalSession:session];					
	[win orderOut:0];
	OfflineAudioDirtyResamplerParams* params = dynamic_cast<OfflineAudioDirtyResamplerParams*>(offlineParams);
	bool wasCancelled = [dlgController wasCancelled];
	params->fNewDuration = [dlgController getDuration];
    params->fAlgo = (OfflineAudioDirtyResamplerAlgo)[dlgController getAlgoType];
	[dlgController release];
	if(wasCancelled)
		return false;
	return true;
}

OfflineAudioProcessorParams* OfflineAudioDirtyResampler::CreateDefaultParams()
{
    OfflineAudioDirtyResamplerParams* p = new OfflineAudioDirtyResamplerParams;
    p->fAlgo = WSOLA;
    return p;
}

ArmeErr OfflineAudioDirtyResampler::PerformAudioProcessing(
	OfflineAudioProcessorParams* offlineParams)
{
    OfflineAudioDirtyResamplerParams* params = dynamic_cast<OfflineAudioDirtyResamplerParams*>(offlineParams);
	ZzAssert(params != NULL);
	ArmeErr err;

    switch (params->fAlgo)
    {
        case WSOLA:
        case OLA:
            err = PerformAudioProcessingWSOLA(offlineParams);
            break;
        case Resample:
            err =  PerformAudioProcessingResample(offlineParams);
            break;
        case DirtyResampler:
            err = PerformAudioProcessingDirtyResampler(offlineParams);
            break;
        default:
            ZzAssert(0); // this really shouldnt happen
			break;
    }
	offlineParams->SetProgress(1.0);
	return err;
}

ArmeErr OfflineAudioDirtyResampler::PerformAudioProcessingResample(
	OfflineAudioProcessorParams* offlineParams)
{
	ArmeErr err;
    OfflineAudioDirtyResamplerParams* params = dynamic_cast<OfflineAudioDirtyResamplerParams*>(offlineParams);
	float scaleProgress = 1.0/(float)offlineParams->GetNbClips();
	ZzAssert(params != NULL);
   	for( SInt32 clip = 0; clip < offlineParams->GetNbClips(); clip++)
	{
		OfflineAudioReader* reader = offlineParams->GetReader(clip);
		OfflineAudioWriter* writer = offlineParams->GetWriter(clip);
		SInt64 N = 4096;        // buffer length
        SInt32 nbChannels = reader->GetNbChannels();
		SInt64 readDuration = reader->GetDuration();
		SInt64 writeDuration = GET_VAL(params->fNewDuration);
        ABOTinterp8* resampler = new ABOTinterp8;
        ABOTfilterSeries** filter = new ABOTfilterSeries*[nbChannels];
        int i;
        for(i = 0; i < nbChannels; i++)
        {
        	filter[i] = new ABOTfilterSeries(10);
        }
        
        float cutoff = (5.0f/12.0f) * (float)(writeDuration)/(float)(readDuration);
        
        for (i = 0; i < nbChannels; i++)
        {
            filter[i]->setButterLP(cutoff); 
            //this will set the cutoff freq to 5/6 of the destination band to reduce aliasing
        }
        float* Buffer = new float[nbChannels * N];
    
		SInt64 Offset = 0;
        
        
        // PHASE 1: filtering if lowering sample rate
        if (writeDuration < readDuration)
        { 
			writer->SetDuration(readDuration);
    
            // forward filtering.... 
            Offset = 0;
            while(Offset < readDuration)
            {
				// this will be from scaleProgress*clip -> scaleProgress*(clip+0.33)
				float progress = scaleProgress*(clip+0.33*((float)Offset/readDuration));
				offlineParams->SetProgress(progress);
			
				if(offlineParams->IsCancelled())
					return ArmeErr(kArmeAudioProcCancelled);
                reader->ReadFramesSafe(Buffer,Offset,N);
                for(int j=0; j<N; j++)
                {
                    for(int i=0; i< nbChannels; i++)
                    {	
                        filter[i]->Bang(Buffer[i + nbChannels * j], Buffer[i + nbChannels * j]);
                    }
                }
                writer->WriteFramesSafe(Buffer, Offset, N);
            
                Offset += N;
            
            }
    
            // backward filtering
            Offset = readDuration;
            while (Offset > 0)
            {
				// this will be from scaleProgress*(clip+0.33) -> scaleProgress*(clip+0.66)
				float progress = scaleProgress*(clip+0.33+0.33*((float)(readDuration-Offset)/readDuration));
				offlineParams->SetProgress(progress);

				if(offlineParams->IsCancelled())
					return ArmeErr(kArmeAudioProcCancelled);
                Offset -= N;

                reader->ReadFramesSafe(Buffer, Offset,N);
                for(int j=N-1; j>=0; j--)
                {
                    for(int i=0; i< nbChannels; i++)
                    {	
                        filter[i]->Bang(Buffer[i + nbChannels * j], Buffer[i + nbChannels * j]);
                    }
                }
                writer->WriteFramesSafe(Buffer, Offset, N);
            
                            
            }

        }
        // PHASE 2: resampling
		writer->SetDuration(writeDuration);
        SInt64 inFrames = 8 + (N * readDuration) / writeDuration;
        SInt64 inOffset, localOffset;
        double fracOffset;
        double ratio = (double)(readDuration)/(double)(writeDuration);
        float* inBuffer = new float[nbChannels * inFrames];
        
        Offset = 0;
        inOffset = 0;
        
        while (Offset < writeDuration)
        {
			// this will be from scaleProgress*(clip+0.66) -> scaleProgress*(clip+0.99)
			float progress = scaleProgress*(clip+0.66+0.33*((float)(Offset)/writeDuration));
			offlineParams->SetProgress(progress);
			if(offlineParams->IsCancelled())
				return ArmeErr(kArmeAudioProcCancelled);
            fracOffset = ratio * (double)Offset;
            inOffset = (SInt64)fracOffset;
            reader->ReadFramesSafe(inBuffer, inOffset-3, inFrames);
            
            

            for (int j=0; j<N; j++)
            {
                fracOffset = ratio * (double)(Offset + j);
                fracOffset -= (double)inOffset;
                localOffset = (SInt64)fracOffset;
                fracOffset -= (double)localOffset;
                
                for (int i=0; i<nbChannels; i++)
                {
                    resampler->Bang(
                        (float)(fracOffset+3.0f), 
                        &(inBuffer[i + nbChannels * localOffset]), 
                        nbChannels,
                        Buffer[i + nbChannels * j]
                    );
                }
            }
            
            writer->WriteFramesSafe(Buffer, Offset, N);
            Offset += N;
        }
		delete [] Buffer;
        delete [] inBuffer;
        
		for(i = 0; i < nbChannels; i++)
		{
			delete filter[i];
		}
		delete [] filter;
	}
	return err;
}


// the progress in this function is sent to the OfflineAudioProcessorParams
// instance, not to the boptask

ArmeErr OfflineAudioDirtyResampler::PerformAudioProcessingDirtyResampler(
	OfflineAudioProcessorParams* offlineParams)
{
	ArmeErr err;
	OfflineAudioDirtyResamplerParams* params = dynamic_cast<OfflineAudioDirtyResamplerParams*>(offlineParams);
	ZzAssert(params != NULL);
	for( SInt32 clip = 0; clip < offlineParams->GetNbClips(); clip++)
	{
		float baseProgress = ((float)clip/(float)offlineParams->GetNbClips());
		float scaleProgress = 1.0/(float)offlineParams->GetNbClips();
		OfflineAudioReader* reader = offlineParams->GetReader(clip);
		OfflineAudioWriter* writer = offlineParams->GetWriter(clip);
		const SInt32 kTmpWriteBufferNbFrames = 32768;
		SInt32 nbChannels = reader->GetNbChannels();
		SInt64 readDuration = reader->GetDuration();
		SInt64 writeDuration = GET_VAL(params->fNewDuration);
		writer->SetDuration(writeDuration);
	
		// the conversion factor
		double newToOldFactor = (double)(readDuration-1)/(double)(writeDuration-1);
		
		SInt32 tmpReadBufferNbFrames = (SInt32)(kTmpWriteBufferNbFrames*newToOldFactor) + 2; // "+1" to take rounding errors into account
		
		float* tmpWriteBuffer = new float[nbChannels*kTmpWriteBufferNbFrames];
		float* tmpReadBuffer = new float[nbChannels*tmpReadBufferNbFrames];
			
	//	SInt64 readPos = 0;
		SInt64 writePos = 0;
		
		while(writeDuration - writePos > 0)
		{
			if(offlineParams->IsCancelled())
				return ArmeErr(kArmeAudioProcCancelled);
			// when pos = duration (minus something) and clip = getNbClip - 1, 
			// SetProgress(baseProgress + scaleProgress - epsilon)
			// = SetProgress((NBCLIPS-1)/NBCLIPS + 1/NBCLIPS - epsilon) = SetProgress(1-epsilon)
			offlineParams->SetProgress(baseProgress + scaleProgress*(float)writePos/(float)writeDuration);
			SInt32 writeNbFrames;
			if( (writeDuration-writePos) >= kTmpWriteBufferNbFrames )
				writeNbFrames = kTmpWriteBufferNbFrames;
			else
				writeNbFrames = writeDuration-writePos;
	
			// now we are going to compute the pos and size of the
			// buffer we need to read in order to write frames from
			// writePos to writePos + writeNbFrames (excluding end)
			SInt64 readPos = (SInt64)(writePos*newToOldFactor);
			SInt64 lastReadPos = (SInt64)((writePos+writeNbFrames-1)*newToOldFactor);
			SInt32 readNbFrames = lastReadPos-readPos+1;
			ZzAssert(readNbFrames <= tmpReadBufferNbFrames);
			reader->ReadFrames(tmpReadBuffer,readPos,readNbFrames);
			
			// here we are going to write the destination from
			// writePos to writePos + writeNbFrames (excluding end)
			if(nbChannels == 1)
			{
				for(SInt32 writeFrameInBuffer = 0; writeFrameInBuffer < writeNbFrames; writeFrameInBuffer++)
				{
					SInt32 readFrameInBuffer = ((SInt32)((writePos+writeFrameInBuffer)*newToOldFactor)) - readPos;
					tmpWriteBuffer[writeFrameInBuffer] = tmpReadBuffer[readFrameInBuffer];
				}
			}
			else if(nbChannels == 2)
			{
				for(SInt32 writeFrameInBuffer = 0; writeFrameInBuffer < writeNbFrames; writeFrameInBuffer++)
				{
					SInt32 readFrameInBuffer = ((SInt32)((writePos+writeFrameInBuffer)*newToOldFactor)) - readPos;
					tmpWriteBuffer[2*writeFrameInBuffer] = tmpReadBuffer[2*readFrameInBuffer];
					tmpWriteBuffer[2*writeFrameInBuffer+1] = tmpReadBuffer[2*readFrameInBuffer+1];
				}
			}
			else
			{
				ZzAssert(0);
			}
			writer->WriteFrames(tmpWriteBuffer,writePos,writeNbFrames);
			readPos += readNbFrames;
			writePos += writeNbFrames;
		}
		delete [] tmpReadBuffer;
		delete [] tmpWriteBuffer;
	}
	return err;
}

// this is the tobe WSOLA timestretcher
ArmeErr OfflineAudioDirtyResampler::PerformAudioProcessingWSOLA(
	OfflineAudioProcessorParams* offlineParams)
{
	ArmeErr err;
	OfflineAudioDirtyResamplerParams* params = dynamic_cast<OfflineAudioDirtyResamplerParams*>(offlineParams);
	float scaleProgress = 1.0/(float)offlineParams->GetNbClips();
	ZzAssert(params != NULL);
	for( SInt32 clip = 0; clip < offlineParams->GetNbClips(); clip++)
	{
		OfflineAudioReader* reader = offlineParams->GetReader(clip);
		OfflineAudioWriter* writer = offlineParams->GetWriter(clip);
		SInt64 N = 1024;        // overlap add buffer length
        SInt32 nbChannels = reader->GetNbChannels();
		SInt64 readDuration = reader->GetDuration();
		SInt64 writeDuration = GET_VAL(params->fNewDuration);
        SInt32 syncEnable = (params->fAlgo == OLA) ? (0) : (1);
		writer->SetDuration(writeDuration);
        ABOTwavesync * sync = new ABOTwavesync(N, nbChannels);
	
		// the conversion factor (inverse rico of timewarp line)
		double s_inv = (double)(readDuration)/(double)(writeDuration);

        // the buffers
		float* outputBuffer = new float[nbChannels * N];
		float* regionBuffer = new float[nbChannels * 2 * N];
			
		SInt64 inputOffset = 0;
		SInt64 outputOffset = 0;
        
//      SInt64 	leftOverRead;
		SInt64	i, j, k, l;
        float a,b;
        SInt32 d=0;
	//	SInt32 dd=0;
        
        float myfreq = ArboPi / (float)N;
        float mycos = cos(myfreq);
        float mysin = sin(myfreq);
        float p,q,pp,qq;
        
        //loop invariants: output_offset == offset of the next output buffer fragment to write
        //				   input_offset == offset of the next input buffer fragment to read
        while(outputOffset < writeDuration)
        {
			// this will be from scaleProgress*clip -> scaleProgress*(clip+0.33)
			float progress = scaleProgress*(clip+((float)outputOffset/writeDuration));
			offlineParams->SetProgress(progress);
			
			if(offlineParams->IsCancelled())
				return ArmeErr(kArmeAudioProcCancelled);
            // 1. read continuation of last output buffer into output_buffer[];
            reader->ReadFramesSafe(outputBuffer,inputOffset,N);
            
            // 2. compute desired offset from time stretch function
            inputOffset = (SInt64)(s_inv * (double)(outputOffset));
            // subtract maximum tolerance interval
//            inputOffset -= N/2;
            
                                    
            // 3. read region around desired offset into region_buffer[];
            reader->ReadFramesSafe(regionBuffer, inputOffset, 2*N);

            // 4. get synchronisation offset


            d = sync->GetSyncOffset(outputBuffer, regionBuffer);
            d *= syncEnable;

            //d = 0; // disable sync
            
            // 5. crossfade output_buffer[0:N-1] with region_buffer[d:N-1+d];
            p = 0.5f;
            q = 0.0f;
            
            for (i = 0; i < N ; i ++)
            {
                // k is base index in outputBuffer
                // l is base index in regionBuffer
                k = nbChannels * (i);
                l = nbChannels * (i + d);
                
                // linear interpolation constants a and b
                a = 0.5f + p;
                b = 0.5f - p;


                // real dirty debug hack:  REMOVE THIS!
                // first channel is idea;
                // second channel is new segment
//                outputBuffer[k+1]=regionBuffer[l+1];
                

                for (j=0; j<nbChannels; j++)
                {	//a = 0.0f; b = 1.0f;  // windowing off
                    outputBuffer[k+j] = a * outputBuffer[k+j] + b * regionBuffer[l+j];
                }
                
                pp = mycos * p + mysin * q;
                qq = mycos * q - mysin * p;
                p = pp;
                q = qq;


            }
            
            // 6. write the outputbuffer
            writer->WriteFramesSafe(outputBuffer, outputOffset, N);
            
            // 7. advance outputOffset, inputOffset
            inputOffset += N + d;
            outputOffset += N;
            
        }

		delete [] outputBuffer;
		delete [] regionBuffer;
        delete sync;
	}
	return err;
}

		
bool OfflineAudioDirtyResampler::IsLengthy(OfflineAudioProcessorParams* offlineParams)
{	
	// we consider a long operation when the total number of samples frames to resample is 
	// greater than 50000
	SInt64 totalNbSamples = 0;
	for( SInt32 clip = 0; clip < offlineParams->GetNbClips(); clip++)
	{
		totalNbSamples += offlineParams->GetReader(clip)->GetNbChannels() *
			offlineParams->GetReader(clip)->GetDuration();
	}
	return(totalNbSamples > 50000);
}
	
