/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#ifndef OfflineAudioProcessing_h
#include "OfflineAudioProcessing.h"
#endif //OfflineAudioProcessing_h

#ifdef TRACK_BASED_AUDIO_CLIPS

#ifndef MEAudioClip_h
#include "MEAudioClip.h"
#endif //MEAudioClip_h

#else //#ifdef TRACK_BASED_AUDIO_CLIPS

#ifndef TAudioClip2_h
#include "TAudioClip2.h"
#endif //TAudioClip2_h

#endif //#ifdef TRACK_BASED_AUDIO_CLIPS

#ifndef MEMTGAAudioFile_h
#include "MEMTGAAudioFile.h"
#endif //MEMTGAAudioFile_h

#ifndef TArmeDoc_h
#include "TArmeDoc.h"
#endif //TArmeDoc_h

#undef VER_DEF_TOPIC
#define VER_DEF_TOPIC offlineProcBenjamin

/*******************************************************************************************************

OfflineAudioClipReader

*******************************************************************************************************/

OfflineAudioClipReader::OfflineAudioClipReader(
	MEAudioClip*	clip, 
	ArmeTime 		clipStartTime, 
	ArmeTime 		duration)
{
	fClip = clip;
	fStartTime = GET_VAL(clipStartTime);
	fDuration = GET_VAL(duration);							
}

ArmeErr OfflineAudioClipReader::ReadFrames(
	float*	dst, 
	SInt64	selectionStartTime, 
	SInt64	duration) const
{
	ZzRtAssert(fStartTime + selectionStartTime + duration <= GET_VAL(fClip->GetDuration()));
	return fClip->GetData()->ReadFloatFramesToFlatBuffer(dst,fStartTime + selectionStartTime,duration,kCallingThread_OfflineAudioSplitTaskThread);
}

ArmeErr OfflineAudioReader::ReadFramesSafe(
	float*	dst, 
	SInt64	selectionStartTime, 
	SInt64	duration) const
{
    SInt64 	i;
	ArmeErr	err;
    SInt64 	clipDuration = this->GetDuration();
    SInt32 	nbChannels = this->GetNbChannels();
    
    // take care of left boundary effects
    if (selectionStartTime < 0) 
    {
        SInt32 prePad = -selectionStartTime;

        // make sure prePad does not exceed duration
        if (prePad > duration){
            prePad = duration;
            duration = 0;
        }
        else duration -= prePad;
        
        // pad with zero when reading before start of selection
        for (i=0; i < prePad * nbChannels; i++) dst[i] = 0.0f;

        // advance buffer
        dst += prePad * nbChannels;

        // reset start point
        selectionStartTime = 0;
    }
    
    
    // take care of right boundary effects    
    if (duration + selectionStartTime > clipDuration)
	{
		// get post pad amount
		SInt32 postPad = duration + selectionStartTime - clipDuration;

		// make sure postPad does not exceed duration
		if (postPad > duration)
		{
			postPad = duration;
			duration = 0;
        }
        else 
			duration -= postPad;
        
        // fill the end of the buffer with zeros
        for (i=0; i < postPad * nbChannels; i++)
		{
			dst[i + duration * nbChannels] = 0.0f;
		}
    }
    // perform the actual read if there are still frames to read
    if(duration)
	{
		err = ReadFrames(dst, selectionStartTime, duration);
	}
	return err;
}
	
SInt32 OfflineAudioClipReader::GetNbChannels() const
{
	return fClip->GetData()->GetNbChannels();
}

SInt64 OfflineAudioClipReader::GetDuration() const
{
	return fDuration;
}

OfflineAudioClipReader::~OfflineAudioClipReader()
{
}

/*******************************************************************************************************

OfflineAudioWriter

*******************************************************************************************************/


OfflineAudioClipWriter::OfflineAudioClipWriter(MEAudioClip* clip, ArmeTime clipStartTime, ArmeTime duration) :
	fChunk(clip->GetDoc())
{
	fClip = clip;
	fStartTime = GET_VAL(clipStartTime);
	fChunk = clip->GetDoc()->AudioChunkFileManager()->AllocateNewZeroRefMTGAChunk(); // ref count is 0 at start
	fAudioFile.OpenSaveNewChunkFile(clip->GetDoc(),fChunk,fClip->GetData()->GetNbChannels());
}

OfflineAudioClipWriter::~OfflineAudioClipWriter()
{
}

void OfflineAudioClipWriter::CloseAudioFile()
{
	fAudioFile.Close();
	// reopen the file in read mode and synchronize the chunk duration and the file size written
	ArmeErr err = fAudioFile.OpenReadChunk(fClip->GetDoc(),fChunk);
	// if we went that far, we can't really go back.
	ZzAssert(err.type == kArmeOK);
	fChunk.duration = fAudioFile.GetNbFrames();
	fAudioFile.Close();
}

void OfflineAudioClipWriter::SetDuration(SInt64 duration)
{
	fDuration = duration;
}

SInt64 OfflineAudioClipWriter::GetDuration() const
{
	return fDuration;
}

SInt32 OfflineAudioClipWriter::GetNbChannels() const
{
	return fClip->GetData()->GetNbChannels();
}

const MEAudioChunk&	OfflineAudioClipWriter::GetChunk() const
{
	return fChunk;
}

ArmeErr OfflineAudioClipWriter::WriteFrames(const float* src, SInt64 selectionStartTime, SInt64 duration)
{
	return fAudioFile.WriteFloat32FramesFromFlatBuffer(src,selectionStartTime,duration);
}

ArmeErr OfflineAudioWriter::WriteFramesSafe(const float* src, SInt64 selectionStartTime, SInt64 duration)
{
    SInt64 clipDuration = this->GetDuration();
    SInt32 nbChannels = this->GetNbChannels();
	ArmeErr	err;

    // check left border
    if (selectionStartTime < 0)
    {
        SInt64 preSkip = -selectionStartTime;
        duration -= preSkip;
        src += preSkip * nbChannels;
        selectionStartTime = 0;
    }
    
    // check right border
    if (duration + selectionStartTime > clipDuration) duration = clipDuration - selectionStartTime;
  
    // perform actual write
    err = this->WriteFrames(src, selectionStartTime, duration);
	return err;
}

/*******************************************************************************************************

OfflineAudioProcessorParams

*******************************************************************************************************/

OfflineAudioProcessorParams::OfflineAudioProcessorParams() :
	fIsCancelled(false)
{
}

SInt32 OfflineAudioProcessorParams::GetNbClips()
{	
	// some debug stuff
	SInt32 countReaders = fReaders.CountItems();
	SInt32 countWriters = fWriters.CountItems();
	ZzAssert(countReaders == countWriters);
	return countReaders;
}

OfflineAudioWriter* OfflineAudioProcessorParams::GetWriter(SInt32 clip)
{
	OfflineAudioWriter* writer = fWriters.ItemAt(clip);
	return writer;
}

OfflineAudioReader* OfflineAudioProcessorParams::GetReader(SInt32 clip)
{
	OfflineAudioReader* reader = fReaders.ItemAt(clip);
	return reader;
} 

void OfflineAudioProcessorParams::AdoptWriter(OfflineAudioWriter* writer)
{
	fWriters.AddItem(writer);
}

void OfflineAudioProcessorParams::AdoptReader(OfflineAudioReader* reader)
{
	fReaders.AddItem(reader);
}

OfflineAudioProcessorParams::~OfflineAudioProcessorParams()
{
	for(SInt32  i = 0; i < fReaders.CountItems(); i++)
	{
		delete fReaders[i];
	}
	
	for(SInt32  i = 0; i < fReaders.CountItems(); i++)
	{
		delete fWriters[i];
	}
}

/*
#ifdef TRACK_BASED_AUDIO_CLIPS
MEAudioClip* OfflineAudioProcessorParams::GetClip(SInt32 clip)
{
	// some debug stuff
	MEAudioClip* readerClip = fReaders.ItemAt(clip)->GetClip();
	MEAudioClip* writerClip = fWriters.ItemAt(clip)->GetClip();
	ZzAssert(readerClip == writerClip);
	return readerClip;
}
#else //#ifdef TRACK_BASED_AUDIO_CLIPS
TAudioClip2* OfflineAudioProcessorParams::GetClip(SInt32 clip)
{
	// some debug stuff
	TAudioClip2* readerClip = fReaders.ItemAt(clip)->GetClip();
	TAudioClip2* writerClip = fWriters.ItemAt(clip)->GetClip();
	ZzAssert(readerClip == writerClip);
	return readerClip;
}
#endif //#ifdef TRACK_BASED_AUDIO_CLIPS
*/

/*******************************************************************************************************

OfflineAudioProcessor

*******************************************************************************************************/

OfflineAudioProcessorParams* OfflineAudioProcessor::CreateDefaultParams()
{
	return new OfflineAudioProcessorParams;
}

bool OfflineAudioProcessor::DisplayAudioProcessorConfiguration(
	OfflineAudioProcessorParams* offlineParams)
{
	return true;
}

bool OfflineAudioProcessor::IsLengthy(OfflineAudioProcessorParams* offlineParams)
{
	return true;
}































/* OBSOLETE BECAUSE WE'RE USING THE SoundConverter RESAMPLING NOW*/
OBSOLETE_OfflineAudioFileReader::OBSOLETE_OfflineAudioFileReader(MEMTGAAudioFile* file)
{
	fFile = file;
	ZzAssert(fFile->GetOpenMode() == MEMTGAAudioFile::kOpenRead);
}

/* OBSOLETE BECAUSE WE'RE USING THE SoundConverter RESAMPLING NOW*/
ArmeErr OBSOLETE_OfflineAudioFileReader::ReadFrames(float* dst, SInt64 selectionStartTime, SInt64 duration) const
{
	ArmeErr err;
	err = fFile->Seek(selectionStartTime);
	if(err.type != kArmeOK)
		return err;
	return fFile->ReadFloat32Frames(dst,&duration);
}

/* OBSOLETE BECAUSE WE'RE USING THE SoundConverter RESAMPLING NOW*/
SInt64 OBSOLETE_OfflineAudioFileReader::GetDuration() const
{
	return fFile->GetNbFrames();
}

/* OBSOLETE BECAUSE WE'RE USING THE SoundConverter RESAMPLING NOW*/
SInt32 OBSOLETE_OfflineAudioFileReader::GetNbChannels() const
{
	return fFile->GetNbChannels();
}

/* OBSOLETE BECAUSE WE'RE USING THE SoundConverter RESAMPLING NOW*/
OfflineAudioFileWriter::OfflineAudioFileWriter(MEMTGAAudioFile* file)
{
	fFile = file;
	fDuration = 0;
	ZzAssert(fFile->IsOpenedForWriting());
}

/* OBSOLETE BECAUSE WE'RE USING THE SoundConverter RESAMPLING NOW*/
void OfflineAudioFileWriter::SetDuration(SInt64 duration)
{
	vernl(1,"OfflineAudioFileWriter::SetDuration("<<(signed long long)duration<<") shouldn't be called. fix your code");
	fDuration = duration;
}

/* OBSOLETE BECAUSE WE'RE USING THE SoundConverter RESAMPLING NOW*/
SInt64 OfflineAudioFileWriter::GetDuration() const
{
	return fDuration;
}

/* OBSOLETE BECAUSE WE'RE USING THE SoundConverter RESAMPLING NOW*/
SInt32 OfflineAudioFileWriter::GetNbChannels() const
{
	return fFile->GetNbChannels();
}

/* OBSOLETE BECAUSE WE'RE USING THE SoundConverter RESAMPLING NOW*/
ArmeErr OfflineAudioFileWriter::WriteFrames(const float* src, SInt64 selectionStartTime, SInt64 duration)
{
	ArmeErr err;
	err = fFile->Seek(selectionStartTime);
	if(err.type != kArmeOK)
		return err;
	return fFile->WriteFloat32Frames(src,duration);
}

/* OBSOLETE BECAUSE WE'RE USING THE SoundConverter RESAMPLING NOW*/
void OfflineAudioFileWriter::CloseAudioFile()
{
	fFile->Close();
}

