/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#define OfflineAudioProcessing_h

#ifndef OfflineAudioProcessing_h
#include "OfflineAudioProcessing.h"
#endif //OfflineAudioProcessing_h

#ifndef ArmeBasicTypes_h
#include "ArmeBasicTypes.h"
#endif //ArmeBasicTypes_h

#ifndef ZzArray_h
#include "ZzArray.h"
#endif //ZzArray_h

#ifndef ZzProgressTracker_h
#include "ZzProgressTracker.h"
#endif //ZzProgressTracker_h

#ifndef MEMTGAAudioFile_h
#include "MEMTGAAudioFile.h"
#endif //MEMTGAAudioFile_h

#ifndef MEAudioChunkList_h
#include "MEAudioChunkList.h"
#endif //MEAudioChunkList_h

class MEAudioClip;

/*
	@class OfflineAudioReader 

	Base class to read in objects. concrete classes implement this for
	specific media containers
	
*/

class OfflineAudioReader
{
public :
    /* this is the save version. if read beyond boundaries, *dst is padded with zeros */
	ArmeErr				ReadFramesSafe(float* dst, SInt64 selectionStartTime, SInt64 duration) const;
	
	/*! @function ReadFrames This function reads "duration" frames at position "startTime".
		The requested range is not allowed to extend past clip end */
	virtual ArmeErr		ReadFrames(float* dst, SInt64 selectionStartTime, SInt64 duration) const = 0;

	/*! @function GetDuration() This function returns the number of frames in the source 
	    selection */
	virtual SInt64		GetDuration() const = 0;
	
	/*! @function GetNbChannels This function returns the number of channels in the source 
    selection format */
	virtual SInt32		GetNbChannels() const = 0;
    
    virtual ~OfflineAudioReader() {};
};

/*
	@class OfflineAudioClipReader 
	
	This class allows the offline processor to read samples in audio clips (that is,
	in the chunklist embedded in the MEAudioData of this clip). This object only allows 
	to read up to the size given by the GetDuration() function.
			   
*/

class OfflineAudioClipReader : public OfflineAudioReader
{
public :
	OfflineAudioClipReader(MEAudioClip* clip, ArmeTime clipStartTime, ArmeTime duration);
	virtual ~OfflineAudioClipReader();

	virtual ArmeErr	ReadFrames(float* dst, SInt64 selectionStartTime, SInt64 duration) const;
    
	virtual SInt32	GetNbChannels() const;
	
	/*! @function GetStartTimeInClip() Returns the reader start time in the clip */
	SInt64	GetStartTimeInClip() { return fStartTime; }
	
	virtual SInt64	GetDuration() const;
	
	/*! @function GetClip() Returns the clip associated with the reader */
	MEAudioClip*	GetClip() { return fClip; }
private :
	MEAudioClip*	fClip;
	SInt64			fStartTime;
	SInt64			fDuration;
};

/*
	@class OBSOLETE_OfflineAudioFileReader 
	
	This class allows the offline processor to read samples in a MEMTGAAudioFile.
	This object only allows to read up to the size given by the GetDuration() 
	function. The file must be open in read mode.
			   
*/

class OBSOLETE_OfflineAudioFileReader : public OfflineAudioReader
{
public :
	OBSOLETE_OfflineAudioFileReader(MEMTGAAudioFile* file);
	
	/*! @function ReadFrames This function reads "duration" frames at position "startTime".
		The requested range is not allowed to extend past clip end */
	virtual ArmeErr ReadFrames(float* dst, SInt64 selectionStartTime, SInt64 duration) const;

	/*! @function GetDuration() This function returns the number of frames in the source 
	    selection */
	virtual SInt64	GetDuration() const;
	
	/*! @function GetNbChannels This function returns the number of channels in the source 
    selection format */
	virtual SInt32	GetNbChannels() const;
	
private :
	MEMTGAAudioFile*	fFile;
    virtual ~OBSOLETE_OfflineAudioFileReader() {};
};

/*
	@class OfflineAudioWriter
	
	This class allows the offline processor to write samples. 
	
	Please note that if the GetDuration() function returns < 0, then
	it means it's up to the processor to decide the number of output
	frames. The processor then needs to 
	 - compute the number of output frames according to the input frames
	   and the processing parameters
	 - call SetDuration() on the writer
	 - then, it is allowed to call WriteFrames() on the writer
*/

class OfflineAudioWriter
{
public :
	ArmeErr				WriteFramesSafe(const float* src, SInt64 selectionStartTime, SInt64 duration);
public :
	virtual void		SetDuration(SInt64 duration) = 0;
	virtual SInt64		GetDuration() const = 0;
	virtual SInt32		GetNbChannels() const = 0;
	virtual ArmeErr		WriteFrames(const float* src, SInt64 selectionStartTime, SInt64 duration) = 0;
	virtual void		CloseAudioFile() = 0;
    virtual ~OfflineAudioWriter() {};

};

class MEAudioChunk;

/*!
	@class OfflineAudioClipWriter 
	Concrete subclass OfflineAudioWriter that writes in audio clips 
	(that is, that inserts a new chunk in the chunklist of the audio data of this clip)
*/

class OfflineAudioClipWriter : public OfflineAudioWriter
{
public :
	OfflineAudioClipWriter(MEAudioClip* clip, ArmeTime clipStartTime, ArmeTime duration);
	virtual ~OfflineAudioClipWriter();

	virtual void		SetDuration(SInt64 duration);
	virtual SInt64		GetDuration() const;
	virtual SInt32		GetNbChannels() const;
	virtual ArmeErr		WriteFrames(const float* src, SInt64 selectionStartTime, SInt64 duration);
	virtual void		CloseAudioFile();

	const MEAudioChunk&	GetChunk() const;

	/*! @function GetClip() Returns the clip associated with the reader */
	MEAudioClip*	GetClip() { return fClip; }
private :
	MEAudioClip*	fClip;
	SInt64					fStartTime;
	SInt64					fDuration;
	MEAudioChunk			fChunk;
	MEMTGAAudioFile				fAudioFile;
};

/*!
	@class OfflineAudioFileWriter 
	
	Concrete subclass of OfflineAudioWriter that writes in a MEMTGAAudioFile instance.
	This file must be opened for writing
*/

class OfflineAudioFileWriter : public OfflineAudioWriter
{
public :
	/*! @function OfflineAudioFileWriter The file param needs to be already opened in write mode */
	OfflineAudioFileWriter(MEMTGAAudioFile* file);
public :
	virtual void		SetDuration(SInt64 duration);
	virtual SInt64		GetDuration() const;
	virtual SInt32		GetNbChannels() const;
	virtual ArmeErr		WriteFrames(const float* src, SInt64 selectionStartTime, SInt64 duration);
	virtual void		CloseAudioFile();
	virtual ~OfflineAudioFileWriter() {};

private :
	MEMTGAAudioFile*		fFile;
	SInt64				fDuration;
};

/*
	@class OfflineAudioProcessorParams
	
	This class is the base class for the offline processor 
	parameters. 
	
	It provides an interface to retrieve the number of clips that need
	to be processed by the command, as well as means to get a read/write interface
	to the clip data.
		
	NOTE :
	Parameters are not structures private to the offline processor because
	
	1) later, OfflineAudioProcessorParams will contain a formal interface for offline
	   processors to declare their parameters.
	   
	2) This base class can contain some information shared by all the offline
	   processor (such as number of input frames, number of channels, etc...)
	   
*/

class OfflineAudioProcessorParams : public ZzProgressTracker
{
public :
	/*! @function OfflineAudioProcessorParams Creaate the basic data structures for the params */ 
						OfflineAudioProcessorParams();
	virtual				~OfflineAudioProcessorParams();
	
	/*! @function GetNbClips Returns the nb of clips that needs to be processed */ 
	SInt32				GetNbClips();

	/*! @function GetWriter Returns the writer object for the i-th clip */ 
	OfflineAudioWriter*	GetWriter(SInt32 clip);

	/*! @function GetReader Returns the reader object for the i-th clip */ 
	OfflineAudioReader*	GetReader(SInt32 clip);
	
	bool				IsCancelled();
	
	/*! @function This function needs to be called when the processing needs to stop. The processor should call
	    the IsCancelled function at each loop iteration to check whether it needs to stop processing and return
		a "false" value that means it has been cancelled */
	void				Cancel();
	
	/*! @function SetCancel Tells the processing to stop. Thread-safe */

public : // not meant for client use
	void				AdoptWriter(OfflineAudioWriter* writer);
	void				AdoptReader(OfflineAudioReader* reader);

private :
	ZzArray<OfflineAudioWriter*>	fWriters;
	ZzArray<OfflineAudioReader*>	fReaders;
	bool							fIsCancelled; // no need for a lock on this since it's on/off
};

inline bool OfflineAudioProcessorParams::IsCancelled()
{
	return fIsCancelled;
}

inline void OfflineAudioProcessorParams::Cancel()
{
	fIsCancelled = true;
}

/*
	@class OfflineAudioProcessor
	
	This is the base class for the offline processor. There should be a subclass
	of this class for each offline audio processor. Sngletons for these classes needs to
	be created at application launch time.
	
	These are mixed MI/Arme classes in that they're responsible for creating their own
	interface. Maybe later we can have a formal method for the offline processors to report
	their capabilities and parameters so that a dialog box asking for the params can be built
	on the fly, so that scripting can be used (in other words, to clearly separate view, model 
	and vontroller). 
	
	Here's what you need to do to create an offline processor :
	
	- subclass this for the new processor (example : ReverseOfflineAudioProcessor)
	- subclass OfflineAudioProcessorParams to add processing parameter data, setters and getters
	  (for example, ReverseOfflineAudioProcessorParams)
	- implement ReverseOfflineAudioProcessor::CreateDefaultParams() to create a default instance
      of your parameters (the nb of clips, channels, readers, writers and the like will be set
	  later by the framework, don't worry). You can defer parameter initialisation to 
	  DisplayAudioProcessorConfiguration() if they depend upon the duration and nb of channels.
	  Return YES if processing should continue or NO if the user clicked on cancel.
	- implement ReverseOfflineAudioProcessor::DisplayAudioProcessorConfiguration() to display a
	  Cocoa dlg that fills offlineParams. Services for "preview result" audio output will be 
	  offered later
	- implement ReverseOfflineAudioProcessor::PerformAudioProcessing() to actually fill the 
	  writers object with the resulting data
	  
	IMPORTANT NOTE : PLEASE MAKE ALL THESE FUNCTIONS THREAD-SAFE !!!!!!!!!! 
	THE SAME OfflineAudioProcessor INSTANCE CAN AND WILL BE USED BY SEVERAL THREADS TO PROCESS 
	AUDIO CHUNKS !!! YOU HAVE BEEN WARNED !!!
		  	
	(btw, please put all these subclasses into MI (Montage folder), not in Arme since there 
	 will be resources in this stuff)
*/

class OfflineAudioProcessor
{
public :
	/*! @function CreateDefaultParams
		This function should create an instance of a subclass of OfflineAudioProcessorParams that
		represent the parameter set for this processing algorithm. Deletion of this instance is handled
		by the framework
	*/
	virtual OfflineAudioProcessorParams*	CreateDefaultParams();

	/*! @function DisplayAudioProcessorConfiguration
	    This function should display a Cocoa window that allows the user to setup the processor
	    parameters. Please note this function may return immediatly if no setup is needed (example :
		normalize). 
	*/
	virtual	bool							DisplayAudioProcessorConfiguration(
												OfflineAudioProcessorParams* offlineParams);

	/*! @function PerformAudioProcessing This function performs the actual processing according to
		the OfflineAudioProcessorParams subclass instance. Deletion of the offlineParams instance is
		being handled by the host.
	*/
	virtual ArmeErr							PerformAudioProcessing(
												OfflineAudioProcessorParams* offlineParams) = 0;
												
	/*! @function IsLengthy This function must return true if the processing, being given the passed
        parameters, is going to be lengthy (which means a progress bar must be displayed) 
	*/
	virtual bool							IsLengthy(OfflineAudioProcessorParams* offlineParams);
	virtual ~OfflineAudioProcessor() {};

};

template <class DerivedType> class OfflineAudioProcessorBase : public OfflineAudioProcessor
{
public :
	static DerivedType*						GetInstance();
	virtual ~OfflineAudioProcessorBase() {};

private :
	static DerivedType*	fsInstance;		
};

template <class DerivedType> inline DerivedType* OfflineAudioProcessorBase<DerivedType>::GetInstance()
{
	if(!fsInstance)
		fsInstance = new DerivedType;
	return fsInstance;
}
	   	
template <class DerivedType> DerivedType* OfflineAudioProcessorBase<DerivedType>::fsInstance = NULL;
