/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#ifndef OfflineAudioProcessorSilence_h
#include "OfflineAudioProcessorSilence.h"
#endif //OfflineAudioProcessorSilence_h

#import "MIInsertSilenceDlgCtrller.h"

OfflineAudioProcessorParams* OfflineAudioProcessorSilence::CreateDefaultParams()
{
	OfflineSilenceParams* params = new OfflineSilenceParams;
	params->fDuration = kOfflineSilenceUninitializedTime;
	return params;
}

// the progress in this function is sent to the OfflineAudioProcessorParams
// instance
ArmeErr OfflineAudioProcessorSilence::PerformAudioProcessing(
	OfflineAudioProcessorParams* offlineParams)
{	
	ArmeErr err;
	OfflineSilenceParams* params = dynamic_cast<OfflineSilenceParams*>(offlineParams);
	ZzAssert(params);
	const SInt32 kTmpBufferNbFrames = 32768;
	for( SInt32 clip = 0; clip < offlineParams->GetNbClips(); clip++)
	{
		float baseProgress = ((float)clip/(float)offlineParams->GetNbClips());
		float scaleProgress = 1.0/(float)offlineParams->GetNbClips();
		OfflineAudioReader* reader = offlineParams->GetReader(clip);
		OfflineAudioWriter* writer = offlineParams->GetWriter(clip);
		SInt32 nbChannels = reader->GetNbChannels();
		// the sprecial kOfflineSilenceReplaceSelWithSilence value for the duration means we're 
		// replacing the selection with silence. otherwise, the value is the duration chosen with
		// the dlg...
		SInt64 duration;
		if(params->fDuration == kOfflineSilenceReplaceSelWithSilence)
		{
			duration = reader->GetDuration();
		}
		else
		{
			ZzAssert(GET_VAL(params->fDuration) > 0);
			duration = GET_VAL(params->fDuration);
		}
		writer->SetDuration(duration);
		float* tmpBuffer = new float[nbChannels*kTmpBufferNbFrames];

        // zero out buffer
        float zero = 0.0f;
		ZzAssert((kTmpBufferNbFrames % 8) == 0);
        for (int i=0; i < kTmpBufferNbFrames*nbChannels; i+= 8)
        {
            tmpBuffer[i] =  zero;
            tmpBuffer[i+1] =  zero;
            tmpBuffer[i+2] =  zero;
            tmpBuffer[i+3] =  zero;
            tmpBuffer[i+4] =  zero;
            tmpBuffer[i+5] =  zero;
            tmpBuffer[i+6] =  zero;
            tmpBuffer[i+7] =  zero;
        }
		SInt64 pos = 0;
		while(duration - pos > 0)
		{
			offlineParams->SetProgress(baseProgress + scaleProgress*(float)pos/(float)duration);
			SInt32 writeNbFrames;
			if( (duration-pos) >= kTmpBufferNbFrames )
				writeNbFrames = kTmpBufferNbFrames;
			else
				writeNbFrames = duration-pos;

			writer->WriteFrames(tmpBuffer,pos,writeNbFrames);
			pos += writeNbFrames;
		}
		delete [] tmpBuffer;
	}
    offlineParams->SetProgress(1);
	return err;
}
	
bool OfflineAudioProcessorSilence::IsLengthy(OfflineAudioProcessorParams* offlineParams)
{	
	return true;
}
	
bool OfflineAudioProcessorSilence::DisplayAudioProcessorConfiguration(
												OfflineAudioProcessorParams* offlineParams)
{

	OfflineSilenceParams* params = dynamic_cast<OfflineSilenceParams*>(offlineParams);
	ZzAssert(params->fDuration != kOfflineSilenceUninitializedTime);
	
	// if the special kOfflineSilenceReplaceSelWithSilence is assigned to the sel, it means
	// the length will be set by the caller and we don't need to display that dialog
	if(params->fDuration == kOfflineSilenceReplaceSelWithSilence)
		return true;
	
	NSModalSession session;
	NSWindow* 	win;
	
	MIInsertSilenceDlgCtrller* ctrller =
		[MIInsertSilenceDlgCtrller alloc];
	[ctrller initWithWindowNibName:@"MIInsertSilenceDlg" owner:ctrller];
	win = [ctrller window];
	[ctrller setMontageDocument:params->fMontageDoc];
	[ctrller setInitialDuration:_ArmeTime(0)];
	
	session = [NSApp beginModalSessionForWindow:win];
	while([NSApp runModalSession:session] == NSRunContinuesResponse)
	{
	}
	[NSApp endModalSession:session];					
	bool ret;
	if([ctrller duration] <= _ArmeTime(0))
	{
		ret = false; // duration 0 OR cancelled dialog, which is the same.... we don't insert anythin
	}
	else
	{
		params->fDuration = [ctrller duration];
		ret = true;
	}
	[ctrller release];
	[win orderOut:0];
	return ret;
}
	
