/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#ifndef OfflineAudioTimeStretch_h
#include "OfflineAudioTimeStretch.h"
#endif //OfflineAudioTimeStretch_h

#include "ABOTwavesync.h"
#include "ABOTinterp8.h"
#include "ABOTfilters.h"

#import "MIOfflineAudioTimeStretchController.h"
#import "MIOfflineAudioProcessingNormalizeDlgController.h"

bool OfflineAudioTimeStretch::DisplayAudioProcessorConfiguration(
												OfflineAudioProcessorParams* offlineParams)
{
	NSModalSession session;
	NSWindow* 	win;
    SInt64 duration, maxDuration=0;
    OfflineAudioReader* reader;
    SInt32 nbClips = offlineParams->GetNbClips();
    
	id dlgController = [MIOfflineAudioTimeStretchController alloc];
	[dlgController initWithWindowNibName:@"MIOfflineAudioTimeStretch" owner:dlgController];
	win = [dlgController window];
    
    for (int i=0; i<nbClips; i++)
    {
        reader = offlineParams->GetReader(i);
        duration = reader->GetDuration();
        if (duration > maxDuration) maxDuration = duration;
    }

    [dlgController setOrigDuration:_ArmeTime(maxDuration)];
    [dlgController setAlgoType:0];
	
	session = [NSApp beginModalSessionForWindow:win];
	while([NSApp runModalSession:session] == NSRunContinuesResponse)
	{
	}
	[NSApp endModalSession:session];					
	[win orderOut:0];
	OfflineAudioTimeStretchParams* params = dynamic_cast<OfflineAudioTimeStretchParams*>(offlineParams);
	bool wasCancelled = [dlgController wasCancelled];
	params->fNewDuration = [dlgController getDuration];
    params->fAlgo = (OfflineAudioTimeStretchAlgo)[dlgController getAlgoType];
	[dlgController release];
	if(wasCancelled)
		return false;
	return true;
}

OfflineAudioProcessorParams* OfflineAudioTimeStretch::CreateDefaultParams()
{
    OfflineAudioTimeStretchParams* p = new OfflineAudioTimeStretchParams;
    p->fAlgo = WSOLA; /* set the default algorithm to high quality */
    return p;
}

ArmeErr OfflineAudioTimeStretch::PerformAudioProcessing(
	OfflineAudioProcessorParams* offlineParams)
{
    OfflineAudioTimeStretchParams* params = dynamic_cast<OfflineAudioTimeStretchParams*>(offlineParams);
	ZzAssert(params != NULL);
	ArmeErr err;

    switch (params->fAlgo)
    {
        case WSOLA:
        case OLA:
            err = PerformAudioProcessingWSOLA(offlineParams);
            break;
        default:
            ZzAssert(0); // this really shouldnt happen
			break;
    }
	offlineParams->SetProgress(1.0);
	return err;
}


// this is the WSOLA (waveform synchronous overlap add) timestretcher
ArmeErr OfflineAudioTimeStretch::PerformAudioProcessingWSOLA(
	OfflineAudioProcessorParams* offlineParams)
{
	ArmeErr err;
	OfflineAudioTimeStretchParams* params = dynamic_cast<OfflineAudioTimeStretchParams*>(offlineParams);
	float scaleProgress = 1.0/(float)offlineParams->GetNbClips();
	ZzAssert(params != NULL);
	for( SInt32 clip = 0; clip < offlineParams->GetNbClips(); clip++)
	{
		OfflineAudioReader* reader = offlineParams->GetReader(clip);
		OfflineAudioWriter* writer = offlineParams->GetWriter(clip);
		SInt64 N = 1024;        // overlap add buffer length
        SInt32 nbChannels = reader->GetNbChannels();
		SInt64 readDuration = reader->GetDuration();
		SInt64 writeDuration = GET_VAL(params->fNewDuration);
        bool syncEnable = (params->fAlgo == WSOLA);
		writer->SetDuration(writeDuration);
        ABOTwavesync * sync = new ABOTwavesync(N, nbChannels);
	
		// the conversion factor (inverse rico of timewarp line)
		double s_inv = (double)(readDuration)/(double)(writeDuration);

        // the buffers
		float* outputBuffer = new float[nbChannels * N];
		float* regionBuffer = new float[nbChannels * 2 * N];
			
		SInt64 inputOffset = 0;
		SInt64 outputOffset = 0;
        
//      SInt64 	leftOverRead;
		SInt64	i, j, k, l;
        float a,b;
        SInt32 d=0;
	//	SInt32 dd=0;
        
        float myfreq = ArboPi / (float)N;
        float mycos = cos(myfreq);
        float mysin = sin(myfreq);
        float p,q,pp,qq;
        
        //loop invariants: output_offset == offset of the next output buffer fragment to write
        //				   input_offset == offset of the next input buffer fragment to read
        while(outputOffset < writeDuration)
        {
			// this will be from scaleProgress*clip -> scaleProgress*(clip+0.33)
			float progress = scaleProgress*(clip+((float)outputOffset/writeDuration));
			offlineParams->SetProgress(progress);
			
			if(offlineParams->IsCancelled())
				return ArmeErr(kArmeAudioProcCancelled);
            // 1. read continuation of last output buffer into output_buffer[];
            reader->ReadFramesSafe(outputBuffer,inputOffset,N);
            
            // 2. compute desired offset from time stretch function
            inputOffset = (SInt64)(s_inv * (double)(outputOffset));

            // 3. read region around desired offset into region_buffer[];
            reader->ReadFramesSafe(regionBuffer, inputOffset, 2*N);

            // 4. get synchronisation offset if we are doing a high quality time stretch
            if (syncEnable) d = sync->GetSyncOffset(outputBuffer, regionBuffer);
            else d = 0;
            
            // 5. crossfade output_buffer[0:N-1] with region_buffer[d:N-1+d];
            p = 0.5f;
            q = 0.0f;
            
            for (i = 0; i < N ; i ++)
            {

            
                // k is base index in outputBuffer
                // l is base index in regionBuffer
                k = nbChannels * (i);
                l = nbChannels * (i + d);
                
                // hanning window interpolation constants a and b  (a + b == 1)
                a = 0.5f + p;
                b = 0.5f - p;


                for (j=0; j<nbChannels; j++)
                {
                    outputBuffer[k+j] = a * outputBuffer[k+j] + b * regionBuffer[l+j];
                }
                
                pp = mycos * p + mysin * q;
                qq = mycos * q - mysin * p;
                p = pp;
                q = qq;


            }
            
            // 6. write the outputbuffer
            writer->WriteFramesSafe(outputBuffer, outputOffset, N);
            
            // 7. advance outputOffset, inputOffset
            inputOffset += N + d;
            outputOffset += N;
            
        }

		delete [] outputBuffer;
		delete [] regionBuffer;
        delete sync;
	}
	return err;
}

		
bool OfflineAudioTimeStretch::IsLengthy(OfflineAudioProcessorParams* offlineParams)
{	
	// we consider a long operation when the total number of samples frames to resample is 
	// greater than 50000
	SInt64 totalNbSamples = 0;
	for( SInt32 clip = 0; clip < offlineParams->GetNbClips(); clip++)
	{
		totalNbSamples += offlineParams->GetReader(clip)->GetNbChannels() *
			offlineParams->GetReader(clip)->GetDuration();
	}
	return(totalNbSamples > 50000);
}
	
