/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#ifndef ABOTcomplex_h
#define ABOTcomplex_h

#include <math.h>

#ifndef ZzVerbose_h
#include "ZzVerbose.h"
#endif //ZzVerbose_h

#ifndef DSP_Globals_h
#include "DSP_Globals.h"
#endif //DSP_Globals_h

/* 	complex data type class
    note this is not terribly optimized so dont use in loops */
class ABOTcomplex
{
    public:
      inline ABOTcomplex() {_r = _i = 0;}
        inline ABOTcomplex(const float &a, const float &b) {setCart(a, b);}
        inline ABOTcomplex(const float &phasor) {setAngle(phasor);}
        
        inline void setAngle(const float &angle) {_r = cos(angle); _i = sin(angle);}
        inline void setPolar(const float &phasor, const float &norm) 
        {_r = norm * cos(phasor); _i = norm * sin(phasor);}
        inline void setCart(const float &a, const float &b) {_r = a; _i = b;}
        
        inline const float& r() const {return _r;}
        inline const float& i() const {return _i;}
        
        inline float norm2() const {return _r*_r+_i*_i;}
        inline float norm() const {return sqrt(norm2());}
        inline void normalize() {float n = 1.0f / norm(); _r *= n; _i *= n;}
        
        inline ABOTcomplex conj() const {return ABOTcomplex(_r, -_i);}

        inline float angle() const {return atan2(_i, _r);}


        inline ABOTcomplex operator+ (const ABOTcomplex &a) const
        {
            return ABOTcomplex(_r + a.r(), _i + a.i());
        }
        inline ABOTcomplex operator+ (float f) const
        {
            return ABOTcomplex(_r + f, _i);
        }
        inline ABOTcomplex operator- (const ABOTcomplex &a) const
        {
            return ABOTcomplex(_r - a.r(), _i - a.i());
        }
        inline ABOTcomplex operator- (float f) const
        {
            return ABOTcomplex(_r - f, _i);
        }

        inline ABOTcomplex operator* (const ABOTcomplex &a) const 
        {
            return ABOTcomplex(_r * a.r() - _i * a.i(), _i * a.r() + _r * a.i());
        }
        inline ABOTcomplex operator* (float f) const
        {
            return ABOTcomplex(_r * f, _i * f);
        }
        inline ABOTcomplex operator/ (const ABOTcomplex &a) const 
        {
            float n_t = 1.0f / a.norm2();
            return ABOTcomplex(n_t * (_r * a.r() + _i * a.i()), n_t * (_i * a.r() - _r * a.i()));
        }
        inline ABOTcomplex operator/ (float f) const 
        {
            float n_t = 1.0f / f;
            return ABOTcomplex(n_t * _r, n_t * _i);
        }
        
        inline friend std::ostream& operator<< (std::ostream& o, ABOTcomplex& a)
        {
            return o << "(" << a.r() << "," << a.i() << ")";
        }

        inline friend ABOTcomplex operator+ (float f, ABOTcomplex& a)
        {
            return(ABOTcomplex(a.r() + f, a.i()));
        }
        
        inline friend ABOTcomplex operator- (float f, ABOTcomplex& a)
        {
            return(ABOTcomplex(f - a.r(), - a.i()));
        }
        
        inline friend ABOTcomplex operator/ (float f, ABOTcomplex& a)
        {
            return(ABOTcomplex(f,0) / a);
        }
        
        // hier is iets mis mee
        inline friend ABOTcomplex operator* (float f, ABOTcomplex& a)
        {
            return(ABOTcomplex(f*a.r(), f*a.i()));
        }
        
        
        inline ABOTcomplex& operator *= (float f)
        {
            _r *= f;
            _i *= f;
            return *this;
        }

        inline ABOTcomplex& operator /= (float f)
        {
            _r /= f;
            _i /= f;
            return *this;
        }

        inline ABOTcomplex& operator *= (ABOTcomplex& a)
        {
            float r_t = _r * a.r() - _i * a.i();
                   _i = _r * a.i() + _i * a.r();
                   _r = r_t;
                   
            return *this;
        }

        inline ABOTcomplex& operator /= (ABOTcomplex& a)
        {
            float n_t = a.norm2();
            float r_t = n_t * (_r * a.r() + _i * a.i());
                   _i = n_t * (_i * a.r() - _r * a.i());
                   _r = r_t;
                   
            return *this;
        }

        
        float _r;
        float _i;
};


// COMPLEX LOG

inline ABOTcomplex abotlog(ABOTcomplex a) /* complex log */
{
    float r_t = log(a.norm());
    float i_t = a.angle();
    return ABOTcomplex(r_t, i_t);
}

// COMPLEX EXP

inline ABOTcomplex abotexp(ABOTcomplex a) /* complex exp */
{
    return (ABOTcomplex(a.i()) * exp(a.r()));
}

// BILINEAR TRANSFORM analog -> digital

inline ABOTcomplex bilin_stoz(ABOTcomplex a)
{
    ABOTcomplex a2 = a * 0.5f;
    return((1.0f + a2)/(1.0f - a2));
}    

// BILINEAR TRANSFORM digital -> analog

inline ABOTcomplex bilin_ztos(ABOTcomplex a)
{
    return ((a - 1.0f) / (a + 1.0f))*2.0f;
}

// not really a complex function but a nice complement to the bilinear routines

inline float bilin_prewarp(float freq)
{
    return 2.0f * tan(kPi * freq);
}    

#endif //ABOTcomplex_h