/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/
 
/* code for dynwav pd class */

#ifndef ABOTinterp8_h
#define ABOTinterp8_h

#include <math.h>

static const float N1 =   0.59814453125;
static const float N2 =  -0.11962890625;
static const float N3 =   0.02392578125;
static const float N4 =  -0.00244140625;
static const float oneover3 = (1.0f/3.0f);
static const float oneover5 = (1.0f/5.0f);
static const float oneover7 = (1.0f/7.0f);
   
static const float __align = 0.0f;

class ABOTinterp8
{

public:
    inline ABOTinterp8(){}
    inline ~ABOTinterp8(){}

	/* feeder routines */

	/* feeder: stride=1 circular buffer */
	inline float BangCircular(
		float frac,
		float* buffer,
		const int &index,
		const int &mask
	)
	{
		return BangCircularStride(
			frac,
			buffer,
			index,
			mask,
			1);
	}
    
	/* feeder: strided circular buffer */
	inline float BangCircularStride(
		float frac,
		float* buffer,
		const int &index,
		const int &mask,
		const int &stride
	)
	{
		return BangRef(
			frac,
			buffer[(index+0*stride) & mask],
			buffer[(index+1*stride) & mask],
			buffer[(index-1*stride) & mask],
			buffer[(index+2*stride) & mask],
			buffer[(index-2*stride) & mask],
			buffer[(index+3*stride) & mask],
			buffer[(index-3*stride) & mask],
			buffer[(index+4*stride) & mask]
			);
		
	}
	
	/* default feeder: strided flat buffer */
    inline void Bang(
        float findex,	// fractional index of output sample \in [3. 4[ for best results
        float *buffer, 	// buffer[0 .. 7] will be used as interpolands
        int	stride,		// buffer stride
        float &output	// reference to output sample
    )
	{

		buffer += 3*stride;
		
		output = BangRef(
			findex - 3.0f,
			buffer[+0*stride],
			buffer[+1*stride],
			buffer[-1*stride],
			buffer[+2*stride],
			buffer[-2*stride],
			buffer[+3*stride],
			buffer[-3*stride],
			buffer[+4*stride]
			);
	}

	/* feeder: circular buffer */


	/* the actual process routines, works on references provided by the feeder routine
		or the user of the class. */
	
	inline float BangRef
	(
		float frac, /* fractional index, starting from sample 0 */
		
					/* the samples are ordered symmetricly around 1/2 */
		float &fm1, /* sample 0, 1 */ 
		float &fp1, 
		float &fm2, /* sample -1, 2 */
		float &fp2, 
		float &fm3, /* sample -2, 3 */
		float &fp3, 
		float &fm4, /* sample -3, 4 */
		float &fp4
	)
    {
        
      // 8 point symmetric polynomial interpolation routine (lagrange)
      // the use of symmetry allows some optimizations

      
//	  float frac
	  float fm, fp, fe, fo;
//	  float r, q;
	  float x1, x2, x3, x4;
	  float g1, g2, g3, g4;
	  float gg, g2g3g4, g1g3g4, g1g2g4, g1g2g3;
	  float acc;
	  
	  frac = 2 *frac - 1; 	// frac = 2 * shifted fractional index
      
	  /* remark: the algoritm uses a (virtual) fractional index in [-1,1]
		and (virtual) sample points at [-7, -5, -3, -1, 1, 3, 5, 7]
		this is to use the avaiable symmetry and not have annoying 1/2 factors */

	  x1 = frac;
	  x2 = frac * oneover3;
	  x3 = frac * oneover5;
	  x4 = frac * oneover7;

	  g1 = 1 - x1*x1;
	  g2 = 1 - x2*x2;
	  g3 = 1 - x3*x3;
	  g4 = 1 - x4*x4;

	  gg       = g3 * g4;
	  g2g3g4   = g2 * gg;       /* 1 */
	  g1g3g4   = g1 * gg;       /* 2 */
	  gg       = g1 * g2;
	  g1g2g4   = g4 * gg;       /* 3 */
	  g1g2g3   = g3 * gg;       /* 4 */


	  /* 1, -1*/
	  fm = fm1;
	  fp = fp1;
	  fe = fp + fm;
	  fo = fp - fm;

	  acc = N1 * g2g3g4 * (fe + x1*fo);

	  /* 2, -2 */
	  fm = fm2;
	  fp = fp2;
	  fe = fp + fm;
	  fo = fp - fm;

	  acc += N2 * g1g3g4 * (fe + x2*fo);

	  /* 3, -3 */
	  fm = fm3;
	  fp = fp3;
	  fe = fp + fm;
	  fo = fp - fm;

	  acc += N3 * g1g2g4 * (fe + x3*fo);

	  /* 4, -4 */
	  fm = fm4;
	  fp = fp4;
	  fe = fp + fm;
	  fo = fp - fm;

	  acc += N4 * g1g2g3 * (fe + x4*fo);


	  return acc;

	}    



private:
    /* const float N1 = 1 / (   2      * (1-(1/9))  * (1-(1/25))  * (1-(1/49))  );
    ** const float N2 = 1 / ( (1-(9))  *      2     * (1-(9/25))  * (1-(9/49))  );
    ** const float N3 = 1 / ( (1-(25)) * (1-(25/9)) *      2      * (1-(25/49)) );
    ** const float N4 = 1 / ( (1-(49)) * (1-(49/9)) * (1-(49/25)) *    2        );
    */
};


#endif // ABOTinterp8_h
