/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/
 
/*
 * ABOTwavesync: A class for correlation based waveform synchronization
 *
 * instantiation arguments: 
 *
 * nbFrames = max nb of frames that will be processed
 * the nbFrames can be changed later but it won't realloc buffers
 *
 * nbChannels = max nb of channels that will be processed
 * the input will be summed to a mono block and then correlated
 *
 * usage: offset = GetSyncOffset(float* ideal, float* region);
 * ideal will be the block you are looking for (length == NbFrames)
 * region will be the block in which is searched (length == NbFrames * 2)
 * the method returns the offset of the maximum correlation offset
 * so region[offset*NbChannels] will be the start of NbFrames size block that 
 * is most similar to the NbFrames size block starting at ideal[0]
 */


#ifndef ABOTwavesync_veclib_h
#define ABOTwavesync_veclib_h


#ifndef ArmeBasicTypes_h
#include "ArmeBasicTypes.h"
#endif //ArmeBasicTypes_h

#include <vecLib/vDSP.h>


class ABOTwavesync
{

public:
	// declare the platform independent interface
	DECLARE_INTERFACE___ABOTwavesync
 
private:

    SInt32		nbFrames;
    SInt32		nbChannels;

	// this defines the bias window
	float 		weightCenter;
	float 		weightTail;
	SInt32		biasOffset;

    // split complex (see vecLib documentation) data representation
    float*		RegionBuffer;
    float*		IdealBuffer;

};

	inline void ABOTwavesync::SetNbFrames(SInt32 i) {nbFrames = i;}
	inline void ABOTwavesync::SetNbChannels(SInt32 i) {nbChannels = i;}
	inline void ABOTwavesync::SetWeightCenter(float w) {weightCenter = w;}
	inline void ABOTwavesync::SetWeightTail(float w) {weightTail = w;}
	inline void ABOTwavesync::SetBiasOffset(SInt32 o) {biasOffset = o;}


	inline ABOTwavesync::ABOTwavesync(SInt32 nbFrames, SInt32 nbChannels)
	{
		this->nbFrames = nbFrames;
		this->nbChannels = nbChannels;
		this->weightCenter = 1.0f;
		this->weightTail = 1.0f;
		this->biasOffset = nbFrames / 2;
    
		// create correlation buffers
		this->IdealBuffer = new float[nbFrames]; 
		this->RegionBuffer = new float[2*nbFrames];
	}


    inline ABOTwavesync::~ABOTwavesync()
	{
		delete [] this->IdealBuffer;
		delete [] this->RegionBuffer;
	}


    inline SInt32 ABOTwavesync::GetSyncOffset(float* ideal, float* region)
	{
		// call the circular buffer routine from here
		return GetSyncOffsetCircular(ideal, region, 0, 0, -1, -1);
	}

    inline SInt32 ABOTwavesync::GetSyncOffsetCircular(float* ideal_buf, float* region_buf,
		SInt32 ideal_index, SInt32 region_index, SInt32 ideal_mask, SInt32 region_mask)
	{

		int i;
		int j;

		int N = this->nbFrames;
		int C = this->nbChannels;
		float sum;

		float max;
		float val;
		SInt32 shift;
    
    
		// 1. GET INPUT
		// copy ideal[] to this->IdealBufferReal/Imag[] + convert to mono + (multiply by window)
 
		for (i = 0; i < N; i++)
		{
 
            sum = ideal_buf[(ideal_index + C*i) & ideal_mask];
            for (j = 1; j < C; j++) sum += ideal_buf[(ideal_index + j + C*i) & ideal_mask];
            
            this->IdealBuffer[i] = sum;
            

            // LATER: 1+cos/2 window
            // it stresses the crossfaded center part, which should be maximally similar
            // edges don't count that much because they are aligned anyway

		}
    
		// copy region[] to my_region[] + convert to mono
    
		for (i = 0; i < 2*N; i++)
		{

            sum = region_buf[(region_index + C*i) & region_mask];
            for (j = 1; j < C; j++) sum += region_buf[(region_index + j+ C*i) & region_mask];
            this->RegionBuffer[i] = sum;

		}
 
    
		// 2. COMPUTE CORRELATION
    
		// see vDSP_book page 70 (veclib doc)
		conv(this->RegionBuffer, 1, this->IdealBuffer, 1, this->RegionBuffer, 1, N, N);
    

		// 2b. APPLY BIAS WINDOW
		float window;
		float window_inc;

		// left tail
		window = weightTail;
		window_inc = (weightCenter - weightTail) / (float)(biasOffset);
		for (int i=0; i<biasOffset; i++)
		{
			this->RegionBuffer[i] *= window;
			window += window_inc;
		}

		// right tail
		window = weightCenter;
		window_inc = (weightTail - weightCenter) / (float)(N - biasOffset);
		for (int i = biasOffset; i<N; i++)
		{
			this->RegionBuffer[i] *= window;
			window += window_inc;
		}
		

		// 3. FIND OFFSET
		// find maximum correlation offset

		shift = 0;
		max = 0.0f;
		for (i=0; i<N; i++)
		{

			val = this->RegionBuffer[i];
			if (val > max)
			{
				max = val;
				shift = i;
			}

		}
		//printf("shift: %d\n", shift);

		return shift;
	
	}


#endif
