/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#ifndef LibraryToggle_h
#include "LibraryToggle.h"
#endif //LibraryToggle_h

#import "MILibraryElement.h"

#import "MIObject.h"
#import "MITime.h"
#import "TArmeWSObj.h"
#import "MIMontageDocument.h"

NSString *MILibraryElementsPboardType = @"MILibraryElementsPboardType";


@implementation MILibraryElement

- (id)init
{
	self = [super init];
	if ( self )
	{
		fPath = nil;
		fObject = nil;
		fParent = nil;
		fChildren = [[NSMutableArray alloc] initWithObjects:nil];
	}
	return self;
}

- (void)dealloc
{
	[fObject release];
	[fChildren release];
	[fPath release];
	[super dealloc];
}

- (id)initWithPath:(NSString*)path
{
	self = [super init];
	if ( self )
	{
		fPath = [path copy];
		fObject = nil;
		fParent = nil;
		fChildren = [[NSMutableArray alloc] initWithObjects:nil];
	}
	return self;
}

- (id)initWithObject:(MIObject*)object
{
	self = [super init];
	if ( self )
	{
		fPath = nil;
		fParent = nil;
		fObject = object;
		fChildren = [[NSMutableArray alloc] initWithObjects:nil];
	}
	return self;
}

#if USE_LIBRARY

-(BOOL)writeToArchive:(ArmeArchiveHandler*)archiver
{
	[super writeToArchive:archiver];
	archiver->WriteObjCObject(fPath);
	archiver->WriteObjCObject(fParent);
	archiver->WriteObjCObject(fObject);
	archiver->WriteObjCObject(fChildren);
	return YES;
}
	
-(id)initFromArchive:(ArmeArchiveHandler*)unarchiver
{
	self = [super initFromArchive:unarchiver];
	if(self)
	{
		fPath = [unarchiver->ReadObjCObject() retain];
		fParent = unarchiver->ReadObjCObject();
		fObject = [unarchiver->ReadObjCObject() retain];
		fChildren = [unarchiver->ReadObjCObject() retain];
	}
	return self;
}

+ (id)elementWithPath:(NSString*)path
{
	return [[[MILibraryElement alloc] initWithPath:path] autorelease];
}

+ (id)elementWithObject:(MIObject*)object
{
	return [[[MILibraryElement alloc] initWithObject:object] autorelease];
}

- (void)setObject:(MIObject*)object
{
	ZzAssert(fObject == nil);
	fObject = [object retain];
	[fPath release];
	fPath = nil;
}

- (void)setPath:(NSString*)path
{
	[fPath release];
	fPath = [path copy];
	fObject = nil;
}

- (void)setParent:(MILibraryElement*)parent
{
	fParent = parent;
}

- (MILibraryElement*)parent
{
	return fParent;
}

- (MIObject*)object
{
	return fObject;
}

- (NSString*)path
{
	return fPath;
}

- (NSArray*)children
{
	return fChildren;
}

- (void)addChild:(MILibraryElement*)child
{
	[fChildren addObject:child];
	[child setParent:self];
}

- (void)removeChild:(MILibraryElement*)child
{
	[child setParent:nil];
	[fChildren removeObject:child];
}

- (NSString*)title
{
	if ( fObject )
		return [fObject title];
	else
		return [fPath lastPathComponent];
}

- (NSString*)startTime
{
	if ( fObject )
		return [[MITime timeWithArmeTime:[fObject WSObject]->GetAbsoluteStartTime() inDocument:[fObject montageDoc]] getSMPTEString];
	else
		return @"-";
}

- (NSString*)endTime
{
	if ( fObject )
		return [[MITime timeWithArmeTime:[fObject WSObject]->GetAbsoluteEndxTime() inDocument:[fObject montageDoc]] getSMPTEString];
	else
		return @"-";
}

- (NSString*)duration
{
	if ( fObject )
		return [[MITime timeWithArmeTime:[fObject WSObject]->GetDuration() inDocument:[fObject montageDoc]] getSMPTEString];
	else
		return @"-";
}

- (NSString*)status
{
	id value = @"";
	if ( fObject )
	{
		// mute
		if ( [fObject isMuted] )
			value = [value stringByAppendingString:@"Muted "];
		else
			value = [value stringByAppendingString:@"On "];
		// solo
		// trim
//		if ( [fObject isTrimmed] )
//			value = [value stringByAppendingString:@"Trimmed "];
		// lock
		if ( [fObject WSObject]->IsTheVideoSolo() || [fObject WSObject]->IsTheAudioSolo() )
			value = [value stringByAppendingString:@"Solo "];
			
		if ( [fObject isLocked] )
			value = [value stringByAppendingString:@"Locked "];
	}
	else
	{
		value = @"Not Imported ";
	}
	
	return value;
}

- (NSString*)type
{
	id value;
	if ( fObject )
	{
		switch( [fObject getType] )
		{
		default:
		case TArmeWSObj::kArmeWSObj:
			value = @"Object";
			break;
		case TArmeWSObj::kContainer:
			value = @"Group";
			break;
		case TArmeWSObj::kAudioProcessor:
			value = @"Audio Effect";
			break;
		case TArmeWSObj::kTLC:
			value = @"Top Level Container";
			break;
		case TArmeWSObj::kAudioClip:
			value = @"Audio Clip";
			break;
		case TArmeWSObj::kVideoClip:
			value = @"Video Clip";
			break;
		case TArmeWSObj::kVideoEffect:
			value = @"Video Effect";
			break;
		case TArmeWSObj::kVideoXFadeEffect:
			value = @"Transition";
			break;
		case TArmeWSObj::kTextClip:
			value = @"Text Clip";
			break;
		case TArmeWSObj::kMIDIClip:
			value = @"MIDI Clip";
			break;
		}
	}
	else
	{
		value = @"-";
	}
	
	return value;
}

- (void)refreshData
{
	NSEnumerator			*myChildrenEnum = [fChildren objectEnumerator];
	MILibraryElement		*elem;
	NSArray					*objChildren;
	MIObject				*anObj;
	NSEnumerator			*objEnum;
	
	if ( !fObject )
		return;

	objChildren = [fObject getChildren]; // children of the MIObject
	
	// check that children of the MILibraryElement are still up to date
	while ( elem = [myChildrenEnum nextObject] )
	{
		if ( elem->fObject )
		{
			// check that the MIObject's children contains my children and delete them if that is not the case
			if ( [objChildren containsObject:elem->fObject] )
				[elem refreshData];
			else
			{
				[(MILibraryElement*)[[[[fObject montageDoc] libraryRoot] children] objectAtIndex:1] addChild:elem];
				[fChildren removeObject:elem];
			}
		}
	}
	
	
	if ( [fChildren count] != [objChildren count] )
	{
		objEnum = [objChildren objectEnumerator];
		// check that the MIObject's children are also children of me
		while (anObj = [objEnum nextObject] )
		{
			if ( [anObj getType] == (TArmeWSObj::kVideoXFadeEffect) )
				continue; // don't put transitions in there
				
			myChildrenEnum = [fChildren objectEnumerator];
			BOOL found = NO;
			while ( elem = [myChildrenEnum nextObject] )
			{
				if ( elem->fObject == anObj )
				{
					found = YES;
					break;
				}
			}
			
			if ( !found )
			{
				elem = [MILibraryElement elementWithObject:anObj];
				[fChildren addObject:elem];
				[elem refreshData];
			}
		}
	}
	
}

- (int)compareTitle:(MILibraryElement*)other
{
	return [[self title] caseInsensitiveCompare:[other title]];
}

- (int)compareStartTime:(MILibraryElement*)other
{
	return [[self startTime] caseInsensitiveCompare:[other startTime]];
}

- (int)compareEndTime:(MILibraryElement*)other
{
	return [[self endTime] caseInsensitiveCompare:[other endTime]];
}

- (int)compareDuration:(MILibraryElement*)other
{
	return [[self duration] caseInsensitiveCompare:[other duration]];
}

- (int)compareStatus:(MILibraryElement*)other
{
	return [[self status] caseInsensitiveCompare:[other status]];
}

- (int)compareType:(MILibraryElement*)other
{
	return [[self type] caseInsensitiveCompare:[other type]];
}




- (void)sortChildrenWithCriteria:(id)criteria
{
	NSEnumerator	*objEnum = [fChildren objectEnumerator];
	id				child;
	SEL				func;
	
	if ( [fChildren count] <= 0 )
		return;
		
	while ( child = [objEnum nextObject] )
		[child sortChildrenWithCriteria:criteria];

	if ( [criteria isEqualToString:@"Title"] )
	{
		func = @selector(compareTitle:);
	}
	else if ( [criteria isEqualToString:@"Type"] )
	{
		func = @selector(compareType:);
	}
	else if ( [criteria isEqualToString:@"Duration"] )
	{
		func = @selector(compareDuration:);
	}
	else if ( [criteria isEqualToString:@"Start Time"] )
	{
		func = @selector(compareStartTime:);
	}
	else if ( [criteria isEqualToString:@"End Time"] )
	{
		func = @selector(compareEndTime:);
	}
	else if ( [criteria isEqualToString:@"Status"] )
	{
		func = @selector(compareStatus:);
	}
	[fChildren sortUsingSelector:func];
	
}

- (void)encodeWithCoder:(NSCoder *)encoder
{
	[encoder encodeObject:fPath];
	[encoder encodeBytes:&fObject length:sizeof(void*)];
//	[encoder encodeValueOfObjCType:@encode(void*) at:&fObject];
	[encoder encodeObject:fChildren];
}

- (id)initWithCoder:(NSCoder *)decoder
{
	unsigned int length;
	self = [super init];
	if ( self )
	{
		fPath = [[decoder decodeObject] retain];
		fObject = *(MIObject**)([decoder decodeBytesWithReturnedLength:&length]);
//		[decoder decodeValueOfObjCType:@encode(void*) at:&fObject];
		fChildren = [[decoder decodeObject] retain];
	}
		
	return self;
}

#endif //#if USE_LIBRARY

@end

