/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/
#import "MontageInterface.h"

#ifndef ArmeBasicTypes_h
#include "ArmeBasicTypes.h"
#endif //ArmeBasicTypes_h

#ifndef ArmeObjCArchivable_h
#include "ArmeObjCArchivable.h"
#endif //ArmeObjCArchivable_h

/*!
    @header MIObject.h

    defines the MIObject class.
*/

@class MIObjectRep;
@class MIMontageDocument;
@class MITopLevelContainerRep;
@class MIGuide;
@class MIObjectMemento;

//#import "TArmeWSObj.h"
class TArmeWSObj;
class TArmeWSObjMemento;
/*!
  @class MIObject
  
  @abstract     this class represents an object on the user interfance side
  
  @discussion
a MIObject can be can be a container, a clip, a processor, ...
there is a 1 on 1 correspondance between a MIObject and an ArmeWSObject

the MIObject holds data for its representations to query

*/

@interface MIObject : NSObject
{
@public
    /*! @var fRepresentations holds an array of every representation of this object in order to minimize seek
        time for refreshing the data. */
    NSMutableArray		*fRepresentations;	

    /*! @var fParentObject pointer to the parent object */
	MIObject			*fParentObject;

    /*! @var fOldParentObject pointer to the parent object before being  */
	MIObject			*fOldParentObject;

    /*! @var fFreezeSource pointer to the container that was the source of the freeze operation  */
	MIObject			*fFreezeSource;

    /*! @var fFrozenObject pointer to the object that was the result of the freeze operation of this object  */
	MIObject			*fFrozenObject;

	BOOL				fContentHasBeenModifiedSinceLastFreezeOperation;
	
	BOOL				fIsBeingDealloced;
	
    /*! @var fChildren the array of children this object has */
	NSMutableArray		*fChildren;

    /*! @var fIsOpaque (OLD) should always be yes, had a use before, could have one still */
	BOOL				fIsOpaque;
    
    /*! @var fWSObject pointer to the C++ object representing this MIObject */
	TArmeWSObj			*fWSObject;
    
    /*! @var fMontageDoc pointer to the (MIMontage)document in which this object sits */
	MIMontageDocument	*fMontageDoc;

    /*! @var fVerticalPos the vertical position of the object (parent relative) in VPosUnit*/
	MIVPos				fVerticalPos;  
	
    /*! @var fHeight the height of the object in VPosUnit */
	MIVPos				fHeight;	

    /*! @var fIsCollapsed whether the representation is collapsed or not */
    BOOL				fIsCollapsed;

    /*! @var fOldFrameHeight
		used for collapsing, this variable holds the height of the frame before it was collapsed */
    MIVPos				fOldHeight;

	/*! @var fBackgroundColor defines the object's background color */
	NSColor				*fBackgroundColor;
	
	ArmeTime			fLockedToTime;
	
	BOOL				fClipIsLocked;
}

// arch stuff
- (id)initFromArchive:(ArmeArchiveHandler*)unarchiver;
- (BOOL)writeToArchive:(ArmeArchiveHandler*)archiver;

/*! @function initWithArmeWSObj prefered initializer for the class, argument passed is the corresponding
    TArmeWSObj in the media engine */
- (id)initWithArmeWSObj:(TArmeWSObj*)armeObj;

/*! @function getCopy gets a copy of the object and paste it in the clipboard.  This functions also copies the TArmeWSObj referenced but not the MIObjectRep*/
- (MIObject*)copyInClipBoard:(MIObject*)parentInClipBoard;

/*! @function getSliceCopy gets a copy of a slice of the object and paste it in the clipboard.  This functions also copies the TArmeWSObj referenced but not the MIObjectRep*/
- (MIObject*)sliceCopyInClipBoard:(MIObject*)parentInClipBoard from:(ArmeTime)startTime to:(ArmeTime)endTime;

- (BOOL)deleteSelfAndChildren;


/*! @function putAllFamilyInFlatArrayIncludingSelf This function adds all the object's family 
              (including the object itself) to an array passed as parameter) */
-(void)putAllFamilyInFlatArrayIncludingSelf:(NSMutableArray*)whereToAdd;

/*! @function invalidateRep invalidates all the representations of the object and the ones of its children*/
- (void)invalidateRep;

	/*! @function isCollapsed returns YES if the object is collapsed */
- (BOOL)isCollapsed;

	/*! @function toggleCollapsed toggle the collapsed state of the object */
- (void)toggleCollapsed;

	/*! @function setCollapse sets the collapsed state of the object */
- (void)setCollapsed:(BOOL)flag;


- (ArmeTime)lockedToTime;

- (BOOL)isLocked;

- (void)setLocked:(BOOL)flag;

- (MIVPos)getTitleHeight;
- (MIVPos)getAutomationCurvesHeight;
- (MIVPos)getBodyHeight;
	/*! @function addRepresentation add a representation to the representation array.
    one representation per top level container */
- (void)addRepresentation:(id)newRepresentation;

/*! @function removeRepresentation remove a representation of the representation array. */

- (void)removeRepresentation:(id)representation;
/*! @function getRepresentations get this object's array of representations
    there is one representation per top level container.
    The object is not responsible for drawing itself, the representations are.
    see the drawRect method to see how an object is drawn.
    */
- (NSArray*)getRepresentations;

/*
	creates the representations of the object in all TLCs of the document
*/
- (void)createAllRepresentations;

/*! @function createRepresentationInView create a representation for this object in a given TLC view
    value returned is a pointer to the newly created representation */
- (MIObjectRep*)createRepresentationInView:(MITopLevelContainerRep*)parentView;

/*! @function createRepresentationInView create a representation for this object in a given MIObjectRep
    view and calls the function recursively on its children */
- (MIObjectRep*)createRepresentationsRecursively:(id)parentView;

- (void)addToSelections;
- (void)setSelectionsFromTime:(ArmeTime)from toTime:(ArmeTime)to;

- (void)moveChildToTopOfStack:(MIObject*)aChild;

/*! @function addChild add a child to this object */
- (void)addChild:(MIObject*)aChild;

/*! @function remove a child from this object
    this function only removes the child, it doesn't move it somewhere else, does not deallocate anything
    (although a release is sent to the child object)
    */
- (void)removeChild:(MIObject*)aChild;

/*! @function getChildren gets the array of this object's children */
- (NSArray*)getChildren;

- (NSArray*)getVideoChildrenAtTime:(ArmeTime)time;
- (NSArray*)getAudioChildrenAtTime:(ArmeTime)time;
- (NSArray*)getVideoChildren;
- (NSArray*)getAudioChildren;

- (MIObject*)findObjectRecursively:(TArmeWSObj*)wsObj;

- (void)setMuted:(BOOL)flag;
- (BOOL)isMuted;

/*! @function isOpaque whether this object is opaque or not */
- (BOOL)isOpaque;

/*! @function setIsOpaque set the object's opacity */
- (void)setIsOpaque:(BOOL)flag;

/*! @function isContainer tells whether this object is a container or not. */
- (BOOL)isContainer;

/*! @function WSObject return the C++ object corresponding to this object */
- (TArmeWSObj*)WSObject;

/*! @function setArmeWSObject function used by the initializer to set the C++ object corresponding
    to this object, should NEVER be used until you have a VERY GOOD reason */
- (void)setArmeWSObject:(TArmeWSObj*)WSObject;

/*! @function montageDoc returns the MIMontageDocument this object is in */
- (MIMontageDocument*)montageDoc;

/*! @function setMontageDoc this function specifies the document in which this object is.
    It DOESN'T actually put the object in the new document hierarchy though */
- (void)setMontageDoc:(MIMontageDocument*)montageDoc;

/*! @function backgroundColor returns the background color of this object */
- (NSColor*)backgroundColor;

/*! @function setBackgroundColor: sets the background color of this object */
- (void)setBackgroundColor:(NSColor*)color;

/*! @function getLength returns the ArmeTime length of the corresponding TArmeWSObject */
- (ArmeTime)getLength;

/*! @function getHorizontalPos return the ArmeTime start position of the corresponding TArmeWSObject */
- (ArmeTime)getHorizontalPos;

/*! @function getMaxHorizontalPos returns the end position of the object*/
- (ArmeTime)getMaxHorizontalPos;

/*! @function getMaxHorizontalPos returns the end position of the object*/
- (MIVPos)getMaxVerticalPos;

/*! @function setVerticalPos sets the vertical position of this object in VPosUnit (parent relative) */
- (void)setVerticalPos:(MIVPos)verticalPos;

/*! @function returns the vertical position of the object relative to its parent */
- (MIVPos)getVerticalPos;

/*! @function returns the absolute position of the object*/
- (MIVPos)getAbsoluteVerticalPos;

/*! @function setHeight sets the height of this object in VPosUnit */
- (void)setHeight:(MIVPos)height;

/*! @function getHeight returns the height of this object */
- (MIVPos)getHeight;

- (MIRect)modelFrame;

- (MIRect)absoluteModelFrame;

/*! @function returns this object's parent object */
- (MIObject*)parentObject;

- (MIPoint)convertPointFromTLCCoords:(MIPoint)p;
- (MIRect)convertRectToTLCCoords:(MIRect)rect;
- (MIPoint)convertPointToTLCCoords:(MIPoint)point;
- (MIRect)convertRectFromTLCCoords:(MIRect)rect;

/*! @function sets this object's parent object.
    it DOESN'T put this object in the new parent's children array, use the moveToObject method for that */
- (void)setParentObject:(MIObject*)parentObject;

/*! @function freezeSource returns the object as it was before beign frozen */
- (MIObject*)freezeSource;

/*! @function setFreezeSource sets this object's freezeSource.*/
- (void)setFreezeSource:(MIObject*)freezeSource;

- (MIObject*)frozenObject;
- (void)setFrozenObject:(MIObject*)frozenObject;
- (BOOL)isFrozenObject;
- (BOOL)hasParentFrozenObject;


- (void)markModified;
- (void)markParentModified;
- (void)cleanModifiedFlag;
- (void)_fillArrayWithModifiedParentsFlag:(ZzArray<bool>*)parentsFlag;
- (ZzArray<bool>*)getCurrentModifiedParentsFlag;
- (void)_resetModifiedFlag:(ZzArray<bool>*)parentsFlag indexInArray:(unsigned int)index;
- (void)resetModifiedFlag:(ZzArray<bool>*)parentsFlag;
- (BOOL)hasBeenModified;

/*! @function title get the name of the object*/
- (NSString*)title;

- (BOOL)isTLC;

/*! @function isDescendantOf Tells whether the receiver is a descendant of an object */
- (BOOL)isDescendantOf:(MIObject*)object;

/*! @function getType easy accessor to the type of the object */
- (SInt32)getType;

/*! @function recursivelyRecreateFrame
	recreates the object's frame as well as the object's children's, and the object's children's children's...
	*/
- (void)recursivelyRecreateFrame;

/*! @function recreateFrame
    recreates the frame of the representations according to the fHeight, fVerticalPos
    and the WSObject data
    this function does not ask the object to redraw itself though*/
- (void)recreateFrame;

/* @function minimalRectToFitContent returns the smallest rectangle in the current view that fits the content */
- (MIRect)minimalRectToFitContent;

/*! @function resizeFrameToFullyIncludeChild
    function to resize the object's frame as well as the origin of the children
    so that the object fully encloses the child.
    this function returns true if the child origin has changed. */
// DEPRECATED - (BOOL)resizeFrameToFullyIncludeChild:(MIObject*)theChild;

// DEPRECATED - (void)shrinkToFitContent;


/*! @function resizeFrameToFullyIncludeChildren
    function to resize the object's frame as well as the origin of the children
    so that the object fully encloses all of its children.
    this function returns true if the child origin has changed. */
// DEPRECATED - (BOOL)resizeFrameToFullyIncludeChildren;

/*! @function preserveChildrenPositionsWithNewFrame
	when the object's representation frame is changing, the global position on the screen shouldn't change
	this function makes sure that whatever the new frame size/origin is, their children don't visually
	move because of the different coordinate system.
	This function has to be called BEFORE this object's frame actually changes.
*/
// DEPRECATED - (void)preserveChildrenPositionsWithNewFrame:(NSRect)newFrame inTLC:(MITopLevelContainerRep*)theTLC;

/*! @function setModelFrameFromRect
    this function modifies the model (in Arme and here) from the newFrame argument
    this function also calls recreateFrame*/
- (void)setModelFrameFromRect:(NSRect)newFrame inTLC:(MITopLevelContainerRep*)theTLC;

- (void)setModelFrame:(MIRect)newFrame;

/*! @function addCursorRectsInTLC
	function to create cursor rectangles from the sizes of the representations.
							Note: recursive */
- (void)addCursorRectsInTLC:(MITopLevelContainerRep*)theTLC;

/*! @function drawRect
    draw function, ask the object's representation in theTLC to redraw itself
    with origin at "offset"
    this function effectively calls the representations drawRect method with
    the correct 'clipping rectangle'
    */
- (void)drawRect:(NSRect)aRect inTLC:(MITopLevelContainerRep*)theTLC withOffset:(NSPoint)offset;


/*! @function refreshHierarchy
    TODO make sure the hierarchy is correct and update it if not.
    */
- (void)refreshHierarchy;

/*! @function setRepresentationsNeedDisplay
    sets the TLC method needsDisplayInRect for all representations' frames */
- (void)setRepresentationsNeedDisplay;

/*! @function refreshCachedTitleStringImageInReps calls refreshCachedTitleStringImage for all reps */
- (void)refreshCachedTitleStringImageInReps;

/*! @function setTitlesNeedDisplay
    sets the TLC method needsDisplayInRect for all representations' title rects */
- (void)setTitlesNeedDisplay;

/*! @function invalidateRepresentationsCachedImages This function is called
when clip data has changed. Since the data is different, the image has to
be modified. $$$TODO Maybe only provide a time range that is invalidated  */
- (void)invalidateRepresentationsCachedImages;

/*! @function moveToObject
    this function moves this object into a new parent,
    modifies its frame, sets the parentObject variables,
    asks the object to recreate its frame and resize it to fully include children
    and asks the representations to redraw themselves
    */
- (void)moveToObject:(MIObject*)newParent atTime:(ArmeTime)newPos withVPos:(MIVPos)VPos;

/*! @function findRepresentationInWindow this function returns this object's representation in a given
    window or nil if this object has no representation in the specified window */
- (MIObjectRep*)findRepresentationInWindow:(NSWindow*)theWindow;
/*! @function findRepresentationInWindow this function returns this object's representation in a given
    top level container or nil if this object has no representation in the specified TLC */
- (MIObjectRep*)findRepresentationInTLC:(MITopLevelContainerRep*)theTLC;

/*! @function hitTest find this object's lowest descedant in the specified Top level container */ 
- (MIObjectRep*)hitTest:(NSPoint)aPoint inTLC:(MITopLevelContainerRep*)theTLC;

/*! @function description used by the debugger 'po' command
    returns a string describing the object and its contents
    */
- (NSString*)description;

/*! @function setOuttaHierarchy Visually remove the object from the doc (impl : removes the object 
	reps, removes the rep ptrs from the selections, stores the parent and tells the TArmeWSObj to
	SetOuttaHierarchy */
	
- (void)setOuttaHierarchy;
- (MIObjectMemento*)getMemento;
- (void)setMemento:(MIObjectMemento*)memento;
- (id)initWithMemento:(MIObjectMemento*)memento;

/*! @function putBackInHierarchy Visually replaces the object in the doc, at its old location (beware : its 
   original parent NEEDS TO BE THERE) (impl : creates the object reps, restores the parent, tells the TArmeWSObj to SetBackInHierarchy(oldParent) then clears the oldParent ptr*/
- (void)putBackInHierarchy;

-(BOOL)isInHierarchy;

@end

@interface MIObjectMemento : NSObject
{
@public
	TArmeWSObjMemento*			fWSObjectMemento;
	MIVPos						fVerticalPos;
	MIVPos						fHeight;
	MIVPos						fOldHeight;
	BOOL						fIsMuted;
	BOOL						fIsCollapsed; // from the representations
	MIMontageDocument*			fMontageDoc;
	MIObject*					fFreezeSource;
	MIObject*					fFrozenObject;
	BOOL						fContentHasBeenModifiedSinceLastFreezeOperation;
}

- (void)setVerticalPos:(MIVPos)verticalPos;
- (MIVPos)getVerticalPos;
- (void)setHeight:(MIVPos)height;
- (MIVPos)getHeight;
- (void)setOldHeight:(MIVPos)height;
- (MIVPos)getOldHeight;
- (BOOL)isMuted;
- (void)setMuted:(BOOL)flag;
- (BOOL)isCollapsed;
- (void)setIsCollapsed:(BOOL)flag;
- (void)setWSObjMemento:(TArmeWSObjMemento*)memento;
- (TArmeWSObjMemento*)getWSObjMemento;
- (MIMontageDocument*)montageDoc;
- (void)setMontageDoc:(MIMontageDocument*)doc;

// arch stuff
- (id)initFromArchive:(ArmeArchiveHandler*)unarchiver;
- (BOOL)writeToArchive:(ArmeArchiveHandler*)archiver;

@end
