/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/
 
#import "MIRuler.h"
#import "MITopLevelContainerRep.h"
#import "MIRulerTrackers.h"
#import "MITime.h"
#import "MIMainViewController.h"
#import "MIMontageDocument.h"
#import "MIImageController.h"
#import "MIApplicationController.h"
#import "MICreateMarkerTracker.h"
#import "MIMoveMarkerTracker.h"
#import "MIPreferences.h"
#import "MITimePaletteController.h"

static int	_tickHeights[4] = { 5, 10, 15, 20 };
int			kMinSpaceBetweenTicks = 15;
int			kMinSpaceBetweenCaption = 90;
const float	kPlayMarkerIndicatorPos = -1;
const float kLeftLocatorIndicatorPos = 2;
const float kRightLocatorIndicatorPos = 11;
const float kMarkerIndicatorLength = 10.0;
const float	kHorizontalLinePos = 18.5;

@implementation MIRuler

- (id)initWithFrame:(NSRect)frame
{
	self = [super initWithFrame:frame];
	
	fMouseTracker = nil;
	
	return self;
}

- (void)dealloc
{
    //NSLog(@"MIRuler - dealloc");
	[fCachedRulerImage release];
    [[NSNotificationCenter defaultCenter] removeObserver:self name:NSViewFrameDidChangeNotification object:fTopLevelContainerView];
    [[NSNotificationCenter defaultCenter] removeObserver:self name:NSViewBoundsDidChangeNotification object:[[fTopLevelContainerView enclosingScrollView] contentView]];
    [super dealloc];
}

- (void)setTopLevelContainerView:(MITopLevelContainerRep*)theNewView
{
	[[NSNotificationCenter defaultCenter] removeObserver:self name:NSViewFrameDidChangeNotification object:fTopLevelContainerView];
	[[NSNotificationCenter defaultCenter] removeObserver:self name:NSViewBoundsDidChangeNotification object:[[fTopLevelContainerView enclosingScrollView] contentView]];
	fTopLevelContainerView = theNewView;
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(TLCFrameChanged:)name:NSViewFrameDidChangeNotification object:theNewView];
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(TLCScrolled:)name:NSViewBoundsDidChangeNotification object:[[theNewView enclosingScrollView] contentView]];
		
	[self refreshMarksImage];
}

- (MITopLevelContainerRep*)getTopLevelContainerView
{
	return fTopLevelContainerView;
}

extern bool gDisableRulerDrawing;

- (void)drawRect:(NSRect)rect
{
	if(gDisableRulerDrawing)
		return;
	//MTGStartTimeLog
	if([[MIApplicationController sharedApplicationController] mainViewRedrawDisabled])
		return;
	// the frame of the ruler is kept in sync with that of the TLC, so if the visible rectangle of the TLC is starting at pixel 1100,
	// the bounds origin of the ruler (=visible part) is also 1100.
	/*[self drawBackgroundInRect:rect];
	[self drawMarksInRect:rect];*/
	[self drawRulerInRect:rect];
	[self drawMarkersInRect:rect];
	[self drawDottedLinesForSelection:rect];

	[fMouseTracker draw:rect];
	[fTopLevelContainerView->fMouseTracker alternateDraw:rect];
	//MTGEndTimeLog(@"MIRuler",@"drawRect:")
}

- (void)drawRulerInRect:(NSRect)rect
{
	//MTGStartTimeLog
	ArmeTime minTime = [fTopLevelContainerView pixelToArmeTime:NSMinX(rect)];
	ArmeTime maxTime = [fTopLevelContainerView pixelToArmeTime:NSMaxX(rect)];
	UInt32 hZoomIndex = [fMainViewController hZoomIndex];

	if ( ( fCachedRulerImage ) &&
	  ( minTime >= fMinTime ) &&
	  ( maxTime <= fMaxTime ) &&
	  ( hZoomIndex == fHZoomIndex ) )
	{
		[fCachedRulerImage compositeToPoint:rect.origin fromRect:NSMakeRect( NSMinX(rect) - fFirstPixel, NSMinY(rect), NSWidth(rect), NSHeight(rect) ) operation:NSCompositeCopy];
	}
	else
	{
		fMinTime = minTime;
		fMaxTime = maxTime;
		fHZoomIndex = hZoomIndex;
		[fCachedRulerImage release];

		[self drawBackgroundInRect:rect];
		[self drawMarksInRect:rect];

		NSBitmapImageRep *imageRep  = [[NSBitmapImageRep alloc] initWithFocusedViewRect:rect];
		fCachedRulerImage = [[NSImage alloc] initWithSize:rect.size];
		[fCachedRulerImage addRepresentation:imageRep];
		[imageRep release];
		fFirstPixel = NSMinX(rect);
	}
	//MTGEndTimeLog(@"MIRuler",@"drawRulerInRect:")
}

- (void)drawBackgroundInRect:(NSRect)rect
{
	//MTGStartTimeLog
	[[NSColor whiteColor] set];
	[NSBezierPath fillRect:rect];
	
	// draw the horizontal line
	[[NSColor blackColor] set];
	[NSBezierPath strokeLineFromPoint:NSMakePoint(NSMinX(rect), kHorizontalLinePos) toPoint:NSMakePoint(NSMaxX(rect), kHorizontalLinePos)];
	//MTGEndTimeLog(@"MIRuler",@"drawBackgroundInRect:")
}


- (void)drawMarksInRect:(NSRect)rect
{
	//MTGStartTimeLog
	MITime				*aTime;
	MITime				*aSubTime;
	MITime				*firstTime;
	ArmeTime			maxTime;
	int					tickHeight;
	float				lastCaptionPos = (-kMinSpaceBetweenCaption)-1;
	BOOL				bCaptionBetweenMinTime = NO;
	NSColor				*subTickColor = [NSColor magentaColor];
//	NSColor				*tickCaptionColor = [NSColor grayColor];
	NSColor				*tickColor = [NSColor blackColor];
	NSColor				*tickIndicatorColor = [NSColor blackColor];
	MIMontageDocument	*document = [fMainViewController document];
	BOOL				displaySMPTE = ([[MIPreferences sharedPrefs] timeDisplay] & kTimeDisplaySMPTE);
	int					largest = 0;
	float 				tickPos;
	
	
	aTime = 	[[MITime alloc] initWithDocument:document];
	aSubTime =	[[MITime alloc] initWithDocument:document];
	firstTime = [[MITime alloc] initWithDocument:document];
	
	if ( !aTime )
		return;
		
	maxTime = [fTopLevelContainerView pixelToArmeTime:NSMaxX(rect)];
	
	// determine first time unit.
	[firstTime setDoubleTime:[fTopLevelContainerView pixelToDoubleTime:NSMinX(rect)]];

//	[firstTime setDoubleTime:( (((int)([firstTime getDoubleTime]/[fMinTimeUnit getDoubleTime]))+1)*[fMinTimeUnit getDoubleTime])];
	[firstTime setDoubleTime:( (((int)([firstTime getDoubleTime]/[fMinTimeUnit getDoubleTime])))*[fMinTimeUnit getDoubleTime])];

	// look for "largest" tick mark in the first "kMinSpaceBetweenCaption" pixels
	[aTime setDoubleTime:[firstTime getDoubleTime]];
	do
	{
		// determine tick height

		tickHeight = 0;
		if ( displaySMPTE )
		{
			if ( ![aTime getVideoFrames] )
			{
				tickHeight++;
				if ( ![aTime getSeconds] )
				{
					tickHeight++;
					if ( ![aTime getMinutes] )
						tickHeight++;
				}
			}
		}
		else
		{
			if ( [aTime getPPQs] == 0)
			{
				tickHeight++;
				if ( [aTime get16ths] == 1)
				{
					tickHeight++;
					if ( [aTime getBeats] == 1 )
						tickHeight++;
				}
			}
		}
		
		tickPos = [fTopLevelContainerView doubleTimeToPixel:[aTime getDoubleTime]];
		if ( tickHeight > largest )
		{
			largest = tickHeight;
			// get the tick position in pixels
			lastCaptionPos = tickPos - (kMinSpaceBetweenCaption+1);
		}
		[aTime addTime:fMinTimeUnit];
		
	} while (tickPos < (NSMinX(rect) + kMinSpaceBetweenCaption) );
/*
	if ( displaySMPTE )
		[firstTime setH:[firstTime getHours] m:[firstTime getMinutes] s:[firstTime getSeconds]];
	else
		[firstTime setBars:[firstTime getBars] beats:[firstTime getBeats] nb16ths:[firstTime get16ths] ppq:0];
*/
	// draw the hash marks
	[aTime setDoubleTime:[firstTime getDoubleTime]];
	do
	{
		float tickPos;
		// determine tick height

		tickHeight = 0;
		if ( displaySMPTE )
		{
			if ( ![aTime getVideoFrames] )
			{
				tickHeight++;
				if ( ![aTime getSeconds] )
				{
					tickHeight++;
					if ( ![aTime getMinutes] )
						tickHeight++;
				}
			}
		}
		else
		{
			if ( ![aTime getPPQs] )
			{
				tickHeight++;
				if ( ![aTime get16ths] )
				{
					tickHeight++;
					if ( ![aTime getBeats] )
						tickHeight++;
				}
			}
		}

		// get the tick position in pixels
		tickPos = [fTopLevelContainerView doubleTimeToPixel:[aTime getDoubleTime]];

		if ( tickPos - lastCaptionPos > kMinSpaceBetweenCaption )
		{
			// there's a caption here
			NSString	*caption;
			NSRect		captionRect;

			lastCaptionPos = tickPos;
			
			if ( displaySMPTE )
				caption = [aTime getSMPTEString];
			else
				caption = [aTime getBPMString];
				
			captionRect.size = [caption sizeWithAttributes:nil];
			captionRect.origin = NSMakePoint(tickPos, kHorizontalLinePos+2);
			
			[tickIndicatorColor set];
			[NSBezierPath strokeLineFromPoint:NSMakePoint(tickPos, NSMaxY([self bounds])/*kHorizontalLinePos+15*/) toPoint:NSMakePoint(tickPos, kHorizontalLinePos-_tickHeights[tickHeight])];

			// draw the caption to the left
			[caption drawAtPoint:NSMakePoint(tickPos - captionRect.size.width , kHorizontalLinePos+2) withAttributes:nil];
		}
		else
		{		
			[tickColor set];
//			[NSBezierPath strokeLineFromPoint:NSMakePoint(tickPos, NSMaxY([self bounds])/*kHorizontalLinePos+15*/) toPoint:NSMakePoint(tickPos, kHorizontalLinePos-_tickHeights[tickHeight])];
			[NSBezierPath strokeLineFromPoint:NSMakePoint(tickPos, kHorizontalLinePos+4) toPoint:NSMakePoint(tickPos, kHorizontalLinePos-_tickHeights[tickHeight])];
		}
		[aTime addTime:fMinTimeUnit];
		
	} while ( [aTime getDoubleTime] <= GET_VALUE(maxTime) );
	
	// draw the submarsk (video frames) or PPQs
	
	if ( [fTopLevelContainerView doubleTimeToPixel:[fMinTimeUnit getDoubleTime]] > (kMinSpaceBetweenCaption * 2) )
		bCaptionBetweenMinTime = YES;
		
	if ( fMinSubTimeUnit )
	{
		float tickPos;
		[aTime setDoubleTime:[firstTime getDoubleTime]];
		do
		{
			float nextCaptionPos = 0;
			 
			[aSubTime setDoubleTime:[aTime getDoubleTime]];
			if ( bCaptionBetweenMinTime )
			{
				lastCaptionPos = [fTopLevelContainerView doubleTimeToPixel:[aTime getDoubleTime]];
				nextCaptionPos = lastCaptionPos + [fTopLevelContainerView armeTimeToPixel:_ArmeTime([aTime getAudioFrameRate])];
			}
			[aSubTime addTime:fMinSubTimeUnit];
			do
			{
				tickPos = [fTopLevelContainerView doubleTimeToPixel:[aSubTime getDoubleTime]];
				if ( tickPos - lastCaptionPos > kMinSpaceBetweenCaption &&
					 nextCaptionPos - tickPos > kMinSpaceBetweenCaption )
				{
					// there's a caption
					NSString	*caption;
					NSRect		captionRect;
		
					lastCaptionPos = tickPos;
					
					if ( displaySMPTE )
						caption = [aSubTime getSMPTEString];
					else
						caption = [aSubTime getBPMString];
						
					captionRect.size = [caption sizeWithAttributes:nil];
					captionRect.origin = NSMakePoint(tickPos, kHorizontalLinePos+2);

//					[NSBezierPath strokeRect:NSInsetRect(captionRect,-2,0)];
					[tickIndicatorColor set];
					[NSBezierPath strokeLineFromPoint:NSMakePoint(tickPos, NSMaxY([self bounds])/*kHorizontalLinePos+15*/) toPoint:NSMakePoint(tickPos, kHorizontalLinePos-_tickHeights[0])];
					[subTickColor set];
					
					// draw the caption to the left					
					[caption drawAtPoint:NSMakePoint(tickPos - captionRect.size.width , kHorizontalLinePos+2) withAttributes:nil];
				}
				else
				{
					// tick mark without caption
					[subTickColor set];
					[NSBezierPath strokeLineFromPoint:NSMakePoint(tickPos, kHorizontalLinePos+4) toPoint:NSMakePoint(tickPos, kHorizontalLinePos-_tickHeights[0])];
				}
				[aSubTime addTime:fMinSubTimeUnit];
			} while ( /*[aTime getSeconds] == [aSubTime getSeconds]*/ [aTime getDoubleTime] + [fMinTimeUnit getDoubleTime] > [aSubTime getDoubleTime] );
			
			//[aTime addH:0 m:0 s:1];	
			[aTime addDoubleTime:[fMinTimeUnit getDoubleTime]];
		} while ( [aTime getDoubleTime] <= GET_VALUE(maxTime) );
	}
	
	[aTime release];
	[aSubTime release];
	[firstTime release];
	//MTGEndTimeLog(@"MIRuler",@"drawMarksInRect:")
}

- (void)drawMarkersInRect:(NSRect)rect
{
	//MTGStartTimeLog
//	NSRect		markerRect = NSMakeRect(0,26,10,10);
	int			markerPos = 0;
	unsigned			i;
	NSImage		*img = nil;
	float		indicatorVPos = 0;
	MIMontageDocument *document = [fMainViewController document];
	
	[NSBezierPath setDefaultLineWidth:2.0f];

	for ( i = 0 ; i < 3 + [document getMarkerCount]; i++ )
	{
		switch( i )
		{
		  case 0:
			// left locator
			[[NSColor greenColor] set];
			markerPos = (int)[fTopLevelContainerView getLeftLocatorPosition];
			img = mainImageController->fRulerLeftLocator;
			indicatorVPos = kLeftLocatorIndicatorPos;
			break;
		  case 1:
			// right locator
			[[NSColor redColor] set];
			markerPos = (int)[fTopLevelContainerView getRightLocatorPosition];
			img = mainImageController->fRulerRightLocator;
			indicatorVPos = kRightLocatorIndicatorPos;
			break;
		  case 2: 
			// play marker
			[[NSColor blueColor] set];
			markerPos = (int)[fTopLevelContainerView getCurrentPlayBarPosition];
			img = mainImageController->fRulerPlayLocator;
			indicatorVPos = kPlayMarkerIndicatorPos;
			break;
		  default:
			  // plain markers
			  [[NSColor grayColor] set];
			  markerPos = (int)[fTopLevelContainerView armeTimeToPixel:[document getMarker:i-3]];
			  img = mainImageController->fRulerGenericLocator;
			  indicatorVPos = -1;
		}

		
		if ( markerPos >= NSMinX([self bounds]) && markerPos <= NSMaxX([self bounds]) )
		{
			[NSBezierPath strokeLineFromPoint:NSMakePoint(markerPos, NSMinY([self bounds])) toPoint:NSMakePoint(markerPos, NSMaxY([self bounds]))];
//			[img compositeToPoint:NSMakePoint(markerPos-([img size].width/2),kHorizontalLinePos-[img size].height) operation:NSCompositeSourceAtop];
			[img compositeToPoint:NSMakePoint(markerPos-([img size].width/2),NSMinY([self bounds])-2) operation:NSCompositeSourceAtop];
		}
		else if ( indicatorVPos != -1)
		{
			if ( markerPos < NSMinX([self bounds]) )
//				[NSBezierPath strokeLineFromPoint:NSMakePoint(NSMinX([self bounds]), indicatorVPos) toPoint:NSMakePoint(NSMinX([self bounds])+kMarkerIndicatorLength, indicatorVPos)];
				[NSBezierPath fillRect:NSMakeRect(NSMinX([self bounds]), indicatorVPos, kMarkerIndicatorLength, 5)];
			else
//				[NSBezierPath strokeLineFromPoint:NSMakePoint(NSMaxX([self bounds]), indicatorVPos) toPoint:NSMakePoint(NSMaxX([self bounds])-kMarkerIndicatorLength, indicatorVPos)];
				[NSBezierPath fillRect:NSMakeRect(NSMaxX([self bounds])- kMarkerIndicatorLength, indicatorVPos, kMarkerIndicatorLength, 5)];
			
		}
		
/*		markerRect.origin.x -= 5;
		[NSBezierPath fillRect:markerRect];*/
	}
	[NSBezierPath setDefaultLineWidth:0.0f];

	//MTGEndTimeLog(@"MIRuler",@"drawMarkersInRect:")
}

- (void)drawDottedLinesForSelection:(NSRect)rect
{
	NSImage	*dots = mainImageController->fRulerDottedLine;
	float p1;
	float p2;
	if ( [((MIWSSelection*)[fMainViewController selection]) type] == kMIWSSelectionTypeTimeSlice )
	{
		p1 = [fMainViewController armeTimeToPixel:[[fMainViewController selection] startTime]];
		p2 = [fMainViewController armeTimeToPixel:[[fMainViewController selection] endxTime]];
	}
	else if ( [((MIWSSelection*)[fMainViewController selection]) type] == kMIWSSelectionTypeObjects )
	{
		NSRect			allSelRect;
		//NSEnumerator	*selObjectsEnum;
		MIObjectRep		*anObjectRep;
		NSRect			anObjectFrame;
		//selObjectsEnum = [fMainViewController selectionObjectEnumerator];
	
		allSelRect = NSZeroRect;
	
		for(unsigned int i = 0; i < [[fMainViewController selection] count]; i++)
		{
			anObjectRep = [[fMainViewController selection] objectAtIndex:i];
			anObjectFrame = [anObjectRep getTLCFrame];
	
			if ( NSEqualRects( allSelRect, NSZeroRect) )
			{
				allSelRect = anObjectFrame;
			}
			else
			{
				allSelRect = NSUnionRect( allSelRect, anObjectFrame);
			}
		}
		
		p1 = NSMinX(allSelRect);
		p2 = NSMaxX(allSelRect);
	}
	p1 -= [mainImageController->fRulerDottedLine size].width/2;
	p2 -= [mainImageController->fRulerDottedLine size].width/2;
	[dots compositeToPoint:NSMakePoint(p1, 0) operation:NSCompositeSourceAtop];
	[dots compositeToPoint:NSMakePoint(p2, 0) operation:NSCompositeSourceAtop];
}

- (void)playbarNeedsDisplay
{
    int markerPos = (int)[fTopLevelContainerView getCurrentPlayBarPosition];
    NSImage *img = mainImageController->fRulerPlayLocator;
//    NSRect rect = NSMakeRect(markerPos-([img size].width/2),NSMinY([self bounds])-2,[img size].width, [img size].height);
    NSRect rect = NSMakeRect(markerPos-([img size].width/2),NSMinY([self bounds])-2,[img size].width, NSHeight([self bounds]));
    [self setNeedsDisplayInRect:rect];
}

-(BOOL)isOpaque
{
	// this function is used to tell the window manager that there's no need to draw anything behind.
	return YES;
}

- (void)refreshMarksImage
{
	int 			i;
	MITime			*aTime = nil;
	int				unitSteps[4] = { 1, 2, 5, 10};
	bool			minUnitFound = NO;
	MITime			*minTimeUnit = nil ;// = [MITime timeWithDocument:[fMainViewController document]];
	bool			minSubUnitFound = NO;
	MITime			*minSubTimeUnit = nil;//  = [MITime timeWithDocument:[fMainViewController document]];
	BOOL			displaySMPTE = ([[MIPreferences sharedPrefs] timeDisplay] & kTimeDisplaySMPTE);
	
	// check the minimum unit we can use.
	// take 5 pixels
	// transform to ArmeTime
	// feed that to an MITime
	// look how many seconds/minutes/hours are in that and take something above.
	
	[fMinSubTimeUnit release];
	fMinSubTimeUnit = nil;
	[fMinTimeUnit release];
	fMinTimeUnit = nil;
	
	aTime = 	[[[MITime alloc] initWithDocument:[fMainViewController document]] autorelease];
	
	if ( displaySMPTE )
	{
		// SMPTE
//		for ( i = 0; i < 4; i++ )
//		{
//			[aTime setH:0 m:0 s:0 vFr:unitSteps[i] ms:0];
			[aTime setH:0 m:0 s:0 vFr:1 ms:0];
			if ( [fTopLevelContainerView doubleTimeToPixel:[aTime getDoubleTime]] >= kMinSpaceBetweenTicks )
			{
				minSubTimeUnit = [[MITime alloc] initWithDocument:[fMainViewController document]];
				[minSubTimeUnit setDoubleTime:[aTime getDoubleTime]];
				minSubUnitFound = YES;
//				break;
			}
//		}
		
		for ( i = 0; i < 4; i++ )
		{
			[aTime setH:0 m:0 s:unitSteps[i]];
			if ( [fTopLevelContainerView doubleTimeToPixel:[aTime getDoubleTime]] >= kMinSpaceBetweenTicks )
			{
				minTimeUnit = [aTime retain];
				minUnitFound = YES;
				break;
			}		
		}
		
		if ( !minUnitFound )
		{
			for ( i = 0; i < 4; i++ )
			{
				[aTime setH:0 m:unitSteps[i] s:0];
				if ( [fTopLevelContainerView doubleTimeToPixel:[aTime getDoubleTime]] >= kMinSpaceBetweenTicks )
				{
					minTimeUnit = [aTime retain];
					minUnitFound = YES;
					break;
				}		
			}
		}
		
		if ( !minUnitFound )
		{
			for ( i = 0; i < 4; i++ )
			{
				[aTime setH:unitSteps[i] m:0 s:0];
				if ( [fTopLevelContainerView doubleTimeToPixel:[aTime getDoubleTime]] >= kMinSpaceBetweenTicks )
				{
					minTimeUnit = [aTime retain];
					minUnitFound = YES;
					break;
				}		
			}
		}
	}
	else
	{
		// BPM
		minTimeUnit = [MITime timeWithDocument:[fMainViewController document]];
		MITime *t = [MITime timeWithDocument:[fMainViewController document]];
		[t setBars:2 beats:1 nb16ths:1 ppq:0];
		
		if ( [fTopLevelContainerView doubleTimeToPixel:[t getDoubleTime]] >= 2*kMinSpaceBetweenTicks )
		{
			[minTimeUnit setDoubleTime:[t getDoubleTime]];
			for ( int i = 128 ; i >= 1; i /=2 )
			{
				if ( [fTopLevelContainerView doubleTimeToPixel:([t getDoubleTime]/i)] >= kMinSpaceBetweenTicks )
				{
					[minTimeUnit setDoubleTime:[t getDoubleTime]/i];
					minUnitFound = YES;
					break;
				}
			}
		}
		else
		{
			// look for a "nice" multiple of bars
			minUnitFound = NO;
			int mult = 1;
			
			while ( !minUnitFound )
			{
				for ( i = 0; i < 4; i++ )
				{
					if ( [fTopLevelContainerView doubleTimeToPixel:([t getDoubleTime]*unitSteps[i]*mult)] >= kMinSpaceBetweenTicks )
					{
						[minTimeUnit setDoubleTime:[t getDoubleTime]*unitSteps[i]*mult];
						minUnitFound = YES;
						break;
					}
				} 
				mult *= 10;
			}
		}
		
/*
		for ( i = 1; i < 5; i++ )
		{
			[aTime setBars:1 beats:1 nb16ths:i ppq:0];
			if ( [fTopLevelContainerView doubleTimeToPixel:[aTime getDoubleTime]] >= kMinSpaceBetweenTicks )
			{
				minSubTimeUnit = [aTime retain];
				minSubUnitFound = YES;
				break;
			}
		}
		if ( !minUnitFound )
		{
			for ( i = 1; i < 3; i++ )
			{
				[aTime setBars:1 beats:i nb16ths:1 ppq:0];
				if ( [fTopLevelContainerView doubleTimeToPixel:[aTime getDoubleTime]] >= kMinSpaceBetweenTicks )
				{
					minTimeUnit = [aTime retain];
					minUnitFound = YES;
					break;
				}
			}
		}
		if ( !minUnitFound )
		{
			for ( i = 0; i < 4; i++ )
			{
				[aTime setBars:unitSteps[i] beats:1 nb16ths:1 ppq:0];
				if ( [fTopLevelContainerView doubleTimeToPixel:[aTime getDoubleTime]] >= kMinSpaceBetweenTicks )
				{
					minTimeUnit = [aTime retain];
					minUnitFound = YES;
					break;
				}		
			}
		}
*/
	}
	
	if(!minUnitFound)
	{
		minTimeUnit = [MITime timeWithDocument:[fMainViewController document]];
		[minTimeUnit setDoubleTime:441.0]; //totally arbitrary. shouldn't happen
	}

	[fMinTimeUnit release];
	fMinTimeUnit = [[MITime timeWithDocument:[fMainViewController document]] retain];
	[fMinTimeUnit setDoubleTime:[minTimeUnit getDoubleTime]];
//	fMinTimeUnit = minTimeUnit;
	if ( minSubUnitFound )
		fMinSubTimeUnit = minSubTimeUnit;
	else
		fMinSubTimeUnit = nil;
}

-(BOOL)acceptsFirstMouse:(NSEvent *)theEvent
{
	// this function is used to tell that the view receives the first 'mousedown' event
	return YES;
}

- (void)mouseDown:(NSEvent*)theEvent;
{
	NSPoint			mouseLoc;
	unsigned int 	theModifier;
	float			markerPos;
	NSRect			myBounds;
	BOOL			bCreateTracker = NO;
	int				i;
	
	mouseLoc = [self convertPoint:[theEvent locationInWindow] fromView:nil];
	theModifier = [theEvent modifierFlags];
	myBounds = [self bounds];

	if ( !fMouseTracker )
	{
		// is mouse down in top part of ruler ?
		bCreateTracker = NO;

		if ( mouseLoc.y > kHorizontalLinePos ) bCreateTracker = YES;

		if ( bCreateTracker )
		{
//			fMouseTracker = [[MICreateMarkerTracker alloc] initWithView:self atPos:mouseLoc withModifiers:theModifier];
			fMouseTracker = [[MIMoveLocatorTracker alloc] initWithView:self atPos:mouseLoc withModifiers:theModifier];
			[fMouseTracker setLocatorType:MICustomMarker index:(UInt32)-1];
		}
		
	}
	
	// check for the 3 locators
	if ( !fMouseTracker )
	{
		for ( i = 0; i < 3; i++ )
		{
			bCreateTracker = NO;
			float indicatorVPos;
			
			switch( i )
			{
				case 0:
					markerPos = [fTopLevelContainerView getCurrentPlayBarPosition];
					indicatorVPos = -50;
					break;
				case 1:
					markerPos = [fTopLevelContainerView getLeftLocatorPosition];
					indicatorVPos = kLeftLocatorIndicatorPos;
					break;
				case 2:
					markerPos = [fTopLevelContainerView getRightLocatorPosition];
					indicatorVPos = kRightLocatorIndicatorPos;
					break;
			}
	
			if ( mouseLoc.x > markerPos - 6 && mouseLoc.x < markerPos + 6 ) // marker icon hit
			{
				bCreateTracker = YES;
			}
			else if ( i != 0 && mouseLoc.y >= indicatorVPos && mouseLoc.y <= indicatorVPos + 5 )
			{
				if ( markerPos <= NSMinX(myBounds) && mouseLoc.x <= NSMinX(myBounds) + kMarkerIndicatorLength + 2 )
				{
					bCreateTracker = YES;
				}
				else if ( markerPos >= NSMaxX(myBounds) && mouseLoc.x >= NSMaxX(myBounds) - (kMarkerIndicatorLength +2) )
				{
					bCreateTracker = YES;
				}
			}
	
			if ( bCreateTracker )
			{
				fMouseTracker = [[MIMoveLocatorTracker alloc] initWithView:self atPos:mouseLoc withModifiers:theModifier];
				switch ( i )
				{
					case 0:
						[fMouseTracker setLocatorType:MIPlayBarLocator index:0];
						break;
					case 1:
						[fMouseTracker setLocatorType:MILeftLocator index:0];
						break;
					case 2:
						[fMouseTracker setLocatorType:MIRightLocator index:0];
						break;
				}
				break; 
			}
		} // for
	} // if

	// check for generic locators
	if ( !fMouseTracker )
	{
		NSImage *img = mainImageController->fRulerGenericLocator;
		MIMontageDocument *document = [fMainViewController document];
		ArmeTime time = [fMainViewController pixelToArmeTime:mouseLoc.x];
		ArmeTime maxTime = [fMainViewController pixelToArmeTime:mouseLoc.x + [img size].width];
		ArmeTime markerTime;
		bCreateTracker = NO;

		UInt32 closestMarker = [document getClosestMarkerForTime:time];
		if ( closestMarker != (UInt32)-1 )
		{
			markerTime = [document getMarker:closestMarker];
			if ( [document ArmeTimeWidth:time withArmeTime:markerTime] < [document ArmeTimeWidth:time withArmeTime:maxTime] )
			{
				bCreateTracker = YES;
			}
		}

		if ( bCreateTracker )
		{
			if ( theModifier & NSAlternateKeyMask )
			{
//			fMouseTracker = [[MICreateMarkerTracker alloc] initWithView:self atPos:mouseLoc withModifiers:theModifier];
				fMouseTracker = [[MIMoveLocatorTracker alloc] initWithView:self atPos:mouseLoc withModifiers:theModifier];
				[fMouseTracker setLocatorType:MICustomMarker index:closestMarker];
			}
			else
			{
//			fMouseTracker = [[MIMoveMarkerTracker alloc] initWithView:self atPos:mouseLoc withModifiers:theModifier andMarker:i];
				fMouseTracker = [[MIMoveLocatorTracker alloc] initWithView:self atPos:mouseLoc withModifiers:theModifier];
				[fMouseTracker setLocatorType:MICustomMarker index:closestMarker];
			}
		}
	}
		
	// start the mouse tracker
	if(fMouseTracker)
		[fMouseTracker start:mouseLoc withModifiers:theModifier];

	if ( !fMouseTracker )
	{
		// If no tracker exit, assume we move the play bar		
		fMouseTracker = [[MIMoveLocatorTracker alloc] initWithView:self atPos:mouseLoc withModifiers:theModifier];
		[fMouseTracker setLocatorType:MIPlayBarLocator index:0];
		[fMouseTracker start:mouseLoc withModifiers:theModifier];
		[fMouseTracker mouseMoved:mouseLoc withModifiers:theModifier];
		[fMouseTracker stop:mouseLoc withModifiers:theModifier]; // this is needed so that Preroll is called
		[[MITimePaletteController sharedTimePaletteController] setFeedbackType:MITimePalettePlayLocatorFeedback];
	}
	[[[MIApplicationController sharedApplicationController] timePaletteController] getValuesFromMainView];
}

- (void)mouseDragged:(NSEvent *)theEvent
{
	NSPoint	mouseLoc;
	
	if ( [theEvent locationInWindow].x > NSMaxX([self frame]) || [theEvent locationInWindow].x < NSMinX([self frame]) )
	{
		//// **** HACK ****
		// the following hack is to be sure the scrollview won't scroll all the way up.
		// event is in window coordinates, not view
        /*float y;
		y = theEvent->_location.y;
		theEvent->_location.y = 100; this has been commented out on */
		[fTopLevelContainerView autoscroll:theEvent];

	}
	
	mouseLoc = [self convertPoint:[theEvent locationInWindow] fromView:nil];	
	
	[fMouseTracker mouseMoved:mouseLoc withModifiers:[theEvent modifierFlags]];
}

- (void)mouseUp:(NSEvent *)theEvent
{
	[fMouseTracker stop:[self convertPoint:[theEvent locationInWindow] fromView:nil] withModifiers:[theEvent modifierFlags]];
	[fMouseTracker release];
	fMouseTracker = nil;
	
	if ( [[fMainViewController selection] count] )
	{
		[[MITimePaletteController sharedTimePaletteController] setFeedbackType:MITimePaletteSelectionFeedback];
		[[MITimePaletteController sharedTimePaletteController] setFeedbackObject:[fMainViewController selection]];
	}
	else
		[[MITimePaletteController sharedTimePaletteController] setFeedbackType:MITimePaletteLocatorsFeedback];

	[[[MIApplicationController sharedApplicationController] timePaletteController] getValuesFromMainView];
	[self setNeedsDisplay:YES];

}

- (void)TLCFrameChanged:(NSNotification*)theNotification
{
/*	NSRect	newTLCFrame;
	NSRect	myFrame;
	
	newTLCFrame = [fTopLevelContainerView frame];
	myFrame = [self frame];
	if ( NSWidth(myFrame) == NSWidth(newTLCFrame) )
		return;
	
//	myFrame.origin.x = NSMinX(newTLCFrame);
	myFrame.origin.x = 0;
	myFrame.size.height = 40;
	myFrame.size.width = NSWidth(newTLCFrame);

	[self setFrame:myFrame];
	
	if ( fCachedMarksImage )
	{
		[fCachedMarksImage release];
	}
	
	fCachedMarksImage = [[NSImage alloc] initWithSize:myFrame.size];*/
	[self refreshMarksImage];
}

- (void)TLCScrolled:(NSNotification*)theNotification
{
	[self setBoundsOrigin:NSMakePoint(NSMinX([fTopLevelContainerView visibleRect]), 0)];
	[self setNeedsDisplay:YES];
}

- (NSImage*)createOrRefreshImage:(NSImage*)image ofSize:(NSSize)size
{
	NSImage	*theImage = nil;
	
	theImage = image;
	
	if ( theImage )
	{
		if ( !NSEqualSizes([theImage size], size) )
		{
			[theImage release];
			theImage = [[NSImage alloc] initWithSize: size];
		}
	}
	else
	{
		theImage = [[NSImage alloc] initWithSize:size];
	}

	return theImage;
}

- (MITime*)minTimeUnit
{
	return fMinTimeUnit;
}

- (MITime*)minSubTimeUnit
{
	return fMinSubTimeUnit;
}

- (NSRect)rulerTopRect {
	NSRect visibleRect = [self visibleRect];
	return NSMakeRect( NSMinX( visibleRect ), kHorizontalLinePos, NSWidth( visibleRect ), NSHeight ( visibleRect ) - kHorizontalLinePos ) ;
}

- (NSRect)rulerBottomRect {
	NSRect visibleRect = [self visibleRect];
	return NSMakeRect( NSMinX( visibleRect ), NSMinY( visibleRect ), NSWidth( visibleRect ), kHorizontalLinePos ) ;
}


@end
