/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import "MISequenceGrabberView.h"
#import "MontageInterface.h"

#ifndef TArmeMediaEngine_h
#include "TArmeMediaEngine.h"
#endif //TArmeMediaEngine_h

#ifndef MEAudioMonitoring_h
#include "MEAudioMonitoring.h"
#endif //MEAudioMonitoring_h

#ifndef ZzMath_h
#include "ZzMath.h"
#endif //ZzMath_h

#if !defined (__MWERKS__)
#import <ScreenSaver/ScreenSaverView.h>
#else //#if !defined (__MWERKS__)
static __inline__ NSRect SSCenteredRectInRect(NSRect innerRect, NSRect outerRect)
{
    innerRect.origin.x = floor((outerRect.size.width - innerRect.size.width) / 2);
    innerRect.origin.y = floor((outerRect.size.height - innerRect.size.height) / 2);
    return innerRect;
}
#endif //#if !defined (__MWERKS__)

#undef VER_DEF_TOPIC
#undef VER_DEF_TOPIC videoAlex

@implementation MISequenceGrabberView

- (id)initWithFrame:(NSRect)frame
{
    if (( self = [super initWithFrame:frame] ))
    {
        fDecompressor = 0;
		fPreviewing = NO;
		fVideoPreviewChannel = nil;
		fAudioPreviewChannel = nil;
		fLastAudioLevel = 0.;
		fPreviewBounds = NSRect2Rect([self bounds]);
		fDecompressFrames = YES;
		fBlueScreen = YES;
		fMonitoringObject = NULL;
    }
    return self;
}

- (void)awakeFromNib {
	fPreviewBounds = NSRect2Rect([self bounds]);
}

- (void)drawRect:(NSRect)rect
{
    if (fBlueScreen)
    {
		NSString* legend;
		[[NSColor blueColor] set];
        [NSBezierPath fillRect:[self bounds]];
		if (fConnected) legend = @"Camcorder connected";
		else legend = @"no input selected";
		NSRect bounds = NSZeroRect;
		bounds.size = [legend sizeWithAttributes:nil];
		bounds = SSCenteredRectInRect(bounds, [self bounds]);
		[legend drawAtPoint:bounds.origin withAttributes:nil];
    }
}

- (BOOL)previewing
{
	return fPreviewing;
}

- (void)setConnected:(BOOL)connected {
	fConnected = connected;
}

- (BOOL)isConnected {
	return fConnected;
}

- (void)setShowBlueScreen:(BOOL)flag {
	fBlueScreen = flag;
	if ( flag ) [self setNeedsDisplay:YES];
}

- (void)startPreviewing:(SeqGrabComponent)grabber video:(SGChannel)vchannel audio:(SGChannel)achannel
{
    ZzAssert(!fPreviewing);
	fVideoPreviewChannel = vchannel;
	fAudioPreviewChannel = achannel;
	fLastAudioLevel = 0.;

	fGrabber = grabber;
	[self lockFocus];
    ComponentResult result = ::SGSetGWorld(grabber, (CGrafPtr)[self qdPort], (GDHandle)NULL);
	[self unlockFocus];
	
	// now let's retrieve the sound record chunk size
	
	// first, the audio channel params. please note some returned field are not valid
	SoundDescriptionHandle soundDesc = (SoundDescriptionHandle)NewHandle(0);
	OSStatus err = ::SGGetChannelSampleDescription(achannel, (Handle)soundDesc);
	if(err != noErr)
	{
		const char* errText = ArmeGetQTError(err);
		std::cerr << "::SGGetChannelSampleDescription returned " << errText << " (" << err << ")" << std::endl;
	}

	// now let's assign the sound data paramters and pick reasonable defaults in case 
	// of problems. i'm doing this because i've read that some vdigs cannot report correct
	// data in ::SGGetChannelSampleDescription unless ...StartRecord has ALREADY been called.
	UInt32 audioFrameRate;
	UInt32 nbChannels;
		
	if(err != noErr)
	{
		audioFrameRate = 48000; // HARDCODED IN CASE OF CAMERA PROBLEMS
		nbChannels = 2;
	}
	else
	{
		audioFrameRate = (**soundDesc).sampleRate >> 16;
		nbChannels = (**soundDesc).numChannels;
	}
		
	// now let's retrieve the number of seconds in a packet. please
	// note this is weird since if thenumber is negative, then it means
	// it's a fixed point fraction and otherwise it's a plain integer number
	// of seconds.
	long soundChunkSize;
	double chunkNbSecs;
	soundChunkSize = ::SGGetSoundRecordChunkSize(achannel);      
	if(soundChunkSize >= 0)
	{
		chunkNbSecs = (double)soundChunkSize;
	}
	else
	{
		chunkNbSecs = (-soundChunkSize)/65536.0;
	}
		
	UInt32 bufferNbFrames = (UInt32)(chunkNbSecs * (double)audioFrameRate);
		
	// we store 8 full buffers in the ring buffer
	UInt32 ringBufferNbFramesNbBits = ZzMathLog2Cap(8*bufferNbFrames);
//	UInt32 ringBufferNbFramesNbBits = 16; // 64K frames.... let's try and see...
	fMonitoringObject = TArmeMediaEngine::CreateAudioMonitor(ringBufferNbFramesNbBits,nbChannels); // "2" means stereo.

	result = ::SGStartRecord(grabber);
	
	fPreviewing = YES;
}

- (void)stopPreviewing
{
    if (fDecompressor)
        ::CDSequenceEnd(fDecompressor);
    fDecompressor = 0;
	fPreviewing = NO;
	fVideoPreviewChannel = nil;
	fAudioPreviewChannel = nil;
	TArmeMediaEngine::RemoveAudioMonitor(fMonitoringObject);
	fMonitoringObject = NULL;
    [self setNeedsDisplay:TRUE];
}

- (float)getMeterValue:(id)sender
{
	return fLastAudioLevel;
}

- (OSErr)receiveSGData:(SGChannel)channel data:(Ptr)data length:(long)len
{
    ComponentResult err = noErr;
	if (!data or !len)
		return err;
	if (fDecompressFrames)
	{
		if (channel == fAudioPreviewChannel)
		{
//			SoundDescriptionHandle soundDesc = (SoundDescriptionHandle)NewHandle(0);
//			err = ::SGGetChannelSampleDescription(channel, (Handle)soundDesc);

			// data format is set in MIRecordPreviewController -openChannels
   // 44.1 x 16 x 2 uncompressed at time of writing
			if( (len % 8) !=  0)
				std::cerr << "Sequence Grabber capture audio length is not % 8 == 0" << std::endl;
			
			float* soundSamples = (float*)data;
			float channelAverage = (soundSamples[0] + soundSamples[1]) / 2.;
			fLastAudioLevel = channelAverage / (float)SHRT_MAX;
			
			fMonitoringObject->WriteFrames(soundSamples,(len / sizeof(float))/2);

			

			// in the mixer window now apparently?
   //[fSGVUMeter setNeedsDisplay:TRUE];
		}
		else if (channel == fVideoPreviewChannel)
		{
			if (!fDecompressor)
			{
				// Set up getting grabbed data into the GWorld
				ImageDescriptionHandle imageDesc = (ImageDescriptionHandle)NewHandle(0);
				err = ::SGGetChannelSampleDescription(channel,(Handle)imageDesc);

				MatrixRecord matrix;
				Rect sourceRect;

				SetRect( &sourceRect, 0, 0, (**imageDesc).width, (**imageDesc).height );
				RectMatrix( &matrix, &sourceRect, &fPreviewBounds );

				[self lockFocus];
				err = ::DecompressSequenceBeginS(&fDecompressor, imageDesc, data, len, (CGrafPtr)[self qdPort], NULL,
									 &sourceRect, &matrix, srcCopy, NULL, 0, codecNormalQuality, bestSpeedCodec);
				::DisposeHandle((Handle)imageDesc);
				[self unlockFocus];
			}

			[self lockFocusIfCanDraw];
			// decompress a frame into the GWorld
			CodecFlags ignore;
			err = ::DecompressSequenceFrameS(fDecompressor, data, len, 0, &ignore,  NULL);

			[[NSGraphicsContext currentContext] flushGraphics];
			[self unlockFocus];
		}
	}

    return err;
}

- (BOOL)isOpaque {
	return YES;
}

- (BOOL)shouldDecompressFrames
{
	return fDecompressFrames;
}

- (void)setShouldDecompressFrames:(BOOL)flag
{
	fDecompressFrames = flag;
	if ( !fDecompressFrames ) {
		if (fDecompressor)
			::CDSequenceEnd(fDecompressor);
		fDecompressor = 0;
	} else {
		[self lockFocus];
		ComponentResult result = SGSetGWorld(fGrabber, (CGrafPtr)[self qdPort], NULL);
		[self unlockFocus];

		result = ::SGStartRecord(fGrabber);
		fPreviewing = YES;
	}
}

- (void)dealloc
{
	if(fMonitoringObject)
		delete fMonitoringObject;
	fMonitoringObject = NULL;
	[super dealloc];
}

@end

