/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import "MIViewerController.h"
#import "MenuTags.h"
#import "MIDocumentController.h"
#import "MIMontageDocument.h"
#import "MIPickScreenController.h"
#import "MIApplication.h"

#ifndef TArmeMasterMovie_h
#include "TArmeMasterMovie.h"
#endif //TArmeMasterMovie_h

#ifndef TArmeDoc_h
#include "TArmeDoc.h"
#endif //TArmeDoc_h

@implementation MIViewerController

+ (MIViewerController*)sharedViewerController {
    static MIViewerController *sharedViewerController = nil;
    if ( sharedViewerController == nil ) {
        sharedViewerController = [[self alloc] init];
    }
    return sharedViewerController;
}

+ (BOOL)factoryVisibleDefault {
    return YES;
}

+ (NSString*)visibleDefaultName {
    return @"ViewerVisible";
}

- (NSString*)nibName {
#ifndef USE_NSQUICKDRAWVIEW
    return @"Viewer";
#else	
    return @"ViewerQD";
#endif
}

- (id)init {
    if (( self = [super init] )) {
		fShouldIgnoreNotifications = NO;
		fMagnification = 1;
		// setup menus
		fHalfSizeMenuItem = [NSApp searchForMenuItemWithTag:MTGHalfSizeVideoMenuTag];
		fNormalSizeMenuItem = [NSApp searchForMenuItemWithTag:MTGNormalSizeVideoMenuTag];
		fDoubleSizeMenuItem = [NSApp searchForMenuItemWithTag:MTGDoubleSizeVideoMenuTag];
		fFullScreenMenuItem = [NSApp searchForMenuItemWithTag:MTGFullScreenVideoMenuTag];
		fHideMenuItem = [NSApp searchForMenuItemWithTag:MTGCloseVideoMenuTag];
		[fHalfSizeMenuItem setTarget:self];
		[fHalfSizeMenuItem setAction:@selector(viewBackstageHalfx:)];
		[fNormalSizeMenuItem setTarget:self];
		[fNormalSizeMenuItem setAction:@selector(viewBackstage1x:)];
		[fDoubleSizeMenuItem setTarget:self];
		[fDoubleSizeMenuItem setAction:@selector(viewBackstage2x:)];
		[fFullScreenMenuItem setTarget:self];
		[fFullScreenMenuItem setAction:@selector(viewFullScreen:)];
		[fHideMenuItem setTarget:self];

		// setup full screen window
		fFullScreenVideoWindow = [[MIFullScreenVideoWindow alloc] initWithContentRect:NSMakeRect(0, 0, 640, 480)
			styleMask:NSBorderlessWindowMask
			backing:NSBackingStoreBuffered
			defer:NO];
		
		[fFullScreenVideoWindow setBackgroundColor:[NSColor blackColor]];		

		// register for notifications
        NSDocumentController* documentController = [NSDocumentController sharedDocumentController];
		
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(currentDocumentChanged:) name:MTGCurrentDocumentChangedNotification object:documentController];
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(documentAdded:) name:MTGDocumentAddedNotification object:documentController];
        [[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(documentRemoved:) name:MTGDocumentRemovedNotification object:documentController];
    }
    return self;
}

- (void)dealloc {
	vernl(6,"[MIViewerController dealloc]");
	NSDocumentController* documentController = [NSDocumentController sharedDocumentController];
	
    [[NSNotificationCenter defaultCenter] removeObserver:self name:MTGCurrentDocumentChangedNotification object:documentController];
	[[NSNotificationCenter defaultCenter] removeObserver:self name:MTGDocumentAddedNotification object:documentController];
	[[NSNotificationCenter defaultCenter] removeObserver:self name:MTGDocumentRemovedNotification object:documentController];

	if ( [self isWindowLoaded] ) [[NSNotificationCenter defaultCenter] removeObserver:self name:NSWindowDidMoveNotification object:[self window]];
	if ( [self isWindowLoaded] ) [[NSNotificationCenter defaultCenter] removeObserver:self name:NSWindowWillCloseNotification object:[self window]];
	
    [fFullScreenVideoWindow release];
	[fFullScreenMovieView release];
	[self setMovie:nil];
#ifdef USE_NSQUICKDRAWVIEW
	[fQDView release];
#else
	//[fMovieView release];
#endif
    [super dealloc];
}

- (void)windowWasLoaded {
#ifdef USE_NSQUICKDRAWVIEW
	[fQDView retain];
	[[self window] setTitle:@"Viewer (NSQuickDrawView)"];
#else
	//[fMovieView retain];
#endif
	
	[[self window] setLevel:NSFloatingWindowLevel];

	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(windowDidMove:) name:NSWindowDidMoveNotification object:[self window]];
	[[NSNotificationCenter defaultCenter] addObserver:self selector:@selector(windowWillClose:) name:NSWindowWillCloseNotification object:[self window]];
	
}

- (void)setActiveDocument:(MIMontageDocument*)activeDocument {
	vernl(6,"[MIViewerController setActiveDocument]");
	if ( [self isWindowLoaded] ) {
		if (activeDocument) {
			TArmeDoc* armeDoc = [activeDocument armeDocument];
			ZzAssert(armeDoc);
			TArmeMasterMovie* masterMovie = armeDoc->fMasterMovie;
			ZzAssert(masterMovie);
			[self setMovie:masterMovie->GetQTMovie()];
		} else [self setMovie:nil];
	}
}

- (void)currentDocumentChanged:(NSNotification*)notification {
	vernl(6,"[MIViewerController currentDocumentChanged]");
	BOOL wasLoaded = NO;
	//BOOL visibleDefault = [[NSUserDefaults standardUserDefaults] boolForKey:[[self class] visibleDefaultName]];
	if ( ( ![self isWindowLoaded] ) ) {
		[self window]; // load window
		wasLoaded = YES;
	}
	
	if ( [self isWindowLoaded] ) {
		if ( !fShouldIgnoreNotifications ) {
			NSDictionary* notificationDictionary = [notification userInfo];
			MIMontageDocument* activeDoc = nil;
			if ( notificationDictionary ) {
				activeDoc = [notificationDictionary objectForKey:MTGCurrentDocumentKey];
			}
			
			[self setActiveDocument:activeDoc];

			NSPoint point;
			id value = [activeDoc documentSettingForKey:@"viewerTopLeftLocation"];
			if ( value ) {
				point = NSPointFromString( value );
			} else {
				NSRect windowFrame;
				NSString *windowFrameString = [[NSUserDefaults standardUserDefaults] objectForKey:[NSString stringWithFormat:@"NSWindow Frame %@",[self nibName]]];
				if ( windowFrameString ) windowFrame = NSRectFromString(windowFrameString);
				else windowFrame = [[[MIViewerController sharedViewerController] window] frame];

				point = NSMakePoint( NSMinX( windowFrame ), NSMaxY( windowFrame ));
			}

			[self setWindowFrameTopLeftCorner:point];

			float magnification;
			id number = [activeDoc documentSettingForKey:@"viewerMagnification"];
			if ( number )
				magnification = [number floatValue];
			else {
				TArmeDoc* armeDoc = [activeDoc armeDocument];
				WorkingCodecType codec = armeDoc->fMasterMovie->GetWorkingCodecType();

				if ( ( codec == kDvPal720_576) || ( codec == kDvNtsc720_480) || ( codec == kDvPal16_9) || ( codec == kDvNtsc16_9) ) magnification = 0.5;
				else magnification = 1;
			}

			[self setMagnification:magnification forDocument:activeDoc];
			/*if ( [self isVisible] ) [self show:nil];
			else if ( ( wasLoaded ) && ( visibleDefault ) ) [self show:nil];*/
			[super show:nil];
		}
	}
}

- (void)documentAdded:(NSNotification*)notification {
	/*vernl(6,"[MIViewerController documentAdded]");

    NSDictionary* notificationDictionary = [notification userInfo];

    if ( notificationDictionary ) {
		MIMontageDocument* addedDoc = nil;

        addedDoc = [notificationDictionary objectForKey:MTGAddedDocumentKey];

		[self setMagnification:addedDoc->fModel->fViewerMagnification forDocument:addedDoc];
		[self show:nil];
    }*/
}

bool g_MIViewerController_documentRemoved_bodystarted = false;
bool g_MIViewerController_documentRemoved_bodyfinished = false;

- (void)documentRemoved:(NSNotification*)notification 
{
	g_MIViewerController_documentRemoved_bodystarted = true;
	vernl(6,"[MIViewerController documentRemoved] event time stamp = %f" << [[NSApp currentEvent] timestamp]);
	[self setMovie:nil]; 
	g_MIViewerController_documentRemoved_bodyfinished = true;
}


- (void)windowDidMove:(NSNotification*)notification {
	NSRect windowFrame = [[self window] frame];
	NSPoint point = NSMakePoint( NSMinX( windowFrame ), NSMaxY( windowFrame ));
	[[[NSDocumentController sharedDocumentController] currentDocument] setDocumentSetting:NSStringFromPoint(point) forKey:@"viewerTopLeftLocation"];
}

- (IBAction)show:(id)sender {
    if ( [self isFullScreen] ) {
        [self setFullScreen:NO];
    }
    [super show:sender];
	MIMontageDocument *currentDocument = [[MIDocumentController sharedDocumentController] currentDocument];
	[self setActiveDocument:currentDocument];
	if ([currentDocument armeDocument])
		[currentDocument armeDocument]->fViewerIsVisible = true;
}

- (void)windowWillClose:(NSNotification*)notification {
	[self setMovie:nil];
}

- (void)setWindowFrameTopLeftCorner:(NSPoint)point {
	NSWindow *window = [self window];
	NSRect frame = [window frame];
	frame.origin.x = point.x;
	frame.origin.y = point.y - NSHeight( frame );
	[window setFrame:frame display:NO];
}

- (float)magnification {
	return fMagnification;
}

- (void)setMagnification:(float)magnification {
	[self setMagnification:magnification forDocument:[[NSDocumentController sharedDocumentController] currentDocument]];
}

- (void)setMagnification:(float)magnification forDocument:(MIMontageDocument*)document {
	if (document) {
		TArmeDoc* armeDoc = [document armeDocument];
		ZzAssert(armeDoc);
		TArmeMasterMovie* masterMovie = armeDoc->fMasterMovie;
		ZzAssert(masterMovie);

		NSWindow* window = [self window];

		NSScreen *screen = [window screen];
		NSRect screenFrame = [screen visibleFrame];
		
		fMagnification = magnification;

		[document setDocumentSetting:[NSNumber numberWithFloat:fMagnification] forKey:@"viewerMagnification"];

		NSRect windowFrame = [window frame];
		NSRect contentRect = [NSPanel contentRectForFrameRect:windowFrame styleMask:[window styleMask]];

		float top = NSMaxY( windowFrame ) ;

		contentRect.size.width = masterMovie->GetViewerWidth() * fMagnification;
		contentRect.size.height = masterMovie->GetViewerHeight() * fMagnification;

		NSRect newWindowframe = [NSPanel frameRectForContentRect:contentRect styleMask:[window styleMask]];

		newWindowframe.origin.y = top - NSHeight( newWindowframe ) ;

		if ( NSMinX( newWindowframe ) < NSMinX( screenFrame ) )
			newWindowframe.origin.x = NSMinX( screenFrame );

		if ( NSMaxX( newWindowframe ) > NSMaxX( screenFrame ) )
			newWindowframe.origin.x = NSMaxX( screenFrame ) - NSWidth( newWindowframe ) ;

		if ( NSMinY( newWindowframe ) < NSMinY( screenFrame ) )
			newWindowframe.origin.y = NSMinY( screenFrame ) ;
		
		if ( NSMaxY( newWindowframe ) > NSMaxY( screenFrame ) )
			newWindowframe.origin.y = NSMaxY( screenFrame ) - NSHeight( newWindowframe );

		

		if ( NSWidth(newWindowframe) < 180 ) {
			if ( [window styleMask] & NSClosableWindowMask ) {
				BOOL isVisible = [window isVisible];
				if ( isVisible ) [window orderOut:nil];
				[window setFrame:newWindowframe display:NO];
				NSPanel *panel = [[NSPanel alloc] initWithContentRect:[NSWindow contentRectForFrameRect:newWindowframe styleMask:NSTitledWindowMask] styleMask:NSTitledWindowMask backing:NSBackingStoreBuffered defer:YES];
				id contentView = [[window contentView] retain];
				[window setContentView:nil];
				[panel setContentView:contentView];
				[panel setTitle:[window title]];
				if ( isVisible ) {
					[panel orderFront:nil];
				}
				[contentView release];
				[self setWindow:panel];
				window = panel;
			} else [window setFrame:newWindowframe display:[window isVisible]];
		} else {
			if ( !( [window styleMask] & NSClosableWindowMask ) ) {
				BOOL isVisible = [window isVisible];
				if ( isVisible ) [window orderOut:nil];
				[window setFrame:newWindowframe display:NO];
				NSPanel *panel = [[NSPanel alloc] initWithContentRect:[NSWindow contentRectForFrameRect:newWindowframe styleMask:NSTitledWindowMask | NSClosableWindowMask] styleMask:NSTitledWindowMask | NSClosableWindowMask backing:NSBackingStoreBuffered defer:YES];
				id contentView = [[window contentView] retain];
				[window setContentView:nil];
				[panel setContentView:contentView];
				[panel setTitle:[window title]];
				if ( isVisible ) {
					[panel orderFront:nil];
				}
				[contentView release];
				[self setWindow:panel];
				window = panel;
			} else [window setFrame:newWindowframe display:[window isVisible]];
		}
		
#ifdef USE_NSQUICKDRAWVIEW
		[[self window] setTitle:@"Viewer (NSQuickDrawView)"];
		if ( fQTMovie ) {
			[fQDView lockFocus];
			Rect movieBox;
			GetPortBounds((CGrafPtr)[fQDView qdPort], &movieBox);
			NSLog( @"left = %d, top = %d, right = %d, bottom = %d", movieBox.left, movieBox.top, movieBox.right, movieBox.bottom );
			SetMovieBox ( fQTMovie, &movieBox);
			[fQDView unlockFocus];
		}
#endif
		
	}
}

- (void)setFullScreen:(BOOL)fullScreen {
	if ( fullScreen ) {
		NSScreen* screen = [[MIPickScreenController sharedPickScreenController] pickScreen];

        if (screen) {
			[fFullScreenVideoWindow setFrame:[screen frame] display:YES];
			[fFullScreenVideoWindow setLevel:NSScreenSaverWindowLevel];
			[fFullScreenVideoWindow makeKeyAndOrderFront:nil];

#ifdef USE_NSQUICKDRAWVIEW
			Movie QTMovie = fQTMovie;
#else
			Movie QTMovie = (Movie)([[fMovieView movie] QTMovie]);
#endif
			
			fCurrentMagnification = fMagnification;
			[self setMagnification:1];

#ifdef USE_NSQUICKDRAWVIEW
			[self setMovie:nil];
#else
			[fMovieView setMovie:nil];
#endif
			
			Rect movieBounds;
			::GetMovieBox(QTMovie, &movieBounds);
			
			NSRect movieFrame = ScaleNSRectToNSRect(NSMakeRect(movieBounds.left, movieBounds.top, movieBounds.right-movieBounds.left, movieBounds.bottom-movieBounds.top), [screen frame], true);
			float x = ( NSWidth( [screen frame] ) - NSWidth( movieFrame ) ) / 2.0;
			float y = ( NSHeight( [screen frame] ) - NSHeight( movieFrame ) ) / 2.0;
			movieFrame = NSMakeRect( x, y, NSWidth( movieFrame ), NSHeight( movieFrame ) ) ;

			fFullScreenMovieView = [[NSMovieView alloc] initWithFrame:movieFrame];
			[fFullScreenMovieView showController:NO adjustingSize:YES];
			
			[fFullScreenMovieView setMovie:fMovie];
			[[fFullScreenVideoWindow contentView] addSubview:fFullScreenMovieView];
			[[fFullScreenVideoWindow contentView] setNeedsDisplay:YES];
			// Georges wants cursor hidden until mouse moved
			[NSCursor setHiddenUntilMouseMoves:YES];
			// and menu bar to show when moved over
			if (screen == [NSScreen menubarScreen])
				[fFullScreenVideoWindow showMenuBarUnderMouse:YES];
		}
	} else {
		[self setMagnification:fCurrentMagnification];
		[fFullScreenVideoWindow showMenuBarUnderMouse:NO];
		[fFullScreenVideoWindow orderOut:nil];
#ifdef USE_NSQUICKDRAWVIEW
		[self setMovie:(Movie)[[fFullScreenMovieView movie] QTMovie]];
#else
		[fMovieView setMovie:[fFullScreenMovieView movie]];
#endif
		[fFullScreenMovieView setMovie:nil];
		[fFullScreenMovieView removeFromSuperview];
		[fFullScreenMovieView release];
		fFullScreenMovieView = nil;
	}
}

- (BOOL)isFullScreen {
	return [fFullScreenVideoWindow isVisible];
}

- (IBAction)viewBackstageHalfx:(id)sender {
	if ( [self isFullScreen] ) {
        [self setFullScreen:NO];
    }	
    [self setMagnification:0.5];
	[self show:sender];
}

- (IBAction)viewBackstage1x:(id)sender {
	if ( [self isFullScreen] ) {
        [self setFullScreen:NO];
    }	
    [self setMagnification:1];
	[self show:sender];
}

- (IBAction)viewBackstage2x:(id)sender {
	if ( [self isFullScreen] ) {
        [self setFullScreen:NO];
    }	
    [self setMagnification:2];
	[self show:sender];
}

- (IBAction)viewFullScreen:(id)sender
{
	[self setFullScreen:YES];
}


- (void)setMovie:(Movie)movie 
{
	// case 1 : we want to set the viewer "movie-less"
	if(movie == nil)
	{
#ifndef USE_NSQUICKDRAWVIEW
		if ( fMovieView ) {
			[fMovieView setMovie:nil];
			[fMovieView removeFromSuperviewWithoutNeedingDisplay];
			fMovieView = nil;
		}
		// maybe there is no movie assigned yet (for example, if the doc window has never got the
		// focus !
		if(fMovie)
		{
			[fMovie release];
			fMovie = nil;
		}
#else
		SetMovieGWorld( movie, nil, nil ); // this sets the movie to use the current port. What is the current port in a Cocoa App ? I don't know.
		fQTMovie = nil;
#endif
	}
	else // case 2 : we want to set a real movie
	{
/*#ifndef USE_NSQUICKDRAWVIEW			
		if( (fMovie != nil) && ([fMovie QTMovie] == movie) )
#else
		if (fQTMovie == movie)
#endif
		{
			// it's the same old movie, just don't do anything
		}
		else // otherwise it means it's a really new movie
		{*/
#ifndef USE_NSQUICKDRAWVIEW
		if ( fMovieView ) {
			[fMovieView setMovie:nil];
			[fMovieView removeFromSuperviewWithoutNeedingDisplay];
			fMovieView = nil;
		}
		
		if ( fMovie ) {
			[fMovie release];
			fMovie = nil;
		}

		fMovieView = [[NSMovieView alloc] initWithFrame:[fMovieViewContainer frame]];
		[fMovieView showController:NO adjustingSize:NO];
		[fMovieView setEditable:NO];
		[fMovieView setAutoresizingMask:NSViewWidthSizable | NSViewHeightSizable];
		[fMovieViewContainer addSubview:fMovieView];
		[fMovieView release];
		
		TimeValue timeNow;
		Fixed playRate;
		fMovie = [[MIMovie alloc] initWithMovie:movie shouldReleaseMovie:NO];
		[fMovieView setMovie:fMovie];
		timeNow = GetMovieTime(movie, nil);
		playRate = GetMoviePreferredRate(movie);
//		PrePrerollMovie(movie, timeNow, playRate, nil, nil);
//		PrerollMovie(movie, timeNow, playRate);
#else
		TimeValue timeNow;
		Fixed playRate;
		[fQDView lockFocus];
		SetMovieGWorld( movie, (CGrafPtr)[fQDView qdPort], nil );
		Rect movieBox;
		GetPortBounds((CGrafPtr)[fQDView qdPort], &movieBox);
		NSLog( @"left = %d, top = %d, right = %d, bottom = %d", movieBox.left, movieBox.top, movieBox.right, movieBox.bottom );
		SetMovieBox ( movie, &movieBox);
		[fQDView unlockFocus];
		fQTMovie = movie;
		timeNow = GetMovieTime(movie, nil);
		playRate = GetMoviePreferredRate(movie);
		PrePrerollMovie(movie, timeNow, playRate, nil, nil);
		PrerollMovie(movie, timeNow, playRate);
#endif
		//}
	}
}

- (Movie)movie {
#ifndef USE_NSQUICKDRAWVIEW	
	return (Movie)[fMovie QTMovie];
#else
	return fQTMovie;
#endif	
}

- (NSMovieView*)movieView
{
	return fMovieView;
}

- (void)applicationWillTerminate:(NSNotification*)theNotification {
	vernl(6,"[MIViewerController applicationWillTerminate]");
	
	// OVERRIDE THIS TO NOT DELETE SELF, BECAUSE THAT DESTROYS THE MOVIE,
	// AND THE DOCUMENT DEALLOCATION CODE NEEDS THE MOVIE !!!!!!
	//[self release];
}


- (BOOL)validateMenuItem:(NSMenuItem *)anItem {
	if ( [[NSDocumentController sharedDocumentController] currentDocument] != NULL ) {
		[anItem setState:NSOffState];
		if ( [self isFullScreen] ) {
			if ( anItem == fFullScreenMenuItem ) [fFullScreenMenuItem setState:NSOnState];
		} else {
			if ( [self isVisible] ) {
				if ( ( anItem == fHalfSizeMenuItem ) && ( fMagnification == 0.5 ) )
					[fHalfSizeMenuItem setState:NSOnState];
				if ( ( anItem == fNormalSizeMenuItem ) && ( fMagnification == 1 ) )
					[fNormalSizeMenuItem setState:NSOnState];
				if ( ( anItem == fDoubleSizeMenuItem ) && ( fMagnification == 2 ) )
					[fDoubleSizeMenuItem setState:NSOnState];
			} else if ( anItem == fHideMenuItem ) return NO;
		}
		return YES;
	} else return NO;
}

- (BOOL)shouldIgnoreNotifications {
	return fShouldIgnoreNotifications;
}

- (void)setShouldIgnoreNotifications:(BOOL)flag {
	fShouldIgnoreNotifications = flag;
}

- (void)registerSelf {
	[[MIPaletteController sharedPaletteController] registerPaletteController:self visibleAfterLaunch:NO];
}

- (BOOL)ignoreArrowKeys {
	return YES;
}

- (BOOL)ignoresUndoRedoActions {
	return YES;
}


@end

@implementation MIFullScreenVideoWindow

#define kDelayBeforeHidingMouseCursor 5

- (IBAction)makeKeyAndOrderFront:(id)sender {
	[super makeKeyAndOrderFront:sender];
	[self setAcceptsMouseMovedEvents:YES];
	[NSCursor setHiddenUntilMouseMoves:YES];
	fTimer = [NSTimer scheduledTimerWithTimeInterval:kDelayBeforeHidingMouseCursor target:self selector:@selector(hideCursor:) userInfo:nil repeats:YES];
}

- (IBAction)orderOut:(id)sender {
	[self setAcceptsMouseMovedEvents:NO];
	[fTimer invalidate];
	fTimer = nil;
	[super orderOut:sender];
}

- (void)showMenuBarUnderMouse:(BOOL)flag {
	if ( flag ) {
		if (fTrackingRect) {
			[[self contentView] removeTrackingRect:fTrackingRect];
			fTrackingRect = nil;
		}
		NSRect trackingRect = [self frame];
		trackingRect.origin.y = NSMaxY(trackingRect) - [NSMenuView menuBarHeight];
		trackingRect.size.height = [NSMenuView menuBarHeight];
		fTrackingRect = [[self contentView] addTrackingRect:trackingRect owner:self userData:nil assumeInside:NO];
	} else {
		if ( fTrackingRect ) {
			[[self contentView] removeTrackingRect:fTrackingRect];
			fTrackingRect = nil;
			[self setLevel:NSScreenSaverWindowLevel];
		}
	}
}

- (void)mouseEntered:(NSEvent*)event {
	if (fTrackingRect) {
		[[self contentView] removeTrackingRect:fTrackingRect];
		fTrackingRect = nil;
	}

	if ( [self level] == NSScreenSaverWindowLevel ) {
		[self setLevel:NSMainMenuWindowLevel - 1];
		NSRect trackingRect = [self frame];
		trackingRect.size.height -= [NSMenuView menuBarHeight];
		fTrackingRect = [[self contentView] addTrackingRect:trackingRect owner:self userData:nil assumeInside:NO];
	} else {
		[self setLevel:NSScreenSaverWindowLevel];
		NSRect trackingRect = [self frame];
		trackingRect.origin.y = NSMaxY(trackingRect) - [NSMenuView menuBarHeight];
		trackingRect.size.height = [NSMenuView menuBarHeight];
		fTrackingRect = [[self contentView] addTrackingRect:trackingRect owner:self userData:nil assumeInside:NO];
		[[self contentView] setNeedsDisplay:YES];
	}
}

- (void)mouseExited:(NSEvent *)theEvent {
}

- (void)mouseMoved:(NSEvent *)theEvent {
	[fTimer setFireDate:[NSDate dateWithTimeIntervalSinceNow:kDelayBeforeHidingMouseCursor]];
}

- (void)hideCursor:(NSTimer*)timer {
	[NSCursor setHiddenUntilMouseMoves:YES];
}

- (BOOL)canBecomeKeyWindow {
    return YES;
}

- (void)keyDown:(NSEvent *)theEvent {
    [[MIViewerController sharedViewerController] setFullScreen:NO];
}


@end
