/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import "MIAllOnCommand.h"
#import "MIMainViewController.h"
#import "MIObjectRep.h"
#import "MIObject.h"
#import "MIDocumentController.h"
#import "MITopLevelContainerRep.h"

#import "MIMontageDocument.h"
#include "TVideoObject.h"
#include "TArmeDoc.h"


// these functions are in "MIMuteCommand.mm"
extern void GetMutedStatusBeforeCommand(MIObject* obj, ZzArray<SInt32>* mutedStatusBeforeCommand);
extern void SetMutedStatusAsTheyWereBeforeCommand(MIObject* obj, ZzArray<SInt32>* mutedStatusBeforeCommand, int& index);

/****************************************************************************************************

MIAllOnCommand

****************************************************************************************************/


@implementation MIAllOnCommand

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
	self = [super initFromArchive:handler];
	if ( self )
	{
		SInt32 nbStatus = handler->ReadSInt32();
		fObjectsStatusBeforeCommand = new ZzArray<SInt32>(nbStatus);
		for (int i=0; i < nbStatus; ++i)
			fObjectsStatusBeforeCommand->AddItem(handler->ReadSInt32());

		fPreviousVideoSoloObject = dynamic_cast<TArmeWSObj*>(handler->ReadCPPObject());
		fPreviousAudioSoloObject = dynamic_cast<TArmeWSObj*>(handler->ReadCPPObject());
		fMontageDoc = handler->ReadObjCObject();
	}
	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
	if ( [super writeToArchive:handler] )
	{
		handler->WriteSInt32(fObjectsStatusBeforeCommand->CountItems());
		for (int i=0; i < fObjectsStatusBeforeCommand->CountItems(); ++i)
			handler->WriteSInt32(fObjectsStatusBeforeCommand->ItemAt(i));

		handler->WriteCPPObject(fPreviousVideoSoloObject);
		handler->WriteCPPObject(fPreviousAudioSoloObject);
		handler->WriteObjCObject(fMontageDoc);
		return YES;
	}
	else
		return NO;
}

- (void)dealloc
{
	delete fObjectsStatusBeforeCommand;
	[super dealloc];
}

- (NSString*)commandName
{
	return @"All On";
}
- (id)initWithSelection:(MIWSSelection*)sel
{
	self = [super initWithSelection:sel];

	if ( self )
	{
		fObjectsStatusBeforeCommand = new ZzArray<SInt32>(32);
		fMontageDoc = [[NSDocumentController sharedDocumentController] currentDocument];
		
		[self modifiedVideoHierarchy];
		[self modifiedAudioHierarchy];
		[self overviewNeedsUpdate];
	}
	return self;
}

- (BOOL)displaySetupDialog
{
	return YES;
}

- (MICmdStatus)doIt
{
	fPreviousVideoSoloObject = [fMontageDoc armeDocument]->GetVideoSoloObject();
	fPreviousAudioSoloObject = [fMontageDoc armeDocument]->GetAudioSoloObject();
	
	NSEnumerator* childEnum = [[[fMontageDoc topLevelContainer] getChildren] objectEnumerator];
	MIObject* child;
	while (child = [childEnum nextObject])
	{
		if ([child WSObject]->GetType() != TArmeWSObj::kVideoXFadeEffect)
		{
			GetMutedStatusBeforeCommand(child, fObjectsStatusBeforeCommand);
			[child setMuted:false];
		}
	}

	[fMontageDoc armeDocument]->SetVideoSoloObject(NULL);
	[fMontageDoc armeDocument]->SetAudioSoloObject(NULL);


	//update all the views
	NSEnumerator* docEnum = [[[MIDocumentController sharedDocumentController] documents] objectEnumerator];
	MIMontageDocument* doc;
	while(doc = [docEnum nextObject])
	{
		NSEnumerator* mainViewEnum = [[doc windowControllers] objectEnumerator];
		MIMainViewController* mainViewController;
		while(mainViewController = [mainViewEnum nextObject])
		{
			[[mainViewController topLevelContainerView] setNeedsDisplay:YES];
		}
	}
	return kMICmdStatusDone;
}

- (void)undoIt
{
	int i=0;
	
	NSEnumerator* childEnum = [[[fMontageDoc topLevelContainer] getChildren] objectEnumerator];
	MIObject* child;
	while (child = [childEnum nextObject])
	{
		if ([child WSObject]->GetType() != TArmeWSObj::kVideoXFadeEffect)
			SetMutedStatusAsTheyWereBeforeCommand(child, fObjectsStatusBeforeCommand, i);
	}

	[fMontageDoc armeDocument]->SetVideoSoloObject(fPreviousVideoSoloObject);
	[fMontageDoc armeDocument]->SetAudioSoloObject(fPreviousAudioSoloObject);

	//update all the views
	NSEnumerator* docEnum = [[[MIDocumentController sharedDocumentController] documents] objectEnumerator];
	MIMontageDocument* doc;
	while(doc = [docEnum nextObject])
	{
		NSEnumerator* mainViewEnum = [[doc windowControllers] objectEnumerator];
		MIMainViewController* mainViewController;
		while(mainViewController = [mainViewEnum nextObject])
		{
			[[mainViewController topLevelContainerView] setNeedsDisplay:YES];
		}
	}
}

- (void)redoIt
{
	[self doIt];
}

+ (BOOL)supportsSelection:(MIWSSelection*)sel
{
	return [[[NSDocumentController sharedDocumentController] currentDocument] hasContent];
}

@end

