/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import <Cocoa/Cocoa.h>

#import "MIWSSelection.h"
#import "MIObject.h"

#ifndef ArmeObjCArchivable_h
#include "ArmeObjCArchivable.h"
#endif //ArmeObjCArchivable_h

#ifndef ArmeArchiveHandler_h
#include "ArmeArchiveHandler.h"
#endif //ArmeArchiveHandler_h

class ArmeArchiveHandler;

#ifndef MIProgressBarDlgManager_h
#include "MIProgressBarDlgManager.h"
#endif //MIProgressBarDlgManager_h

enum MICommandState
{
	kMICommandStateCanOnlyBeDone = 0,
	kMICommandStateCanOnlyBeUndone,
	kMICommandStateCanOnlyBeRedone
};

enum MICmdStatus
{
	kMICmdStatusCancelled = 1967,
	kMICmdStatusDone
};

/*!
	@class MICommand
	
	An MICommand represents an object that performs a model
	modification and knows how to undo this modification.
*/

@interface MICommand : NSObject 
{
@public
	SInt32			fState;
	MIWSSelection	*fSelection;
	BOOL			fModifiedAudioHierarchy;
	BOOL			fModifiedVideoHierarchy;
	BOOL			fOverviewUpdate;
}

- (id)initFromArchive:(ArmeArchiveHandler*)handler;
- (BOOL)writeToArchive:(ArmeArchiveHandler*)handler;


- (void)modifiedAudioHierarchy;
- (void)modifiedVideoHierarchy;
- (void)overviewNeedsUpdate;

- (BOOL)hasModifiedAudioHierarchy;
- (BOOL)hasModifiedVideoHierarchy;

- (BOOL)causedNeedOfOverviewUpdate;

/*!
	@function setSelection This function sets the (list of) object(s) this command will be applied on
*/ 
- (void)setSelection:(MIWSSelection*)sel;

- (id)initWithSelection:(MIWSSelection*)sel;

- (void)selectOriginalSelection;

/*! 
	@function doIt This function must be overridden and should perform 
    the actual model change. A status can be returned if the command was
	cancelled by the user (for lengthy commands).
	MANDATORY OVERRIDDING
*/
- (MICmdStatus)doIt;

/*! 
	@function undoIt This function must be overridden and should reverse
    the model to the state it was in before command execution.
	MANDATORY OVERRIDDING
*/
- (void)undoIt;

/*! 
	@function redoIt This function must be overridden and should perform 
    the actual model change when it has already been undone once. 
	It may be different from doIt if objects have been created by doIt and
	haven't been deleted by undoIt so they can be reused here.
	MANDATORY OVERRIDDING
*/
- (void)redoIt;

/*! 
	@function displaySetupDialog This function should be overridden if 
	the command needs user-supplied parameters.
	The function should return NO if the user has clicked on cancel. YES
	otherwise. $$$TODO think about scripting and how to provide parameters
	with another method.
	OPTIONAL OVERRIDDING
*/
- (BOOL)displaySetupDialog;

/*! @function CheckFrozenObjectInSelectionAndAskIfUserWantsToContinue This function checks if the selection contains
	any object that is the result of a freeze operation.  If yes, it asks the user if he wants
	to continue and loose the possibility to unfreeze the object or if he wants to cancel*/
- (BOOL)CheckFrozenObjectInSelectionAndAskIfUserWantsToContinue:(MIWSSelection*)sel;

- (BOOL)CheckFrozenObjectAndAskIfUserWantsToContinue:(MIObject*)obj;

/*! @function init (private) */
- (id)init;

/*! @function _doIt (private) */
- (MICmdStatus)_doIt;

/*! @function _undoIt (private) */
- (void)_undoIt;

/*! @function _redoIt (private) */
- (void)_redoIt;

/*! @function canDoIt (private) */
- (BOOL)canDoIt;

/*! @function canUndoIt (private) */
- (BOOL)canUndoIt;

/*! @function canRedoIt (private) */
- (BOOL)canRedoIt;

/*! @function commandName returns the name of the command for display in the edit->uno/redo menu */
- (NSString*)commandName;

+ (BOOL)supportsSelection:(MIWSSelection*)sel;

/*Progress Bar dialog methods*/

+ (MIProgressBarDlgManager*)getCurrentProgressDialog;
+ (void)setCurrentProgressDialog:(MIProgressBarDlgManager*)dlog;

- (int)getNbLengthyTasks;
- (BOOL)isCancelable;
@end

