/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import "MICopyCommand.h"
#import "MIObject.h"
#import "MIApplicationController.h"
#import "MIMainViewController.h"
#import "MIObjectRep.h"


#undef VER_DEF_TOPIC
#define VER_DEF_TOPIC MICopyCommand

/****************************************************************************************************

MICopyCommand

****************************************************************************************************/

@implementation MICopyCommand

- (void)dealloc
{
	[fObjectsToCopy release];
	[super dealloc];
}

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
	self = [super initFromArchive:handler];
	if(self)
	{
		// nothing to do (there is no undo/redo)
	}
	fObjectsToCopy = nil;
	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
	[super writeToArchive:handler];
	// nothing to do (there is no undo/redo)
	return YES;
}


- (id)initWithSelection:(MIWSSelection*)sel usingClipboard:(MIObject*)clipboard
{
	vernl(6,"initWithSelection");
	self = [super initWithSelection:sel];
	
	[fSelection adjustSelectionTimesToVideo];
		
	if ( self )
	{
		[fSelection discardVideoTransitions];

		fObjectsToCopy = [[NSMutableArray alloc] initWithCapacity:1];
	
		//if no custom clipboard is defined, use the application clipboard
		if (clipboard)
		{
			fClipboard = clipboard;
		}
		else
		{
			fClipboard = [[MIApplicationController sharedApplicationController] clipBoard];
		}
		
		//copy the selection
		//******************
		MIObjectRep* objectRep;
		NSEnumerator* objEnum = nil;
		
		switch ([fSelection type])
		{
			case kMIWSSelectionTypeTimeSlice :
			{
				fIsSliceCopy = true;
				fStartTime = [fSelection startTime];
				fEndTime = [fSelection endxTime];
				objEnum = [fSelection objectEnumerator];
				
			};break;
			case kMIWSSelectionTypeObjects :
			{
				fIsSliceCopy = false;
				//'flatten' the selection
				objEnum = [[fSelection putRepsInArrayAndRemoveDescendants] objectEnumerator];
			};break;
		}
	
		// let's enumerate all the selection objects.
		while ( objectRep = [objEnum nextObject] )
			[fObjectsToCopy addObject:[objectRep representedObject]];
	}
	return self;
}

- (id)initWithObject:(MIObject*)object from:(ArmeTime)startTime to:(ArmeTime)endTime usingClipboard:(MIObject*)clipboard
{
	self = [super initWithSelection:nil];
	if ( self )
	{
		fObjectsToCopy = [[NSMutableArray alloc] initWithCapacity:1];
		[fObjectsToCopy addObject:object];
		fStartTime = startTime;
		fEndTime = endTime;
		fIsSliceCopy = true;
		fClipboard = clipboard;
	}
	return self;
}

- (BOOL)displaySetupDialog
{
	return YES;
}

- (MICmdStatus)doIt
{
	vernl(6,"doIt");

	// clear clipboard
	//****************
	while ([[fClipboard getChildren] count] > 0)
	{
		[fClipboard removeChild:[[fClipboard getChildren] objectAtIndex:0]];
	}
	vernl(6,"doIt-");
	
	NSEnumerator* objEnum = [fObjectsToCopy objectEnumerator];
	MIObject* object;

	//copy the selection
	//******************
	if (fIsSliceCopy)
	{
		// let's enumerate all the selection objects.
		while ( object = [objEnum nextObject] )
		{
			// and call recursively the sliceCopy
			MIObject* copy = [object sliceCopyInClipBoard:fClipboard from:fStartTime to:fEndTime];
			[copy setVerticalPos:[object getAbsoluteVerticalPos]];
			vernl(6,"doIt-");
		}
	}
	else
	{
		// let's enumerate all the selection objects.
		while ( object = [objEnum nextObject] )
		{
			// and call recursively the copy
			MIObject* copy = [object copyInClipBoard:fClipboard];
			[copy setVerticalPos:[object getAbsoluteVerticalPos]];
			vernl(6,"doIt-");
		}
	}
	vernl(6,"doIt-");
	return kMICmdStatusDone;
}

- (void)undoIt
{
// nothing to do
}

- (void)redoIt
{
// nothing to do
}

+ (BOOL)supportsSelection:(MIWSSelection*)sel
{
	if ( [sel containsOnlyVideoTransitions])
		return NO;

	if ( [sel count] == 0 ) return NO;

	return YES;
}

@end

