/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import "MICreateAudioClipFromChunkCommand.h"
#import "MIObject.h"
#import "MIDeleteCommand.h"
#import "MIResizeCommand.h"
#import "MIApplicationController.h"

#ifndef MEAudioClip_h
#include "MEAudioClip.h"
#endif //MEAudioClip_h

@implementation MICreateAudioClipFromChunkCommand

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
	self = [super initFromArchive:handler];
	if(self)
	{
		fChunk = NULL; // it's only needed in the doIt call, and this call won't ever be made after archiving/unarchiving
		fCreatedObject = [handler->ReadObjCObject() retain];
		fEnlargeCommand = [handler->ReadObjCObject() retain];
		fTargetParent = nil; // it's only needed in the doIt call, and this call won't ever be made after archiving/unarchiving
		fTargetPosition = MIZeroPoint; // it's only needed in the doIt call, and this call won't ever be made after archiving/unarchiving
		fDeleteCommand = nil; // it's only needed in the doIt call, and this call won't ever be made after archiving/unarchiving
		fSplitTask = NULL; // it's only needed in the doIt call, and this call won't ever be made after archiving/unarchiving
		fNbChannels = 0; // it's only needed in the doIt call, and this call won't ever be made after archiving/unarchiving
		fMontageDoc = nil; // it's only needed in the doIt call, and this call won't ever be made after archiving/unarchiving
	}
	[self modifiedAudioHierarchy];
	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
	[super writeToArchive:handler];
	ZzAssert(fCreatedObject); // command should't get archived before a doIt call
	handler->WriteObjCObject(fCreatedObject);
	handler->WriteObjCObject(fEnlargeCommand);
	return YES;
}

- (id)initWithChunk:(const MEAudioChunk&)chunk montageDoc:(MIMontageDocument*)montageDoc
{
	self = [super initWithSelection:nil];
	if(self)
	{
		fChunk = new MEAudioChunk;
		*fChunk = chunk;
		fCreatedObject = nil;
		fTargetParent = nil;
		fTargetPosition = MIZeroPoint;
		fDeleteCommand = nil;
		fSplitTask = NULL;
		fNbChannels = 0;
		fMontageDoc = montageDoc;
		fEnlargeCommand = nil;
	}
	return self;
}

- (void)setParent:(MIObject*)parent
{	
	fTargetParent = parent;
}

- (void)setPosition:(MIPoint)position
{
	fTargetPosition = position;
}

- (void)setNbChannels:(SInt32)nbChannels
{
	fNbChannels = nbChannels;
}

- (MIObject*)getCreatedObject
{
	return fCreatedObject;
}

- (MICmdStatus)doIt
{
	ZzAssert(fNbChannels > 0);
	MEAudioClip* audioClip = new MEAudioClip(*fChunk,fNbChannels,[MICommand getCurrentProgressDialog]);
	fCreatedObject = [[MIObject alloc] initWithArmeWSObj:audioClip];
	
    [fCreatedObject setMontageDoc:fMontageDoc];
	[fCreatedObject moveToObject:fTargetParent	atTime:fTargetPosition.time withVPos:fTargetPosition.vpos];
	[fCreatedObject createAllRepresentations];
	
	// check if parent size is big enough
	if ( !MIContainsRect( [fTargetParent modelFrame], [fTargetParent minimalRectToFitContent] ) )
	{
		fEnlargeCommand = [[MIResizeCommand alloc] initWithObject:fTargetParent];
		[fEnlargeCommand setObject:fTargetParent newFrame:MIUnionRect([fTargetParent modelFrame], [fTargetParent minimalRectToFitContent])];
		[fEnlargeCommand _doIt];
	}

	BOOL notCancelled = [[MIApplicationController sharedApplicationController] executeAndDeleteSplitTask:fSplitTask forMainViewController:NULL];
	ZzAssert(notCancelled);
	
	return kMICmdStatusDone;
}

- (void)undoIt
{
	[fEnlargeCommand _undoIt];

	if ( !fDeleteCommand )
	{
		fDeleteCommand = [[MIDeleteCommand alloc] initWithObject:fCreatedObject];
		[fDeleteCommand _doIt];
	}
	else
	{
		[fDeleteCommand _redoIt];
	}
}

- (void)redoIt
{
	[fDeleteCommand _undoIt];
	[fEnlargeCommand _redoIt];
}

- (void)dealloc
{
	[fDeleteCommand release];
	[fEnlargeCommand release];
	delete fChunk;
    [super dealloc];
}

@end

