/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#include "MICreateAudioClipsFromRecordedTakesCommand.h"
#import "MIObject.h"
#import "MIDeleteCommand.h"
#import "MIResizeCommand.h"
#import "MIApplicationController.h"

#ifndef MEAudioClip_h
#include "MEAudioClip.h"
#endif //MEAudioClip_h

#ifndef ZzSplitTask_h
#include "ZzSplitTask.h"
#endif //ZzSplitTask_h

#import "MIMontageDocument.h"

@implementation MICreateAudioClipsFromRecordedTakesCommand

- (NSString*)commandName
{
	return @"Create Audio Clips From Capture";
}

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
	self = [super initFromArchive:handler];
	if(self)
	{
		fIndicesZzArray = nil; // only needed at doIt time
		fNbChannelsZzArray = nil; // only needed at doIt time
		fPositionZzArray = nil; // only needed at doIt time
		fDurationZzArray = nil; // only needed at doIt time
		fCreatedObjects = [handler->ReadObjCObject() retain];
		fEnlargeCommand = [handler->ReadObjCObject() retain];
		fDeleteCommands = nil; // it's only needed in the doIt call, and this call won't ever be made after archiving/unarchiving
		fMontageDoc = nil; // it's only needed in the doIt call, and this call won't ever be made after archiving/unarchiving
		fClipNames = nil; // it's only needed in the doIt call, and this call won't ever be made after archiving/unarchiving
	}
	[self modifiedAudioHierarchy];
	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
	[super writeToArchive:handler];
	ZzAssert(fCreatedObjects); // command should't get archived before a doIt call
	handler->WriteObjCObject(fCreatedObjects);
	handler->WriteObjCObject(fEnlargeCommand);
	return YES;
}

-(id)initWithDoc:(MIMontageDocument*)mtgDoc
{
	self = [super initWithSelection:nil];
	if(self)
	{
		fIndicesZzArray = new ZzArray<MEAudioChunkFileIndex>;
		fNbChannelsZzArray = new ZzArray<SInt32>;
		fPositionZzArray = new ZzArray<MIPoint>;
		fDurationZzArray = new ZzArray<ArmeTime>;
		fCreatedObjects = nil;
		fDeleteCommands = nil;
		fMontageDoc = mtgDoc;
		fEnlargeCommand = nil;
		fClipNames = [[NSMutableArray alloc] initWithCapacity:5];
	}
	[self modifiedAudioHierarchy];
	return self;
}
		
- (MICmdStatus)doIt
{
	SInt32 nbOfIndices = fIndicesZzArray->CountItems();
	SInt32 nbOfNbchannels = fNbChannelsZzArray->CountItems();
	ZzAssert(nbOfNbchannels == nbOfIndices);
	SInt32 nbOfPositions = fPositionZzArray->CountItems();
	ZzAssert(nbOfPositions == nbOfIndices);
	SInt32 nbOfDuration = fDurationZzArray->CountItems();
	ZzAssert(nbOfDuration == nbOfIndices);
	SInt32 nbOfNames = [fClipNames count];
	ZzAssert(nbOfNames == nbOfIndices);
	
	TArmeDoc* armeDoc = [fMontageDoc armeDocument];
	MIObject* miTLC = [fMontageDoc topLevelContainer];
	fCreatedObjects = [[NSMutableArray alloc] init];
	MIObject* createdObject = nil;
	[MICommand getCurrentProgressDialog]->SetNbSubTasks(2);

	[MICommand getCurrentProgressDialog]->EnterSubTask();
	[MICommand getCurrentProgressDialog]->SetNbSubTasks(nbOfIndices);
	for(SInt32 i = 0; i < nbOfIndices; i++)
	{
		// let's build a chunk that describes the audio file
		MEAudioChunk chunk;
		chunk.duration = 0; //SetFileIndex crashes if duration is != 0.
		chunk.doc = armeDoc;
		chunk.startInFile = 0;
		chunk.SetFileIndex(fIndicesZzArray->ItemAt(i));
		chunk.duration = GET_VAL(fDurationZzArray->ItemAt(i));
		[MICommand getCurrentProgressDialog]->EnterSubTask();
		MEAudioClip* audioClip = new MEAudioClip(chunk,fNbChannelsZzArray->ItemAt(i),[MICommand getCurrentProgressDialog]);
		[MICommand getCurrentProgressDialog]->LeaveSubTask();
		createdObject = [[[MIObject alloc] initWithArmeWSObj:audioClip] autorelease];

		[createdObject WSObject]->SetNameCopy( [[fClipNames objectAtIndex:i] UTF8String] );
		[createdObject setMontageDoc:fMontageDoc];
		[createdObject moveToObject:miTLC	atTime:fPositionZzArray->ItemAt(i).time withVPos:fPositionZzArray->ItemAt(i).vpos];
		[createdObject createAllRepresentations];
		[createdObject setRepresentationsNeedDisplay];
		[fCreatedObjects addObject:createdObject];
	}
	[MICommand getCurrentProgressDialog]->LeaveSubTask();

	[MICommand getCurrentProgressDialog]->EnterSubTask();
	// check if parent size is big enough
	if ( !MIContainsRect( [miTLC modelFrame], [miTLC minimalRectToFitContent] ) )
	{
		fEnlargeCommand = [[MIResizeCommand alloc] initWithObject:miTLC];
		[fEnlargeCommand setObject:miTLC newFrame:MIUnionRect([miTLC modelFrame], [miTLC minimalRectToFitContent])];
		[fEnlargeCommand _doIt];
	}
	[MICommand getCurrentProgressDialog]->LeaveSubTask();

	return kMICmdStatusDone;
}

- (void)undoIt
{
	[MICommand getCurrentProgressDialog]->SetNbSubTasks(2);
	[MICommand getCurrentProgressDialog]->EnterSubTask();
	[fEnlargeCommand _undoIt];
	[MICommand getCurrentProgressDialog]->LeaveSubTask();

	if ( !fDeleteCommands )
	{
		[MICommand getCurrentProgressDialog]->EnterSubTask();
		fDeleteCommands = [[NSMutableArray alloc] init];
		[MICommand getCurrentProgressDialog]->SetNbSubTasks([fCreatedObjects count]);
		for(UInt32 i = 0; i < [fCreatedObjects count]; i++)
		{
			MIDeleteCommand* deleteCommand = [[[MIDeleteCommand alloc] initWithObject:[fCreatedObjects objectAtIndex:i]] autorelease];
			[fDeleteCommands addObject:deleteCommand];
			[MICommand getCurrentProgressDialog]->EnterSubTask();
			[deleteCommand _doIt];
			[MICommand getCurrentProgressDialog]->LeaveSubTask();
		}
		[MICommand getCurrentProgressDialog]->LeaveSubTask();
	}
	else
	{
		[MICommand getCurrentProgressDialog]->EnterSubTask();
		[MICommand getCurrentProgressDialog]->SetNbSubTasks([fDeleteCommands count]);
		for(UInt32 i = 0; i < [fDeleteCommands count]; i++)
		{
			[MICommand getCurrentProgressDialog]->EnterSubTask();
			[[fDeleteCommands objectAtIndex:i] _redoIt];
			[MICommand getCurrentProgressDialog]->LeaveSubTask();
		}
		[MICommand getCurrentProgressDialog]->LeaveSubTask();
	}
}

- (void)redoIt
{
	[MICommand getCurrentProgressDialog]->SetNbSubTasks([fDeleteCommands count]+1);
	for(UInt32 i = 0; i < [fDeleteCommands count]; i++)
	{
		[MICommand getCurrentProgressDialog]->EnterSubTask();
		[[fDeleteCommands objectAtIndex:i] _undoIt];
		[MICommand getCurrentProgressDialog]->LeaveSubTask();
	}
	[MICommand getCurrentProgressDialog]->EnterSubTask();
	[fEnlargeCommand _redoIt];
	[MICommand getCurrentProgressDialog]->LeaveSubTask();
}

- (void)dealloc
{
	[fDeleteCommands release];
	[fEnlargeCommand release];
	[fCreatedObjects release];
	[fClipNames release];
	delete fIndicesZzArray;
	delete fNbChannelsZzArray;
	delete fPositionZzArray;
	delete fDurationZzArray;
    [super dealloc];
}

- (void)addFile:(MEAudioChunkFileIndex)index nbChannels:(SInt32)nbChannels position:(MIPoint)position duration:(ArmeTime)takeDuration name:(NSString*)name
{
	fIndicesZzArray->AddItem(index);
	fNbChannelsZzArray->AddItem(nbChannels);
	fPositionZzArray->AddItem(position);
	fDurationZzArray->AddItem(takeDuration);
	[fClipNames addObject:name];
}


@end
