/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import "MICreateContainerCommand.h"
#import "MIMainViewController.h"
#import "MISingleMoveCommand.h"
#import "MIEncloseSelectionCommand.h"
#import "MIMontageDocument.h"
#import "MITopLevelContainerRep.h"
#import "MIDeleteCommand.h"
#import "MIUpdateVideoTransitionsCommand.h"
#import "MITime.h"

#ifndef TArmeDoc_h
#include "TArmeDoc.h"
#endif //TArmeDoc_h

#ifndef TArmeWSObj_h
#include "TArmeWSObj.h"
#endif //TArmeWSObj_h

#ifndef TArmeContainer_h
#include "TArmeContainer.h"
#endif //TArmeContainer_h

/****************************************************************************************************

MICreateContainerCommand

****************************************************************************************************/

@implementation MICreateContainerCommand : MICommand

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
	self = [super initFromArchive:handler];
	if(self)
	{
		fMoveCommand = [handler->ReadObjCObject() retain];
		fEncloseCommand = [handler->ReadObjCObject() retain];
		fDeleteCommand = [handler->ReadObjCObject() retain];
		fCreatedObject = [handler->ReadObjCObject() retain];
	}
	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
	if ( [super writeToArchive:handler] )
	{
		handler->WriteObjCObject( fMoveCommand );
		handler->WriteObjCObject( fEncloseCommand );
		handler->WriteObjCObject( fDeleteCommand );
		handler->WriteObjCObject( fCreatedObject );
		return YES;
	}
	else
		return NO;
}

- (void)dealloc
{
	[fMoveCommand release];
	[fEncloseCommand release];
#ifdef DEBUG
//    printf("MICreateContainerCommand releasing object %p in -dealloc\n",(void*)fCreatedObject);
#endif //DEBUG
	[fCreatedObject release];
	[fDeleteCommand release];
	[super dealloc];
}

- (NSString*)commandName
{
	return @"Create Group";
}

- (MIObject*)getCreatedObject
{
	return fCreatedObject;
}

- (id)initWithSelection:(MIWSSelection*)sel
{
	self = [super initWithSelection:sel];

	[fSelection discardVideoTransitions];
	[self modifiedVideoHierarchy];
	return self;
}

- (void)setPosition:(MIPoint)point
{
	fPosition = point;
}

- (void)setParent:(MIObject*)parent
{
	fParent = parent;
}

- (MICmdStatus)doIt
{
	TArmeDoc				*armeDoc;
	MIMontageDocument		*mDoc;
	TArmeWSObj				*newArmeObj;
	MIObject				*newObj;
	static int				groupCount = 0;
	MITime					*t;

	mDoc = [[NSDocumentController sharedDocumentController] currentDocument];
    armeDoc = [mDoc armeDocument];

	newArmeObj = new TArmeContainer( [[NSString stringWithFormat:@"group %d", ++groupCount] cString], armeDoc );
	
	t = [MITime timeWithDocument:mDoc];
	
	[t setSeconds:10];
	
	newArmeObj->_SetLength([t getArmeTime]);
	
	// copy the children routing, if any
	MEVAORouting newRouting = MIGetSelectionVAORouting(fSelection);
	newArmeObj->SetVAORouting(newRouting);

	if ( !newArmeObj )
	{
		verntl(interfaceStephan,0,"Null object returned while trying to create a container\n\r");
		groupCount--;
		return kMICmdStatusCancelled;
	}

	newObj = [[MIObject alloc] initWithArmeWSObj:newArmeObj];
#ifdef DEBUG
//    printf("MICreateContainerCommand creating object %p in -doIt\n",(void*)newObj);
#endif //#ifdef DEBUG
    [newObj setMontageDoc:mDoc];
	[newObj createAllRepresentations];

	fMoveCommand = [[MISingleMoveCommand alloc] initWithSelection:nil];
	[fMoveCommand setObject:newObj];
	if ( fParent )
		[fMoveCommand setNewParent:fParent];
	else
		[fMoveCommand setNewParent:[mDoc topLevelContainer]];
	[fMoveCommand setNewOrigin:fPosition];

	[fMoveCommand _doIt];

//#ifdef DEBUG
//	printf("MICreateContainerCommand retaining newObj %p by (fCreatedObject = [newObj retain])\n",(void*)newObj);
//#endif //#ifdef DEBUG
	fCreatedObject = newObj; //created with ref count == 1 so we don't retain it...
	
	// make the new container enclose the current selection, if it is possible
//	[[newObj findRepresentationInTLC:[fCurrentMainViewController topLevelContainerView]] makeSelfEncloseCurrentSelection];

	fEncloseCommand = [[MIEncloseSelectionCommand alloc] initWithSelection:fSelection];
	[fEncloseCommand setContainer:fCreatedObject];
	if ( ![MIEncloseSelectionCommand supportsSelection:fSelection] )
	{
		[fEncloseCommand release];
		fEncloseCommand = nil;
	}
	[fEncloseCommand _doIt];
	[self overviewNeedsUpdate];
	
	return kMICmdStatusDone;
}

- (void)undoIt
{
	[fEncloseCommand _undoIt];
	
	//the enclose command has modified the transitions, we need to update them before deleting the container
	MIUpdateVideoTransitionsCommand* update = [[MIUpdateVideoTransitionsCommand alloc] initWithDocument:[[NSDocumentController sharedDocumentController] currentDocument] from:ARME_TIME_UNDEF to:ARME_TIME_UNDEF];
	[update doIt];
	[update release];

	fDeleteCommand = [[MIDeleteCommand alloc] initWithObject:fCreatedObject];
	[fDeleteCommand _doIt];
}

- (void)redoIt
{
	 [fDeleteCommand _undoIt];
	 [fDeleteCommand release];
	 fDeleteCommand = nil;
	 
	 [fEncloseCommand _redoIt];
}

+ (BOOL)supportsSelection:(MIWSSelection*)sel
{
	// it works also without a selection
	if (!sel)
		return YES;

	if ( [sel containsOnlyVideoTransitions])
		return NO;

	if ( [sel type] != kMIWSSelectionTypeObjects )
		return NO;
		
	return YES;
}

@end


