/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import "MICreatePictFromFrame.h"
#import "MIMainViewController.h"
#import "MIObjectRep.h"
#import "MIObject.h"
#import "MIMontageDocument.h"
#import "MICutCommand.h"
#import "MIPasteLowLevelCommand.h"
#import "MIFreezeCommand.h"
#import "MIResizeCommand.h"
#include "TVideoObject.h"
#include "TArmeDoc.h"


#ifndef MIProgressBarDlgManager_h
#include "MIProgressBarDlgManager.h"
#endif //MIProgressBarDlgManager_h


/****************************************************************************************************

MICreatePictFromFrame

****************************************************************************************************/


@implementation MICreatePictFromFrameCommand

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
	[super initFromArchive:handler];
	fResult = [handler->ReadObjCObject() retain];
	
	if (handler->GetFileVer() >= 0x00010029)
		fEnlargeCommand = [handler->ReadObjCObject() retain];
	else
		fEnlargeCommand = nil;
	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
	[super writeToArchive:handler];
	handler->WriteObjCObject(fResult);
	
	//from version 0x00010029
	handler->WriteObjCObject(fEnlargeCommand);
	return YES;
}

- (void)dealloc
{
	[fResult release];
	[fEnlargeCommand release];
	[super dealloc];
}

- (NSString*)commandName
{
	return @"Create Still Clip";
}

- (id)initWithSelection:(MIWSSelection*)sel inDoc:(MIMontageDocument*)doc
{
	self = [super initWithSelection:sel];
	
	[fSelection adjustSelectionTimesToVideo];
	fEnlargeCommand = nil;
		
	fObjectsInSelection = [[NSMutableArray alloc] initWithCapacity:1];
	if ([sel type] == kMIWSSelectionTypeTimeSlice)
	{
		fPictTime = [sel startTime];
	}
	else
		fPictTime = [doc armeDocument]->GetPlayPos();

	if ( self )
	{
		[fSelection discardVideoTransitions];

		[self modifiedVideoHierarchy];
		[self overviewNeedsUpdate];

		if (fSelection && [fSelection count]>0)
		{
			ArmeTime startTime = ARME_TIME_MAX;
			ArmeTime endTime = ARME_TIME_MIN;
		
			NSEnumerator*	objEnum = [fSelection objectEnumerator];
			MIObjectRep* obj;
			while (obj = [objEnum nextObject])
			{
				if ([[obj representedObject] WSObject]->ContainsVideo())
				{
					[fObjectsInSelection addObject:[obj representedObject]];
					if ([[obj representedObject] WSObject]->GetAbsoluteStartTime() < startTime)
						startTime = [[obj representedObject] WSObject]->GetAbsoluteStartTime();
					if ([[obj representedObject] WSObject]->GetAbsoluteEndxTime() > endTime)
						endTime = [[obj representedObject] WSObject]->GetAbsoluteEndxTime();
				}
			}
			
			if ((fPictTime > startTime) && (fPictTime < endTime))
			{
				fParentObject = [[fObjectsInSelection objectAtIndex:0] parentObject];
				fVPos = [[fObjectsInSelection objectAtIndex:0] getVerticalPos];
				GET_VALUE(fVPos) -= 100;
				[[fObjectsInSelection objectAtIndex:0] WSObject]->GetNameForStillImage(fName, kMaxFileLength);
	
				return self;
			}
		}
		// else there is no selection or play bar is out of the selection range, we take the whole movie
		[fObjectsInSelection addObject:[doc topLevelContainer]];
		fParentObject = [doc topLevelContainer];
		fVPos = [[doc findObject:[[doc topLevelContainer] WSObject]->FindVideoObjectAtTime(fPictTime)] getVerticalPos];
		GET_VALUE(fVPos) -= 100;
		
		sprintf(fName, "Master Movie.still(%ld)", [doc armeDocument]->GetStillImageIndex());
	}
	
	return self;
}

- (MICmdStatus)doIt
{
	ArmeErr err;

	// render one image in a file.  The image will be imported and as a single image, its duration will be set to X seconds
	fResult = [Render(err, fObjectsInSelection, fParentObject, fVPos, _MIVPos(0), [MICommand getCurrentProgressDialog], false, fPictTime, fPictTime, true, false, false, true) retain];
	if ((fResult == nil) || (err.type != kArmeOK))
	{
		NSRunAlertPanel(@"Error", @"Error during the create still operation.  Command Failed.", @"OK", nil, nil);
	
		return kMICmdStatusCancelled;
	}
	
	[fResult moveToObject:fParentObject atTime:fPictTime-[fParentObject WSObject]->GetAbsoluteStartTime() withVPos:fVPos];

//	[[fResult montageDoc] armeDocument]->GetTopContainer()->UpdateAffectedHierarchy();

	// give a name to the new clip
	[fResult WSObject]->SetNameCopy(fName);

	[fResult setRepresentationsNeedDisplay];

	// check if parent size is big enough
	if ( !MIContainsRect( [fParentObject modelFrame], [fParentObject minimalRectToFitContent] ) )
	{
		fEnlargeCommand = [[MIResizeCommand alloc] initWithSelection:nil];
		[fEnlargeCommand setObject:fParentObject newFrame:MIUnionRect([fParentObject modelFrame], [fParentObject minimalRectToFitContent])];
		[fEnlargeCommand _doIt];
	}
	
	return kMICmdStatusDone;
}

- (int)getNbLengthyTasks
{
	return 1;
}
- (void)undoIt
{
	if (fEnlargeCommand)
		[fEnlargeCommand _undoIt];
	[fResult setOuttaHierarchy];

}

- (void)redoIt
{
	[fResult putBackInHierarchy];
	[fResult setRepresentationsNeedDisplay];

	if (fEnlargeCommand)
		[fEnlargeCommand _undoIt];
}

- (BOOL)displaySetupDialog
{
	return YES;
}

- (ArmeTime)pictTime
{
	return fPictTime;
}
+ (BOOL)supportsSelection:(MIWSSelection*)sel
{
	TArmeDoc* doc = [[[NSDocumentController sharedDocumentController] currentDocument] armeDocument];
	// if there is no selection
	if (!sel || ([sel count] == 0))
	{
		//if there is video below the playber, create a still clip
		if (doc->GetTopContainer()->ContainsVideoAtTime(doc->GetPlayPos()))
		{
			return YES;
		}
		else
			return NO;
	}

	// doesn't work on slice selection
	if ([sel type] != kMIWSSelectionTypeObjects)
		return NO;

	if ([sel containsVideo:NO andTransparentTextClip:NO])
	{
		if ([sel containsOnlyVideoTransitions])
			return NO;
		
		//ensure that there is some video below the playbar
		NSEnumerator*	objEnum = [sel objectEnumerator];
		MIObjectRep* obj;
		while (obj = [objEnum nextObject])
		{
			if ([[obj representedObject] WSObject]->IsVideoObject() && ((TVideoObject*)[[obj representedObject] WSObject])->ContainsVideoAtTime(doc->GetPlayPos()))
			{
				if (((TVideoObject*)[[obj representedObject] WSObject])->IsVisible())
					return YES;
			}
		}
		
		return NO;
	}
	else
		return NO;
}
@end
