/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import "MIDJSwitchCommand.h"
#import "MIObject.h"
#import "MIMainViewController.h"
#import "MIMontageDocument.h"
#import "MISingleMoveCommand.h"
#import "MIResizeCommand.h"
#import "MIObjectRep.h"

#ifndef TArmeWSObj_h
#include "TArmeWSObj.h"
#endif //TArmeWSObj_h

#ifndef TArmeDoc_h
#include "TArmeDoc.h"
#endif //TArmeDoc_h

@implementation MIDJSwitchCommand

- (void)dealloc
{
	[fSingleMoveCommands release];
	[fEnlargeCommands release];

	[super dealloc];
}

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
	self = [super initFromArchive:handler];
	if ( self )
	{
		fSingleMoveCommands = [handler->ReadObjCObject() retain];
		fEnlargeCommands = [handler->ReadObjCObject() retain];
	}
	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
	if ( [super writeToArchive:handler] )
	{
		handler->WriteObjCObject( fSingleMoveCommands );
		handler->WriteObjCObject( fEnlargeCommands );
		return YES;
	}
	else
		return NO;
}

- (NSString*)commandName
{
	return @"DJ Switch";
}

- (id)initWithSelection:(MIWSSelection*)sel
{
	self = [super initWithSelection:sel];

	if ( self )
	{
		fSingleMoveCommands = [[NSMutableArray alloc] init];
		fEnlargeCommands = [[NSMutableArray alloc] init];
	}
	return self;
}

+ (BOOL)supportsSelection:(MIWSSelection*)sel
{
	if ( [sel type] == kMIWSSelectionTypeObjects
		 )
		return YES;
	else
		return NO;
}

- (BOOL)displaySetupDialog
{	
	// let's check if at least one object will be moved by the command
	BOOL cmdWillDoSomethingRelevant = NO;
	NSArray* repArray = [fSelection putRepsInArrayAndRemoveDescendants];
	NSEnumerator* objEnum = [repArray objectEnumerator];
	MIObjectRep* objectRep;
	
	while ( objectRep = [objEnum nextObject] )
	{
		TArmeWSObj* wsobj = [[objectRep representedObject] WSObject];
		wsobj->GetDoc()->LockRead();
		ArmeTime lloc = wsobj->GetDoc()->GetLLocator();
		ArmeTime rloc = wsobj->GetDoc()->GetRLocator();
		ArmeTime objStart = wsobj->GetAbsoluteStartTime();
		if(objStart == lloc || objStart == rloc)
			cmdWillDoSomethingRelevant = YES;
		wsobj->GetDoc()->UnlockRead();
	}
	return cmdWillDoSomethingRelevant;
}

- (MICmdStatus)doIt
{
	NSEnumerator* objEnum = [[fSelection putRepsInArrayAndRemoveDescendants] objectEnumerator];
	MIObjectRep* objectRep;
	
	while ( objectRep = [objEnum nextObject] )
	{
		MIObject* obj = [objectRep representedObject];
		TArmeWSObj* wsobj = [obj WSObject];
		wsobj->GetDoc()->LockRead();
		ArmeTime lloc = wsobj->GetDoc()->GetLLocator();
		ArmeTime rloc = wsobj->GetDoc()->GetRLocator();
		wsobj->GetDoc()->UnlockRead();
		ArmeTime loopLen = _ArmeTime(GET_VAL(rloc)-GET_VAL(lloc));
		ArmeTime offset;
		
		ArmeTime objStart = wsobj->GetAbsoluteStartTime();
		if(objStart == lloc)
			offset = loopLen;
		else if(objStart == rloc)
			offset = _ArmeTime(-GET_VAL(loopLen));
		else
			continue;

		// we need to add "offset" to the object xpos, without changing its parent
		MISingleMoveCommand* singleMove = [[[MISingleMoveCommand alloc] initWithSelection:nil] autorelease];
		ArmeTime objLocalPosX = wsobj->GetPos().fStartTime;
		MIVPos objLocalPosY = [obj getVerticalPos];
		[singleMove setObject:obj];
		[singleMove setNewParent:obj->fParentObject];
		MIPoint pt;
		pt.time = objLocalPosX+offset;
		pt.vpos = objLocalPosY;
		[singleMove setNewOrigin:pt];
		[fSingleMoveCommands addObject:singleMove];

//		MIObjectRep* parentRep = [obj->fParentObject findRepresentationInTLC:[fController topLevelContainerView]];
		// let's enlarge the parent if it needs
		if ( !MIContainsRect( [obj->fParentObject modelFrame], [obj->fParentObject minimalRectToFitContent] ) )
		{
			MIResizeCommand* enlargeCommand = [[[MIResizeCommand alloc] initWithSelection:nil] autorelease];
			[enlargeCommand setObject:obj->fParentObject newFrame:
				MIUnionRect([obj->fParentObject modelFrame], [obj->fParentObject minimalRectToFitContent])];
			[fEnlargeCommands addObject:enlargeCommand];
		}
	}
	
	// now exec the commands
	NSEnumerator* singleMoveCommandsEnum = [fSingleMoveCommands objectEnumerator];
	NSEnumerator* enlargeCommandsEnum = [fEnlargeCommands objectEnumerator];
	MISingleMoveCommand* singleMoveCommand;
	MIResizeCommand* enlargeCommand;
	
	while ( singleMoveCommand = [singleMoveCommandsEnum nextObject] )
	{
		enlargeCommand = [enlargeCommandsEnum nextObject];
		[singleMoveCommand _doIt];
		[enlargeCommand _doIt];
	}
	return kMICmdStatusDone;
}

- (void)undoIt
{
	NSEnumerator* enlargeCommandsEnum = [fEnlargeCommands reverseObjectEnumerator];
	NSEnumerator* singleMoveCommandsEnum = [fSingleMoveCommands reverseObjectEnumerator];
	MIResizeCommand* enlargeCommand;
	MISingleMoveCommand* singleMoveCommand;
	
	while ( enlargeCommand = [enlargeCommandsEnum nextObject] )
	{
		singleMoveCommand = [singleMoveCommandsEnum nextObject];
		[enlargeCommand _undoIt];
		[singleMoveCommand _undoIt];
	}
}

- (void)redoIt
{
	NSEnumerator* singleMoveCommandsEnum = [fSingleMoveCommands objectEnumerator];
	NSEnumerator* enlargeCommandsEnum = [fEnlargeCommands objectEnumerator];
	MISingleMoveCommand* singleMoveCommand;
	MIResizeCommand* enlargeCommand;
	
	while ( singleMoveCommand = [singleMoveCommandsEnum nextObject] )
	{
		enlargeCommand = [enlargeCommandsEnum nextObject];
		[singleMoveCommand _redoIt];
		[enlargeCommand _redoIt];
	}
}

@end

