/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#include "MIDeleteContainerCommand.h"

#ifndef TArmeDoc_h
#include "TArmeDoc.h"
#endif //TArmeDoc_h

#import "MIObject.h"
#import "MIObjectRep.h"
#import "MIWSSelection.h"
#import "MIMoveCommand.h"
#import "MIMontageDocument.h"
#import "MIApplicationController.h"

/****************************************************************************************************

MIDeleteContainerCommand

****************************************************************************************************/

@implementation MIDeleteContainerCommand
- (void)dealloc
{
	[fMemento release];
	[fDeleteOtherContainersCommand release];
	
	if ([self canUndoIt])
	{
		// if the command has been done or redone, we can safely delete the container
		// please note that since the command has been applied, the deletion is very simple :
		// there are no representations, and the parent is nil.
		
		// let's verify this.
		ZzAssertMsg( 
			(	([[fObjectToDelete getRepresentations] count] == 0) && 
				([fObjectToDelete parentObject] == nil) && 
				([fObjectToDelete WSObject]->GetPos().fParent == NULL)	),
				"Please save your project under a different name and restart HyperEngine-AV. Sorry for the inconvenience.");
		{
//			TArmeWSObj* obj = [fObjectToDelete WSObject];
//			delete obj;
		}
	}
	// anyway, the cmd had a ref on the objectToDelete and it doesn't need it anymore
	[fObjectToDelete release];
	
	[fMoveCommand release];
	[super dealloc];
}

-(id)initFromArchive:(ArmeArchiveHandler*)unarchiver
{
	self = [super initFromArchive:unarchiver];
	if(self)
	{
		fObjectToDelete = [unarchiver->ReadObjCObject() retain];
		fMemento = [unarchiver->ReadObjCObject() retain];
		fIsTheVideoSolo = unarchiver->ReadBool();
		fIsTheAudioSolo = unarchiver->ReadBool();
		fMoveCommand = [unarchiver->ReadObjCObject() retain];
		
		if (unarchiver->GetFileVer() >= 0x0001002C)
		{
			fDeleteOtherContainersCommand = [unarchiver->ReadObjCObject() retain];
		}
		else
		{
			fDeleteOtherContainersCommand = nil;
		}
	}
	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)archiver
{
	[super writeToArchive:archiver];
	archiver->WriteObjCObject(fObjectToDelete);
	archiver->WriteObjCObject(fMemento);
	archiver->WriteBool(fIsTheVideoSolo);
	archiver->WriteBool(fIsTheAudioSolo);
	archiver->WriteObjCObject(fMoveCommand);
	
	//from version 0x0001002C
	archiver->WriteObjCObject(fDeleteOtherContainersCommand);
	
	return YES;
}

- (id)initWithSelection:(MIWSSelection*)sel
{
	BOOL supported = [MIDeleteContainerCommand supportsSelection:sel];
	if(!supported)
	{
		ZzAssertMsg(supported,"The Delete Container command does not support this selection... You can safely continue.");
		return nil;
	}
	
	self = [super initWithSelection:nil];
	[self setObjectToDelete:[[sel objectAtIndex:0] representedObject]];
	
	if ([sel count] > 1)
	{
		fDeleteOtherContainersCommand = [[NSMutableArray alloc] initWithCapacity:([sel count] - 1)];
		for (unsigned int i=1; i < [sel count]; ++i)
		{
			[fDeleteOtherContainersCommand addObject:[[[MIDeleteContainerCommand alloc] initWithObjectToDelete:[[sel objectAtIndex:i] representedObject]] autorelease]];
		}
	}
	
	return self;
}

- (id)initWithObjectToDelete:(MIObject*)obj
{	
	self = [super initWithSelection:nil];
	if(self)
	{
		[self setObjectToDelete:obj];
	}
	return self;
}

- (void)setObjectToDelete:(MIObject*)obj
{
	fObjectToDelete = [obj retain];
	fIsTheVideoSolo = false;
	fIsTheAudioSolo = false;
	fMoveCommand = nil;
	//first, make a copy of what we are going to delete
	fMemento = [[fObjectToDelete getMemento] retain];

	if ([fObjectToDelete WSObject]->ContainsVideo(false, true))
		[self modifiedVideoHierarchy];
	if ([fObjectToDelete WSObject]->ContainsAudio())
		[self modifiedAudioHierarchy];
	[self overviewNeedsUpdate];
}
- (BOOL)displaySetupDialog
{
	return YES;
}

- (MICmdStatus)doIt
{
	MIMontageDocument	*doc;
	
	doc = [fObjectToDelete montageDoc];
	[fObjectToDelete invalidateRep];
	[fObjectToDelete setRepresentationsNeedDisplay];
	
	
	/***************************
	
	1 : let's move the children one step to the top
	
	***************************/
		
	// let's create the array holding the "top level children"
	NSMutableArray* topChildren = [[[NSMutableArray alloc] init] autorelease];
	
	NSEnumerator* childrenEnum = [[fObjectToDelete getChildren] objectEnumerator];
	MIObject* child;
	while(child = [childrenEnum nextObject])
	{
		if ([child WSObject]->GetType() != TArmeWSObj::kVideoXFadeEffect)
			[topChildren addObject:child];
	}
	
	// create a move command, init it with the top children, and tell the
	// command what the new parent is......
	fMoveCommand = [[MIMoveCommand alloc] initWithObjArray:topChildren];
	[fMoveCommand setNewParent:[fObjectToDelete parentObject]];
	
	[fMoveCommand _doIt];
										
	// now let's actually delete the container (set it outta hierarchy)
	[fObjectToDelete setOuttaHierarchy];

	fIsTheVideoSolo = [fObjectToDelete WSObject]->IsTheVideoSolo();

	if (fIsTheVideoSolo)
		[[fObjectToDelete montageDoc] armeDocument]->SetVideoSoloObject(NULL);
	fIsTheAudioSolo = [fObjectToDelete WSObject]->IsTheAudioSolo();
	if (fIsTheAudioSolo)
		[[fObjectToDelete montageDoc] armeDocument]->SetAudioSoloObject(NULL);

	for (unsigned int i = 0; i < [fDeleteOtherContainersCommand count]; ++i)
	{
		if ([[fDeleteOtherContainersCommand objectAtIndex:i] state] == kMICommandStateCanOnlyBeRedone)
			[[fDeleteOtherContainersCommand objectAtIndex:i] _redoIt];
		else
			[[fDeleteOtherContainersCommand objectAtIndex:i] _doIt];
	}

	[[MIApplicationController sharedApplicationController] modelChangedForDocument:doc];
		
	return kMICmdStatusDone;
}

- (void)undoIt
{
	for (int i = [fDeleteOtherContainersCommand count]-1; i >= 0 ; --i)
	{
		[[fDeleteOtherContainersCommand objectAtIndex:i] _undoIt];
	}

	[fObjectToDelete putBackInHierarchy];
	[fObjectToDelete setRepresentationsNeedDisplay];
	if (fIsTheVideoSolo)
		[[fObjectToDelete montageDoc] armeDocument]->SetVideoSoloObject([fObjectToDelete WSObject]);
	if (fIsTheAudioSolo)
		[[fObjectToDelete montageDoc] armeDocument]->SetAudioSoloObject([fObjectToDelete WSObject]);
		
	[fMoveCommand _undoIt];
	
	[fObjectToDelete setRepresentationsNeedDisplay];
	[[MIApplicationController sharedApplicationController] modelChangedForDocument:[fObjectToDelete montageDoc]];
}

- (void)redoIt
{
	[self doIt];
}

+ (BOOL)supportsSelection:(MIWSSelection*)sel
{
	// we return true only if the selection is a single container,
	// and if the whole object is selected
	if (([sel type] != kMIWSSelectionTypeObjects) || ([sel count] == 0))
	{
		return NO;
	}
	else
	{
		for (unsigned int i = 0; i < [sel count]; ++i)
		{
			TArmeWSObj* wsobj = [[[sel objectAtIndex:i] representedObject] WSObject];
			SInt32 type = wsobj->GetType();
			if(	(type != TArmeWSObj::kVideoEffect) && (type != TArmeWSObj::kAudioProcessor) && (type != TArmeWSObj::kContainer) )
				return NO;
		}
		return YES;
	}
}

@end

