/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/
#import "MIDeleteLowLevelCommand.h"
#import "MIObject.h"
#import "MIMontageDocument.h"
#include "TArmeDoc.h"
#import "MIApplicationController.h"
#import "MIMainViewController.h"
#import "MIApplication.h"
#import "MITopLevelContainerRep.h"

#ifndef TArmeWSObj_h
#include "TArmeWSObj.h"
#endif //TArmeWSObj_h

/****************************************************************************************************

MIDeleteLowLevelCommand

****************************************************************************************************/

@implementation MIDeleteLowLevelCommand
- (void)dealloc
{
	[fMemento release];
	
	if ([self canUndoIt])
	{
		// if the command can be undone, it means that has been done or redone; in this case,
		// the object is no longer visible in the document and, since the command is being deallocated now,
		// it will never be undone anymore so, we can definitely delete the object (if it has been deleted completely)
		if (fObjectHasBeenCompletelyDeleted)
			[fObjectToDelete deleteSelfAndChildren];
	}

	// anyway, the cmd had a ref on the objectToDelete and it doesn't need it anymore
#ifdef DEBUG
//	printf("MIDeleteLowLevelCommand about to release obj %p in [fObjectToDelete release]\n",(void*)fObjectToDelete);
#endif //#ifdef DEBUG
	[fObjectToDelete release];
	[super dealloc];
}

-(id)initFromArchive:(ArmeArchiveHandler*)unarchiver
{
	self = [super initFromArchive:unarchiver];
	if(self)
	{
		fMemento = [unarchiver->ReadObjCObject() retain];
		fObjectToDelete = [unarchiver->ReadObjCObject() retain];
		fStartTime = unarchiver->ReadArmeTime();
		fEndTime = unarchiver->ReadArmeTime();
		fIsTheVideoSolo = unarchiver->ReadBool();
		fIsTheAudioSolo = unarchiver->ReadBool();
		
		fObjectHasBeenCompletelyDeleted = unarchiver->ReadBool();
	}
	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)archiver
{
	[super writeToArchive:archiver];
	archiver->WriteObjCObject(fMemento);
	archiver->WriteObjCObject(fObjectToDelete);
	archiver->WriteArmeTime(fStartTime);
	archiver->WriteArmeTime(fEndTime);
	archiver->WriteBool(fIsTheVideoSolo);
	archiver->WriteBool(fIsTheAudioSolo);
	
	archiver->WriteBool(fObjectHasBeenCompletelyDeleted);
	return YES;
}


- (id)initWithObjectToDelete:(MIObject*)objectToDelete from:(ArmeTime)startTime to:(ArmeTime)endTime
{
	self = [super initWithSelection:nil];
#ifdef DEBUG
//	printf("MIDeleteLowLevelCommand about to retain obj %p in initWithObjectToDelete ([objectToDelete retain])\n",(void*)objectToDelete);
#endif //#ifdef DEBUG
	fObjectToDelete = [objectToDelete retain];
	fStartTime = startTime;
	fEndTime = endTime;
	fIsTheVideoSolo = false;
	fIsTheAudioSolo = false;
	//first, make a copy of what we are going to delete
	fMemento = [[fObjectToDelete getMemento] retain];

	if ([fObjectToDelete WSObject]->ContainsVideo(false, true))
		[self modifiedVideoHierarchy];
	if ([fObjectToDelete WSObject]->ContainsAudio())
		[self modifiedAudioHierarchy];
	[self overviewNeedsUpdate];


	return self;
}

- (BOOL)displaySetupDialog
{
	return YES;
}

- (MICmdStatus)doIt
{
	MIMontageDocument	*doc;
	MIRect				fr;
	
	doc = [fObjectToDelete montageDoc];
	[fObjectToDelete invalidateRep];
	fr = [fObjectToDelete modelFrame];
	
//	[fObjectToDelete setRepresentationsNeedDisplay];

	if ([fObjectToDelete WSObject]->DeleteSlice(fStartTime, fEndTime, [MICommand getCurrentProgressDialog]) == YES)
	{
		//if the WSObject has been completely deleted,
		//delete the MIObject also
		[fObjectToDelete setOuttaHierarchy];
		fObjectHasBeenCompletelyDeleted = YES;
		
		// if the object we have deleted was the solo object, 
		fIsTheVideoSolo = [fObjectToDelete WSObject]->IsTheVideoSolo();
		if (fIsTheVideoSolo)
			[[fObjectToDelete montageDoc] armeDocument]->SetVideoSoloObject(NULL);
		fIsTheAudioSolo = [fObjectToDelete WSObject]->IsTheAudioSolo();
		if (fIsTheAudioSolo)
			[[fObjectToDelete montageDoc] armeDocument]->SetAudioSoloObject(NULL);
	}
	else
	{
		NSRect	r;
		[fObjectToDelete recreateFrame];
		[fObjectToDelete invalidateRepresentationsCachedImages];
		[fObjectToDelete invalidateRep];
		[fObjectToDelete setRepresentationsNeedDisplay];
		
		r = [[NSApp mainViewController] MIRectToNSRect:[[fObjectToDelete parentObject] convertRectToTLCCoords:fr]];
		[[[NSApp mainViewController] topLevelContainerView] setNeedsDisplayInRect:r];
		
		[[NSApp mainViewController] clearSelection];
		fObjectHasBeenCompletelyDeleted = NO;
		[fObjectToDelete markModified];
	}
	
	
	
	[[MIApplicationController sharedApplicationController] modelChangedForDocument:doc];
	return kMICmdStatusDone;
}

- (void)undoIt
{
	MIMontageDocument	*doc = [fObjectToDelete montageDoc];

	if (fObjectHasBeenCompletelyDeleted)
	{
		[fObjectToDelete putBackInHierarchy];
		[fObjectToDelete setRepresentationsNeedDisplay];
		if (fIsTheVideoSolo)
			[[fObjectToDelete montageDoc] armeDocument]->SetVideoSoloObject([fObjectToDelete WSObject]);
		if (fIsTheAudioSolo)
			[[fObjectToDelete montageDoc] armeDocument]->SetAudioSoloObject([fObjectToDelete WSObject]);
	}
	else
	{
		//update the MIObject
		[fObjectToDelete setMemento:fMemento];

		[fObjectToDelete invalidateRep];
		[fObjectToDelete recreateFrame];
		[fObjectToDelete invalidateRepresentationsCachedImages];
		[fObjectToDelete invalidateRep];
	}
	[fObjectToDelete setRepresentationsNeedDisplay];
	[[MIApplicationController sharedApplicationController] modelChangedForDocument:doc];
}

- (void)redoIt
{
	[self doIt];
}

+ (BOOL)supportsSelection:(MIWSSelection*)sel
{
	//this function should never be called on a low level command
	return NO;
}

@end

