/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import "MIDuplicateManyCommand.h"
#import "MIDuplicateManyController.h"
#import "MICopyCommand.h"
#import "MITime.h"
#import "MIObjectRep.h"
#import "MIMontageDocument.h"
#import "MIObject.h"
#import "MIPasteLowLevelCommand.h"
#import "MIApplicationController.h"
#import "MIInsertLowLevelCommand.h"

#ifndef TArmeDoc_h
#include "TArmeDoc.h"
#endif //TArmeDoc_h

/****************************************************************************************************

MIDuplicateManyCommand

****************************************************************************************************/

@implementation MIDuplicateManyCommand

- (void)dealloc
{
	[fClipboard release];
	[super dealloc];
}

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
	self = [super initFromArchive:handler];
	fMontageDoc = handler->ReadObjCObject();

	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
	[super writeToArchive:handler];
	handler->WriteObjCObject(fMontageDoc);

	return YES;
}

- (NSString*)commandName
{
	return @"Duplicate Many";
}

+ (BOOL)supportsSelection:(MIWSSelection*)sel
{
	if (!sel) return NO;
	
	// you can't duplicate video transitions
	if ([sel containsOnlyVideoTransitions])
		return NO;

	if ([sel count] == 0)
		return NO;

	if ( [sel type] == kMIWSSelectionTypeObjects
#ifdef AUTOMATION_DATA
		|| [sel type] == kMIWSSelectionTypeAutomationPoints
#endif
		 )
		return YES;
	else if ( [sel type] == kMIWSSelectionTypeTimeSlice )
	{
		if ([sel duration] > ARME_TIME_ZERO)
			return YES;
	}
	return NO;
}

- (id)initWithSelection:(MIWSSelection*)sel
{
	self = [super initWithSelection:sel];
	[fSelection discardVideoTransitions];

	fMontageDoc = [[[fSelection objectAtIndex:0] representedObject] montageDoc];
	fContainerObject = nil;
	
	return self;
}

- (id)initWithSelectionForSimpleDuplicate:(MIWSSelection*)sel
{
	self = [super initWithSelection:sel];

	fMontageDoc = [[[sel objectAtIndex:0] representedObject] montageDoc];

	fNumberOfDuplications = 1;

	if ([sel type] == kMIWSSelectionTypeTimeSlice)
	{
		fShouldKeepInsideContainer = YES;
		fHorizontalEvery = NO;
		fHorizontalImmediately = YES;
	
		fHorizontalTimeInc = ARME_TIME_ZERO;
	
		fVerticalImmediately = YES;
		fVerticalPixels = NO;
		fPixels = 0;
	
		fContainerObject = nil;
	}
	else
	{
		fShouldKeepInsideContainer = YES;
		fHorizontalEvery = NO;
		fHorizontalImmediately = YES;
	
		fHorizontalTimeInc = ARME_TIME_ZERO;
	
		fVerticalImmediately = NO;
		fVerticalPixels = YES;
		fPixels = 0;
	
		fContainerObject = nil;
	}
	
	return self;
}

- (int)getNumberOfDuplications {return [self getNbSubCommands];}; // note : fNumberOfDuplications is set to zero in createSubCommands
- (ArmeTime)getHorizontalTimeInc {return fHorizontalTimeInc;};
- (BOOL)isChainedHorizontaly {return fHorizontalImmediately;};
- (BOOL)isStackedVerticaly {return fVerticalImmediately;};
- (int)getVerticalPixelsInc {return fPixels;};
- (BOOL)isKeepInsideContainer {return fShouldKeepInsideContainer;};

- (BOOL)displaySetupDialog
{
	MIDuplicateManyController* dlog = nil;
	
	if ([fSelection type] == kMIWSSelectionTypeObjects)
		dlog = [MIDuplicateManyController duplicateManyPanel:YES];
	else
		dlog = [MIDuplicateManyController duplicateManyPanel:NO];
	
	if ( [dlog runModal] == NSOKButton ) 
	{
		fNumberOfDuplications = [dlog numberOfDuplications];
		fShouldKeepInsideContainer = [dlog shouldKeepInsideContainer];

		fHorizontalEvery = [dlog horizontalEvery];
		fHorizontalImmediately = [dlog horizontalImmediately];

		MITime* time = [dlog time];
		fHorizontalTimeInc = [time getArmeTime];

		fVerticalImmediately = [dlog verticalImmediately];
		fVerticalPixels = [dlog verticalNone];
		fPixels = 0;

		[self createSubCommands];

		return YES;
	}
	return NO;
}


-(void)createSubCommands
{
	// init the command with these parameters
	NSMutableArray* objectsToDuplicate = [[[NSMutableArray alloc] initWithCapacity:1] autorelease];
	NSArray*	objectsInSelection = [fSelection putRepsInArrayAndRemoveDescendants];
	MIVPos		smallestVPos = VPOS_MAX;
	MIVPos		highestVPos = VPOS_MIN;
	
	for (unsigned int i=0; i < [objectsInSelection count]; ++i)
	{
		[objectsToDuplicate addObject:[[objectsInSelection objectAtIndex:i] representedObject]];

		if ([[objectsToDuplicate objectAtIndex:i] getAbsoluteVerticalPos] < smallestVPos)
			smallestVPos = [[objectsToDuplicate objectAtIndex:i] getAbsoluteVerticalPos];

		if (([[objectsToDuplicate objectAtIndex:i] getAbsoluteVerticalPos] + [[objectsToDuplicate objectAtIndex:i] getHeight]) > highestVPos)
			highestVPos = [[objectsToDuplicate objectAtIndex:i] getAbsoluteVerticalPos] + [[objectsToDuplicate objectAtIndex:i] getHeight];
	}
	
	ArmeTime 	timeInc;
	MIVPos		vposInc;
	ArmeTime 	pastePositionH;
	MIVPos		pastePositionV;
	
	if (fHorizontalEvery) //every X HH::MM::SS::FF
		timeInc = fHorizontalTimeInc;
	else	//every selection length sec
		timeInc = [fSelection length];
	
	if (fVerticalImmediately)	// every selection height sec
		vposInc = smallestVPos - highestVPos;
	else
		GET_VALUE(vposInc) = -fPixels;
		
	if ([fSelection type] == kMIWSSelectionTypeTimeSlice)
		pastePositionH = [fSelection startTime] + timeInc;
	else
		pastePositionH = [[fSelection getLeftMostObject] WSObject]->GetAbsoluteStartTime() + timeInc;

	pastePositionV = smallestVPos + vposInc;

	MIObject* pasteInObject;
	if ( fContainerObject )
		pasteInObject = fContainerObject;
	else if (fShouldKeepInsideContainer)
	{
		if ([fSelection type] == kMIWSSelectionTypeTimeSlice)
			pasteInObject = [objectsToDuplicate objectAtIndex:0];
		else
			pasteInObject = [[objectsToDuplicate objectAtIndex:0] parentObject];
	}
	else 
	{
		pasteInObject = [fMontageDoc topLevelContainer];
	}

	//create the clipboard
	fClipboard = [[MIObject alloc] initWithArmeWSObj:(new TArmeContainer("temp clipboard for Duplicate"))];

	// copy the selected part
	MICopyCommand* copy = [[[MICopyCommand alloc] initWithSelection:fSelection usingClipboard:fClipboard] autorelease];
	[copy _doIt];

	int duplicationIndex = 1; // the suffix we add to the object name is built using this var
	bool onlyOneDup = false;
	if(fNumberOfDuplications == 1)
		onlyOneDup = true;
		
	while (fNumberOfDuplications > 0)
	{
		if (([fSelection type] == kMIWSSelectionTypeTimeSlice) && (fShouldKeepInsideContainer))
		{
		
			[self addCommand:[[[MIInsertLowLevelCommand alloc] initWithSelection:fSelection
								objectToInsert:[[fClipboard getChildren] objectAtIndex:0]
								atPosition:pastePositionH 
								inObject:pasteInObject] autorelease]];
		}
		else
		{
			SInt32 suffix; // it's an enum
			if(onlyOneDup)
			{
				suffix = kSuffixCopy;
			}
			else
			{
				suffix = kSuffixNumberZero+duplicationIndex;
			}
				
			MIPasteLowLevelCommand* cmd = [[[MIPasteLowLevelCommand alloc] initWithDocument:fMontageDoc
				objectsToPaste:[fClipboard getChildren]
				inObject:pasteInObject
				atTime:(pastePositionH - [pasteInObject WSObject]->GetAbsoluteStartTime())
				atVPos:pastePositionV] autorelease];
			[cmd setSuffix:suffix];
			[self addCommand:cmd];
			duplicationIndex++;
		}
		pastePositionH += timeInc;
		pastePositionV += vposInc;

		--fNumberOfDuplications;
	}
}
@end


