/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import "MIEncloseSelectionCommand.h"
#import "MIResizeCommand.h"
#import "MIMoveCommand.h"
#import "MISingleMoveCommand.h"

#ifndef ArmeArchiveHandler_h
#include "ArmeArchiveHandler.h"
#endif //ArmeArchiveHandler_h

#import "MIObject.h"
#import "MIObjectRep.h"
#import "MIMainViewController.h"
#import "MITopLevelContainerRep.h"

/****************************************************************************************************

MIEncloseSelectionCommand

****************************************************************************************************/

@implementation MIEncloseSelectionCommand : MICommand
- (void)dealloc
{
	[fResizeSelfCommand release];
	[fMoveSelectionCommand release];
	[fMoveSelfInParentCommand release];
	[super dealloc];
}

- (NSString*)commandName
{
	return @"EncloseSelection";
}

- (id)initWithSelection:(MIWSSelection*)sel
{
	self = [super initWithSelection:sel];
	if ( self )
	{
		[fSelection discardVideoTransitions];

		fResizeSelfCommand = nil;
		fMoveSelectionCommand = nil;
		fMoveSelfInParentCommand = nil;

		[self modifiedVideoHierarchy];
		[self modifiedAudioHierarchy];
		[self overviewNeedsUpdate];
	}
	return self;
}

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
    if (( self = [super initFromArchive:handler] ))
    {
		fResizeSelfCommand = [handler->ReadObjCObject() retain];
		fMoveSelectionCommand = [handler->ReadObjCObject() retain];
		fMoveSelfInParentCommand = [handler->ReadObjCObject() retain];
    }
    return self ;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
    if ( [super writeToArchive:handler] )
    {
		handler->WriteObjCObject(fResizeSelfCommand);
		handler->WriteObjCObject(fMoveSelectionCommand);
		handler->WriteObjCObject(fMoveSelfInParentCommand);
		return YES;
    }
    else
		return NO;
}

- (void)setContainer:(MIObject*)container
{
	fContainer = container;
}

+ (BOOL)supportsSelection:(MIWSSelection*)sel
{
	if ( [sel containsOnlyVideoTransitions])
		return NO;

	if ( [sel type] != kMIWSSelectionTypeObjects )
		return NO;
	else
		return YES;	
}

- (MICmdStatus)doIt
{
	NSArray*		selArray;
	NSEnumerator	*objEnum;
	MIObject		*parentObject = nil;
	MIObjectRep		*aRep;
	MIRect			enclosingRect = MIZeroRect;
//	MICommand		*cmd;

	// can do this only if self is a container
	if ( ![fContainer isContainer] )
		return kMICmdStatusCancelled;

	// can do this only if object doesn't already have children
	if ( [[fContainer getChildren] count] )
		return kMICmdStatusCancelled;

	selArray = [fSelection putRepsInArrayAndRemoveDescendants];

	if ( ![selArray count] )
		return kMICmdStatusCancelled;

	// can do this only if all the objects share the same parent
	objEnum = [selArray objectEnumerator];

	while ( aRep = [objEnum nextObject] )
	{
		if ( MIEqualRects(enclosingRect, MIZeroRect) )
		{
			enclosingRect = [[aRep representedObject] modelFrame];
		}
		else
		{
			enclosingRect = MIUnionRect(enclosingRect, [[aRep representedObject] modelFrame] );
		}

		if ( !parentObject )
			parentObject = [[aRep representedObject] parentObject];
		else if ( parentObject != [[aRep representedObject] parentObject] )
			return kMICmdStatusCancelled;
	}


	// create a resize command to become of the appropriate size
	// create a move command to move the current selection in self
	// create a singleMove command to move the container in the new parent

	// set my frame to the frame we just calculated
	// [[self representedObject] setModelFrameFromRect:enclosingRect inTLC:[self getTopLevelContainerRep]];
	fResizeSelfCommand = [[MIResizeCommand alloc] initWithSelection:nil];
	[fResizeSelfCommand setObject:fContainer newFrame:enclosingRect];
	[fResizeSelfCommand _doIt];
	
	// insert ourselves as child of the old parent
	if (  parentObject && ![parentObject isTLC] )
	{
		fMoveSelfInParentCommand = [[MISingleMoveCommand alloc] initWithSelection:nil];
		[fMoveSelfInParentCommand setObject:fContainer];
		[fMoveSelfInParentCommand setNewParent:parentObject];
		[fMoveSelfInParentCommand setNewOrigin:enclosingRect.origin];
		[fMoveSelfInParentCommand _doIt];
	}

	// move each object in self
	fMoveSelectionCommand = [[MIMoveCommand alloc] initWithSelection:fSelection];
	[fMoveSelectionCommand setOffset:MIZeroPoint];
	[fMoveSelectionCommand setNewParent:fContainer];

	[fMoveSelectionCommand _doIt];

	// make sure we're big enough to hold our children
	//fShrinkToFitCommand = [[MIShrinkToFitCommand alloc] initWithSelection:nil];
	//[fShrinkToFitCommand setObject:fContainer];
	//[fShrinkToFitCommand _doIt];
	
	return kMICmdStatusDone;
}

- (void)undoIt
{
	[fMoveSelfInParentCommand _undoIt];
	[fMoveSelectionCommand _undoIt];
	[fResizeSelfCommand _undoIt];
}

- (void)redoIt
{
	[fResizeSelfCommand _redoIt];
	[fMoveSelectionCommand _redoIt];
	[fMoveSelfInParentCommand _redoIt];
}
@end


