/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import "MIFreezeCommand.h"
#import "MIObject.h"
#import "MIObjectRep.h"
#import "MIMainViewController.h"
#import "MISoloCommand.h"
#import "MIMontageDocument.h"
#import "MIApplicationController.h"
#import "MIUpdateVideoTransitionsCommand.h"
#import "MIResizeCommand.h"

#ifndef TArmeMediaEngine_h
#include "TArmeMediaEngine.h"
#endif //TArmeMediaEngine_h

#ifndef TVideoClip_h
#include "TVideoClip.h"
#endif //TVideoClip_h

#ifndef Zz_PathHelpers_h
#include "Zz_PathHelpers.h"
#endif //Zz_PathHelpers_h

#ifndef MEAudioClip_h
#include "MEAudioClip.h"
#endif //MEAudioClip_h

#ifndef ArmeTimeRange_h
#include "ArmeTimeRange.h"
#endif //ArmeTimeRange_h

#ifndef MovieHelper_h
#include "MovieHelper.h"
#endif //MovieHelper_h

#ifndef ZzHelpers_h
#include "ZzHelpers.h"
#endif //ZzHelpers_h

#ifndef MIProgressBarDlgManager_h
#include "MIProgressBarDlgManager.h"
#endif //MIProgressBarDlgManager_h


/****************************************************************************************************

MIFreezeCommand

****************************************************************************************************/

@implementation MIFreezeCommand

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
	[super initFromArchive:handler];
	fPreviousVideoSoloObject = dynamic_cast<TArmeWSObj*>(handler->ReadCPPObject());
	fPreviousAudioSoloObject = dynamic_cast<TArmeWSObj*>(handler->ReadCPPObject());
	fMontageDoc = handler->ReadObjCObject();
	if (handler->GetFileVer() >= 0x00010016)
		fIsRenderCommand = handler->ReadBOOL();
	else
		fIsRenderCommand = YES;
	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
	[super writeToArchive:handler];
	handler->WriteCPPObject(fPreviousVideoSoloObject);
	handler->WriteCPPObject(fPreviousAudioSoloObject);
	handler->WriteObjCObject(fMontageDoc);
	
	// from 0x00010016
	handler->WriteBOOL(fIsRenderCommand);
	return YES;
}

- (void)dealloc
{
	[super dealloc];
}

- (NSString*)commandName
{
	if (fIsRenderCommand)
		return @"Render Clip";
	else
		return @"Un-Render Clip";
}

- (id)initWithSelection:(MIWSSelection*)sel
{
	self = [super initWithSelection:sel];
	
	if ( self )
	{
		[fSelection discardVideoTransitions];

		fMontageDoc = [[NSDocumentController sharedDocumentController] currentDocument];
		
		NSArray* repArray = [fSelection putRepsInArrayAndRemoveDescendants];
		NSEnumerator* objRepEnum = [repArray objectEnumerator];
		MIObjectRep* rep;
		MIObject* object;
		while(rep = [objRepEnum nextObject])
		{
			object = [rep representedObject];
			if ([object freezeSource] && ![object hasBeenModified]) // the object is already a frozen object, we need to unfreeze it
			{
				[self addCommand:[[[MIUnFreezeLowLevelCommand alloc] initWithObjectToUnFreeze:object] autorelease]];
				fIsRenderCommand = NO;
			}
			else // the object needs to be frozen
			{
				[self addCommand:[[[MIFreezeLowLevelCommand alloc] initWithObjectToFreeze:object] autorelease]];
				fIsRenderCommand = YES;
			}
		}
	
		fPreviousVideoSoloObject = [fMontageDoc armeDocument]->GetVideoSoloObject();
		fPreviousAudioSoloObject = [fMontageDoc armeDocument]->GetAudioSoloObject();

		if ([fSelection containsVideo:true andTransparentTextClip:NO])
			[self modifiedVideoHierarchy];
		if ([fSelection containsAudio])
			[self modifiedAudioHierarchy];
	}
	return self;
}

- (BOOL)displaySetupDialog
{
	return YES;
}

- (MICmdStatus)doIt
{
	MICmdStatus st = [super doIt];
/*	if (st == kMICmdStatusCancelled)
	{
		NSRunAlertPanel(@"Error", @"Error during the render operation.  Command Failed.", @"OK", nil, nil);
	}*/
	
	return st;
}

- (void)undoIt
{
	[super undoIt];
	
	[fMontageDoc armeDocument]->SetVideoSoloObject(fPreviousVideoSoloObject);
	[fMontageDoc armeDocument]->SetAudioSoloObject(fPreviousAudioSoloObject);
}

- (void)redoIt
{
	[super redoIt];
}


+ (BOOL)supportsSelection:(MIWSSelection*)sel
{
	if ( [sel containsOnlyVideoTransitions])
		return NO;

	if ( [sel count] < 1)
		return NO;
	if ( [sel type] == kMIWSSelectionTypeObjects )
	{
		return ([MIFreezeCommand isRenderCommand:sel] || [MIFreezeCommand isUnRenderCommand:sel]);
	
	}
	else
		return NO;
}
+ (BOOL)isRenderCommand:(MIWSSelection*)sel
{
	NSEnumerator* selEnum = [sel objectEnumerator];
	MIObject* selObj;
	while (selObj = [[selEnum nextObject] representedObject])
	{
		switch([selObj WSObject]->GetType())
		{
			//any single clip can not be rendered
			case TArmeWSObj::kVideoClip : 
			case TArmeWSObj::kAudioClip :
			case TArmeWSObj::kTextClip : 
				return NO;
				break;
			case TArmeWSObj::kAudioProcessor :
			case TArmeWSObj::kContainer :
			case TArmeWSObj::kVideoEffect :
				//if an audio Processor does not contain audio or video, reject the command
				if (!([selObj WSObject]->ContainsAudio() || [selObj WSObject]->ContainsVideo(false, true)))
					return NO;
				// if one of the group is the result of a render, it cannot be rendered anymore
				if ([selObj freezeSource] && ![selObj hasBeenModified])
					return NO;
				break;
			case TArmeWSObj::kVideoXFadeEffect :
				break;
		}
	}
	return YES;
}
+ (BOOL)isUnRenderCommand:(MIWSSelection*)sel
{
	NSEnumerator* selEnum = [sel objectEnumerator];
	MIObject* selObj;
	while (selObj = [[selEnum nextObject] representedObject])
	{
		switch([selObj WSObject]->GetType())
		{
			case TArmeWSObj::kVideoClip : 
			case TArmeWSObj::kAudioClip :
			case TArmeWSObj::kContainer :
				//if the clip is the result of a render and it has not been modified, the command can work otherwise not
				if (!([selObj freezeSource] && ![selObj hasBeenModified]))
					return NO;
				break;
			case TArmeWSObj::kTextClip : 
			case TArmeWSObj::kAudioProcessor :

			case TArmeWSObj::kVideoEffect :
				// a group can not be the result of a render operation
				return NO;
			case TArmeWSObj::kVideoXFadeEffect :
				break;
		}
	}
	return YES;
}


@end

@implementation MIBaseFreezeLowLevelCommand

- (void)toggleToSourceObject:(MIObject*)object
{
	// Note : some commands like insertSlice, deleteSlice, ... will mark the object as modified
	[[object freezeSource] cleanModifiedFlag];

	[[object freezeSource] putBackInHierarchy];
	[object setOuttaHierarchy];

	[[object freezeSource] setRepresentationsNeedDisplay];
}

- (void)toggleToFrozenObject:(MIObject*)object
{
	// the object had already been frozen before and has not been modified since : we can reuse the old frozen object
	[[object frozenObject] putBackInHierarchy];
	[[object frozenObject] cleanModifiedFlag];
	[object setOuttaHierarchy];
	[[object frozenObject] setRepresentationsNeedDisplay];
}

@end

@implementation MIFreezeLowLevelCommand

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
	self = [super initFromArchive:handler];
	if(self)
	{
		fObjectToFreeze = [handler->ReadObjCObject("MIObject") retain];
		fFreezingResultObject = [handler->ReadObjCObject("MIObject") retain];
	
		//not serialized members :
		fRenderReverse = NO;
		fRenderUsingCustomAlgo = NO;
	}
	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
	[super writeToArchive:handler];
	handler->WriteObjCObject(fObjectToFreeze);
	handler->WriteObjCObject(fFreezingResultObject);
	return YES;
}

- (void)dealloc
{
	[fObjectToFreeze release];
	[fFreezingResultObject release];
	[super dealloc];
}

- (NSString*)commandName
{
	return @"Freeze Selection (Low Level)";
}

- (void)setRenderReverse
{
	fRenderReverse = YES;
}

- (id)initWithObjectToFreeze:(MIObject*)object
{
	self = [super initWithSelection:nil];
	fObjectToFreeze = [object retain];
	fFreezingResultObject = nil;
	fRenderReverse = NO;
	fRenderUsingCustomAlgo = NO;
	return self;
}

- (BOOL)displaySetupDialog
{
	return YES;
}

- (int)getNbLengthyTasks
{
	return 1;
}

- (MICmdStatus)doIt
{
	MIObject* object = fObjectToFreeze;

	// the object has already been frozen before and it has not been modified since
/*	NOT for version 1.0
	if ([object frozenObject] && ![object hasBeenModified])
	{
		[self toggleToFrozenObject:object];
		
		fFreezingResultObject = [object frozenObject];
	}
	else*/
	{
		// if the solo object is a descendant of this object, set the solo object to NULL
		TArmeWSObj* currentVideoSolo = [object WSObject]->GetDoc()->GetVideoSoloObject();
		if (currentVideoSolo && currentVideoSolo->IsDescendantOf([object WSObject]) && ![object WSObject]->IsTheVideoSolo())
			[[object montageDoc] armeDocument]->SetVideoSoloObject(NULL);
		TArmeWSObj* currentAudioSolo = [object WSObject]->GetDoc()->GetAudioSoloObject();
		if (currentAudioSolo && currentAudioSolo->IsDescendantOf([object WSObject]) && ![object WSObject]->IsTheAudioSolo())
			[[object montageDoc] armeDocument]->SetAudioSoloObject(NULL);
	
		// first, release the old frozen object attached to this object
		[[object frozenObject] release];
		[object cleanModifiedFlag];
	
		MIObject* frozenObjectsParent = [object parentObject];
		MIVPos videoVPos = [object getVerticalPos];
		MIVPos audioVPos = [object getVerticalPos];
		
		// if the object is a clip, there is nothing to do (unless it is a reverse)
/*		if ((([object WSObject]->GetType() == TArmeWSObj::kVideoClip)
				|| ([object WSObject]->GetType() == TArmeWSObj::kAudioClip)
				|| ([object WSObject]->GetType() == TArmeWSObj::kTextClip)) && !fRenderReverse)
			return kMICmdStatusCancelled;*/

		NSMutableArray* objectsToFreeze = [[[NSMutableArray alloc] initWithCapacity:1] autorelease];
		[objectsToFreeze addObject:object];
		
		ArmeErr err;
		fFreezingResultObject = [Render(err, objectsToFreeze, frozenObjectsParent, videoVPos, audioVPos, [MICommand getCurrentProgressDialog], false, ARME_TIME_UNDEF, ARME_TIME_UNDEF, false, fRenderReverse, fRenderUsingCustomAlgo) retain];
		if ((fFreezingResultObject == nil) || (err.type != kArmeOK))
		{
			[fFreezingResultObject release];
			return kMICmdStatusCancelled;
		}

		[fFreezingResultObject setFreezeSource:object];
		[object setFrozenObject:fFreezingResultObject];

		//remove the object to freeze from hierarchy
		[object setOuttaHierarchy];

		//if the solo object was in the selection, it has been frozen, we need to update this
		if ([object WSObject]->IsTheVideoSolo())
			[[object montageDoc] armeDocument]->SetVideoSoloObject([[object frozenObject] WSObject]);
		if ([object WSObject]->IsTheAudioSolo())
			[[object montageDoc] armeDocument]->SetAudioSoloObject([[object frozenObject] WSObject]);
	}
	return kMICmdStatusDone;
}

- (void)undoIt
{
	[self toggleToSourceObject:fFreezingResultObject];
}

- (void)redoIt
{
	if (!fObjectToFreeze->fFrozenObject) // it can happen if the render was in a macro command and the undo of another command has delete the fFrozenObject
	{
		[self doIt];
	}
	else
	{
		//if the solo object was in the selection, it has been frozen, we need to update this
		if ([fObjectToFreeze WSObject]->IsTheVideoSolo())
		{
			[[fObjectToFreeze montageDoc] armeDocument]->SetVideoSoloObject([[fObjectToFreeze frozenObject] WSObject]);
		}
		else if ([fObjectToFreeze WSObject]->GetDoc()->GetVideoSoloObject() && [fObjectToFreeze WSObject]->GetDoc()->GetVideoSoloObject()->IsDescendantOf([fObjectToFreeze WSObject]))
		{
			[[fObjectToFreeze montageDoc] armeDocument]->SetVideoSoloObject(NULL);
		}
		if ([fObjectToFreeze WSObject]->IsTheAudioSolo())
		{
			[[fObjectToFreeze montageDoc] armeDocument]->SetAudioSoloObject([[fObjectToFreeze frozenObject] WSObject]);
		}
		else if ([fObjectToFreeze WSObject]->GetDoc()->GetAudioSoloObject() && [fObjectToFreeze WSObject]->GetDoc()->GetAudioSoloObject()->IsDescendantOf([fObjectToFreeze WSObject]))
		{
			[[fObjectToFreeze montageDoc] armeDocument]->SetAudioSoloObject(NULL);
		}
	
		[self toggleToFrozenObject:fObjectToFreeze];
	}
}

@end

@implementation MIUnFreezeLowLevelCommand

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
	[super initFromArchive:handler];
	
	fObjectToUnFreeze = [handler->ReadObjCObject() retain];
	fParent = [handler->ReadObjCObject() retain];
	fStartTime = handler->ReadArmeTime();
	GET_VALUE(fVerticalPos) = handler->ReadSInt32();
	
	if (handler->GetFileVer() >= 0x00010029)
		fEnlargeCommand = [handler->ReadObjCObject() retain];
	else
		fEnlargeCommand = nil;
	
	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
	[super writeToArchive:handler];
	handler->WriteObjCObject(fObjectToUnFreeze);
	handler->WriteObjCObject(fParent);
	handler->WriteArmeTime(fStartTime);
	handler->WriteSInt32(GET_VALUE(fVerticalPos));
	
	//from version 0x00010029
	handler->WriteObjCObject(fEnlargeCommand);
	return YES;
}

- (void)dealloc
{
	[fObjectToUnFreeze release];
	[fParent release];
	[fEnlargeCommand release];
	[super dealloc];
}

- (NSString*)commandName
{
	return @"UnFreeze Selection (Low Level)";
}

- (id)initWithObjectToUnFreeze:(MIObject*)object
{
	self = [super initWithSelection:nil];
	fObjectToUnFreeze = [object retain];
	fParent = nil;
	fEnlargeCommand = nil;
	
	return self;
}

- (BOOL)displaySetupDialog
{
	return YES;
}

- (MICmdStatus)doIt
{
	if (fEnlargeCommand)
	{
		[fEnlargeCommand release];
		fEnlargeCommand = nil;
	}


	//if the solo object was in the selection, it has been frozen, we need to update this
	if ([fObjectToUnFreeze WSObject]->IsTheVideoSolo())
		[[fObjectToUnFreeze montageDoc] armeDocument]->SetVideoSoloObject([[fObjectToUnFreeze freezeSource] WSObject]);
	if ([fObjectToUnFreeze WSObject]->IsTheAudioSolo())
		[[fObjectToUnFreeze montageDoc] armeDocument]->SetAudioSoloObject([[fObjectToUnFreeze freezeSource] WSObject]);

	//we are going to move the source object to the same pos as the frozen object
	MIObject* parent = [fObjectToUnFreeze parentObject];
	ArmeTime startTime = [fObjectToUnFreeze WSObject]->GetPos().fStartTime;
	MIVPos vPos = [fObjectToUnFreeze getVerticalPos];

	[self toggleToSourceObject:fObjectToUnFreeze];
	
	//save the old pos of the source object (for undo)
	fParent = [[[fObjectToUnFreeze freezeSource] parentObject] retain];
	fStartTime = [[fObjectToUnFreeze freezeSource] WSObject]->GetPos().fStartTime;
	fVerticalPos = [[fObjectToUnFreeze freezeSource] getVerticalPos];
		
	[[fObjectToUnFreeze freezeSource] moveToObject:parent atTime:startTime withVPos:vPos];
	
	// check if parent size is big enough
	if ( !MIContainsRect( [parent modelFrame], [parent minimalRectToFitContent] ) )
	{
		fEnlargeCommand = [[MIResizeCommand alloc] initWithSelection:nil];
		[fEnlargeCommand setObject:parent newFrame:MIUnionRect([parent modelFrame], [parent minimalRectToFitContent])];
		[fEnlargeCommand _doIt];
	}
	
	return kMICmdStatusDone;
}

- (void)undoIt
{
	if (fEnlargeCommand)
		[fEnlargeCommand _undoIt];

	[[fObjectToUnFreeze freezeSource] moveToObject:fParent atTime:fStartTime withVPos:fVerticalPos];

	[self toggleToFrozenObject:[fObjectToUnFreeze freezeSource]];
}

- (void)redoIt
{
	[self doIt];
}

@end


MIObject* getLeftMostObjectInArray(NSArray* array)
{
	NSEnumerator* 	objEnumerator = [array objectEnumerator];
	MIObject* 		obj;
	ArmeTime 		leftMostStartTime = ARME_TIME_MAX;
	MIObject*		leftMostObject = nil;

	// no sense to call this function if there are no object in the array
	ZzAssert([array count] > 0);

	while(obj = [objEnumerator nextObject])
	{
		if ([obj WSObject]->GetAbsoluteStartTime() < leftMostStartTime)
		{
			leftMostObject = obj;
			leftMostStartTime = [leftMostObject WSObject]->GetAbsoluteStartTime();
		}
	}
	
	return leftMostObject;
}

extern void GetMutedStatusBeforeCommand(MIObject* obj, ZzArray<SInt32>* mutedStatusBeforeCommand);
extern void SetMutedStatusAsTheyWereBeforeCommand(MIObject* obj, ZzArray<SInt32>* mutedStatusBeforeCommand, int& index);


/*
This function is used by the Freeze and the Merge commands
*/
MIObject* Render(ArmeErr &err, NSArray* objectsToFreeze, MIObject* frozenObjectsParent, MIVPos videoVPos, MIVPos audioVPos, MIProgressBarDlgManager* progressMngr, bool bConstantMixTransition, ArmeTime startTime, ArmeTime endTime, bool onlyVideo, bool reverse, bool useCustomAlgo, bool stillClipToScale)
{
	unsigned int i=0;
	bool bSelectionContainsVideo = false;
	bool bSelectionContainsAudio = false;
	TArmeDoc* armeDoc = [[objectsToFreeze objectAtIndex:0] WSObject]->GetDoc();
	MIMontageDocument* montageDoc = [[objectsToFreeze objectAtIndex:0] montageDoc];
	const char* newName = [[objectsToFreeze objectAtIndex:0] WSObject]->GetName();
	TArmeWSObjPos pos = [getLeftMostObjectInArray(objectsToFreeze) WSObject]->GetPos(); 
	bool bCreatedContainer = false;
	MIObject* fFreezingResultObject = NULL;
	TArmeContainer* container = NULL;

	MIObject* miContainer = nil;
	MIObject* miAudioClip = nil;
	MIObject* miVideoClip = nil;

	ZzAssert(progressMngr);  //you need a progress bar dlog

	try
	{
	
		//check if selection contains video and/or audio
		while (i < [objectsToFreeze count])
		{
			if ([[objectsToFreeze objectAtIndex:i] WSObject]->ContainsVideo(false, true))
				bSelectionContainsVideo = true;
			if ([[objectsToFreeze objectAtIndex:i] WSObject]->ContainsAudio( true ))
				bSelectionContainsAudio = true;
				i++;
		}
		
		//compute the start and end time of the render
		if ((startTime == ARME_TIME_UNDEF) || (endTime == ARME_TIME_UNDEF))
		{
			startTime = ARME_TIME_MAX;
			endTime = ARME_TIME_MIN;
	
			i = 0;
			while (i < [objectsToFreeze count])
			{
				if ([[objectsToFreeze objectAtIndex:i] WSObject]->GetAbsoluteStartTime() < startTime)
					startTime = [[objectsToFreeze objectAtIndex:i] WSObject]->GetAbsoluteStartTime();
				if ([[objectsToFreeze objectAtIndex:i] WSObject]->GetAbsoluteEndxTime() > endTime)
					endTime = [[objectsToFreeze objectAtIndex:i] WSObject]->GetAbsoluteEndxTime();
				i++;
			}
		}
		
		//compute the number of sub-tasks
		SInt32 nbSubTasks = 0;
		if (bSelectionContainsVideo)
			nbSubTasks += 1;
		if( bSelectionContainsAudio && !onlyVideo)
			nbSubTasks += 1;
		
		progressMngr->SetNbSubTasks(nbSubTasks);
		
		// if the object contains both video and audio, we need to create a group
		if (bSelectionContainsVideo && bSelectionContainsAudio && !onlyVideo)
		{
			//create a TArmeContainer
			NSString	*name;
			MIRect		containerFrame = MIZeroRect;
			
			name = [NSString stringWithFormat:@"%s (rendered)", newName];
			container = new TArmeContainer([name UTF8String], armeDoc);
			container->_SetPos(pos);
			container->_SetLength(endTime - startTime);
			
			//create the MIObject
			miContainer = [[[MIObject alloc] initWithArmeWSObj:container] autorelease];
			[miContainer setMontageDoc:montageDoc];
			[miContainer setVerticalPos:videoVPos];
			videoVPos = kRepFooterHeight;
			audioVPos = kRepFooterHeight;
			
			// calculate new size
			NSEnumerator* objEnum = [objectsToFreeze objectEnumerator];
			MIObject* obj;
			while (obj = [objEnum nextObject])
			{
				if ( !MIEqualRects( MIZeroRect, containerFrame ) )
					containerFrame = MIUnionRect(containerFrame, [obj modelFrame]);
				else
					containerFrame = [obj modelFrame];
			}
			[miContainer setModelFrame:containerFrame];
	
			//add it in the hierarchy
			[frozenObjectsParent addChild:miContainer];
			[miContainer setParentObject:frozenObjectsParent];
		
			//sets some variables to include the audio and video object at the right place
	//		audioVPos = MIMinY(containerFrame);
	//		videoVPos = 
	//		GET_VALUE(audioVPos) = [[objectsToFreeze objectAtIndex:0] getVerticalPos].data + kRepTitleHeight;
	//		GET_VALUE(videoVPos) += kDefaultClipRepresentationHeight;
			frozenObjectsParent = miContainer;
			bCreatedContainer = true;
			
	/*		[miContainer setFreezeSource:object];
			[object setFrozenObject:miContainer];*/
			fFreezingResultObject = miContainer;
		}

		if (bSelectionContainsAudio && !onlyVideo)
		{
			// let's copy the MI array into a ZzArchPtrArray
			// please note the reps array is ALREADY remove-descendented
			ZzArchPtrArray wsObjsToFreeze;
			NSEnumerator* objEnum = [objectsToFreeze objectEnumerator];
			MIObject* obj;
			
			while (obj = [objEnum nextObject])
			{
				TArmeWSObj* wsObj = [obj WSObject];
				wsObjsToFreeze.AddItem(wsObj);
			}
			
			// let's actually build the audio clip
			progressMngr->EnterSubTask();
			ArmeErr err;
			MEAudioClip* audioClip = new MEAudioClip(armeDoc,&wsObjsToFreeze,progressMngr);
			progressMngr->LeaveSubTask();
			
			if(err.type == kArmeOK)
			{
				// give the new clip a name
				char name[kMaxFileLength];
				strcpy(name, newName);
				ZzRemoveFileExtensionFromPath(name, kMaxFileLength);
				sprintf(name, "%s (rendered)", name); 
				audioClip->SetNameCopy(name);
		
				if (bCreatedContainer)
				{
					pos.fParent = container;
					pos.fStartTime = ARME_TIME_ZERO;
				}
				audioClip->_SetPos(pos); //_SetPos takes care of LockWrite
		
				//create the MIObject
				miAudioClip = [[[MIObject alloc] initWithArmeWSObj:audioClip] autorelease];
				[miAudioClip setMontageDoc:montageDoc];
				[miAudioClip setVerticalPos:audioVPos];
		
				//add it in the hierarchy
				[frozenObjectsParent addChild:miAudioClip];
				[miAudioClip setParentObject:frozenObjectsParent];
		
				if (!bCreatedContainer)
				{
					fFreezingResultObject = miAudioClip;
				}
		
				videoVPos += [miAudioClip getHeight] + kBetweenClipsHeight;
			}
		}
	
		if (bSelectionContainsVideo)
		{
			// Solo the objectToFreeze
			//---------------------------
			TArmeWSObj* previousVideoSoloObject = [montageDoc armeDocument]->GetVideoSoloObject();
			[montageDoc armeDocument]->SetVideoSoloObject(NULL);
	
			ZzArray<SInt32> mutedStatusBeforeCommand;
			GetMutedStatusBeforeCommand([montageDoc topLevelContainer], &mutedStatusBeforeCommand);
	
			//mute all video objects
			[montageDoc armeDocument]->GetTopContainer()->SetMuted(true);
		
			NSEnumerator* objEnum = [objectsToFreeze objectEnumerator];
			MIObject* obj;
			while (obj = [objEnum nextObject])
			{
				TVideoObject* video = dynamic_cast<TVideoObject*>([obj WSObject]);
				if (video)
					video->SetMuted(false);
			}
			
			// Render the movie into a file
			//----------------------------
			Movie movie = armeDoc->fMasterMovie->GetQTMovie();
			ArmeErr err;
			ArmeDataFileRef* dataFileRefLowRes;
	
			progressMngr->EnterSubTask();
			try
			{
				dataFileRefLowRes = RenderMovieIntoSingleTrackMovie(err, movie, armeDoc, progressMngr, startTime, endTime, reverse, useCustomAlgo, true, stillClipToScale); //false);
			}
			catch (ZzException* e)
			{
				int index = 0;
				SetMutedStatusAsTheyWereBeforeCommand([montageDoc topLevelContainer], &mutedStatusBeforeCommand, index);
	
				throw;
			}
			progressMngr->LeaveSubTask();
	
			//export to high res movie
	/*		progressMngr->EnterSubTask();
			Movie highRes = NewMovie(newMovieActive);
					SetMovieTimeScale(highRes, armeDoc->fMasterMovie->GetVideoTimeScale());
			armeDoc->GetTopContainer()->ExportInHighResMovie(highRes);
			ArmeDataFileRef* dataFileRefHighRes = RenderMovieIntoSingleTrackMovie(err, highRes, armeDoc, progressMngr, startTime, endTime, reverse, useCustomAlgo, true);		
			armeDoc->GetTopContainer()->CleanAfterExport();
			dataFileRefLowRes->SetHighResSourceFile(dataFileRefHighRes);*/
			
			//this function keeps a ref on the created file (during the object creation)
			dataFileRefLowRes->AddRef();
	//		dataFileRefHighRes->AddRef();
			
	//		progressMngr->LeaveSubTask();
	
			// Un-solo the objectToFreeze
			//------------------------------
	
			int index = 0;
			SetMutedStatusAsTheyWereBeforeCommand([montageDoc topLevelContainer], &mutedStatusBeforeCommand, index);
			
			[montageDoc armeDocument]->SetVideoSoloObject(previousVideoSoloObject);
			
			// rebuild the master movie as it was before (without constant mix transition)
	//		[montageDoc armeDocument]->GetTopContainer()->UpdateAffectedHierarchy(false, NULL, NULL);
	/*		update = [[MIUpdateVideoTransitionsCommand alloc] initWithDocument:montageDoc from:ARME_TIME_UNDEF to:ARME_TIME_UNDEF];
			[update doIt];
			[update release];*/
	
			// Create a video clip from the rendered movie file
			//-------------------------------------------------
			if (bCreatedContainer)
			{
				pos.fParent = container;
				pos.fStartTime = ARME_TIME_ZERO;
			}

			TVideoClip* videoClip = new TVideoClip(dataFileRefLowRes, armeDoc, progressMngr, &err, pos, stillClipToScale);
	/*		char renderedFile[1024];
			dataFileRefLowRes->GetFilePath(renderedFile);
			FSSpec renderedFSSpec;
			Path2FSSpecViaFSRef(renderedFSSpec, renderedFile, false);
			TVideoClip* videoClip = new TVideoClip(renderedFSSpec, 1, armeDoc, progressMngr, &err, pos);*/
	
			// give the new clip a name
			char name[kMaxFileLength];
			strcpy(name, newName);
			ZzRemoveFileExtensionFromPath(name, kMaxFileLength);
			sprintf(name, "%s (rendered)", name); 
			videoClip->SetNameCopy(name);
				
			//create the MIObject
			miVideoClip = [[[MIObject alloc] initWithArmeWSObj:videoClip] autorelease]; 
			[miVideoClip setMontageDoc:montageDoc];
			[miVideoClip setVerticalPos:videoVPos];
	
	/*		NSEnumerator* mainViewEnumerator = [[montageDoc windowControllers] objectEnumerator];
			MIMainViewController	*aMainViewController;
			MIObjectRep* newObjectRep;
			while ( aMainViewController = [mainViewEnumerator nextObject] )
			{
				newObjectRep = [miVideoClip createRepresentationInView:[aMainViewController topLevelContainerView]];
			}*/
	
			//add it in the hierarchy
			[frozenObjectsParent addChild:miVideoClip];
			[miVideoClip setParentObject:frozenObjectsParent];
	
			if (!bCreatedContainer)
			{
				//keep relation between source and frozen object
			//	[miVideoClip setFreezeSource:object];
			//	[object setFrozenObject:miVideoClip];
				
				fFreezingResultObject = miVideoClip;
			}
	
			//this function kept a ref on the created file (during the object creation)
			dataFileRefLowRes->Release();
	//		dataFileRefHighRes->Release();
		}
	
	
		// create the representations
		[miContainer createAllRepresentations];
		[miContainer setRepresentationsNeedDisplay];
	
		// create the representations
		[miAudioClip createAllRepresentations];
		[miAudioClip setRepresentationsNeedDisplay];
		
		// create the representations
		[miVideoClip createAllRepresentations];
		[miVideoClip setRepresentationsNeedDisplay];
		
		if (bCreatedContainer)
		{
			[frozenObjectsParent setModelFrame:[frozenObjectsParent minimalRectToFitContent]];
		}
	}
	catch (ZzException* e)
	{
		if (miVideoClip)
		{
			[[miVideoClip parentObject] removeChild:miVideoClip];
		}
		if (miAudioClip)
		{
			[[miAudioClip parentObject] removeChild:miAudioClip];
		}
		if (miContainer)
		{
			[[miContainer parentObject] removeChild:miContainer];
		}
		throw;
	}

	return fFreezingResultObject;
}