/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import "MIInsertLowLevelCommand.h"
#import "MIMontageDocument.h"
#import "MIDeleteLowLevelCommand.h"
#import "MIMainViewController.h"
#import "MIObject.h"
#import "MIApplicationController.h"
#import "MIResizeCommand.h"

#ifndef TArmeWSObj_h
#include "TArmeWSObj.h"
#endif //TArmeWSObj_h

/****************************************************************************************************

MIInsertLowLevelCommand

****************************************************************************************************/

@implementation MIInsertLowLevelCommand

- (void)dealloc
{
	[fEnlargeCommand release];
	[fInsertInObject release];
	[fMemento release];
	[fObjectToInsert release];
	[super dealloc];
}

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
	self = [super initFromArchive:handler];
	if(self)
	{
		fInsertInObject = [handler->ReadObjCObject() retain];
		fObjectToInsert = [handler->ReadObjCObject() retain];
		fMemento = [handler->ReadObjCObject() retain];
		fInsertPosition = handler->ReadArmeTime();
		fEndOfInsertedObject = handler->ReadArmeTime();
		fEnlargeCommand = [handler->ReadObjCObject() retain];
	}
	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
	[super writeToArchive:handler];
	handler->WriteObjCObject(fInsertInObject);
	handler->WriteObjCObject(fObjectToInsert);
	handler->WriteObjCObject(fMemento);
	handler->WriteArmeTime(fInsertPosition);
	handler->WriteArmeTime(fEndOfInsertedObject);
	handler->WriteObjCObject(fEnlargeCommand);

	return YES;
}

- (id)initWithSelection:(MIWSSelection*)sel objectToInsert:(MIObject*)objectToInsert atPosition:(ArmeTime)insertPosition inObject:(MIObject*)insertInObject
{
	self = [super initWithSelection:sel];
	
	[fSelection adjustSelectionTimesToVideo];
		
	if ( self )
	{
		fInsertInObject = [insertInObject retain];
		fInsertPosition = insertPosition;
		fObjectToInsert = [objectToInsert retain];
		fEnlargeCommand = nil;
	
		fMemento = [[insertInObject getMemento] retain];
		
		if ([sel containsVideo:true andTransparentTextClip:NO])
			[self modifiedVideoHierarchy];
		if ([sel containsAudio])
			[self modifiedAudioHierarchy];
	
	}
	return self;
}

- (MICmdStatus)doIt
{
	// insert
	fEndOfInsertedObject = [fInsertInObject WSObject]->InsertSlice(fInsertPosition, [fObjectToInsert WSObject], [MICommand getCurrentProgressDialog]); 
	[fInsertInObject markModified];
	[fInsertInObject invalidateRep];
	
	// check if parent size is big enough
	MIObject* parent = [fInsertInObject parentObject];
	if (fEnlargeCommand)
		[fEnlargeCommand _redoIt];
	else if ( !MIContainsRect( [parent modelFrame], [parent minimalRectToFitContent] ))
	{
		fEnlargeCommand = [[MIResizeCommand alloc] initWithSelection:nil];
		[fEnlargeCommand setObject:parent newFrame:MIUnionRect([parent modelFrame], [parent minimalRectToFitContent])];
		[fEnlargeCommand _doIt];
	}

	[fInsertInObject setSelectionsFromTime:fInsertPosition toTime:(fInsertPosition + [fObjectToInsert WSObject]->GetDuration())];

	[fInsertInObject invalidateRep];
	[fInsertInObject invalidateRepresentationsCachedImages]; // this deletes the cached image for audio (for example) and sets it to null
	[fInsertInObject setRepresentationsNeedDisplay]; // huh

	
	return kMICmdStatusDone;
}

- (void)undoIt
{
	if (fEnlargeCommand)
		[fEnlargeCommand _undoIt];

	[fInsertInObject invalidateRep];

	switch ([fInsertInObject WSObject]->GetType())
	{
		case TArmeWSObj::kAudioClip :
		{
			// undo = do of delete
			MICommand* undoCommand = [[[MIDeleteLowLevelCommand alloc] initWithObjectToDelete:fInsertInObject from:fInsertPosition to:fEndOfInsertedObject] autorelease];
			[undoCommand doIt];
		};break;
		default :
		[fInsertInObject setMemento:fMemento];
	}

}

- (void)redoIt
{
	[self doIt];
	// redo = undo of delete
//	[fUndoCommand undoIt];
}
@end

