/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import "MIMacroCommand.h"

/****************************************************************************************************

MIMacroCommand

****************************************************************************************************/
@implementation MIMacroCommand

#ifndef ArmeArchiveHandler_h
#include "ArmeArchiveHandler.h"
#endif //ArmeArchiveHandler_h

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
	self = [super initFromArchive:handler];
	if(self)
	{
		fSubCommands = [handler->ReadObjCObject() retain];
		fCommandName = [[self className] retain];
	}
	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
	[super writeToArchive:handler];
	handler->WriteObjCObject(fSubCommands);
	return YES;
}
	
- (id)init
{
	self = [super init];
	if(self)
	{
		fSubCommands = [[NSMutableArray alloc] init];
		fCommandName = [[self className] retain];
	}
	return self;
}

- (id)initWithSelection:(MIWSSelection*)sel
{
	self = [super initWithSelection:sel];
	if(self)
	{
		fSubCommands = [[NSMutableArray alloc] init];
		fCommandName = [[self className] retain];
	}
	return self;
}

- (void)dealloc
{
	[fSubCommands release];
	[fCommandName release];
	[super dealloc];
}


- (void)addCommand:(MICommand*)command
{	
	// make sure the command hasn't been done yet
	ZzAssert(command->fState == kMICommandStateCanOnlyBeDone);
	[fSubCommands addObject:command];
}

- (BOOL)displaySetupDialog
{
	unsigned int count = [fSubCommands count];
	unsigned int i;
	for( i = 0; i < count; i++ )
	{
		if( ![[fSubCommands objectAtIndex:i] displaySetupDialog])
			return NO;
	}
	return YES;
}

- (int)getNbLengthyTasks
{
	unsigned int count = [fSubCommands count];
	unsigned int i;
	int	nbLengthyTasks = 0;
	for( i = 0; i < count; i++ )
	{
		nbLengthyTasks += [[fSubCommands objectAtIndex:i] getNbLengthyTasks];
	}

	return nbLengthyTasks;
}

- (BOOL)isCancelable
{
	return NO;
}

- (int)getNbSubCommands
{
	return [fSubCommands count];
}
- (MICmdStatus)doIt
{
	unsigned int count = [fSubCommands count];
	unsigned int i;

	try
	{
	[MICommand getCurrentProgressDialog]->SetNbSubTasks(count);
	for( i = 0; i < count; i++ )
	{
		vverntl(commandsEngine,2,"Macro command : executing sub command " << i);

		[MICommand getCurrentProgressDialog]->EnterSubTask();
		if ([[fSubCommands objectAtIndex:i] _doIt] == kMICmdStatusCancelled)
		{
			if (i > 0)
			{
				for (int j=i-1; j >= 0; --j)
				{
					[[fSubCommands objectAtIndex:j] _undoIt];
				}
			}
			if ([[fSubCommands objectAtIndex:i] hasModifiedAudioHierarchy])
				[self modifiedAudioHierarchy];
			if ([[fSubCommands objectAtIndex:i] hasModifiedVideoHierarchy])
				[self modifiedVideoHierarchy];
			if ([[fSubCommands objectAtIndex:i] causedNeedOfOverviewUpdate])
				[self overviewNeedsUpdate];
			[MICommand getCurrentProgressDialog]->LeaveSubTask();
			return kMICmdStatusCancelled;
		}
		[MICommand getCurrentProgressDialog]->LeaveSubTask();

		if ([[fSubCommands objectAtIndex:i] hasModifiedAudioHierarchy])
			[self modifiedAudioHierarchy];
		if ([[fSubCommands objectAtIndex:i] hasModifiedVideoHierarchy])
			[self modifiedVideoHierarchy];
		if ([[fSubCommands objectAtIndex:i] causedNeedOfOverviewUpdate])
			[self overviewNeedsUpdate];
	}
	}
	catch(ZzException* e)
	{
		e->DefaultCatcher();
		delete e;
		
		//cancel the commands that have been done already
		if (i > 0)
		{
			for (int j=i; j >= 0; --j)
			{
				if ([[fSubCommands objectAtIndex:j] state] == kMICommandStateCanOnlyBeUndone)
					[[fSubCommands objectAtIndex:j] _undoIt];
			}
		}
		[MICommand getCurrentProgressDialog]->LeaveSubTask();
		return kMICmdStatusCancelled;
	}
	return kMICmdStatusDone;
}

- (void)undoIt
{
	unsigned int count = [fSubCommands count];
	int i;

	[MICommand getCurrentProgressDialog]->SetNbSubTasks(count);
	for( i = count-1; i >= 0; i-- )
	{
		[MICommand getCurrentProgressDialog]->EnterSubTask();
		[[fSubCommands objectAtIndex:i] _undoIt];
		[MICommand getCurrentProgressDialog]->LeaveSubTask();
	}
}

- (void)redoIt
{
	unsigned int count = [fSubCommands count];
	unsigned int i;
	[MICommand getCurrentProgressDialog]->SetNbSubTasks(count);
	for( i = 0; i < count; i++ )
	{
		[MICommand getCurrentProgressDialog]->EnterSubTask();
		[[fSubCommands objectAtIndex:i] _redoIt];
		[MICommand getCurrentProgressDialog]->LeaveSubTask();
	}
}

- (NSString*)commandName
{
	return fCommandName;
}

- (void)setCommandName:(NSString*)commandName
{
	[commandName retain];
	[fCommandName release];
	fCommandName = commandName;
}

@end

