/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import "MIPasteCommand.h"
#import "MIObject.h"
#import "MIApplicationController.h"
#import "MIMainViewController.h"
#import "MIMontageDocument.h"
#import "MIDeleteLowLevelCommand.h"
#import "MIInsertLowLevelCommand.h"
#import "MIPasteLowLevelCommand.h"

#ifndef TArmeDoc_h
#include "TArmeDoc.h"
#endif //TArmeDoc_h

#ifndef TArmeWSObj_h
#include "TArmeWSObj.h"
#endif //TArmeWSObj_h

#ifndef MEAudioClip_h
#include "MEAudioClip.h"
#endif //MEAudioClip_h

/****************************************************************************************************

MIPasteCommand

****************************************************************************************************/

@implementation MIPasteCommand
- (NSString*)commandName
{
	return @"Paste";
}

- (void)dealloc
{
	[super dealloc];
}

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
	self = [super initFromArchive:handler]; //it's a macro command
	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
	[super writeToArchive:handler]; //it's a macro command
	return YES;
}


- (id)initWithSelection:(MIWSSelection*)sel inDoc:(MIMontageDocument*)doc
{
	MIObject	*tlcObject;
	self = [super initWithSelection:sel];
	[fSelection adjustSelectionTimesToVideo];
		
	tlcObject = [doc topLevelContainer];

	if ( self )
	{
		NSEnumerator* clipboardObjEnum = [[[[MIApplicationController sharedApplicationController] clipBoard] getChildren] objectEnumerator];
		MIObject* whereToPaste;
		ArmeTime whenToPaste;
	
		//try to determine where we are going to paste the clipboard.
		// if there is a slice selection and that this selection has the same 'shape' as the clipboard, we insert the clipboard at
		// the selection place (only if selection and clipboard are both a video clip, a sound clip or a group with one video clip and one sound clip).
		if (fSelection)
		{
			switch ([fSelection type])
			{
				case kMIWSSelectionTypeTimeSlice :
				{
					MIObject* selObject;
					MIObject* objectToPaste;
					
					// take only one object into account (the first one).
					selObject = [[[fSelection objectEnumerator] nextObject] representedObject];
	
					// and take a reference on the first object in the clipboard
					objectToPaste = [clipboardObjEnum nextObject];
			
					// is pasting compatible ? objects need to be of the same type AND same nb of channels for audio
					bool pasteIncompatible = false;
					
					if ([objectToPaste WSObject]->GetType() != [selObject WSObject]->GetType() )
						pasteIncompatible = true;
						
					if( ([objectToPaste WSObject]->GetType() == TArmeWSObj::kAudioClip) && ([selObject WSObject]->GetType() == TArmeWSObj::kAudioClip) )
					{
						// same NB of channels ?
						MEAudioClip* getPastedInto = dynamic_cast<MEAudioClip*>([objectToPaste WSObject]);
						MEAudioClip* pastedObj = dynamic_cast<MEAudioClip*>([selObject WSObject]);
						ZzAssert(getPastedInto);
						ZzAssert(pastedObj);
						if( getPastedInto->GetNbChannels() != pastedObj->GetNbChannels() )
							pasteIncompatible = true;
					}
					
					// if object in clipboard and target are not the same type, paste in the TLC at the current play pos
					if (pasteIncompatible)
					{
						whereToPaste = [[objectToPaste montageDoc] topLevelContainer];
						whenToPaste = [[[NSDocumentController sharedDocumentController] currentDocument] armeDocument]->GetPlayPos();
						break;
					}

					switch ([selObject WSObject]->GetType())
					{
						case TArmeWSObj::kVideoClip :
						case TArmeWSObj::kAudioClip :
						case TArmeWSObj::kTextClip :
						{
							whereToPaste = selObject;
							whenToPaste = [fSelection startTime];
	
							// if the clipboard content and the selection have the same type,  and this is a slice selection,
							// we will replace the selection by the clipboard.  For this, we will build a macro command with
							// a MIDeleteLowLevelCommand to delete the current selection,
							[self addCommand:[[[MIDeleteLowLevelCommand alloc] initWithObjectToDelete:selObject from:[fSelection startTime] to:[fSelection endxTime]] autorelease]];
							
							// and a MIInsertLowLevelCommand to insert the clipboard at the current selection position
							[self addCommand:[[[MIInsertLowLevelCommand alloc] initWithSelection:fSelection objectToInsert:objectToPaste atPosition:[fSelection startTime] inObject:whereToPaste] autorelease]];
	
							return self;
						};break;
						case TArmeWSObj::kContainer :
						{
							NSArray* selVideoChildren = [selObject getVideoChildrenAtTime:[fSelection startTime]];
							NSArray* pasteVideoChildren = [objectToPaste getVideoChildren];
							NSArray* selAudioChildren = [selObject getAudioChildrenAtTime:[fSelection startTime]];
							NSArray* pasteAudioChildren = [objectToPaste getAudioChildren];
							
						
							if (([selVideoChildren count] == [pasteVideoChildren count])
								&& ([selAudioChildren count] == [pasteAudioChildren count]))
							// test other types also
							{
								// if the clipboard content and the selection have the same type,  and this is a slice selection,
								// we will replace the selection by the clipboard.  For this, we will build a macro command with
								// a MIDeleteLowLevelCommand to delete the current selection,
								[self addCommand:[[[MIDeleteLowLevelCommand alloc] initWithObjectToDelete:selObject from:[fSelection startTime] to:[fSelection endxTime]] autorelease]];
								
								// insert an empty slice in the container
								[self addCommand:[[[MIInsertLowLevelCommand alloc] initWithSelection:fSelection objectToInsert:objectToPaste atPosition:[fSelection startTime] inObject:selObject] autorelease]];
								// and a MIInsertLowLevelCommand for each of the children
								for (UInt32 i=0; i < [selVideoChildren count]; ++i)
								{
									[self addCommand:[[[MIInsertLowLevelCommand alloc] initWithSelection:fSelection objectToInsert:[pasteVideoChildren objectAtIndex:i] atPosition:[fSelection startTime] inObject:[selVideoChildren objectAtIndex:i]] autorelease]];
								}
								//same for audio
								for (UInt32 i=0; i < [selAudioChildren count]; ++i)
								{
									[self addCommand:[[[MIInsertLowLevelCommand alloc] initWithSelection:fSelection objectToInsert:[pasteAudioChildren objectAtIndex:i] atPosition:[fSelection startTime] inObject:[selAudioChildren objectAtIndex:i]] autorelease]];
								}
							}

							[selVideoChildren release];
							[pasteVideoChildren release];
							[selAudioChildren release];
							[pasteAudioChildren release];
							
							return self;
						
						};break;
						default :
						{
							whereToPaste = [[objectToPaste montageDoc] topLevelContainer];
							whenToPaste = [[[NSDocumentController sharedDocumentController] currentDocument] armeDocument]->GetPlayPos();
						}
					};
				};break;
				default :
					whereToPaste = tlcObject;
					whenToPaste = [[[NSDocumentController sharedDocumentController] currentDocument] armeDocument]->GetPlayPos();
			}
		}
		else
		{
			whereToPaste = tlcObject;
			whenToPaste = [[[NSDocumentController sharedDocumentController] currentDocument] armeDocument]->GetPlayPos();
		}
	
		// otherwise, the Paste command is a simple MIPasteLowLevelCommand
		MIPasteLowLevelCommand* cmd = [[[MIPasteLowLevelCommand alloc]
							initWithDocument:[[NSDocumentController sharedDocumentController] currentDocument]
							objectsToPaste:[[[MIApplicationController sharedApplicationController] clipBoard] getChildren]
							inObject:tlcObject atTime:whenToPaste atVPos:VPOS_UNDEF ] autorelease];
		[cmd setSuffix:kSuffixCopy];
		[self addCommand:cmd];
	}
	return self;
}

- (BOOL)displaySetupDialog
{
	if ([fSelection type] == kMIWSSelectionTypeTimeSlice)
	{
		if ([self CheckFrozenObjectInSelectionAndAskIfUserWantsToContinue:fSelection] == NO)
			return NO;
	}
	return YES;
}

+ (BOOL)supportsSelection:(MIWSSelection*)sel
{
	//there must be something in the clipboard
	if ([[[[MIApplicationController sharedApplicationController] clipBoard] getChildren] count] >= 1)
		return YES;
	else
		return NO;
}

@end

