/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import "MIRenameCommand.h"
#import "MIObject.h"
#import "MIObjectRep.h"
#import "MIRenameCommandDlgController.h"

#ifndef TArmeWSObj_h
#include "TArmeWSObj.h"
#endif //TArmeWSObj_h

@implementation MIRenameCommand

- (NSString*)commandName
{
	return @"Rename";
}

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
	self = [super initFromArchive:handler];
	if(self)
	{
		fModifiedObjects = [handler->ReadObjCObject() retain];
		fOldNames = [handler->ReadObjCObject() retain];
		ZzAssert(fOldNames);
		fNewBaseName = [handler->ReadObjCObject() retain];
		ZzAssert(fNewBaseName);
	}
	return self;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
	[super writeToArchive:handler];
	handler->WriteObjCObject(fModifiedObjects);
	handler->WriteObjCObject(fOldNames);
	handler->WriteObjCObject(fNewBaseName);
	return YES;
}

- (id)initWithSelection:(MIWSSelection*)sel
{
	self = [super initWithSelection:sel];
	if(self)
	{
		ZzAssertMsg([sel type] == kMIWSSelectionTypeObjects,"Command should not be allowed to work on this kind of selection");
	
		// create the members
		fModifiedObjects = [[NSMutableArray alloc] init];
		fOldNames = [[NSMutableArray alloc] init];
		fNewBaseName = nil;
		
		// copy the selection members in the array
		NSEnumerator* selEnum = [sel objectEnumerator];
		MIObjectRep* rep;
		MIObject* obj;
		while(rep = [selEnum nextObject])
		{
			obj = [rep representedObject];
			[fModifiedObjects addObject:obj];
		}

		// then, let's save the old names
		NSEnumerator* objEnum = [fModifiedObjects objectEnumerator];
		while(obj = [objEnum nextObject])
		{
			[fOldNames addObject:[NSString stringWithUTF8String:[obj WSObject]->GetName()]];
		}
	}
	return self;
}

- (id)initWithObject:(MIObject*)obj
{
	self = [super initWithSelection:nil];
	if(self)
	{
		// create the members
		fModifiedObjects = [[NSMutableArray alloc] init];
		fOldNames = [[NSMutableArray alloc] init];
		fNewBaseName = nil;
		
		// copy the selection members in the array
		[fModifiedObjects addObject:obj];

		// then, let's save the old names
		NSEnumerator* objEnum = [fModifiedObjects objectEnumerator];
		while(obj = [objEnum nextObject])
		{
			[fOldNames addObject:[NSString stringWithUTF8String:[obj WSObject]->GetName()]];
		}
			
	}
	return self;
}

- (void)setBaseName:(NSString*)baseName
{
	fNewBaseName = [[NSString alloc] initWithString:baseName];
}

- (NSString*)oldName
{
	return [fOldNames objectAtIndex:0];
}

- (NSString*)newBaseName
{
	return fNewBaseName;
}

+ (BOOL)supportsSelection:(MIWSSelection*)sel
{
	if( ([sel type] == kMIWSSelectionTypeObjects) && ([sel count] == 1) )
		return YES;
	else
		return NO;
}

- (MICmdStatus)doIt
{
	// apply the new names (same as redoIt)
	[self redoIt];
	return kMICmdStatusDone;
}

- (void)undoIt
{
	// set the new names
	if([fModifiedObjects count] == 1)
	{
		[[fModifiedObjects objectAtIndex:0] WSObject]->SetNameCopy([[fOldNames objectAtIndex:0] UTF8String]);
		[[fModifiedObjects objectAtIndex:0] refreshCachedTitleStringImageInReps];
		[[fModifiedObjects objectAtIndex:0] setTitlesNeedDisplay];
	}
	else
	{
		ZzAssertMsg(0,"Cannot rename more than one object. This command should be disabled");
		//for(SInt32 i = 0; i < [fModifiedObjects count]; i++)
		//{
		//	NSString* name = [NSString stringWithFormat:@"%@ %d",fNewBaseName,i];
		//	[obj WSObject]->SetNameCopy([name cString]);
		//}
	}
}

- (void)redoIt
{
	// set the old names
	if([fModifiedObjects count] == 1)
	{
		[[fModifiedObjects objectAtIndex:0] WSObject]->SetNameCopy([fNewBaseName UTF8String]);
		[[fModifiedObjects objectAtIndex:0] refreshCachedTitleStringImageInReps];
		[[fModifiedObjects objectAtIndex:0] setTitlesNeedDisplay];
	}
	else
	{
		ZzAssertMsg(0,"Cannot rename more than one object. This command should be disabled");
		//for(SInt32 i = 0; i < [fModifiedObjects count]; i++)
		//{
		//	NSString* name = [NSString stringWithFormat:@"%@ %d",fNewBaseName,i];
		//	[obj WSObject]->SetNameCopy([name cString]);
		//}
	}
}


- (void)dealloc
{
	[fModifiedObjects release];
	[fOldNames release];
	[fNewBaseName release];
    
    [super dealloc];
}

- (BOOL)displaySetupDialog
{
	NSModalSession session;
	NSWindow* 	win;

	MIRenameCommandDlgController* dlgController = [MIRenameCommandDlgController alloc];
	[dlgController initWithWindowNibName:@"MIRenameCommandDlg" owner:dlgController];
	win = [dlgController window];
	
	if([fModifiedObjects count] == 1)
	{
		[dlgController setText:[NSString stringWithUTF8String:[[fModifiedObjects objectAtIndex:0] WSObject]->GetName()]];
	}
	else
	{
		ZzAssertMsg(0,"Cannot rename more than one object. This command should be disabled");
		return NO;
	}
		
	session = [NSApp beginModalSessionForWindow:win];
	while([NSApp runModalSession:session] == NSRunContinuesResponse)
	{
	}
	[NSApp endModalSession:session];					
	[win orderOut:0];
	if([dlgController cancelled])
	{
		[dlgController release];
		return NO;
	}
	else
	{
		[self setBaseName:[dlgController text]];
		[dlgController release];
		return YES;
	}
}

@end













