/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#import "MIShrinkToFitCommand.h"
#import "MIResizeCommand.h"

#ifndef ArmeArchiveHandler_h
#include "ArmeArchiveHandler.h"
#endif //ArmeArchiveHandler_h

#ifndef TArmeWSObj_h
#include "TArmeWSObj.h"
#endif //TArmeWSObj_h

#import "MIMainViewController.h"
#import "MIObject.h"
#import "MIObjectRep.h"

/****************************************************************************************************

MIShrinkToFitCommand

****************************************************************************************************/

@implementation MIShrinkToFitCommand
- (void)dealloc
{
	[fResizeCommands release];
	[super dealloc];
}

- (NSString*)commandName
{
	return @"Shrink To Fit";
}

- (id)initWithSelection:(MIWSSelection*)sel
{
	self = [super initWithSelection:sel];

	if ( self )
	{
		[fSelection discardVideoTransitions];

		fResizeCommands = [[NSMutableArray alloc] initWithObjects:nil];
		
		[self modifiedVideoHierarchy];
		[self modifiedAudioHierarchy];
		[self overviewNeedsUpdate];
		
	}
	return self;
}

-(id)initFromArchive:(ArmeArchiveHandler*)handler
{
    if (( self = [super initFromArchive:handler] ))
    {
		fResizeCommands = [handler->ReadObjCObject() retain];
    }
    return self ;
}

-(BOOL)writeToArchive:(ArmeArchiveHandler*)handler
{
    if ( [super writeToArchive:handler] )
    {
	//handler->WriteObjCObject(fController);
	handler->WriteObjCObject(fResizeCommands);
	//handler->WriteObjCObject(fRep);
	return YES;
    }
    else return NO;
}

- (BOOL)displaySetupDialog
{
	return YES;
}

+ (BOOL)supportsSelection:(MIWSSelection*)sel
{
	if ( !sel) return NO;
	if ( [sel type] != kMIWSSelectionTypeObjects ) return NO;
	
	if ([sel count] > 0)
	{
		MIObjectRep *objectRep;
		NSEnumerator *enumerator = [sel objectEnumerator];

		while (( objectRep = [enumerator nextObject] )) {

		TArmeWSObj* wsobj = [[objectRep representedObject] WSObject];
		SInt32 type = wsobj->GetType();
		if(	(type != TArmeWSObj::kVideoEffect) && (type != TArmeWSObj::kAudioProcessor) && (type != TArmeWSObj::kContainer) )
			return NO;
			else if ( wsobj->GetChildrenCount() < 1 ) return NO;
			if ( [[objectRep representedObject] isCollapsed] ) return NO;
		}

		return YES;
	}
	
	return NO;
}

- (void)setObject:(MIObject*)obj
{
	fObject = obj;
}


- (MICmdStatus)doIt
{
	MIWSSelection*		sel = fSelection;
	NSEnumerator*		objEnum = [sel objectEnumerator];
	MIObjectRep*		objectRep;
	MIResizeCommand*	cmd;
	NSEnumerator		*cmdEnum;

	if ( fObject )
	{
		// check if parent size is big enough
		if ( !MIEqualRects( [fObject modelFrame], [fObject minimalRectToFitContent] ) )
		{
			cmd = [[[MIResizeCommand alloc] initWithSelection:nil] autorelease];
			[cmd setObject:fObject newFrame:[fObject minimalRectToFitContent]];
			[fResizeCommands addObject:cmd];
		}
	}
	else
	{
		while ( objectRep = [objEnum nextObject] )
		{
			MIRect minimalRect;
			
			if ( ![[objectRep representedObject] isContainer] || ![[[objectRep representedObject] getChildren] count] )
				continue;
			
			minimalRect = [[objectRep representedObject] minimalRectToFitContent];
			// check if parent size is big enough
			if ( !MIEqualRects( [[objectRep representedObject] modelFrame], minimalRect ) &&
				 !MIEqualRects( minimalRect, MIZeroRect ) )
			{
				cmd = [[[MIResizeCommand alloc] initWithSelection:nil] autorelease];
				[cmd setObject:[objectRep representedObject] newFrame:minimalRect];
				[fResizeCommands addObject:cmd];
			}
		}
	}
	
	cmdEnum = [fResizeCommands objectEnumerator];
	[MICommand getCurrentProgressDialog]->SetNbSubTasks([fResizeCommands count]);
	while ( cmd = [cmdEnum nextObject] )
	{
		[MICommand getCurrentProgressDialog]->EnterSubTask();
		[cmd _doIt];
		[MICommand getCurrentProgressDialog]->LeaveSubTask();
	}
	return kMICmdStatusDone;
}

- (void)undoIt
{
	NSEnumerator		*cmdEnum;
	MICommand			*cmd;

	cmdEnum = [fResizeCommands objectEnumerator];
	while ( cmd = [cmdEnum nextObject] )
	{
		[cmd _undoIt];
	}
}

- (void)redoIt
{
	NSEnumerator		*cmdEnum;
	MICommand			*cmd;

	cmdEnum = [fResizeCommands objectEnumerator];
	while ( cmd = [cmdEnum nextObject] )
	{
		[cmd _redoIt];
	}
}

@end
