/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/


#ifndef   Vibrato_h
#include "Vibrato.h"
#endif    Vibrato_h

#ifndef   DSPFunctions_h
#include "DSPFunctions.h"
#endif    DSPFunctions_h


void VibratoPlugin::ProcessFramesAdd(SInt32 nbFrames)
{

    float* ibuf   = GetInputBuffer(0);
    SInt32 ichans = GetInputChannels(0);

    float* obuf   = GetOutputBuffer(0);
    SInt32 ochans = GetOutputChannels(0);
    
	// buffer pointers
    float *inL, *inR, *outL, *outR;
    float outputscale;
	
	// prepare for MONO or STEREO operation
    switch (ichans)
    {
        case 1:		inL = inR = ibuf; 
                    break;
        default:    inL = ibuf; 
                    inR = ibuf + 1; 
                    break;
    }
    switch (ochans)
    {
        case 1:		outL = outR = obuf; 
                    outputscale = 0.5f;
                    break;
        default:    outL = obuf; 
                    outR = obuf + 1; 
                    outputscale = 1.0f;
                    break;
    }
    
    float sr = GetFrameRate();
    float inv_sr = 1.0f / sr;

	// effects params (convert from gui params)
	float depth =			arboSat(fParams.guiParams[kVibrato_Depth] / 100.0, 0.0, 1.0) 
                            * ((float)((kVibrato_BufferLength)/2 - 1 - (float)kVibrato_BufferGuard));
	float rate = 			2.0 * ArboPi * fParams.guiParams[kVibrato_Rate] * inv_sr;
        
    // smoothing params
	float c_depth = 		fParams.c_depth;

    // state information
	float *delayBuffer =	fParams.delayBuffer;
	int delayWriteIndex =	fParams.delayWriteIndex;
	float p =				fParams.p;
	float q =				fParams.q;
	
	// intermediates & constants
	float frac, modfrac;
	float pnext, qnext;
	float x,y,z,  xx,yy,zz;
	int index1, index2, offset, i;
	const float half = 0.5;
    

    // DSP LOOP OVER nbFrames
    for (i=0; i<nbFrames; i++)
    {
		// READ INPUT 
		x  = *inL;
		xx = *inR;
        inL += ichans;
        inR += ichans;

        // GET CURRENT DELAY LENGTH FRACTION & INTEGER OFFSET
		modfrac = half + half * p; 
		frac = (float)kVibrato_BufferGuard + c_depth * modfrac;
		offset = (int)frac;   // integer offset
		frac = frac - offset; // fractional offset
            
        // GET DELAYED VALUE (linear interpolation)
        index1 = (delayWriteIndex - 2*(offset))     & kVibrato_BufferStereoMask;
		index2 = (delayWriteIndex - 2*(offset + 1)) & kVibrato_BufferStereoMask;
		z =  delayBuffer[index2]   * frac + delayBuffer[index1]   * (1 - frac);
		zz = delayBuffer[index2+1] * frac + delayBuffer[index1+1] * (1 - frac);
			
		// WRITE TO DELAYLINE & UPDATE WRITE INDEX
		delayBuffer[delayWriteIndex]   = x;
		delayBuffer[delayWriteIndex+1] = xx;
		delayWriteIndex = (delayWriteIndex + 2) & kVibrato_BufferStereoMask;
		
		// UPDATE LFO (saturated unstable 2-nd order iterative system: nonlinear oscillator)
        pnext = arboSat(p + q * rate);
		qnext = arboSat(q - p * rate);
		p = pnext;
		q = qnext;
			
		// MIX INPUT AND EFFECT OUTPUT
		y  = z;
		yy = zz;
			
		// WRITE OUTPUT
		*outL += y * outputscale;
		*outR += yy * outputscale;
        outL += ochans;
        outR += ochans;
			
		// SMOOTHEN PARAMETERS
		c_depth    = kVibrato_Smooth * depth    + (1.0 - kVibrato_Smooth) * c_depth;
	}   

    // SAVE PLUGIN STATE
	fParams.c_depth 		= c_depth;
	fParams.delayWriteIndex = delayWriteIndex;
	fParams.p 				= p;
    fParams.q 				= q;
}


VibratoPluginClass::VibratoPluginClass()
{



    // define parameter info on the stack
    // params: <name>, <unit>, <absMin>, <absMax>, <Min>, <Max>, <Default>
    ARPIParamInfo p[] = {
        {"Rate (Hz)",		ARPIUnit_Hz,		0.0f, 1000.0f,		0.0f, 50.0f, 	   0.20f},
        {"Depth (%)",		ARPIUnit_percent,	0.0f, 100.0f,		0.0f, 100.0f, 	   20.0f}
    };
    
    // copy info to data member
    assert(sizeof(p) == sizeof(fParamInfoTemplate));
   	memcpy(fParamInfoTemplate, p, sizeof(p));
	
	// pointer used by PRESET_xxx macros
	ARPIBlueWindowPreset* presetPtr = &fFactoryPreset[0];


	// Vibrato preset 1
	PRESET_NAME("Extreme");
	PRESET_PARAM(1.000000, 0.100000, 10.000000, 1); //Rate (Hz)
	PRESET_PARAM(100.000000, 0.000000, 100.000000, 2); //Depth (%)
	PRESET_SOURCE(0.500000); //BW X-axis
	PRESET_SOURCE(1.000000); //BW Y-axis
	PRESET_SOURCE(1.000000); //BW Z-axis
	PRESET_NEXT;

	// Vibrato preset 2
	PRESET_NAME("Slightly Warped");
	PRESET_PARAM(1.000000, 0.100000, 4.000000, 1); //Rate (Hz)
	PRESET_PARAM(4.000000, 0.000000, 10.000000, 2); //Depth (%)
	PRESET_SOURCE(0.623684); //BW X-axis
	PRESET_SOURCE(0.402098); //BW Y-axis
	PRESET_SOURCE(1.000000); //BW Z-axis
	PRESET_NEXT;

	// Vibrato preset 3
	PRESET_NAME("Super Slow");
	PRESET_PARAM(0.100000, 0.100000, 1.000000, 1); //Rate (Hz)
	PRESET_PARAM(100.000000, 0.000000, 100.000000, 2); //Depth (%)
	PRESET_SOURCE(0.000000); //BW X-axis
	PRESET_SOURCE(1.000000); //BW Y-axis
	PRESET_SOURCE(1.000000); //BW Z-axis
	PRESET_NEXT;

	// Vibrato preset 4
	PRESET_NAME("Vibrato Ringer");
	PRESET_PARAM(200.000000, 1.000000, 500.000000, 1); //Rate (Hz)
	PRESET_PARAM(100.000000, 0.000000, 100.000000, 2); //Depth (%)
	PRESET_SOURCE(0.763158); //BW X-axis
	PRESET_SOURCE(1.000000); //BW Y-axis
	PRESET_SOURCE(1.000000); //BW Z-axis
	PRESET_NEXT;



}


SInt32 VibratoPluginClass::GetNbFactoryPresets() const
{
	return kVibrato_NbFactoryPresets;
}

const char* VibratoPluginClass::GetFactoryPresetName(SInt32 preset) const 
{
	return fFactoryPreset[preset].GetName();
}

ARPIBlueWindowPreset* VibratoPluginClass::CreatePresetFromFactoryPreset(SInt32 factoryPreset) const
{
	ARPIBlueWindowPreset* p = new ARPIBlueWindowPreset();
	*p = fFactoryPreset[factoryPreset];
	return p;
}

                VibratoPlugin::VibratoPlugin(const VibratoPluginClass* theClass) : VARPIPlugin(theClass)
{
	// here we need to set the default values and initialize the plugin
    int i;
    
    // clear buffer 
    for (i=0; i<kVibrato_BufferLength; i++) 
       fParams.delayBuffer[i] = 0.0;
       
    // init state variables
    fParams.delayWriteIndex = 0;
    fParams.p = 1;
    fParams.q = 0;
    fParams.c_depth = 0.0;

    // init controller
	fParams.guiParams[kVibrato_Rate] = 0.2;
    fParams.guiParams[kVibrato_Depth] = 20.0;
}


                VibratoPlugin::~VibratoPlugin(){}

bool			VibratoPlugin::SyncParam(SInt32 param)// OBSOLETE
{
	return true;
}	

bool			VibratoPlugin::SetParamValue(SInt32 param, float val)
{
	assert(param >= 0 && param < kVibrato_Params);
	fParams.guiParams[param] = val;
	SyncParam(param);
	return true;
}

float			VibratoPlugin::GetParamValue(SInt32 param) const
{
	assert(param >= 0 && param < kVibrato_Params);
	return fParams.guiParams[param];
}


bool			VibratoPlugin::FormatChange()
{	
	// only for debugging purposes
	vertl(hyperPrismTom,5,"VibratoPlugin::FormatChange()\n");
	return true;
}


const char*		VibratoPluginClass::GetManufacturerName() const	{return "Arboretum Systems, Inc.";}
const char*		VibratoPluginClass::GetPluginName()	const		{return "Vibrato";}
VARPIPlugin*	VibratoPluginClass::CreatePlugin() const 				{return new VibratoPlugin(this);}
SInt32			VibratoPluginClass::GetNbParams() const 		{return kVibrato_Params;}
void*			VibratoPluginClass::CreateView(void* miobject) const 
                                                                {return NULL;}
void			VibratoPluginClass::GetParamInfo(ARPIParamInfo* paramInfo, SInt32 param) const
{
	assert(param >= 0 && param < kVibrato_Params);
	memcpy(paramInfo,&(fParamInfoTemplate[param]),sizeof(ARPIParamInfo));
}

VibratoPluginClass::~VibratoPluginClass()
{
}

//IMPLEMENT_PLUGIN_ARCHIVE_METHODS(VibratoPlugin)

