/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/
 /*
	ZzArchPtrArray is a template class that represents a dynamic array of pointers
	to archivable C++ objects
*/

#define ZzArchPtrArray_h

#ifndef ZzArchPtrArray_h
#include "ZzArchPtrArray.h"
#endif //ZzArchPtrArray_h

#ifndef ArmeArchivable_h
#include "ArmeArchivable.h"
#endif //ArmeArchivable_h

/*!
	@class ZzArchPtrArray
*/

class ZzArchPtrArray : public ArmeArchivable
{
protected :
	// the array of items
	ArmeArchivable**	fData;
	// the current number of items in the array
	SInt32				fCount;
	// the current number of allocated space for items
	SInt32				fMax;

public:
	// returns a a copy of the item at position i and returns 
	ArmeArchivable*	ItemAt( SInt32 i ) const;
	ArmeArchivable*	operator[]( SInt32 i ) const { return ItemAt(i); };
	
	// adds a new item at the end of the array
	void		AddItem( ArmeArchivable* t );
		
	// insert a new item at index i and move the end of the array of one position	
	void		InsertItem(ArmeArchivable* t, SInt32 i);

	// number of items in the array
	SInt32		CountItems() const { return fCount; };

	/*! @function FindIndexForItem Returns the FIRST index of an item inside the array or -1 if item can't be found. */
	SInt32		FindIndexForItem(ArmeArchivable* t) const;

	/*! @function RemoveAll Empties the array. */
	void		RemoveAll();

	/*! @function RemoveItemByVal Remove the FIRST item with the given value */
	void		RemoveItemByValue(ArmeArchivable* t);
	
	/*! @function RemoveItemByIndex Remove an item by index */
	void		RemoveItemByIndex(SInt32 i);
		
	/*! @function SetSize sets the reserved space of the array to "size". If size < CountItems(),
	    the extra items are lost */
	void		SetSize( SInt32 size );
	
	/*! @function ZzArchPtrArray builds a new instance and sets the reserved space of the array to "size" */
				ZzArchPtrArray(SInt32 size = 0) : fData(NULL), fCount(0), fMax(0) { SetSize(size); }
				
				ZzArchPtrArray(ArmeArchiveHandler* handler);
	bool		WriteToArchive(ArmeArchiveHandler* handler) const;

	/*! @function ~ZzArchPtrArray destroys the array and frees the buffer.*/
				virtual ~ZzArchPtrArray();

#ifndef NO_VERBOSE_AT_ALL
	virtual void		DumpToStream(std::ostream& s);
#endif //NO_VERBOSE_AT_ALL

private :
	/*! @function ZzArchPtrArray co copy ctor ! */
				ZzArchPtrArray( const ZzArchPtrArray& array );
				ZzArchPtrArray( ZzArchPtrArray& array );
};

inline ZzArchPtrArray::~ZzArchPtrArray()
{
	free(fData);
}

inline ZzArchPtrArray::ZzArchPtrArray(ArmeArchiveHandler* handler) :
	ArmeArchivable(handler), 
	fData(NULL), fCount(0), fMax(0)
{
	SInt32 count = handler->ReadSInt32();
	SetSize(count);
	for(SInt32 i = 0; i < count; i++)
	{
		ArmeArchivable* newObj = handler->ReadCPPObject();
		AddItem(newObj);
	}
}

inline bool ZzArchPtrArray::WriteToArchive(ArmeArchiveHandler* handler) const
{
	ArmeArchivable::WriteToArchive(handler);
	handler->WriteSInt32(CountItems());
	for(SInt32 i = 0; i < CountItems(); i++)
	{
		handler->WriteCPPObject(ItemAt(i));
	}
	return true;
}

#ifndef NO_VERBOSE_AT_ALL
inline void ZzArchPtrArray::DumpToStream(std::ostream& s)
{
	s << "dump for ZzArchPtrArray : " << std::endl;
	for( SInt32 i = 0; i < CountItems(); i++)
	{
		ArmeArchivable* obj = ItemAt(i);
		s << "Item " << i << " = " << *obj << std::endl;
	}
}

#endif //#ifndef NO_VERBOSE_AT_ALL
	
inline void ZzArchPtrArray::SetSize( SInt32 size )
{
	fMax = size;
	ArmeArchivable** oldData = fData;
	if (fCount > size)
		fCount = size;
	if (fMax)
	{
		fData = (ArmeArchivable**)malloc(sizeof(ArmeArchivable*)*size);
		if (oldData)
			memcpy( fData, oldData, sizeof(ArmeArchivable*)*fCount );
	}
	else
		fData = NULL;
//	vernl(0,"ZzArchPtrArray::SetSize() about to free oldData (" << (void*)oldData << ")");
	free(oldData);
}

inline ArmeArchivable* ZzArchPtrArray::ItemAt( SInt32 i ) const
{
	ZzAssert(i < fCount);
	return fData[i];
};

inline void ZzArchPtrArray::AddItem( ArmeArchivable* t )
{
	if( fCount == fMax )
		SetSize( fCount + 32 );
	fData[fCount] = t;
	fCount++;
}

inline void ZzArchPtrArray::InsertItem(ArmeArchivable* t, SInt32 i)
{
	ZzAssert(i <= fCount);
	
	if( fCount == fMax )
		SetSize( fCount + 32 );

	for(SInt32 j = fCount-1; j >= i; j--)
	{
		fData[j+1] = fData[j];
	}

	fData[i] = t;
	fCount++;
}

inline void ZzArchPtrArray::RemoveItemByIndex(SInt32 i)
{
	ZzAssert(i < fCount);
	// i don't know how memcpy and memove handle overlapping zones
	// so I do it the dumb way now
	for(SInt32 j = i; j < (fCount-1); j++)
	{
		fData[j] = fData[j+1];
	}
	fCount--;
}

inline SInt32 ZzArchPtrArray::FindIndexForItem(ArmeArchivable* t) const
{
	for(SInt32 i = 0; i < fCount; i++)
	{
		if(fData[i] == t)
			return i;
	}
	return -1;
}

inline void ZzArchPtrArray::RemoveItemByValue(ArmeArchivable* t)
{
	SInt32 i = FindIndexForItem(t);
	if(i >= 0)
		RemoveItemByIndex(i);
}

inline void ZzArchPtrArray::RemoveAll()
{
	SetSize(0);
}

