/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/
 /*
	ZzArray.h

	A class that stores whole objects
	and grows if necessary.
	
	Objects are copied when stored and retrieved.
	So better not use this for big long-lifetime objects
	Use ZzPtrArray instead
*/

#define ZzArray_h

#ifndef ZzArray_h
#include "ZzArray.h"
#endif //ZzArray_h

#ifndef ZzTypes_h
#include "ZzTypes.h"
#endif //ZzTypes_h

#ifndef ZzRtAssert_h
#include "ZzRtAssert.h"
#endif //ZzRtAssert_h

#ifndef ZzVerbose_h
#include "ZzVerbose.h"
#endif //ZzVerbose_h

template <class T> class ZzArray
{
public :
	// the array of items
	T*		data;
	// the current number of items in the array
	SInt32	count;
	// the current number of allocated space for items
	SInt32	_max;

public:
	// returns a a copy of the item at position i and returns 
	const T&	ItemAt( SInt32 i ) const;
	const T&	operator[]( SInt32 i ) const { return ItemAt(i); };

	T&			ItemAt( SInt32 i );
	T&			operator[]( SInt32 i ) { return ItemAt(i); };
	
	// adds a new item at the end of the array
	void		AddItem( T t );
		
	// insert a new item at index i and move the end of the array of one position	
	void		InsertItem(T t, SInt32 i);

	// number of items in the array
	SInt32		CountItems() const { return count; };

	/*! @function FindIndexForItem Returns the FIRST index of an item inside the array or -1 if item can't be found. */
	SInt32		FindIndexForItem(T t) const;

	/*! @function RemoveAll Empties the array. */
	void		RemoveAll();

	/*! @function RemoveItemByVal Remove the FIRST item with the given value */
	void		RemoveItemByValue(T t);
	
	/*! @function RemoveItemByIndex Remove an item by index */
	void		RemoveItemByIndex(SInt32 i);
		
	/*! @function SetSize sets the reserved space of the array to "size". If size < CountItems(),
	    the extra items are lost */
	void		SetSize( SInt32 size );
	
	/*! @function ZzArray builds a new instance and sets the reserved space of the array to "size" */
				ZzArray( SInt32 size = 0 ) : data(NULL), count(0), _max(0) { SetSize(size); };

	/*! @function ~ZzArray destroys the array and frees the buffer.*/
				~ZzArray() { delete [] data; };

private :
	/*! @function ZzArray co copy ctor ! */
				ZzArray( const ZzArray& array );
};

template<class T> inline void ZzArray<T>::SetSize( SInt32 size )
{
	_max = size;
	T* oldData = data;
	if (count > size)
		count = size;
	if (_max)
	{
		data = new T[size];
		if (oldData)
		{
			for(SInt32 j = 0; j < count; j++)
			{
				data[j] = oldData[j];
			}
			// WARNING !!! IF size IS GREATER THAN count, THEN THE DEFAULT CTOR IS USED FOR THESE
		}
	}
	else
		data = NULL;
	delete [] oldData;
}

template<class T> inline const T& ZzArray<T>::ItemAt( SInt32 i ) const
{
	ZzRtAssert(i < count);
	return data[i];
}

template<class T> inline T& ZzArray<T>::ItemAt( SInt32 i )
{
	ZzRtAssert(i < count);
	return data[i];
}

template<class T> inline void ZzArray<T>::AddItem( T t )
{
	if( count == _max )
		SetSize( count + 32 );
	data[count] = t;
	count++;
}

template<class T> inline void ZzArray<T>::InsertItem(T t, SInt32 i)
{
	ZzAssert(i <= count);
	
	if( count == _max )
		SetSize( count + 32 );

	for(SInt32 j = count-1; j >= i; j--)
	{
		data[j+1] = data[j];
	}

	data[i] = t;
	count++;
}

template<class T> inline void ZzArray<T>::RemoveItemByIndex(SInt32 i)
{
	assert(i < count);
	// i don't know how memcpy and memove handle overlapping zones
	// so I do it the dumb way now
	for(SInt32 j = i; j < (count-1); j++)
	{
		data[j] = data[j+1];
	}
	count--;
}

template<class T> inline SInt32 ZzArray<T>::FindIndexForItem(T t) const
{
	for(SInt32 i = 0; i < count; i++)
	{
		if(data[i] == t)
			return i;
	}
	return -1;
}

template<class T> inline void ZzArray<T>::RemoveItemByValue(T t)
{
	SInt32 i = FindIndexForItem(t);
	if(i >= 0)
		RemoveItemByIndex(i);
}

template<class T> inline void ZzArray<T>::RemoveAll()
{
	SetSize(0);
}



