/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/
 /*!
 *  ZzPtrArray is a template class that represents a dynamic array of pointers. 
 *  It can be owning or not.
 */

#define ZzPtrArray_h

#ifndef ZzPtrArray_h
#include "ZzPtrArray.h"
#endif //ZzPtrArray_h

#ifndef ZzTypes_h
#include "ZzTypes.h"
#endif //ZzTypes_h

#ifndef ZzVerbose_h
#include "ZzVerbose.h"
#endif //ZzVerbose_h

/*!
	@class ZzPtrArray
	This class defines and implements a dynamic array of pointers
	the ownership of the array is set with the "owning" template param
	If you store pointers to "float", then the T type is "float*"
*/

template <class T, bool owning> class ZzPtrArray
{
protected:
	/*! @var data the array of items */
	T*		data;
	/*! @var count the current number of items in the array */
	int32	count;
	/*! @var max the current number of allocated space for items */
	int32	max;

public:
	/*! @function ItemAt returns the value of the pointer at position i */
	T&			ItemAt( int32 i );
	/*! @function ItemAt returns the value of the pointer at position i */
	//const T&	ItemAt( int32 i ) const { return data[i]; };
	
	void		SetItemAt( int32 i, T val );

	/*! @function AddItem adds and item at the end of the array */
	void		AddItem( T t );
	
	/*! @function AddArray adds a whole array (of the same type) at the end of the array */
	void		AddArray( const ZzPtrArray<T,owning>* other );

	/*! @function RemoveItem Removes the first item which is equal to "t" */
	void		RemoveItem( T t );

	/*! @function RemoveItem Removes i-th item */
	T			RemoveItem( int32 i );

	/*! @function SortItems Sorts the items using a compare function */
	void		SortItems( int (CompareFunc)(const void *, const void *) );

	/*! @function IsEmpty returns whether the array is empty or not */
	bool		IsEmpty() const { return (count==0); }
	
	/*! @function RemoveAll empties the array. $$$ERROR TO BE FIXED : this function does not delete objects pointed even if
		the owning tparam is true */
	void		RemoveAll() { count = 0; }

	/*! @function ReplaceItem replace the i-th item with t. $$$ERROR TO BE FIXED : this function does not delete the old object pointed even if
		the owning tparam is true */
	void		ReplaceItem( int32 i, T t ) { data[i] = t; }

	/*! @function operator[] Shortcut for ItemAt */
	T			operator[]( int32 i ) { return ItemAt(i); };

	/*! @function operator[] Shortcut for ItemAt */
	const T&	operator[]( int32 i ) const { return ItemAt(i); };
	
	/*! @function GetCount Returns the actual number of objects in the array */
	int32		GetCount() const { return count; };
	
	/*! @function SetSize This function modified the current number of ALLOCATED space for items. This can be efficient to 
		avoid re-allocation later. If the given size is greater than the actual count of objects, then the existing objects
		are not affected. On the other hand, if you try to make the size smaller than the count of objects, then the extra 
		objects are lost. $$$ERROR these extra objects pointed are not deleted regardless of the "owning" tparam. */
	void		SetSize( int32 size );
	
	/*! @function Constructor This constructor can be passed an initial size as a clue to possibly avoid later reallocations*/ 
				ZzPtrArray( int32 size = 0 ) : data(NULL) { SetSize(size); };

	/*! @function Destructor This destructor deletes the object pointed by the array elements.*/ 
				~ZzPtrArray() { if(owning) delete [] data; };
};


template<class T,bool owning> inline void ZzPtrArray<T,owning>::SortItems( int (CompareFunc)(const void *, const void *) )
{
	qsort(data, count, sizeof(T), CompareFunc );
}


template<class T,bool owning> inline void AddArray( const ZzPtrArray<T,owning>* other )
{
	// $$$ optimize
	for( int32 i = 0; i < other.GetCount(); i++ )
	{
		AddItem( other->ItemAt(i) );
	}
}

template<class T,bool owning> inline void ZzPtrArray<T,owning>::SetSize( 
	int32 size )
{
	max = size;
	T* oldData = data;
	if (count > size)
		count = size;
	if (max)
	{
		data = new T[size];
		if (oldData)
			memcpy( data, oldData, sizeof(T)*count );
	}
	else
		data = NULL;
	delete [] oldData;
}

template<class T,bool owning>  inline T& ZzPtrArray<T,owning>::ItemAt( int32 i )
{
	if (i < count )
		return data[i];
	else
	{
		ZzAssert(i < count);
		return data[-1];
	}
}

template<class T,bool owning>  inline void ZzPtrArray<T,owning>::SetItemAt( int32 i, T val )
{
	if( i >= count )
		SetSize(i+1);
	data[i] = val;
}

template<class T,bool owning>  inline void ZzPtrArray<T,owning>::AddItem( T t )
{
	if( count == max )
		SetSize( count + 32 );
	data[count] = t;
	count++;
}

template<class T,bool owning> inline void ZzPtrArray<T,owning>::RemoveItem( T t )
{
	int32 count = GetCount();
	for( int32 i =0; i < count; i++ )
	{
		if( ItemAt(i) == t )
		{
			RemoveItem(i);
			return;
		}
	}
}

template<class T,bool owning> inline T ZzPtrArray<T,owning>::RemoveItem( int32 i )
{
	// if this is not the last element, then we move
	// the tail of the array that starts at i+1 one
	// step "left"
	// because if i==count-1 then, there wouldn't be
	// any tail to copy
	T t = ItemAt(i);
	if( i < count-1 )
		memmove( &(data[i]), &(data[i+1]), count-(i-1)*sizeof(T) );
	SetSize(count-1);
	return t;
}

typedef ZzPtrArray<void*,false> ZzVoidArray;


