/*! 

	@header ZzRABI_Generic.h A set of objects that implement the
	
	Resource Acquisition By Initialization 
	
	pattern, as described by Stroustrup
	
	Implement a new class by : 
	1) subclass ZzRABI
	2) store info needed to delete a resource in the ctor
	3) Delete the resources in a void DeleteResource() function
	
	Example usage :
	
	see TArmeQTFileHandler::CreateWSObjsFromFile( ) in HyperEngine-AV
	and ask Benjamin for any questions

*/

#include <vector>
#include <algorithm>
#include <cassert>

/*! @class ZzRABI Basic service that avoids to delete a resource if SaveResource has been called */
class ZzRABI
{
public : 
					  ZzRABI() { fDeleteResource = true; }
	void			SaveResource() { fDeleteResource = false; }
protected :
	bool fDeleteResource;
};

/*
  class ZzRABI_Macro

  Contains a set of RABIs and deletes them all at once
*/
class ZzRABI_Macro : public ZzRABI
{
public :
            ZzRABI_Macro  ( ) : ZzRABI( ) { }
           ~ZzRABI_Macro  ( );
    void    AddRABI       ( ZzRABI* rabi );
private :
  std::vector<ZzRABI*>    _rabis;
};

inline void ZzRABI_Macro::AddRABI( ZzRABI* rabi )
{
#ifdef _DEBUG
  assert(rabi != NULL); // we never know what can happen....
  // make sure the vector doesn't already contain this value
  std::vector<ZzRABI*>::iterator itor = std::find(_rabis.begin( ), _rabis.end( ), rabi);
  assert(itor == _rabis.end( ));
#endif //_DEBUG
  _rabis.push_back(rabi);
}

inline ZzRABI_Macro::~ZzRABI_Macro( )
{
  std::vector<ZzRABI*>::iterator itor = _rabis.begin( );
  while( itor != _rabis.end( ) )
    delete *itor;
}



/*!
	@class ZzRABI_ArrayDeleter
	
	This class calls, in its dtor, "delete []array" on an array supplied in its ctor
	THEN sets the array to NULL. This is why you need to pass the ADDRESS of the pointer
	to the array.

*/

template <typename T> class ZzRABI_ArrayDeleter : public ZzRABI
{
public :
	ZzRABI_ArrayDeleter(T** array) { fArray = array; };
	~ZzRABI_ArrayDeleter() { if(fDeleteResource) { delete [] (*fArray); *fArray = NULL; } };
private :
	T** fArray;
};

/*!
	@class Zz_RABI_ObjectDeleter
	
	This class calls, in its dtor, "delete object" on an object supplied in its ctor
	THEN sets the object ptr to NULL. This is why you need to pass the ADDRESS of the pointer
	to the object.

*/

template <typename T> class ZzRABI_ObjectDeleter : public ZzRABI
{
public :
	ZzRABI_ObjectDeleter(T** obj) { fObj = obj; };
	~ZzRABI_ObjectDeleter() { if(fDeleteResource) { delete (*fObj); *fObj = NULL; } };
private :
	T** fObj;
};

/*
  this RABI is meant to interface with C code 
  upon termination, if fDeleteResource is true,
  it calls the function with the given argument (arg is copied in ctor)
  a possible use could be 
  
  FILE* file = fopen(path,'r');
  if(!file)
    throw SomeException;
  ZzRABI_ResourceFreeingFunction rabiFile(fclose,file)

  .... do some stuff

  .... no need to call fclose( ) .... the rabi will do it
*/

#ifdef _WIN32 

// $$$TODO : check why gcc complains in here

template <typename FunctionPtrType, typename Argument_1_Type> class ZzRABI_ResourceFreeingFunction : public ZzRABI
{
public :
	ZzRABI_ResourceFreeingFunction  (FunctionPtrType func, Argument_1_Type arg1);
	~ZzRABI_ResourceFreeingFunction ( );
private :
  Argument_1_Type _arg1;
  FunctionPtrType _func;
};

template <typename FunctionPtrType, typename Argument_1_Type> inline ZzRABI_ResourceFreeingFunction<typename FunctionPtrType, typename Argument_1_Type>::ZzRABI_ResourceFreeingFunction(
  FunctionPtrType func, Argument_1_Type arg1) :
  _func(func),
  _arg1(arg1)
{
}

template <typename FunctionPtrType, typename Argument_1_Type> inline ZzRABI_ResourceFreeingFunction<typename FunctionPtrType, typename Argument_1_Type>::~ZzRABI_ResourceFreeingFunction( )
{
  if(fDeleteResource)
    (*_func)(_arg1);
}

#endif //_WIN32


