/****************************************************************************
 *   Copyright (C) 2004 by Arboretum Systems, Inc.                          *
 *   Technical administrator: Alain Mazy, hav@mazy.be                       *
 *   Lead program designer: Georges Jaroslaw, gjaro@arboretum.com           *
 *                                                                          *
 *   HyperEngine-AV is a video, audio and text editor that can be used to   *
 *   record, import, edit, process and arrange together video, sound and    *
 *   text. These are all achieved in the program's free-form document       *
 *   window, with no limitations to the number of tracks.                   *
 *   The purpose of this program is to provide to users at all levels of    *
 *   expertise a readily accessible tool enabling them to create movies,    *
 *   slide shows or presentations.                                          *
 *                                                                          *
 *   HyperEngine-AV itself is free software. However, plug-ins that can be  *
 *   used as add-ons to HyperEngine-AV may or may not be free software, at  *
 *   the discretion of each plug-in developer and as specified by these     *
 *   plug-ins' respective licenses.                                         *
 *   You can redistribute HyperEngine-AV itself and/or modify it under the  *
 *   terms of the GNU General Public License as published by the Free       *
 *   Software Foundation; either version 2 of the License, or (at your      *
 *   option) any later version.                                             *
 *                                                                          *
 *   HyperEngine-AV is distributed in the hope that it will be useful, but  *
 *   WITHOUT ANY WARRANTY; without even the implied warranty of             *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
 *   GNU General Public License for all applicable details.                 *
 *                                                                          *
 *   You should have received a copy of the GNU General Public License      *
 *   along with this program; if not, write to:                             *
 *   Free Software Foundation, Inc.,                                        *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.              *
 ****************************************************************************/

#define ZzRingBuffer_h

#ifndef ZzRingBuffer_h
#include "ZzRingBuffer.h"
#endif //ZzRingBuffer_h

#ifndef ArmeBasicTypes_h
#include "ArmeBasicTypes.h"
#endif //ArmeBasicTypes_h

#ifndef ZzRtAssert_h
#include "ZzRtAssert.h"
#endif //ZzRtAssert_h

class ZzRingBuffer
{
public :
	/*! @funtion ZzRingBuffer Builds a new ring buffer containing "1 << rbNbBits" items of size "itemSize" bytes. The write buffer is initially
		         set at pos delta (in items) while the read buffer is set at position 0 */
	ZzRingBuffer(SInt32 itemSize, SInt32 rbNbBits, SInt32 delta);
	
	void	WriteFromFlatBuffer(UInt8* source, SInt32 nbItems);
	void	IncrWritePos(SInt32 nbItems);
	void	ReadToFlatBuffer(UInt8* dst, SInt32 nbItems) const;
	void	IncrReadPos(SInt32 nbItems);
	SInt64 	WritePos() const; // beware of mthread issues !!!!!!!!!!
	SInt64 	ReadPos() const; // beware of mthread issues !!!!!!!!!!
	
	// allow to know where to directly write (useful to avoid tmp bufferz)
	void 	GetWriteFromFlatBufferInfo(SInt32 nbItems, UInt8** part1, SInt32* part1Size, UInt8** part2, SInt32* part2Size );
	
	UInt8*	DebugGetCurrentWriteAddress();
	UInt8*	DebugGetCurrentReadAddress();
	
private :
	SInt64	fReadPos;
	SInt64	fWritePos;
	UInt8*	fRingBuffer;
	SInt32	fRingBufferNbItems;
	SInt32	fItemSize;
};

inline SInt64 ZzRingBuffer::WritePos() const
{
	return fWritePos;
}

inline SInt64 ZzRingBuffer::ReadPos() const
{
	return fReadPos;
}

inline void ZzRingBuffer::WriteFromFlatBuffer(UInt8* source, SInt32 nbItems)
{
	ZzRtAssert(nbItems < fRingBufferNbItems);
	UInt32 itemIndexInRingBuffer = fWritePos & (fRingBufferNbItems - 1);
	UInt8* dstPtr = fRingBuffer + itemIndexInRingBuffer*fItemSize;
//	vernl(4,"writing " << nbItems << " items at dstPtr " << (void*)dstPtr);
	// part 2 length in items
	SInt32 part2NbItems = itemIndexInRingBuffer + nbItems - fRingBufferNbItems;
	if(part2NbItems <= 0)
	{
		memcpy(dstPtr,source,nbItems*fItemSize);
	}
	else
	{
		SInt32 part1NbItems = nbItems - part2NbItems;
		memcpy(dstPtr,source,part1NbItems*fItemSize);
		memcpy(fRingBuffer,source+part1NbItems*fItemSize,part2NbItems*fItemSize);
	}
}

inline void ZzRingBuffer::IncrWritePos(SInt32 nbItems)
{
	fWritePos += nbItems;
}

inline void ZzRingBuffer::ReadToFlatBuffer(UInt8* dst, SInt32 nbItems) const
{
	ZzRtAssert(nbItems < fRingBufferNbItems);
	UInt32 itemIndexInRingBuffer = fReadPos & (fRingBufferNbItems - 1);
	UInt8* srcPtr = fRingBuffer + itemIndexInRingBuffer*fItemSize;
//	vernl(4,"reading " << nbItems << " items from srcPtr " << (void*)srcPtr);
	// part 2 length in items
	SInt32 part2NbItems = itemIndexInRingBuffer + nbItems - fRingBufferNbItems;
	if(part2NbItems <= 0)
	{
		memcpy(dst,srcPtr,nbItems*fItemSize);
	}
	else
	{
		SInt32 part1NbItems = nbItems - part2NbItems;
		memcpy(dst,srcPtr,part1NbItems*fItemSize);
		memcpy(dst+part1NbItems*fItemSize,fRingBuffer,part2NbItems*fItemSize);
	}
}


inline UInt8* ZzRingBuffer::DebugGetCurrentWriteAddress()
{
	UInt32 itemIndexInRingBuffer = fWritePos & (fRingBufferNbItems - 1);
	UInt8* srcPtr = fRingBuffer + itemIndexInRingBuffer*fItemSize;
	return srcPtr;
}

inline UInt8* ZzRingBuffer::DebugGetCurrentReadAddress()
{
	UInt32 itemIndexInRingBuffer = fReadPos & (fRingBufferNbItems - 1);
	UInt8* srcPtr = fRingBuffer + itemIndexInRingBuffer*fItemSize;
	return srcPtr;
}



inline void ZzRingBuffer::IncrReadPos(SInt32 nbItems)
{
	fReadPos += nbItems;
}



